//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  device_file.c
* @brief Use a file as disk device
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     jerome.leveque@parrot.com
* @date       2007-02-08
*******************************************************************************
*/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/io/config_keys.h>
#include <pkgconf/devs_device_file.h>

#define FILE_BLOCK_SIZE			512
#define MEGABYTES				(1024*1024)
#define FILE_NB_BLOCKS			(MEGABYTES * CYGDAT_DEVS_DEVICE_FILE_DISK_SIZE / FILE_BLOCK_SIZE)
#define FILE_MAX_TRANFER_SIZE	(1<<31) // unlimited size (must be a multiple of 512)

static cyg_mutex_t disk_mutex;

typedef struct
{
	int file; //not used cause we will open and close each time the file
	int present;
} file_disk_info_t;

/**
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 */
static Cyg_ErrNo file_disk_disk_read(disk_channel * chan, void *buf, cyg_uint32 len, cyg_uint32 block_num)
{
	Cyg_ErrNo err = ENOERR;
	off_t ret;
	ssize_t ret_read;
	int opened_here = 0;
	file_disk_info_t *file_info = (file_disk_info_t*)chan->dev_priv;

	if (block_num + len > FILE_NB_BLOCKS)
		return -EINVAL;

        cyg_mutex_lock(&disk_mutex);

	if (file_info->file == -1)
	{
		opened_here = 1;
		file_info->file = open(CYGDAT_DEVS_DEVICE_FILE_DISK_NAME, O_RDONLY);
	}

	if (file_info->file != -1)
	{
		ret = lseek(file_info->file, block_num * FILE_BLOCK_SIZE, SEEK_SET);
		if (ret == (off_t)-1)
		{
			err = -EIO;
		}
		else
		{
			ret_read = read(file_info->file, buf, len * FILE_BLOCK_SIZE);
			if (ret_read == (ssize_t)-1)
			{
				err = -EIO;
			}
		}
		if (opened_here == 1)
		{
			if (close(file_info->file) == -1) {
				err = -EIO;
			}
			file_info->file = -1;
		}
	}
	else
	{
		err = -EIO;
	}

        cyg_mutex_unlock(&disk_mutex);

	return err;
}

/**
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 */
static Cyg_ErrNo file_disk_disk_write(disk_channel * chan, const void *buf, cyg_uint32 len, cyg_uint32 block_num)
{
	Cyg_ErrNo err = ENOERR;
	off_t ret;
	ssize_t ret_read;
	int opened_here = 0;
	file_disk_info_t *file_info = (file_disk_info_t*)chan->dev_priv;

	if (block_num + len > FILE_NB_BLOCKS)
		return -EINVAL;

        cyg_mutex_lock(&disk_mutex);

	if (file_info->file == -1)
	{
		opened_here = 1;
		file_info->file = open(CYGDAT_DEVS_DEVICE_FILE_DISK_NAME, O_WRONLY);
	}
	if (file_info->file != -1)
	{
		ret = lseek(file_info->file, block_num * FILE_BLOCK_SIZE, SEEK_SET);
		if (ret == (off_t)-1)
		{
			err = -EIO;
		}
		else
		{
			ret_read = write(file_info->file, buf, len * FILE_BLOCK_SIZE);
			if (ret_read == (ssize_t)-1)
			{
				err = -EIO;
			}
		}
		if (opened_here == 1)
		{
			if (close(file_info->file) == -1)
				err = -EIO;
			file_info->file = -1;
		}
	}
	else
	{
		err = -EIO;
	}

        cyg_mutex_unlock(&disk_mutex);

	return err;
}

/*
 * There are no supported get_config() operations at this time.
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 *
 * return -EINVAL
 */
static Cyg_ErrNo file_disk_get_config(disk_channel * chan, cyg_uint32 key, const void *xbuf, cyg_uint32 *len)
{
	return -EINVAL;
}

/**
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 */
static Cyg_ErrNo file_disk_set_config(disk_channel * chan, cyg_uint32 key, const void *xbuf, cyg_uint32 *len)
{
	file_disk_info_t *file_info = (file_disk_info_t*)chan->dev_priv;
	Cyg_ErrNo result = -EIO;

        cyg_mutex_lock(&disk_mutex);
	switch (key)
	{
		case CYG_IO_SET_CONFIG_DISK_MOUNT:
			if (file_info->file == -1)
			{
				file_info->file = open(CYGDAT_DEVS_DEVICE_FILE_DISK_NAME, O_RDWR | O_SYNC);
				if (file_info->file != -1)
					result = ENOERR;
			}
			break;
		
		case CYG_IO_SET_CONFIG_DISK_UMOUNT:
			if (file_info->file != -1)
			{
				if (close(file_info->file) == 0)
					result = ENOERR;

				file_info->file = -1;
			}
			break;
			
		default:
			result = ENOERR;
			break;
	}

        cyg_mutex_unlock(&disk_mutex);
	return result;
}

/**
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 */
static cyg_bool file_disk_disk_init(struct cyg_devtab_entry *tab)
{
	disk_channel *chan = (disk_channel*) tab->priv;
	cyg_mutex_init(&disk_mutex);
	return (*chan->callbacks->disk_init)(tab);
}

/**
 * @see http://canari/cgi-bin/cvsweb.cgi/~checkout~/SoftParrot5/ecos/ecos-current/doc/html/ref/diskio-driver.html
 */
static Cyg_ErrNo file_disk_disk_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab, const char *name)
{
	Cyg_ErrNo ret = ENOERR;
	cyg_disk_identify_t ident;
	disk_channel *chan = (disk_channel*)(*tab)->priv;
	file_disk_info_t *file_info = (file_disk_info_t*)chan->dev_priv;
	struct stat sstat;

        cyg_mutex_lock(&disk_mutex);

	if (file_info->present) {
		// check that the file is still present
		if (stat(CYGDAT_DEVS_DEVICE_FILE_DISK_NAME, &sstat) == -1) {
			file_info->present = 0;
			(*chan->callbacks->disk_disconnected)(chan);
			if (0 != chan->info->mounts) {
				ret = -ENODEV;
			}
			ret = -ENODEV;
		} else {
            		cyg_mutex_unlock(&disk_mutex);
			ret = (*chan->callbacks->disk_lookup)(tab, sub_tab, name);
			cyg_mutex_lock(&disk_mutex);
			goto END;
		}
	}
    	
	if ((chan->info->mounts != 0) && !file_info->present) {
		// There are still mount points to an old device. We cannot accept
		// new mount requests until those have been cleaned out.
		ret = -ENODEV;
	}
	
	if (ret == ENOERR) {
		ident.serial[0]			= '\0';
		ident.firmware_rev[0]	= '\0';
		ident.model_num[0]		= '\0';
		ident.cylinders_num		= 0; // not used
		ident.heads_num			= 0; // not used
		ident.sectors_num		= 0; // not used
		ident.lba_sectors_num	= FILE_NB_BLOCKS;
		ident.phys_block_size	= FILE_BLOCK_SIZE;
		ident.max_transfer		= FILE_MAX_TRANFER_SIZE;
	
		file_info->present = 1;
        	cyg_mutex_unlock(&disk_mutex);
		(*chan->callbacks->disk_connected)(*tab, &ident);
		// We now have a valid disk and higher-level code knows about it
		ret = (*chan->callbacks->disk_lookup)(tab, sub_tab, name);
        	cyg_mutex_lock(&disk_mutex);
	}
END:
	cyg_mutex_unlock(&disk_mutex);
	return ret;
}

DISK_FUNS(cyg_file_disk_disk_funs, file_disk_disk_read, file_disk_disk_write, file_disk_get_config, file_disk_set_config);

static file_disk_info_t file_disk_info_0 = {
	.file = -1,
	.present = 0
};

DISK_CONTROLLER(cyg_file_disk_disk_controller_0, file_disk_info_0);

DISK_CHANNEL(	cyg_file_disk_disk0_channel,
				cyg_file_disk_disk_funs,
				file_disk_info_0,
				cyg_file_disk_disk_controller_0,
				true,
				1);

BLOCK_DEVTAB_ENTRY(	cyg_file_disk_disk0_devtab_entry,
					CYGDAT_DEVS_DEVICE_FILE_DISK_NAME,
					0,
					&cyg_io_disk_devio,
					&file_disk_disk_init,
					&file_disk_disk_lookup,
					&cyg_file_disk_disk0_channel);

