//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file     mmchs.c
* @brief    SDcard low level driver for OMAP3630
*
* Copyright (C) 2012 Parrot S.A.
*
* @author   jerome.leveque@parrot.com
* @date     2012-04-01
*******************************************************************************
*/

#include <cyg/hal/cpu_types.h>
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_io.h>
#include <cyg/kernel/kapi.h>
#include <cyg/io/sdcard.h>
#include <assert.h>

#include <cyg/hal/omap3630_mmchs.h>

#include "mmchs.h"
#include <pkgconf/devs_disk_sdcard_omap3630.h>
#include <cyg/hal/omap3630_sdma.h>

#define MMCHS_ISR_PRIORITY      (4)
#define MMCHS_MCLK_SPEED        (96000000)          /* 96 MHz */
#define MMCHS_TRANSFERT_TIMEOUT (250)               /* 250 ms */

typedef struct MMCHS_privData {
    cyg_uint32  mmchs_freq;
    cyg_sem_t   mmchs_sem_done;
    cyg_bool    mmchs_data_transfert;
    cyg_uint32  mmchs_interrupt;
    cyg_uint32  mmchs_cmd_reg;
    cyg_uint32  mmchs_dma_tx;
    cyg_uint32  mmchs_dma_rx;
    cyg_uint32  (*wait_end)(cyg_sdcard_disk_info_t *sdcard_info);
} MMCHS_privData;

static MMCHS_privData g_privData;

// interrupt handlers
static cyg_handle_t     sdcard_int_handle;
static cyg_interrupt    intr;
static cyg_uint32       mmchs_isr(cyg_vector_t vector, cyg_addrword_t data);
static void             mmchs_dsr(cyg_vector_t vector, cyg_ucount32 count,
                                  cyg_addrword_t data);


static void
mmchs_reset(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 reg;

    reg = MMCHS_SYSCONFIG_AUTOIDLE              |
        MMCHS_SYSCONFIG_RESET                   |
        MMCHS_SYSCONFIG_ENAWAKEUP               |
        MMCHS_SYSCONFIG_SIDLEMODE_IGNORE        |
        MMCHS_SYSCONFIG_CLOCKACTIVITY_MAINTAIN;

    mmchs_write_reg(sdcard_info->controller_base,
                    OMAP3630_MMCHS_SYSCONFIG, reg);

    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
                             OMAP3630_MMCHS_SYSSTATUS);
    } while (reg != MMCHS_SYSSTATUS_RESET_DONE);
}


static void
mmchs_reset_line(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint32 line)
{
    cyg_uint32 reg;

    reg = mmchs_read_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_SYSCTL);

    //Start reset of the line (dat or command)
    reg |= line;
    mmchs_write_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_SYSCTL, reg);

    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
            OMAP3630_MMCHS_SYSCTL);
    } while (reg & line);
}

/**
 * Set the current frequency of the interface
 *
 * @param freq frequency in Hz
 */
static void
mmchs_set_frequency(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint32 freq)
{
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;
    cyg_uint32 sysctl;
    cyg_uint16 reg;
    cyg_uint32 div, dto = 0;
    cyg_uint32 current_timeout;

    //Disable Clock provided to sdcard
    sysctl = mmchs_read_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_SYSCTL);

    sysctl &= ~(MMCHS_SYSCTL_CEN);
    mmchs_write_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_SYSCTL, sysctl);

    //Select divider for sdclk
    for (div = 1; div < MMCHS_SYSCTL_CLK_DIV_MAX; div++) {
        if ((priv->mmchs_freq / div) <= freq)
            break;
    }

    //Real frequency of the controller
    freq = (priv->mmchs_freq / div);

    sysctl &= ~(MMCHS_SYSCTL_CLK_DIV_MAX << MMCHS_SYSCTL_CLK_DIV_SHIFT);
    sysctl |= (div << MMCHS_SYSCTL_CLK_DIV_SHIFT);
    mmchs_write_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_SYSCTL, sysctl);

    //And wait for stable bit
    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_SYSCTL);
    } while (!(reg & MMCHS_SYSCTL_ICS));

    //Now enable the clock to the card
    sysctl |= MMCHS_SYSCTL_CEN;

    //current_timeout in ms
    current_timeout = ((1 << (13 + MMCHS_SYSCTL_DTO_MAX)) / freq) * 1000;
    dto = MMCHS_SYSCTL_DTO_MAX;
    while (current_timeout > MMCHS_TRANSFERT_TIMEOUT) {
        dto--;
        current_timeout >>= 1;
    }

    if (dto < MMCHS_SYSCTL_DTO_MAX) {
        dto++;
    }

    DEBUG3("Timeout reg: %d\n", dto);

    sysctl &= ~(MMCHS_SYSCTL_DTO_MASK << MMCHS_SYSCTL_DTO_SHIFT);
    sysctl |= dto << MMCHS_SYSCTL_DTO_SHIFT;

    //Write timeout value and enable card clock
    mmchs_write_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_SYSCTL, sysctl);
}


static void
mmchs_prepare_transfert(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_bool read_nWrite, CYG_ADDRESS phys_addr, cyg_uint32 block_number)
{
    cyg_uint32 mmc_fifo, block_reg;
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;

    //Default reg option
    priv->mmchs_cmd_reg = MMCHS_CMD_DE;
    block_reg = (SDCARD_BLOCK_SIZE << MMCHS_BLK_BLEN_SHIFT);

    //Set transfert direction
    priv->mmchs_cmd_reg |= read_nWrite ?
        MMCHS_CMD_DDIR_READ : MMCHS_CMD_DDIR_WRITE;

    //Set block count
    if (block_number > 1) {
        block_reg |= (block_number << MMCHS_BLK_NBLK_SHIFT);
        priv->mmchs_cmd_reg |= (MMCHS_CMD_BCE | MMCHS_CMD_MSBS);
    }

    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_BLK,
            block_reg);

    //Set DMA address
    //TODO: check if without MMU address is the same
    mmc_fifo = (cyg_uint32)sdcard_info->controller_base + OMAP3630_MMCHS_DATA;
    if (read_nWrite) {
        //Read: Card to controller
        set_dma_transfert(priv->mmchs_dma_rx,
            (void*)mmc_fifo, (void*)phys_addr,
            block_number * SDCARD_BLOCK_SIZE, SDCARD_BLOCK_SIZE,
            DATATYPE_32_BITS, ADDRESS_MODE_CONSTANT, ADDRESS_MODE_POST_INCR,
            MODE_SYNC_SOURCE | MODE_SYNC_FRAME_SYNC);
        start_dma_transfert(priv->mmchs_dma_rx, NULL);
    } else {
        //Write: Controller to card
        set_dma_transfert(priv->mmchs_dma_tx,
            (void*)phys_addr, (void*)mmc_fifo,
            block_number * SDCARD_BLOCK_SIZE, SDCARD_BLOCK_SIZE,
            DATATYPE_32_BITS, ADDRESS_MODE_POST_INCR, ADDRESS_MODE_CONSTANT,
            MODE_SYNC_DEST | MODE_SYNC_FRAME_SYNC);
        start_dma_transfert(priv->mmchs_dma_tx, NULL);
    }
}


static SDCARD_eError
mmchs_handle_end_command_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    // check errors and response
    if (status & MMCHS_INT_ERR_CMD_TIMEOUT) {
        err = SDCARD_ERROR_TIMEOUT;
    } else if (status & MMCHS_INT_ERR_CMD_CRC) {
        err = SDCARD_ERROR_CRCFAIL;
    } else if (status & MMCHS_INT_ERR_CMD_ENDBIT) {
        err = SDCARD_ERROR_CMD_HW_ERROR;
    } else if (status & MMCHS_INT_ERR_CMD_INDEX) {
        err = SDCARD_ERROR_INDEX;
    } else if (status & MMCHS_INT_CMD_COMPLETE) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


/**
 * Handle the dpsm
 */
static SDCARD_eError
mmchs_handle_end_data_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    if (status & MMCHS_INT_ERR_DATA_TIMEOUT) {
        err = SDCARD_ERROR_DATA_TIMEOUT;
    } else if (status & MMCHS_INT_ERR_DATA_CRC) {
        err = SDCARD_ERROR_DATA_HW_ERROR;
    } else if (status & MMCHS_INT_ERR_DATA_ENDBIT) {
        err = SDCARD_ERROR_DATA_HW_ERROR;
    } else if (status & MMCHS_INT_ERR_DMA) {
        err = SDCARD_ERROR_DATA_FIFO_UNDERRUN;
    } else if (status & MMCHS_INT_DATA_END) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


static SDCARD_eError
mmchs_handle_end_transfert(cyg_uint32 status)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;

    if (status) {
        //Data interrupt
        if (status & MMCHS_INT_DATA_MASK) {
            err |= mmchs_handle_end_data_transfert(status
                        & MMCHS_INT_DATA_MASK);
        }

        //Command interrupt
        if (status & MMCHS_INT_CMD_MASK) {
            err |= mmchs_handle_end_command_transfert(status
                        & MMCHS_INT_CMD_MASK);
        }
    }

    return err;
}


static cyg_uint32
mmchs_wait_end_polling(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 status = 0;

    //Poll Normal interrupt status
    do {
        status = mmchs_read_reg(sdcard_info->controller_base,
                    OMAP3630_MMCHS_STAT);
    } while (status == 0);

    return status;
}


static cyg_uint32
mmchs_wait_end_irq(cyg_sdcard_disk_info_t *sdcard_info)
{
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;

    cyg_semaphore_wait(&(priv->mmchs_sem_done));
    status = mmchs_read_reg(sdcard_info->controller_base,
                    OMAP3630_MMCHS_STAT);

    return status;
}


static SDCARD_eError
mmchs_wait_end(cyg_sdcard_disk_info_t *sdcard_info)
{
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;
    SDCARD_eError ret;

    status = priv->wait_end(sdcard_info);

    DEBUG3("mmchs_wait_end: status 0x%x\n", status);

    if (status & MMCHS_INT_DATA_END)
        priv->mmchs_data_transfert = false;

    status &= MMCHS_INT_ALL_MASK;

    if (status == MMCHS_INT_ERR_DMA) {
        DEBUG1("DMA error\n");
    }

    ret = mmchs_handle_end_transfert(status);

    if (ret == SDCARD_ERROR_DATA_TIMEOUT) {
        DEBUG1("Data Timeout\n");
        mmchs_reset_line(sdcard_info, MMCHS_SYSCTL_SRD);
    }

    if (ret == SDCARD_ERROR_TIMEOUT) {
        DEBUG1("Command Timeout\n");
        mmchs_reset_line(sdcard_info, MMCHS_SYSCTL_SRC);
    }

    //Clear error and/or status
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_STAT, status);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //We have cleared our interrupt so unmask it
        cyg_drv_interrupt_unmask(priv->mmchs_interrupt);
    }

    return ret;
}

static cyg_uint32
mmchs_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(vector);
    cyg_drv_interrupt_acknowledge(vector);

    return (CYG_ISR_CALL_DSR | CYG_ISR_HANDLED);
}


static void
mmchs_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_sdcard_disk_info_t *sdcard_info = (cyg_sdcard_disk_info_t*)data;
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;

    cyg_semaphore_post(&(priv->mmchs_sem_done));

    //Do not unmask interrupt now
}


SDCARD_eError
ll_sdcard_hw_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 caps, add_caps;
    cyg_uint32 reg;
    cyg_uint32 intmask = MMCHS_INT_CMD_COMPLETE;

    if (sdcard_info->init == true)
        return SDCARD_ERROR_NONE;

	//Set private data
    sdcard_info->priv = (void*)&g_privData;

    //Set base controller address
#if (CYGPKG_DEVS_DISK_SDCARD_OMAP3630_MMCHS_INTERFACE == 1)
    sdcard_info->controller_base = (cyg_uint8*)OMAP3630_MMCHS1_BASE;
    g_privData.mmchs_interrupt = CYGNUM_HAL_INTERRUPT_MMC1_IRQ;
    g_privData.mmchs_dma_tx = OMAP3630_DMA4_MMC1_DMA_TX;
    g_privData.mmchs_dma_rx = OMAP3630_DMA4_MMC1_DMA_RX;

    //Select PIN
    reg = 0;
    //sdmmc1_clk sdmmc1_cmd
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x144, reg);
    //sdmmc1_dat0 sdmmc1_dat1
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x148, reg);
    //sdmmc1_dat2 sdmmc1_dat3
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x14C, reg);
    add_caps = MMCHS_CAPA_VS18 | MMCHS_CAPA_VS30;
#elif (CYGPKG_DEVS_DISK_SDCARD_OMAP3630_MMCHS_INTERFACE == 2)
    sdcard_info->controller_base = (cyg_uint8*)OMAP3630_MMCHS2_BASE;
    g_privData.mmchs_interrupt = CYGNUM_HAL_INTERRUPT_MMC2_IRQ;
    g_privData.mmchs_dma_tx = OMAP3630_DMA4_MMC2_DMA_TX;
    g_privData.mmchs_dma_rx = OMAP3630_DMA4_MMC2_DMA_RX;

    //Select PIN
    reg = 0;
    //sdmmc2_clk sdmmc2_cmd
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x158, reg);
    //sdmmc2_dat0 sdmmc2_dat1
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x15c, reg);
    //sdmmc2_dat2 sdmmc2_dat3
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x160, reg);
    add_caps = MMCHS_CAPA_VS18;
#elif (CYGPKG_DEVS_DISK_SDCARD_OMAP3630_MMCHS_INTERFACE == 3)
    sdcard_info->controller_base = (cyg_uint8*)OMAP3630_MMCHS3_BASE;
    g_privData.mmchs_interrupt = CYGNUM_HAL_INTERRUPT_MMC3_IRQ;
    g_privData.mmchs_dma_tx = OMAP3630_DMA4_MMC3_DMA_TX;
    g_privData.mmchs_dma_rx = OMAP3630_DMA4_MMC3_DMA_RX;

    //Select PIN
    reg = 0x00030003
    //sdmmc3_clk sdmmc3_cmd
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x1D0, reg);
    //sdmmc3_dat0 sdmmc3_dat1
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x164 reg);
    //sdmmc3_dat2 sdmmc3_dat3
    omap3630_write_reg(OMAP3630_SCM_INTERFACE_BASE + 0x168, reg);
    add_caps = MMCHS_CAPA_VS18;
#else
# Warning OMAP3630 MMCHS interface not set
#endif

    //Reset controller for starting clean
    mmchs_reset(sdcard_info);

    //Check capabilities
    caps = mmchs_read_reg(sdcard_info->controller_base,
                         OMAP3630_MMCHS_CAPA);

    CYG_ASSERT(caps & MMCHS_CAPA_DS,
        "DMA NOT supported");
    CYG_ASSERT(caps & MMCHS_CAPA_HSS,
        "High Speed mode NOT supported");
    CYG_ASSERT(caps & MMCHS_CAPA_SRS,
        "Suspend/Resume NOT supported");

    caps |= add_caps;
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CAPA, caps);

    //Set configuration register
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CON, 0);

    //Set Host controller register
    if (add_caps & MMCHS_CAPA_VS30) {
        reg = MMCHS_HCTL_SDVS_3_0_V;
    } else {
        reg = MMCHS_HCTL_SDVS_1_8_V;
    }
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_HCTL, reg);

    reg |= MMCHS_HCTL_SDBP;
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_HCTL, reg);

    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
                             OMAP3630_MMCHS_HCTL);
    } while ((reg & MMCHS_HCTL_SDBP) != MMCHS_HCTL_SDBP);

    //Turn on internal clock
    reg = MMCHS_SYSCTL_ICE;
    reg |= (MMCHS_SYSCTL_CLK_DIV_MAX << MMCHS_SYSCTL_CLK_DIV_SHIFT);
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_SYSCTL, reg);

    //And wait for stable bit
    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_SYSCTL);
    } while (!(reg & MMCHS_SYSCTL_ICS));

    //Set signal we will manage
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_IE, intmask);
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_ISE, intmask);

    //MMCHS Controller INIT Procedure Start
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CON, MMCHS_CON_INIT);
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CMD, 0);

    //Wait for the end of init sequence
    do {
        reg = mmchs_read_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_STAT);
    } while (!(reg & MMCHS_INT_CMD_COMPLETE));

    //Clear status
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_STAT,
        reg);

    //Remove the init bit for sending command
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CON, 0);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //For using interrupt
        g_privData.wait_end = mmchs_wait_end_irq;
        cyg_semaphore_init(&(g_privData.mmchs_sem_done), 0);

        // interrupt init
        cyg_drv_interrupt_create(
                g_privData.mmchs_interrupt,
                MMCHS_ISR_PRIORITY,
                (cyg_addrword_t)sdcard_info,
                (cyg_ISR_t*)&mmchs_isr,
                (cyg_DSR_t*)&mmchs_dsr,
                &sdcard_int_handle,
                &intr);
        cyg_drv_interrupt_attach(sdcard_int_handle);
        cyg_drv_interrupt_unmask(g_privData.mmchs_interrupt);
    } else if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_POLLING) {
        //For using polling
        g_privData.wait_end = mmchs_wait_end_polling;
    } else {
        return SDCARD_ERROR_DATA_HW_ERROR;
    }

    //Set controller base frequency
    g_privData.mmchs_freq = MMCHS_MCLK_SPEED;

    //Init done and OK
    sdcard_info->init = true;

    return SDCARD_ERROR_NONE;
}

SDCARD_eError
ll_sdcard_card_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;
    cyg_uint32 freq;
    cyg_uint32 ctrl;

    //Set frequency according to sdcard connected
    freq = CYG_MIN(priv->mmchs_freq, sdcard_info->sdcard_freq);
    mmchs_set_frequency(sdcard_info, freq);

    if (err == SDCARD_ERROR_NONE) {
        //Set bus width
        ctrl = mmchs_read_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_HCTL);
        ctrl &= ~(MMCHS_HCTL_DTW);
        switch (sdcard_info->sdcard_card_type) {
            case SDCARD_TYPE_MMC:
                //nothing to do
                break;

            case SDCARD_TYPE_SDCARD:
            case SDCARD_TYPE_SDHC:
                ctrl |= MMCHS_HCTL_DTW;
                break;

            default:
                err = SDCARD_ERROR_HARDWARE;
                break;
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        ctrl &= ~(MMCHS_HCTL_SDVS_MASK | MMCHS_HCTL_SDBP);

        //Set voltage
        switch (sdcard_info->sdcard_voltage) {
            default:
            case SDCARD_VOLTAGE_NONE:
                //Turn off
                break;

            case SDCARD_VOLTAGE_1V8:
                ctrl |= (MMCHS_HCTL_SDVS_1_8_V | MMCHS_HCTL_SDBP);
                break;

            case SDCARD_VOLTAGE_3V0:
                ctrl |= (MMCHS_HCTL_SDVS_3_0_V | MMCHS_HCTL_SDBP);
                break;

            case SDCARD_VOLTAGE_3V3:
                ctrl |= (MMCHS_HCTL_SDVS_3_3_V | MMCHS_HCTL_SDBP);
                break;
        }

        mmchs_write_reg(sdcard_info->controller_base,
                OMAP3630_MMCHS_HCTL, ctrl);
    }

    return err;
}

SDCARD_eError
ll_sdcard_send_command(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_uint32 command, cyg_uint32 argument, cyg_uint32 *response)
{
    MMCHS_privData *priv = (MMCHS_privData*)sdcard_info->priv;
    cyg_uint32 intmask = MMCHS_INT_ALL_MASK;
    cyg_uint32 present_state_reg;
    cyg_uint32 present_state_mask = MMCHS_PSTATE_CMDI;
    cyg_uint32 command_reg;
    SDCARD_eError ret;

    //Set command argument
    mmchs_write_reg(sdcard_info->controller_base,
        OMAP3630_MMCHS_ARG, argument);

    command_reg = priv->mmchs_cmd_reg;

    //Set response type
    if (!(command & SDCARD_COMMAND_RESPONSE)) {
        command_reg |= MMCHS_CMD_RSP_NONE;
    } else if (command & SDCARD_COMMAND_LONG_RESPONSE) {
        command_reg |= MMCHS_CMD_RSP_136;
    } else if (command & SDCARD_COMMAND_ASK_BUSY) {
        command_reg |= MMCHS_CMD_RSP_48_BUSY;
    } else {
        command_reg |= MMCHS_CMD_RSP_48;
    }

    if (command & SDCARD_COMMAND_ASK_CRC) {
        command_reg |= MMCHS_CMD_CCCE;
    }
    if (command & SDCARD_COMMAND_ASK_OPCODE) {
        command_reg |= MMCHS_CMD_CICE;
    }
    if (command & SDCARD_COMMAND_WITH_DATA) {
        priv->mmchs_data_transfert = true;
        command_reg |= MMCHS_CMD_DP;
        present_state_mask |= MMCHS_PSTATE_DATI;
    } else {
        // Remove data end bit cause it can appear before
        intmask &= ~(MMCHS_INT_DATA_END);
    }

    //Set signal we will manage
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_IE,
        intmask);
    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_ISE,
        intmask);

    if (command & SDCARD_COMMAND_ASK_ABORT)
        command_reg |= MMCHS_CMD_CMD_FNCT_ABORT;

    command_reg |= ((command & SDCARD_COMMAND_MASK) <<
        MMCHS_CMD_CMD_INDEX_SHIFT);

    //The Host Driver shall check the Command Inhibit (DAT) bit
    //and Command Inhibit (CMD) bit in the Present State register
    //before writing to this register
    do {
        present_state_reg = mmchs_read_reg(sdcard_info->controller_base,
            OMAP3630_MMCHS_PSTATE);
    } while (present_state_reg & present_state_mask);

    mmchs_write_reg(sdcard_info->controller_base, OMAP3630_MMCHS_CMD,
            command_reg);

    // wait for transfer end
    ret = mmchs_wait_end(sdcard_info);

    if (ret == SDCARD_ERROR_NONE) {
        if (response && (command & SDCARD_COMMAND_RESPONSE)) {
            if (command & SDCARD_COMMAND_LONG_RESPONSE) {
                int reg;
                cyg_uint32 resp[4];
                for (reg = 0; reg < 4; reg++) {
                    resp[reg] = mmchs_read_reg(sdcard_info->controller_base,
                                    (OMAP3630_MMCHS_RSP10 + reg * 4));
                }

                for (reg = 0; reg <= 2; reg++) {
                    response[reg] = resp[reg] >> 8;
                    response[reg] |= (resp[reg + 1] & 0xFF) << 24;
                }
                response[3] = resp[3] >> 8;
            } else if (command & SDCARD_COMMAND_RESPONSE) {
                cyg_uint32 resp;
                resp = mmchs_read_reg(sdcard_info->controller_base,
                        OMAP3630_MMCHS_RSP10);
                response[0] = resp;
            }
        }
    } else {
        mmchs_reset(sdcard_info);
    }

    if (priv->mmchs_data_transfert == true) {
        cyg_uint32 sdma = -1;

        if (command_reg & MMCHS_CMD_DDIR_READ) {
            //Read: Card to controller
            sdma = priv->mmchs_dma_rx;
        } else {
            //Write: Controller to card
            sdma = priv->mmchs_dma_tx;
        }

        DEBUG3("Wait transfert end...\n");

        while ((ret == SDCARD_ERROR_NONE) && (priv->mmchs_data_transfert)) {
            ret = mmchs_wait_end(sdcard_info);
        }

        stop_dma_transfert(sdma);
    }

    priv->mmchs_cmd_reg = 0;

    return ret;
}


SDCARD_eError
ll_sdcard_send_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_write)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    mmchs_prepare_transfert(sdcard_info, false, phys_addr, block_to_write);

    return ret;
}


SDCARD_eError
ll_sdcard_receive_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_read)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    mmchs_prepare_transfert(sdcard_info, true, phys_addr, block_to_read);

    return ret;
}
