//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
**************************************************************
* @file parrot5_i2c.c
* @brief I2C master driver for the Parrot5. <br>
* Theses low-level functions are called by the Ecos I2C generic package.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author   gautier.seidel@parrot.fr
* @date     2006-02-22
*
**************************************************************
*/

#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/diag.h>         // diagnostic output
#include <cyg/io/i2c.h>

#include <pkgconf/devs_io_i2c_arm_parrot5.h>

#include "i2c_master.h"

// support for the debug
#if 1
# define DEBUG(_format_, ...)
#else
# define DEBUG(format, ...) diag_printf("%s(%d): " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)
#endif

#if 1
# define DEBUG_ERR(_format_, ...)
#else
# define DEBUG_ERR(format, ...) diag_printf("%s(%d): " format, __FUNCTION__, __LINE__, ## __VA_ARGS__)
#endif

#define I2C_TIMEOUT          (1000000)
#define I2C_EVENT_TIMEOUT    (2)          /* event timeout in ticks  = 20 ms */

static cyg_flag_t parrot5_i2cm_flag;

#ifdef CYGNUM_DEVS_IO_I2C_ARM_PARROT5_USE_IRQS
static const int parrot5_i2cm_use_irqs = 1;
#else
static const int parrot5_i2cm_use_irqs = 0;
#endif

//! a dummy object, we can used it to store bus-specific data if needed
struct parrot5_i2c_data 
{
    
} parrot5_i2c_object;

static void parrot5_i2c_prepare_completion(void)
{
    // do not attempt to sleep in cyg_user_start() context
    if (parrot5_i2cm_use_irqs &&
        (cyg_thread_self() != cyg_thread_idle_thread())) {
        // enable interrupt
        i2cm_write_reg(I2CM_ITEN, I2CM_ITEN_ITEN);
    }
}

static cyg_uint32 parrot5_i2c_wait_completion(void)
{
    cyg_uint32 status, timeout = I2C_TIMEOUT;

    if (parrot5_i2cm_use_irqs && (i2cm_read_reg(I2CM_ITEN) & I2CM_ITEN_ITEN)) {

		// sleep and wait for an event
        cyg_flag_timed_wait(&parrot5_i2cm_flag, 1,
                            CYG_FLAG_WAITMODE_OR|CYG_FLAG_WAITMODE_CLR,
                            cyg_current_time()+I2C_EVENT_TIMEOUT);

        // disable interrupt
        i2cm_write_reg(I2CM_ITEN, 0);
    }

    do {
        status = i2cm_read_reg(I2CM_STATUS);
        timeout--;
    } while ((status & I2CM_STATUS_TIP) && timeout);

    return timeout;
}


/** actually send a byte
 *
 * @param mash
 * @param dev The I2C device you want communicate with. dev is the name of the
 * device, not the address
 * @param data The byte to sent
 * @param start True if a start should be performed at the beginning
 * @param stop True if a stop should be performed at the end
 *
 * @return -I2C_ERR_AL if an arbitrage lost occured, -I2C_ERR_NACK for a NACK, I2C_ERR_OK for an ACK.
 *
 */
static int parrot5_i2c_send_byte ( cyg_i2c_bus* mash,
                                        const cyg_i2c_device* dev,
                                        cyg_uint8 data,
                                        cyg_bool send_start,
                                        cyg_bool send_stop)
{
    cyg_uint32 cmd;
    cyg_uint32 status;
    int ret;

    DEBUG("dev=%08x addr=%02x send_start=%d data=%08x send_stop=%d\n",
          dev, dev->i2c_address, send_start, data, send_stop);
  
    // build the command:
    cmd = I2CM_COMMAND_WR;               // we at least want to write

    if (send_start) {
        cmd |= I2CM_COMMAND_STA ;          // start needed ?
    }

    if (send_stop) {
        cmd |= I2CM_COMMAND_STO;           // stop needed ?
    }

    // put the data to be send in the Tx register
    i2cm_write_reg(I2CM_TRANSMIT, data);

    // enable completion interrupt
    parrot5_i2c_prepare_completion();

    // send the command ( write + others )
    i2cm_write_reg(I2CM_COMMAND, cmd);

    // wait until the end of the tx
    parrot5_i2c_wait_completion();

	// Check STATUS register.
	status = i2cm_read_reg(I2CM_STATUS);

    // return error code.
	if (status & I2CM_STATUS_AL) {
		// return an arbitrage lost error.
		ret = -I2C_ERR_AL;
		DEBUG_ERR("arbitration lost\n");
	}
	else if (status & I2CM_STATUS_RXACK) {
		// return nack.
		ret = -I2C_ERR_NACK;
		DEBUG_ERR("no acknoledge\n");
	}
	else {
		// no error.
		ret = I2C_ERR_OK;
	}
    return ret;
}

/** actually read a byte
 *
 * @param mash
 * @param dev The I2C device you want communicate with. dev is the name of the
 * device, not the address
 * @param nak True if a nak should be performed at the end
 * @param stop True if a stop should be performed at the end
 *
 * @return The byte read or if an error occurred: -I2C_ERR_AL, -I2C_ERR_TIMEOUT.
 */
static int parrot5_i2c_read_byte(cyg_i2c_bus* mash,
                                 const cyg_i2c_device* dev,
                                 cyg_bool nak,
                                 cyg_bool send_stop)
{
    int ret;
    cyg_uint32 cmd, timeout;
    cyg_uint32 status;
  
    DEBUG("dev=%08x nak=%d\n", dev, nak);

    // build the command:
    cmd = I2CM_COMMAND_RD;              // we at least want to read
    if (nak) {                          // nak needed ?
        cmd |= I2CM_COMMAND_NACK;
    }
    if (send_stop) {                     // stop needed ?
        cmd |= I2CM_COMMAND_STO;
    }

    // enable completion interrupt
    parrot5_i2c_prepare_completion();

    // start the reception 
    i2cm_write_reg(I2CM_COMMAND, cmd);

    // wait until the end of the rx
    timeout = parrot5_i2c_wait_completion();

	// Check STATUS register.
	status = i2cm_read_reg(I2CM_STATUS);

	// return error code.
	if (status & I2CM_STATUS_AL) {
		// return an arbitrage lost error.
		ret = -I2C_ERR_AL;
		DEBUG_ERR("arbitration lost\n");
	}
	else if (timeout == 0) {
		// read the result
	    ret = -I2C_ERR_TIMEOUT;
	}	
	else {
		// no error, return byte.
		ret = i2cm_read_reg(I2CM_RECEIVE);
	}

    // trace it if needed
    DEBUG("result %02x\n", ret);

    // return what we read
    return ret;
}

/** send a stop
 *
 * used for compatibility
 *
 * @param dev The device. Not used because the stop is performed for the whole
 * bus
 */
void parrot5_i2c_send_stop(cyg_i2c_bus* bus)
{
    cyg_uint32 timeout;

    DEBUG("parrot5 i2c send stop alone\n");
  
    // enable completion interrupt
    parrot5_i2c_prepare_completion();

    // start the reception 
    i2cm_write_reg(I2CM_COMMAND, I2CM_COMMAND_STO);

    // wait until the end of the processing
    timeout = parrot5_i2c_wait_completion();

	// Check STATUS register.
	if (i2cm_read_reg(I2CM_STATUS) & I2CM_STATUS_AL) {
		// arbitrage lost error.
		DEBUG_ERR("arbitration lost\n");
	}
	else if (timeout == 0) {
		// timeout
        DEBUG_ERR("timeout while sending stop!\n");
	}	
}


/** ************************************
 ** Now the api for upper layers      **
 ** ************************************
 */

//! I2C master interrupt service routine (ISR)
static cyg_uint32 parrot5_i2cm_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    // simply request DSR
    cyg_drv_interrupt_mask(vector);
    return CYG_ISR_CALL_DSR;
}

//! I2C master interrupt deferred service routine (DSR)
static void parrot5_i2cm_dsr(cyg_vector_t vector, cyg_ucount32 count,
                             cyg_addrword_t data)
{
    // acknowledge and disable interrupt
    i2cm_write_reg(I2CM_COMMAND, I2CM_COMMAND_ITACK);
    i2cm_write_reg(I2CM_ITEN, 0);

    // wake-up blocked thread
    cyg_flag_setbits(&parrot5_i2cm_flag, 1);

    cyg_drv_interrupt_unmask(vector);
}

/** i2c master initialization
 *
 * The prescale factor is fixed. Future evolution is to choose it in the
 * configtool
 *
 * @param bus The bus used. There is only one bus in the Parrot5, so not used.
 */
static void parrot5_i2c_init( cyg_i2c_bus* bus)
{
    DEBUG("parrot5 i2c init\n" );

    // enable I2C master clock
    HAL_WRITE_UINT32 ( PARROT5_SYS + _SYS_CEN , SYS_CEN_I2CMCLK );
    i2cm_write_reg(I2CM_PRESCALE, CYGNUM_DEVS_IO_I2C_ARM_PARROT5_PRESCALE );

    // install interrupt handler
    if (parrot5_i2cm_use_irqs) {
        cyg_handle_t handle;
        static cyg_interrupt interrupt;

        cyg_flag_init(&parrot5_i2cm_flag);
        cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_I2CM,
                                 4,
                                 (cyg_addrword_t)0,
                                 parrot5_i2cm_isr,
                                 parrot5_i2cm_dsr,
                                 &handle,
                                 &interrupt);
        cyg_drv_interrupt_attach(handle);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_I2CM);
    }
}


/** perform an i2c transmission
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If true
 * tha I2C address of the device is also sended
 * @param tx_data A pointer to the data to send
 * @param count The number of bytes to send
 * @param send_stop True if a stop should be send at the end of the transmission
 *
 * @return The bytes written or if an error occurred: -I2C_ERR_AL, -I2C_ERR_NACK.
 */
static cyg_uint32 parrot5_i2c_tx(const cyg_i2c_device* dev,
                                 cyg_bool send_start,
                                 const cyg_uint8* tx_data,
                                 cyg_uint32 count,
                                 cyg_bool send_stop)
{
    int status;
    cyg_uint32 result;
  
    CYG_CHECK_DATA_PTR(tx_data,
                       "i2c tx operation requested but no data supplied");
    CYG_PRECONDITION(count > 0, "at least one byte should be sent");

    DEBUG("dev=%08x addr=%02x send_start=%d buf=%08x count=%d "
          "send_stop=%d\n", dev, dev->i2c_address, send_start,
          tx_data, count, send_stop);
  
    // first send the start & the address if needed
    if (send_start) {

        // if we send the start, we also have to send the address shifted left,
        //  complete with 0 to indicate a write. No stop sended
        // the devide should respond an AK, otherwise we stop
        // last args: true for the start, false the stop
        status = parrot5_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       (dev->i2c_address << 1)|0,
                                       true,
                                       false);
        if (status < I2C_ERR_OK) {
            // we don't receive the ack. we stop the tx now
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
            return (cyg_uint32)status;
		}
    }
  
    // now send the data, unless we receive a nak.
    for (result = 0; result < count; result++) {
        // two last args: false for the start, stop if needed
        status = parrot5_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       tx_data[result],
                                       false,
                                       ((result == count-1) && send_stop));

		if (status < I2C_ERR_OK) {
			// return an error code.
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
			return (cyg_uint32)status;
		}
	}
  
    DEBUG("parrot5 i2c tx, %d bytes sent out of %d\n", result, count);
    
    // return the number of bytes sended.
    return result;
}


/** perform an rx from a device (= a read action) 
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If
 * true the I2C address of the device is also sended.
 * @param rx_data A pointer to the buffer where data should be stored
 * @param count The number of bytes to read
 * @param send_nak True if a nack should be send at the end of the transmission.
 * @param send_stop True if a stop should be send at the end of the
 * transmission.
 *
 * @return The number of readed bytes or if an error occurred: -I2C_ERR_AL, -I2C_ERR_NACK, -I2C_ERR_TIMEOUT
 *
 */
static cyg_uint32 parrot5_i2c_rx(const cyg_i2c_device* dev,
                                 cyg_bool send_start,
                                 cyg_uint8* rx_data, cyg_uint32 count,
                                 cyg_bool send_nak,
                                 cyg_bool send_stop)
{
    int status, byte;
    cyg_uint32 result  = 0;

    DEBUG("dev %08x, addr %02x, send_start %d, buf %08x, "
          "count %d, send_nak %d, send_stop %d\n",
          dev, dev->i2c_address, send_start, rx_data, count, send_nak,
          send_stop);
    
    CYG_CHECK_DATA_PTR(rx_data,
                       "i2c rx operation requested but no data supplied");
    CYG_PRECONDITION(count > 0, "at least one byte should be sent");
    CYG_PRECONDITION(send_nak || !send_stop,
                     "a stop can only be generated after rx has been nak'd");
    
    // first send the start & the address if needed
    if (send_start) {

        // if we have to send the start, we also have to send the address
        // shifted left,
        //  complete with 1 to indicate a read.
        // the device should respond an AK, otherwise we stop
        status = parrot5_i2c_send_byte(dev->i2c_bus,
                                       dev,
                                       (dev->i2c_address << 1)|0x01,
                                       true,
                                       false);
        if (status < I2C_ERR_OK) {
            // we don't receive the ack or an arbitrage lost occurred. we stop the tx now.
            DEBUG_ERR("an error (%d) occurred with device at address %02x\n",
                  status, dev->i2c_address);
            return (cyg_uint32)status;
		}
        // The device has ack'd so we can continue sending
    }
    
    // The device has ack'd. Read the number of bytes specified. The
    // device cannot stop sending, although it may clock-stretch. For
    // the last byte we usually want to NAK
    // but if the data stream is variable length with e.g.
    // the first byte indicating the length then we want to support
    // partial reads.
    for (result = 0; result < count; result++) {
        byte = parrot5_i2c_read_byte(dev->i2c_bus, dev,
                                     (send_nak && (result == (count-1))),
                                     (send_stop && (result == (count-1))));
        if (byte < 0) {
        	// return error code.
            return (cyg_uint32)byte;
        }
        rx_data[result] = (cyg_uint8)byte;
    }

    DEBUG("parrot5 i2c rx, read %d bytes\n", result);
    return result;
}

/** set the bus in hte right state, outside tx or rx
 *
 * @param dev The device name, to perform the stop on the right bus.
 */
static void parrot5_i2c_stop(const cyg_i2c_device* dev)
{
    parrot5_i2c_send_stop(dev->i2c_bus);
}


/** Instanciate the bus
 *
 *
 */
CYG_I2C_BUS(parrot5_i2c_bus,
            &parrot5_i2c_init,
            &parrot5_i2c_tx,
            &parrot5_i2c_rx,
            &parrot5_i2c_stop,
            (void*) &parrot5_i2c_object);
