//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 **************************************************************
 * @file s3c2412x_i2c.c
 * @brief I2C master driver for the Parrot5. <br>
 * Theses low-level functions are called by the Ecos I2C generic package.
 *
 * Copyright (C) 2006 Parrot S.A.
 *
 * @author   gautier.seidel@parrot.fr
 * @author   jerome.leveque@parrot.fr
 * @date     2006-02-22
 *
 **************************************************************
 */

#include <cyg/io/i2c.h>
#include <pkgconf/devs_s3c2412x_i2c.h>

#include <cyg/hal/hal_io.h>
#include <cyg/hal/s3c2412x.h>

// Registers bitwise definitions

// Status Register
#define I2CM_STATUS_READ_SLAVE		(0 << 6)
#define I2CM_STATUS_WRITE_SLAVE		(1 << 6)
#define I2CM_STATUS_READ_MASTER		(2 << 6)
#define I2CM_STATUS_WRITE_MASTER	(3 << 6)
#define I2CM_STATUS_START			(1 << 5)
#define I2CM_STATUS_STOP			(0 << 5)
#define I2CM_STATUS_SERIAL_OUTPUT	(1 << 4)
#define I2CM_STATUS_ACK_MASK		(0xFE)

// Command Register
#define I2CM_IICCON_PENDING		(1 << 4)

#define I2CM_COMMAND_ACK        (0 << 1)    // ACK bit command
#define I2CM_COMMAND_NACK       (1 << 1)    // NACK bit command
#define I2CM_COMMAND_WR         (1 << 2)    // Write bit command
#define I2CM_COMMAND_RD         (1 << 3)    // Read bit command
#define I2CM_COMMAND_STO        (1 << 4)    // Stop bit command
#define I2CM_COMMAND_STA        (1 << 5)    // Start bit command

// Interrupt Enable register
#define I2CM_ITEN_ITEN                1  // Interrupt enabled

// Prescale register
#define I2CM_PRESCALE_100k        0x81
#define I2CM_PRESCALE_400k        0x1F

// Transmit register
#define I2CM_TRANSMIT_WRIT            0
#define I2CM_TRANSMIT_READ            1

//! a dummy object, we can used it to store bus-specific data if needed
struct s3c2412x_i2c_data 
{
    
} s3c2412x_i2c_object;

/** actually send a byte
 *
 * @param data The byte to sent
 * @param start True if a start should be performed at the beginning
 * @param stop True if a stop should be performed at the end
 *
 * @return The ACK/NACK status
 *
 */
static cyg_bool s3c2412x_i2c_send_byte(cyg_uint8 data, cyg_bool send_start, cyg_bool send_stop)
{
	cyg_uint32 cmd;
	cyg_uint8 reg_value;

	if (send_start)
	{
		cmd = I2CM_STATUS_SERIAL_OUTPUT;
		HAL_WRITE_UINT8(IICSTAT, cmd);
		// put the data to be send in the Tx register
		//This data should be the address of the slave device on the bus
		HAL_WRITE_UINT8(IICDS, data);
		cmd = I2CM_STATUS_WRITE_MASTER + I2CM_STATUS_SERIAL_OUTPUT + I2CM_STATUS_START; //0xF0
		HAL_WRITE_UINT8(IICSTAT, cmd);
	}
	else if (send_stop)
	{
		cmd = I2CM_STATUS_WRITE_MASTER + I2CM_STATUS_SERIAL_OUTPUT; //0xD0
		HAL_WRITE_UINT8(IICSTAT, cmd);
	}
	else
	{//a middle data
		HAL_WRITE_UINT8(IICDS, data);
	}

	if (!send_start)
	{
		//Resume pending bit
		HAL_READ_UINT8(IICCON, reg_value);
		reg_value &= ~I2CM_IICCON_PENDING;
		HAL_WRITE_UINT8(IICCON, reg_value);
	}

	do
	{
		HAL_READ_UINT8(IICCON, reg_value);
	}
	// wait until the end of the rx (pending bit set)
	while((reg_value & I2CM_IICCON_PENDING) != I2CM_IICCON_PENDING);
	
	// return the ACK/NACK status
	HAL_READ_UINT8(IICSTAT, reg_value);
	return ((reg_value & 1) == 0); //ACK received if reg_value == 0
}

/** actually read a byte
 *
 * @param nak True if a nak should be performed at the end
 * @param stop True if a stop should be performed at the end
 *
 * @return The byte read
 */
static cyg_uint8 s3c2412x_i2c_read_byte(cyg_bool send_start, cyg_bool send_stop)
{
	cyg_uint32 cmd;
	cyg_uint8 reg_value;
	cyg_uint8 data = 0x00; //Slave adress;

	if (send_start)
	{
		cmd = I2CM_STATUS_SERIAL_OUTPUT;
		HAL_WRITE_UINT8(IICSTAT, cmd);
		// put the data to be send in the Tx register
		//This data should be the address of the slave device on the bus
		HAL_WRITE_UINT8(IICDS, data);
		cmd = I2CM_STATUS_READ_MASTER + I2CM_STATUS_SERIAL_OUTPUT + I2CM_STATUS_START; //0xB0
		HAL_WRITE_UINT8(IICSTAT, cmd);
	}
	else if (send_stop)
	{
		cmd = I2CM_STATUS_WRITE_MASTER + I2CM_STATUS_SERIAL_OUTPUT; //0x90
		HAL_WRITE_UINT8(IICSTAT, cmd);
	}
	else
	{//a middle data
		HAL_READ_UINT8(IICDS, data);
	}

	if (!send_start)
	{
		//Resume pending bit
		HAL_READ_UINT8(IICCON, reg_value);
		reg_value &= ~I2CM_IICCON_PENDING;
		HAL_WRITE_UINT8(IICCON, reg_value);
	}

	do
	{
		HAL_READ_UINT8(IICCON, reg_value);
	}
	// wait until the end of the rx (pending bit set)
	while((reg_value & I2CM_IICCON_PENDING) != I2CM_IICCON_PENDING);
	
	return data;
}

/** send a stop
 *
 * used for compatibility
 *
 * @param dev The device. Not used because the stop is performed for the whole bus
 */
void s3c2412x_i2c_send_stop( cyg_i2c_bus* bus )
{
	s3c2412x_i2c_send_byte(0, false, true);
}

/** ************************************
 ** Now the api for upper layers      **
 ** ************************************
 */


/** i2c master initialization
 *
 * The prescale factor is fixed. Future evolution is to choose it in the configtool
 *
 * @param bus The bus used. There is only one bus in the Parrot5, so not used.
 */
static void s3c2412x_i2c_init( cyg_i2c_bus* bus)
{
	cyg_uint8 reg_value = 0xE0; //Ack bit, IICCLK = fPCLK / 512, Enable Interupt
	HAL_WRITE_UINT8(IICCON, reg_value);
}


/** perform an i2c transmission
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If true the I2C address of the device is also sended
 * @param tx_data A pointer to the data to send
 * @param count The number of bytes to send
 * @param send_stop True if a stop should be send at the end of the transmission
 *
 * @return The bytes written
 */
static cyg_uint32 s3c2412x_i2c_tx(const cyg_i2c_device* dev,
             cyg_bool send_start,
             const cyg_uint8* tx_data, cyg_uint32 count,
             cyg_bool send_stop)
{
	cyg_uint32 result  = 0;

	// first send the start & the address if needed
	if (send_start)
	{
		// if we send the start, we also have to send the address shifted left,
		//  complete with 0 to indicate a write. No stop sended
		// the devide should respond an AK, otherwise we stop
		// last args: true for the start, false the stop
		if (!s3c2412x_i2c_send_byte(dev->i2c_address, true , false ))
		{
			// we don't receive the ack. we stop the tx now
			return 0;
		}
    }

	// now send the data, unless we receive a nak.
	do
	{
		cyg_bool stop = (result == (count - 1)) && send_stop;
		result++;
		// two last args: false for the start, stop if needed
		if (!s3c2412x_i2c_send_byte(tx_data[result], false , stop))
		{
			break;
	    }
	}
	while (result < count);
	// return the number of bytes sended
	return result;
}


/** perform a read from a device
 *
 * @param dev The device name
 * @param send_start True if a start should be performed at beginning. If true the I2C address of the device is also sended.
 * @param rx_data A pointer to the buffer where data should be stored
 * @param count The number of bytes to read
 * @param send_nak True if a stop should be send at the end of the transmission.
 * @param send_stop True if a stop should be send at the end of the transmission.
 *
 * @return The number of readed bytes
 *
 */
static cyg_uint32 s3c2412x_i2c_rx(const cyg_i2c_device* dev,
             cyg_bool send_start,
             cyg_uint8* rx_data, cyg_uint32 count,
             cyg_bool send_nak,
             cyg_bool send_stop)
{
	cyg_uint32 result  = 0;

	// first send the start & the address if needed
	if (send_start)
	{
	// if we have to send the start, with the address of the device,
	// the device should respond an ACK, otherwise we stop
		if (!s3c2412x_i2c_send_byte(dev->i2c_address, true , false ))
		{
		// we don't receive the ack. we stop the rx now
			return 0;
		}
		// The device has ack'd so we can continue sending
	}
	// The device has ack'd. Read the number of bytes specified. The
	// device cannot stop sending, although it may clock-stretch. For
	// the last byte we usually we want to NAK
	// but if the data stream is variable length with e.g.
	// the first byte indicating the length then we want to support
	// partial reads.
	for (result = 0; result < count; result++)
	{
		if ((result == (count - 1)) && send_nak)
		{//Clearing acknowledge generation
			cyg_uint8 reg_value;
			HAL_READ_UINT8(IICCON, reg_value);
			reg_value &= 0xEF;
			HAL_WRITE_UINT8(IICCON, reg_value);
		}
		rx_data[result] = s3c2412x_i2c_read_byte(result == 0, send_stop && (result == (count - 1)));
	}
    return result;
}

/** set the bus in hte right state, outside tx or rx
 *
 * @param dev The device name, to perform the stop on the right bus.
 */
static void s3c2412x_i2c_stop(const cyg_i2c_device* dev)
{
     s3c2412x_i2c_send_stop(dev->i2c_bus);
}

/** Instanciate the bus
 *
 *
 */
CYG_I2C_BUS(s3c2412x_i2c_bus,
            &s3c2412x_i2c_init,
            &s3c2412x_i2c_tx,
            &s3c2412x_i2c_rx,
            &s3c2412x_i2c_stop,
            (void*) &s3c2412x_i2c_object);

