/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_evt.h
 * @brief  Management of events triggered by CS42XX devices.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 *  
 * This code has been mainly inspired from the GPIO button driver
 * (by Ivan Djelic, Parrot S.A.).
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */
#ifndef _CS42XX_EVT_H_
#define _CS42XX_EVT_H_

#include <cyg/io/io.h>
#include <cyg/io/i2c.h>
#include <cyg/io/gpio.h>
#include <pkgconf/devices_cs42xx.h>
#include <cyg/devs/cs42xx.h>
#include <cyg/devs/cs42xx_hal.h>


#ifdef CYGPKG_DEVICES_CS42XX_USEINTR


/** Maximal different callbacks that can be associated to events.
 *
 * @def CS42XX_NBCALLBACKS
 */
#define CS42XX_EVT_NBCALLBACKS		CYGPKG_DEVICES_CS42XX_NBCALLBACKSMAX


/*
 * Events.
 * You can combine them using operator '|'.
 */
#define CS42XX_EVT_ADCCLKERR		CS42XX_REGM_INT_ADCCLKERR
#define CS42XX_EVT_DACCLKERR		CS42XX_REGM_INT_DACCLKERR
#define CS42XX_EVT_OVERFLOW			CS42XX_REGM_INT_ADCOVFL
#define CS42XX_EVT_UNDERFLOW		CS42XX_REGM_INT_ADCUNDFL

/*
 * Predefined combined events.
 */
#define CS42XX_EVT_MASK_CLKERR		(CS42XX_EVT_ADCCLKERR|CS42XX_EVT_DACCLKERR)
#define CS42XX_EVT_MASK_OVUNDFLOW	(CS42XX_EVT_OVERFLOW|CS42XX_EVT_UNDERFLOW)
#define CS42XX_EVT_MASK_ADC			(CS42XX_EVT_ADCCLKERR|CS42XX_EVT_OVERFLOW|CS42XX_EVT_UNDERFLOW)
#define CS42XX_EVT_MASK_DAC			(CS42XX_EVT_DACCLKERR)
#define CS42XX_EVT_MASK_ALL			(CS42XX_EVT_MASK_CLKERR|CS42XX_EVT_MASK_OVUNDFLOW)


/** Type for CS42XX (bit-field) events.
 *
 * @typedef CS42XX_EVT_MASK
 */
typedef cyg_uint8 CS42XX_EVT_MASK;

/** Type for event callback.
 *
 * This function will be called after each interrupt caught from a device and
 * corresponding to the event(s) configured.
 *
 * The function will be called with parameters:
 * - the CS42XX device that issued interrupt,
 * - the event(s) which triggered that function call.
 *
 * @typedef CS42XX_EVT_CB
 */
typedef void (*CS42XX_EVT_CB)(CS42XX_DEVID, CS42XX_EVT_MASK);



/*
 * Exported functions.
 */


// --- Events ---

// Associate a callback to a set of events occurring on the given device.
extern Cyg_ErrNo
CS42XX_Evt_Register(CS42XX_DEVID, CS42XX_EVT_MASK, CS42XX_EVT_CB);

// Unregister an event callback for any device.
extern Cyg_ErrNo
CS42XX_Evt_UnregisterCallback(CS42XX_EVT_CB);

// Stop catching some events on the given device.
extern Cyg_ErrNo
CS42XX_Evt_UnregisterEvents(CS42XX_DEVID, CS42XX_EVT_MASK);


// --- Internal to CS42XX driver ---

// Process the interrupt that just happened.
extern void
CS42XX_Evt_ProcessIntr(void);


#endif // CYGPKG_DEVICES_CS42XX_USEINTR


// --- Initialization ---

// Initialize submodule
extern Cyg_ErrNo
CS42XX_Evt_Init(void);



#endif // _CS42XX_EVT_H_

