/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_regs.h
 * @brief  Access to CS42XX registers, with optional cache.
 * 
 * For the complete description of registers, please refer to:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */
#ifndef _CS42XX_REGS_H_
#define _CS42XX_REGS_H_

#include <cyg/devs/cs42xx.h>


/** Define this if you want to use register cache.
 * The CYGPKG_DEVICES_CS42XX_USECACHE package option is likely to be used
 * to select register caching functionality.
 * @def CS42XX_USECACHE
 */
#ifdef CYGPKG_DEVICES_CS42XX_USECACHE
#define CS42XX_USECACHE
#endif


/**
 * @defgroup cs42xx_regs_list  Registers of CS42XX audio codecs
 *
 * Due to the same family of devices, all CS42XX audio codecs are based on
 * the same register mapping. Changes between devices deal with presence
 * or absence of some bits, according to the features delivered with the
 * chip.
 *
 * The following enumerates all registers in a generic way. These are used
 * by the "management of registers" (CS42XX_Regs) part of driver, while
 * specific bits (depending on the kind of codec) require a query to the
 * "Supported features" (CS42XX_Supp) part of driver.
 *
 * @{
 */

/** Type for a register.
 * @typedef CS42XX_REG
 */
typedef enum CS42XX_REG_TAG CS42XX_REG;
enum CS42XX_REG_TAG
{
	CS42XX_REG_NULL = 0x00,     ///< Dummy register (do not use).

	// Common to all CS42xx devices:
	CS42XX_REG_CHIPID,			///< Chip ID.
	CS42XX_REG_POWER_CTRL,		///< Power control.
	CS42XX_REG_DAC_CTRL,		///< DAC control.
	CS42XX_REG_ADC_CTRL,		///< ADC control.
	CS42XX_REG_MCLK_FREQ,		///< MCLK(1,2) frequency.
	CS42XX_REG_SIGNALSEL,		///< Signal selection.
	CS42XX_REG_ADCA_PGA,		///< ADC channel A PGA control.
	CS42XX_REG_ADCB_PGA,		///< ADC channel B PGA control.
	CS42XX_REG_ADC_INCTRL,		///< ADC input control.
	CS42XX_REG_DACA_VOL,		///< DAC channel A volume control.
	CS42XX_REG_DACB_VOL,		///< DAC channel B volume control.
	CS42XX_REG_DAC_OUTCTRL,		///< DAC output control.
	CS42XX_REG_INTSTATUS,		///< Interrupt status.
	CS42XX_REG_INTMASK,			///< Interrupt mask.
	CS42XX_REG_INTMODE_MSB,		///< Interrupt mode (MSB).
	CS42XX_REG_INTMODE_LSB,		///< Interrupt mode (LSB).

	// CS4265 only:
	CS42XX_REG_TXCTRL1,         ///< Transmitter control (1).
	CS42XX_REG_TXCTRL2,         ///< Transmitter control (2).
	CS42XX_REG_CDATA,           ///< C-Data buffer.
	
	// Boundaries
	CS42XX_REG_FIRST   = CS42XX_REG_CHIPID,      ///< First register (all devices).
	CS42XX_REG_LAST    = CS42XX_REG_INTMODE_LSB, ///< Last register (safe).
	CS42XX_REG_LASTMAX = CS42XX_REG_CDATA,       ///< Last register (for init of cache).

	// Codec-specific boundaries (manipulated by CS42XX_Supp)
	CS42XX_REG_LAST_CS4245  = CS42XX_REG_INTMODE_LSB, ///< Last register on CS4245.
	CS42XX_REG_LAST_CS4265  = CS42XX_REG_TXCTRL2,     ///< Last register on CS4265.
	CS42XX_REG_LAST_CS5345  = CS42XX_REG_INTMODE_LSB, ///< Last register on CS5345.
};

/** Last register of device. */
#define CS42XX_REG_LASTOF(devid) \
		CS42XX_Supp_GetLastReg(devid)

/** Number of registers of the given audio codec. */
#define CS42XX_REG_NBOF(devid) \
		(CS42XX_REG_LASTOF(devid) - CS42XX_REG_FIRST + 1)

/** Number of registers in audio codec. */
#define CS42XX_REG_NBMAX \
		(CS42XX_REG_LASTMAX - CS42XX_REG_FIRST + 1)

/** Return true if register exists on this device.
 *
 * @def   CS42XX_REG_ISVALID
 * @param devid  Symbolic device Id
 * @param reg    Offset of register to be tested
 */
#define CS42XX_REG_ISVALID(devid,reg)         \
		(((reg) >= CS42XX_REG_FIRST) &&       \
		 ((reg) <= CS42XX_REG_LASTOF(devid)))

/** Return true if device is valid.
 *
 * @def   CS42XX_REG_ISDEVVALID
 * @param devid  Symbolic device Id
 */
#define CS42XX_REG_ISDEVVALID(devid) \
		((devid) < CS42XX_NBDEVICES)



/**
 * @defgroup cs42xx_regs_descr  Register description
 *
 * Here is the description of register contents, i.e. what does each bit
 * mean in each register.
 *
 * Boolean states are available through a power-of-2 mask (0x1<<(y)) named
 * CS42XX_REGM_xxx_yyy, where xxx recalls the register name and yyy the
 * functionality.
 * When multiple bits are related together (to form an integer), a multi-bit
 * mask (CS42XX_REGM_xxx_yyy_M) and a shift value (CS42XX_REGM_xxx_yyy_SHF)
 * are also defined.
 *
 * @{
 */

/**
 * @defgroup cs42xx_regs_descr_chipid  Chip Id
 * @{
 */
#define CS42XX_REG_CHIPID_PART(x)			(((x) >> 4) & 0x0F) ///< Get the part number from a chip Id.
#define CS42XX_REG_CHIPID_REV(x)			((x) & 0x0F)        ///< Get the revision from a chip Id.

/** @}
 * @defgroup cs42xx_regs_descr_power  Low-power management
 * @{
 */
#define CS42XX_REGM_POWER_FREEZE			(0x01 << 7) ///< Freeze state of control ports.
#define CS42XX_REGM_POWER_MIC				(0x01 << 3) ///< Low-power, mic circuit.
#define CS42XX_REGM_POWER_ADC				(0x01 << 2) ///< Low-power, ADC circuit.
#define CS42XX_REGM_POWER_DAC				(0x01 << 1) ///< Low-power, DAC circuit.
#define CS42XX_REGM_POWER_DEVICE			(0x01 << 0) ///< Low-power for the whole device.

/** @}
 * @defgroup cs42xx_regs_descr_dac  Digital input (DAC)
 * @{
 */
#define CS42XX_REGM_DAC_FM					(0x03 << 6) ///< Speed type for digital input.
#define CS42XX_REGM_DAC_FM_M				(0x03)
#define CS42XX_REGM_DAC_FM_SHF				(6)
#define CS42XX_REGM_DAC_DIF					(0x03 << 4) ///< Frame format for digital input.
#define CS42XX_REGM_DAC_DIF_M				(0x03)
#define CS42XX_REGM_DAC_DIF_SHF				(4)
#define CS42XX_REGM_DAC_MUTE				(0x01 << 2) ///< Mute on digital input.
#define CS42XX_REGM_DAC_DEEMPH				(0x01 << 1) ///< Disabling de-emphasis filter.
#define CS42XX_REGM_DAC_MASTER				(0x01 << 0) ///< Mastering digital input.

/** @}
 * @defgroup cs42xx_regs_descr_adc  Digital output (ADC)
 * @{
 */
#define CS42XX_REGM_ADC_FM					(0x03 << 6) ///< Speed type for digital output.
#define CS42XX_REGM_ADC_FM_M				(0x03)
#define CS42XX_REGM_ADC_FM_SHF				(6)
#define CS42XX_REGM_ADC_DIF					(0x01 << 4) ///< Frame format for digital output.
#define CS42XX_REGM_ADC_DIF_M				(0x01)
#define CS42XX_REGM_ADC_DIF_SHF				(4)
#define CS42XX_REGM_ADC_MUTE				(0x01 << 2) ///< Mute on digital output.
#define CS42XX_REGM_ADC_DISABLEFILTER		(0x01 << 1) ///< Disabling high-pass filter.
#define CS42XX_REGM_ADC_MASTER				(0x01 << 0) ///< Mastering digital output.

/** @}
 * @defgroup cs42xx_regs_descr_mclk  Master Clock frequency
 * @{
 */
#define CS42XX_REGM_MCLK1_FREQ				(0x07 << 4) ///< MCLK1 frequency.
#define CS42XX_REGM_MCLK1_FREQ_M			(0x07)
#define CS42XX_REGM_MCLK1_FREQ_SHF			(4)
#define CS42XX_REGM_MCLK2_FREQ				(0x07 << 0) ///< MCLK2 frequency (CS4245 only).
#define CS42XX_REGM_MCLK2_FREQ_M			(0x07)
#define CS42XX_REGM_MCLK2_FREQ_SHF			(0)

/** @}
 * @defgroup cs42xx_regs_descr_sigsel  Signal selection
 * @{
 */
#define CS42XX_REGM_SIGSEL_SDINSEL			(0x01 << 7) ///< Digital input selected (CS4265 only).
#define CS42XX_REGM_SIGSEL_AOUTSEL			(0x03 << 5) ///< Source for auxiliary output.
#define CS42XX_REGM_SIGSEL_AOUTSEL_M		(0x03)
#define CS42XX_REGM_SIGSEL_AOUTSEL_SHF		(5)
#define CS42XX_REGM_SIGSEL_LOOPBACK			(0x01 << 1) ///< Loopback mode.
#define CS42XX_REGM_SIGSEL_ASYNC			(0x01 << 0) ///< Asynchronous mode (CS4245 only).

/** @}
 * @defgroup cs42xx_regs_descr_pga  ADC input gain (PGA)
 * @{
 */
#define CS42XX_REGM_ADCA_PGAGAIN			(0x3F << 0) ///< Mask for signed PGA gain, channel A.
#define CS42XX_REGM_ADCA_PGAGAIN_NEG			(   1 << 5) ///< Sign bit for PGA gain, channel A.
#define CS42XX_REGM_ADCB_PGAGAIN			(0x3F << 0) ///< Mask for signed PGA gain, channel B.
#define CS42XX_REGM_ADCB_PGAGAIN_NEG			(   1 << 5) ///< Sign bit for PGA gain, channel B.

/** @}
 * @defgroup cs42xx_regs_descr_in  ADC-input related
 * @{
 */
#define CS42XX_REGM_ADC_LVLCHGPOLICY		(0x03 << 3) ///< ADC level change policy.
#define CS42XX_REGM_ADC_LVLCHGPOLICY_M		(0x03)
#define CS42XX_REGM_ADC_LVLCHGPOLICY_SHF	(3)
#define CS42XX_REGM_ADC_SEL					(0x07 << 0) ///< Analog input selected (CS4265: bit 0 only).
#define CS42XX_REGM_ADC_SEL_M				(0x07)
#define CS42XX_REGM_ADC_SEL_SHF				(0)

/** @}
 * @defgroup cs42xx_regs_descr_vol  DAC output volume (attenuation)
 * @{
 */
#define CS42XX_REGM_DACA_VOL				(0xFF) ///< Mask for unsigned output attenuation, channel A.
#define CS42XX_REGM_DACB_VOL				(0xFF) ///< Mask for unsigned output attenuation, channel B.

/** @}
 * @defgroup cs42xx_regs_descr_out  DAC-output related (+ intr output)
 * @{
 */
#define CS42XX_REGM_DAC_LVLCHGPOLICY		(0x03 << 6) ///< DAC level change policy.
#define CS42XX_REGM_DAC_LVLCHGPOLICY_M		(0x03)
#define CS42XX_REGM_DAC_LVLCHGPOLICY_SHF	(6)
#define CS42XX_REGM_DAC_INVERT				(0x01 << 5) ///< Invert DAC analog output.
#define CS42XX_REGM_DAC_INTACTIVEHIGH		(0x01 << 0) ///< "Active" state of interrupt pin.

/** @}
 * @defgroup cs42xx_regs_descr_intr  Interrupt-related registers
 * @{
 */
#define CS42XX_REGM_INT_ADCCLKERR			(0x01 << 3) ///< ADC clock error.
#define CS42XX_REGM_INT_DACCLKERR			(0x01 << 2) ///< DAC clock error.
#define CS42XX_REGM_INT_ADCOVFL				(0x01 << 1) ///< ADC overflow.
#define CS42XX_REGM_INT_ADCUNDFL			(0x01 << 0) ///< ADC underflow.

/** @}
 * @defgroup cs42xx_regs_descr_txctrl1  Transmitter control (1)
 * @{
 */
#define CS42XX_REGM_TXCTRL1_INHIBITE2F		(0x01 << 6) ///< Inhibit C-data E to F transfers.
#define CS42XX_REGM_TXCTRL1_2BYTECDATA		(0x01 << 5) ///< 2-byte control port access.

/** @}
 * @defgroup cs42xx_regs_descr_txctrl2  Transmitter control (2)
 * @{
 */
#define CS42XX_REGM_TXCTRL2_DIF				(0x03 << 6) ///< Frame format for digital interface.
#define CS42XX_REGM_TXCTRL2_DIF_M			(0x03)
#define CS42XX_REGM_TXCTRL2_DIF_SHF			(6)
#define CS42XX_REGM_TXCTRL2_OFF				(0x01 << 5) ///< Set output to constant 0V.
#define CS42XX_REGM_TXCTRL2_MUTE			(0x01 << 4) ///< Mute output.
#define CS42XX_REGM_TXCTRL2_VALIDBIT		(0x01 << 3) ///< Validity bit level.
#define CS42XX_REGM_TXCTRL2_MONO			(0x01 << 2) ///< Mono mode (same channel in both subframes).
#define CS42XX_REGM_TXCTRL2_MONOCHANNEL		(0x01 << 1) ///< 1 source for both output subframes.
#define CS42XX_REGM_TXCTRL2_MONOCHSEL		(0x01 << 0) ///< Source B selected for both output subframes.

/** @} */

// !cs42xx_regs_descr
/** @} */

// !cs42xx_regs_list
/** @} */


/**
 * @addtogroup cs42xx_regcache  Register cache
 *
 * When selected, the register cache allows multiple operations to be
 * applied almost in once, thus slightly reducing traffic on IC.
 *
 * @{
 */
#ifdef CS42XX_USECACHE


/** Initialize register cache (no access to CS42XX devices). */
#define CS42XX_INITCACHE() \
		CS42XX_RegCache_Init()

/** Load the register cache using data available on given device. */
#define CS42XX_LOADCACHE(id) \
		CS42XX_RegCache_Load(id)

/** Peek register directly from device (use: status and int. registers). */
#define CS42XX_PEEK_NOW(id,addr) \
		CS42XX_Peek(id,addr,NULL)

/** Peek register directly from device and report errors in perr. */
#define CS42XX_PEEK_NOW_(id,addr,perr) \
		CS42XX_Peek(id,addr,perr)

/** Peek register of given address (cached). */
#define CS42XX_PEEK(id,addr) \
		CS42XX_RegCache_Peek(id,addr,NULL)

/** Peek register of given address (cached) and report errors in perr. */
#define CS42XX_PEEK_(id,addr,perr) \
		CS42XX_RegCache_Peek(id,addr,perr)

/** Poke value in register (cached). */
#define CS42XX_POKE(id,addr,val) \
		CS42XX_RegCache_Poke(id,addr,val)

/** Poke value in register and commit changes. */
#define CS42XX_POKE_COMMIT(id,addr,val) \
		CS42XX_POKE_COMMITCOND(id,addr,val,true)

/** Poke value in register and if condition is true, commit changes. */
#define CS42XX_POKE_COMMITCOND(id,addr,val,cond)                            \
		({                                                                  \
			Cyg_ErrNo err = CS42XX_RegCache_Poke(id,addr,val);              \
			if ((-ENOERR == err) && (cond))                                 \
			{                                                               \
				err = CS42XX_RegCache_CommitAll(id);                        \
			} /* else an error occurred. */                                 \
			err;                                                            \
		})

/** Commit changes (sync. cache with device registers). */
#define CS42XX_COMMIT(id) \
		CS42XX_RegCache_CommitAll(id)


#else // !CS42XX_USECACHE


#define CS42XX_INITCACHE() \
		ENOERR

#define CS42XX_LOADCACHE(id) \
		CS42XX_Supp_Init(id)

#define CS42XX_PEEK_NOW(id,addr) \
		CS42XX_Peek(id,addr,NULL)

#define CS42XX_PEEK_NOW_(id,addr,perr) \
		CS42XX_Peek(id,addr,perr)

#define CS42XX_PEEK(id,addr) \
		CS42XX_Peek(id,addr,NULL)

#define CS42XX_PEEK_(id,addr,perr) \
		CS42XX_Peek(id,addr,perr)

#define CS42XX_POKE(id,addr,val) \
		CS42XX_Poke(id,addr,val)

#define CS42XX_POKE_COMMIT(id,addr,val) \
		CS42XX_Poke(id,addr,val)

#define CS42XX_POKE_COMMITCOND(id,addr,val,cond) \
		CS42XX_Poke(id,addr,val)

#define CS42XX_COMMIT(id) \
		ENOERR


#endif // CS42XX_USECACHE

// !cs42xx_regcache
/** @} */


/*
 * Exported functions.
 */


#ifdef CS42XX_USECACHE

// Initialize register cache (no access to CS42XX devices).
extern Cyg_ErrNo
CS42XX_RegCache_Init(void);

// Load the register cache using data available on given device.
extern Cyg_ErrNo
CS42XX_RegCache_Load(CS42XX_DEVID);

// Read the content of cached register.
extern CYG_BYTE
CS42XX_RegCache_Peek(CS42XX_DEVID, CS42XX_REG, Cyg_ErrNo *);

// Write the content of register in cache memory (commit will be necessary).
extern Cyg_ErrNo
CS42XX_RegCache_Poke(CS42XX_DEVID, CS42XX_REG, CYG_BYTE val);

// Commit all changed registers to audio codec in once.
extern Cyg_ErrNo
CS42XX_RegCache_CommitAll(CS42XX_DEVID);

#endif // CS42XX_USECACHE


// Read the content of register from audio codec.
extern CYG_BYTE
CS42XX_Peek(CS42XX_DEVID, CS42XX_REG, Cyg_ErrNo *);

// Write the content of register in audio codec.
extern Cyg_ErrNo
CS42XX_Poke(CS42XX_DEVID, CS42XX_REG, CYG_BYTE val);



#endif // _CS42XX_REGS_H_
