//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx.c
 * @brief  Core of driver.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/io/devtab.h>
#include <cyg/devs/cs42xx.h>
#include <cyg/devs/cs42xx_supp.h>
#include <cyg/devs/cs42xx_regs.h>
#include <cyg/devs/cs42xx_i2c.h>
#include <cyg/devs/cs42xx_evt.h>



/*
 * Defines.
 ***********/

// FIXME: This define is hardware specific !!
#define CS42XX_I2C_BUS			parrot5_i2c_bus


/*
 * External symbols.
 ********************/

// Declare external symbols (from HAL_I2C)
HAL_I2C_EXPORTED_DEVICES;


/*
 * Definition of functions.
 ***************************/



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                             GENERAL STUFF                               *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Initialize the driver for all CS42XX devices.
 *
 * This function applies a reset to all CS42XX devices and initializes data
 * internal to driver.
 *
 * @pre  CS42XX_HAL_SetConfig() shall have been called to configure the I/O
 *       pins (or results may be unpredictable).
 *
 * @retval  ENOERR   Successfully initialized
 * @retval -ENXIO    Bad configuration of driver (need to call
 *                   CS42XX_HAL_SetConfig() first, at least for reset)
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Init(void)
{
	Cyg_ErrNo err;

	// Reset all devices
	err = CS42XX_HAL_Reset();

	// Initialize register cache
	CS42XX_INITCACHE();

	// Application has then to declare all devices using:
	// CS42XX_AttachBus(devid, bus, ad0, ad1);

	return err;
}


/** Attach a symbolic device to an I2C bus, and initialize driver for that
 *  device.
 *
 * When I2C bus attaching is successful, register cache is initialized
 * (provided it is enabled). Then, data relative to the kind of device
 * (supported features, etc.) is determined and stored, based on chip Id.
 *
 * @param[in] devid  Symbolic device Id
 * @param[in] bus    Bus to attach device to
 * @param[in] ad0    State of ad0 pin on audio device
 *                   (equal to 'ad' pin if there is only one such pin)
 * @param[in] ad1    State of ad1 pin on audio device
 *                   (set it to 1 if device has only an 'ad' pin)
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   Invalid parameters
 * @retval -ENODEV   No such device
 * @retval -EIO      Error during access to device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Init_AttachBus(CS42XX_DEVID devid, cyg_i2c_bus *bus,
		cyg_bool ad0, cyg_bool ad1)
{
	Cyg_ErrNo err = -EINVAL;

	// Attach device to the bus with the given address (ad0, ad1)
	err = CS42XX_I2C_AttachBus(devid, bus, ad0, ad1);
	if (ENOERR == err)
	{
		// Determine the kind of device by reading the chip Id;
		// also read all registers from device (when cache is enabled).
		err = CS42XX_LOADCACHE(devid);

		// In all cases, CS42XX_Supp_Init() has been called from
		// CS42XX_LOADCACHE().
	}
	// else bus attaching failed

	return err;
}


/** Get the part number Id and revision of audio codec.
 *
 * Part number is on 4 highest bits , while device revision is on the 4
 * lowest bits.
 * Please refer to the constant table cCS42XXSuppChipType for a list of
 * supported chip Ids and the description of each device feature.
 * 
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 * 
 * @return  Part number and revision
 ****************************************************************************/
CYG_BYTE
CS42XX_GetChipId(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	return CS42XX_PEEK_(devid, CS42XX_REG_CHIPID, perr);
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                       CONFIGURATIONS IN ONCE                            *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Fill a config structure according to the current configuration.
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] config  Pointer to global configuration area
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   Invalid parameters
 * @retval -ENODEV   No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Config_Get(CS42XX_DEVID devid, CS42XX_CONFIG *config)
{
	Cyg_ErrNo err;

	if (NULL == config)
	{
		err = -EINVAL;
	}
	else if (devid >= CS42XX_NBDEVICES)
	{
		err = -ENODEV;
	}
	else
	{
		err = ENOERR;

// (Derogation to MACROS_TYPES_DEFINE_1: very local use.)
#define IFOK_OR0(_f_) \
			( ((ENOERR == err)||(-EDEVNOSUPP == err))? _f_(devid, &err): 0 )

		// Get global config
		config->async       = IFOK_OR0(CS42XX_IsAsyncMode       );
		config->auxoutsel   = IFOK_OR0(CS42XX_GetAuxOut         );
		config->inputsel    = IFOK_OR0(CS42XX_GetADCInput       );
		config->loopmode    = IFOK_OR0(CS42XX_GetLoopMode       );

		// Get config of digital output interface
		config->out.format  = IFOK_OR0(CS42XX_GetFrameFormat_Out);
		config->out.master  = IFOK_OR0(CS42XX_IsMaster_Out      );
		config->out.mclkdiv = IFOK_OR0(CS42XX_GetMCLKFreq_Out   );
		config->out.srate   = IFOK_OR0(CS42XX_GetSampleRate_Out );

		// Get config of digital input interface
		config->in.format   = IFOK_OR0(CS42XX_GetFrameFormat_In );
		config->in.master   = IFOK_OR0(CS42XX_IsMaster_In       );
		config->in.mclkdiv  = IFOK_OR0(CS42XX_GetMCLKFreq_In    );
		config->in.srate    = IFOK_OR0(CS42XX_GetSampleRate_In  );

#undef IFOK_OR0

		if (-EDEVNOSUPP == err)
		{
			err = -ENOERR;
		}
	}

	return err;
}


/** Fill a settings structure according to the current settings.
 *
 * @param[in] devid     Symbolic device Id
 * @param[in] settings  Pointer to settings area
 * 
 * @retval  ENOERR   Success
 * @retval -EINVAL   Invalid parameters
 * @retval -ENODEV   No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Settings_Get(CS42XX_DEVID devid, CS42XX_SETTINGS *settings)
{
	Cyg_ErrNo err;

	if (NULL == settings)
	{
		err = -EINVAL;
	}
	else if (devid >= CS42XX_NBDEVICES)
	{
		err = -ENODEV;
	}
	else
	{
		err = ENOERR;

// (Derogation to MACROS_TYPES_DEFINE_1: very local use.)
#define IFOK_OR0(_f_) \
			( ((ENOERR == err)||(-EDEVNOSUPP == err))? _f_(devid, &err): 0 )

		settings->adc.a.gain    = IFOK_OR0(CS42XX_GetGain_ADCA   );
		settings->adc.b.gain    = IFOK_OR0(CS42XX_GetGain_ADCB   );
		settings->adc.mute      = IFOK_OR0(CS42XX_IsMuted_ADC    );
		settings->adc.filter    = IFOK_OR0(CS42XX_IsFiltered_ADC );
		settings->adc.lvlchgpcy = IFOK_OR0(CS42XX_GetLCPolicy_ADC);
		
		settings->dac.a.vol     = IFOK_OR0(CS42XX_GetVolume_DACA );
		settings->dac.b.vol     = IFOK_OR0(CS42XX_GetVolume_DACB );
		settings->dac.mute      = IFOK_OR0(CS42XX_IsMuted_DAC    );
		settings->dac.filter    = IFOK_OR0(CS42XX_IsFiltered_DAC );
		settings->dac.lvlchgpcy = IFOK_OR0(CS42XX_GetLCPolicy_DAC);

#undef IFOK_OR0

		if (-EDEVNOSUPP == err)
		{
			err = -ENOERR;
		}
	}

	return err;
}


/** Apply all global configuration.
 *
 * @warning  If one field is invalid whatever the kind of supported
 *           device, application of configuration stops, i.e. following
 *           configuration fields are *not applied* and an error code is
 *           returned.
 * @warning  If this device does not support one value which however could
 *           be accepted by another supported chip (e.g. shorter range, or
 *           no dedicated pin on that device), the present configuration
 *           field is *ignored* but the other ones are applied: thus, this
 *           function never returns -EDEVNOSUPP.
 *
 * @param[in] devid    Symbolic device Id
 * @param[in] config   Pointer to global configuration area
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   Invalid parameters
 * @retval -ENODEV   No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Config_Set(CS42XX_DEVID devid, CS42XX_CONFIG *config)
{
	Cyg_ErrNo err;

	if (NULL == config)
	{
		err = -EINVAL;
	}
	else if (devid >= CS42XX_NBDEVICES)
	{
		err = -ENODEV;
	}
	else
	{
		err = ENOERR;

// (Derogations to MACROS_TYPES_DEFINE_1: very local use.)
#define IFOK1(_f_,_p1_)												\
			if ((ENOERR == err) || (-EDEVNOSUPP == err))			\
				{ err = _f_(devid,_p1_,CS42XX_CACHE_NOSYNC); }
#define IFOK2(_f_,_p1_,_p2_)										\
			if ((ENOERR == err) || (-EDEVNOSUPP == err))			\
				{ err = _f_(devid,_p1_,_p2_,CS42XX_CACHE_NOSYNC); }

		IFOK1(CS42XX_ConfigAsyncMode, config->async    );
		IFOK1(CS42XX_ConfigAuxOut,    config->auxoutsel);
		IFOK1(CS42XX_ConfigADCInput,  config->inputsel );
		IFOK1(CS42XX_ConfigLoopMode,  config->loopmode );

		// Configure the format of frames on digital interfaces
		IFOK1(CS42XX_ConfigFrameFormat_Out, config->out.format);
		IFOK1(CS42XX_ConfigFrameFormat_In,  config->in .format);

		// Configure who is master on interfaces
		IFOK1(CS42XX_ConfigMaster_Out, config->out.master);
		IFOK1(CS42XX_ConfigMaster_In,  config->in .master);

		// Configure frequency dividers on MCLK(1,2)
		IFOK2(CS42XX_ConfigMCLKFreq, config->out.mclkdiv, config->in.mclkdiv);

		// Configure sample rate
		IFOK1(CS42XX_ConfigSampleRate_Out, config->in.srate);
		IFOK1(CS42XX_ConfigSampleRate_In,  config->in.srate);

		// Synchronize cache
		if ((ENOERR == err) || (-EDEVNOSUPP == err))
		{
			err = CS42XX_COMMIT(devid);
		}
		// else previous failure: do not commit.

#undef IFOK2
#undef IFOK1

	}

	return err;
}


/** Apply all settings in once.
 *
 * @warning  If one setting is invalid whatever the kind of supported
 *           device, application of settings stops, i.e. following settings
 *           are *not applied* and an error code is returned.
 * @warning  If this device does not support one value which however could
 *           be accepted by another supported chip (e.g. shorter range, or
 *           no dedicated pin on that device), the present setting is
 *           *ignored* but the other settings are applied: thus, this
 *           function never returns -EDEVNOSUPP.
 *
 * @param[in] devid    Symbolic device Id
 * @param[in] config   Pointer to global configuration area
 *
 * @retval  ENOERR   Success
 * @retval -EINVAL   Invalid parameters
 * @retval -ENODEV   No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Settings_Set(CS42XX_DEVID devid, CS42XX_SETTINGS *settings)
{
	Cyg_ErrNo err;

	if (NULL == settings)
	{
		err = -EINVAL;
	}
	else if (devid >= CS42XX_NBDEVICES)
	{
		err = -ENODEV;
	}
	else
	{
		err = ENOERR;

// (Derogations to MACROS_TYPES_DEFINE_1: very local use.)
#define IFOK1(_f_,_p1_)												\
			if ((ENOERR == err) || (-EDEVNOSUPP == err))			\
				{ err = _f_(devid,_p1_,CS42XX_CACHE_NOSYNC); }
#define IFOK2(_f_,_p1_,_p2_)										\
			if ((ENOERR == err) || (-EDEVNOSUPP == err))			\
				{ err = _f_(devid,_p1_,_p2_,CS42XX_CACHE_NOSYNC); }

		// Mute or unmute
		IFOK1(CS42XX_Mute_ADC, settings->adc.mute);
		IFOK1(CS42XX_Mute_DAC, settings->dac.mute);

		// Apply gain / volume
		IFOK2(CS42XX_SetGain_ADC,   settings->adc.a.vol, settings->adc.b.vol);
		IFOK2(CS42XX_SetVolume_DAC, settings->dac.a.vol, settings->dac.b.vol);

		// Set filters
		IFOK1(CS42XX_SetFilter_ADC, settings->adc.filter);
		IFOK1(CS42XX_SetFilter_DAC, settings->dac.filter);

		// Apply level change policies
		IFOK1(CS42XX_ConfigLCPolicy_ADC, settings->adc.lvlchgpcy);
		IFOK1(CS42XX_ConfigLCPolicy_DAC, settings->dac.lvlchgpcy);

		// Synchronize cache
		if ((ENOERR == err) || (-EDEVNOSUPP == err))
		{
			err = CS42XX_COMMIT(devid);
		}
		// else previous failure: do not commit.

#undef IFOK2
#undef IFOK1		

	}

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                              INTERRUPTS                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Read the current interrupt status register.
 * Value is read directly from the device, and stored in cache.
 *
 * When perr is not NULL, it points to the error status:
 * - ( ENOERR)  Success
 * - (-EINVAL)  Invalid parameters
 * - (-ENODEV)  No such device
 * - (-EIO)     I/O error
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Status of interrupts
 ****************************************************************************/
CYG_BYTE
CS42XX_GetIntStatus(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	return CS42XX_PEEK_NOW_(devid, CS42XX_REG_INTSTATUS, perr);
}


/** Get the interrupt mask as an integer value.
 *
 * When perr is not NULL, it points to the error status:
 * - ( ENOERR)  Success
 * - (-EINVAL)  Invalid parameters
 * - (-ENODEV)  No such device
 * - (-EIO)     I/O error
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Value of interrupt mask
 ****************************************************************************/
CYG_BYTE
CS42XX_GetIntMaskVal(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_REG reg;
	CYG_BYTE   intrmask;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_INTMASK, perr);
	// Get only valid bits
	intrmask = reg &
			( CS42XX_REGM_INT_ADCCLKERR
			| CS42XX_REGM_INT_DACCLKERR
			| CS42XX_REGM_INT_ADCOVFL
			| CS42XX_REGM_INT_ADCUNDFL);

	return intrmask;
}


/** Configure the INT output pin as active high or low.
 *
 * @param[in] devid    Symbolic device Id
 * @param[in] activeh  Active high (true) or low (false)
 * @param[in] sync     Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Device does not support this feature
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigInt_OutputPin(CS42XX_DEVID devid, cyg_bool activeh,
		CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo  err;
	CS42XX_REG reg;

	if (CS42XX_Supp_HasIntrOutput(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_OUTCTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_DAC_INTACTIVEHIGH, activeh);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DAC_OUTCTRL,
						reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No interrupt output pin on that device.
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set the interrupt mask (detailed).
 *
 * @note  This function is OBSOLETE and should be removed soon (too many
 *        parameters).
 *
 * @param[in] devid      Symbolic device Id
 * @param[in] adcclkerr  true to trigger on ADC clock error
 * @param[in] dacclkerr  true to trigger on DAC clock error
 * @param[in] adcovrfl   true to trigger on ADC overflow
 * @param[in] adcundrfl  true to trigger on ADC underflow
 * @param[in] sync       Immediate change
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigInt_Mask(CS42XX_DEVID devid,
		cyg_bool adcclkerr, cyg_bool dacclkerr,
		cyg_bool adcovrfl,  cyg_bool adcundrfl,
		CS42XX_CACHESYNC sync)
{
	CYG_BYTE mask;

	mask = (adcclkerr? CS42XX_REGM_INT_ADCCLKERR: 0x00)
		|  (dacclkerr? CS42XX_REGM_INT_DACCLKERR: 0x00)
		|  (adcovrfl ? CS42XX_REGM_INT_ADCOVFL:   0x00)
		|  (adcundrfl? CS42XX_REGM_INT_ADCUNDFL:  0x00);

	return CS42XX_ConfigInt_MaskVal(devid, mask, sync);
}


/** Set the interrupt mask (mask given as integer).
 *
 * @warning  Unsupported bits are ignored, i.e. -EDEVNOSUPP is never
 *           returned.
 *
 * @param[in] devid     Symbolic device Id
 * @param[in] intrmask  Interrupt mask (bit=1 for activation)
 * @param[in] sync      Immediate change
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigInt_MaskVal(CS42XX_DEVID devid, CYG_BYTE intrmask,
		CS42XX_CACHESYNC sync)
{
	CYG_BYTE mask = 0x00;

	if (CS42XX_Supp_HasADC(devid))
	{
		mask |= CS42XX_REGM_INT_ADCCLKERR
			 | CS42XX_REGM_INT_ADCOVFL | CS42XX_REGM_INT_ADCUNDFL;
	}
	// else ADC is not supported.
	if (CS42XX_Supp_HasDAC(devid))
	{
		mask |= CS42XX_REGM_INT_DACCLKERR;
	}
	// else DAC is not supported.

	return CS42XX_POKE_COMMITCOND(devid,
			CS42XX_REG_INTMASK, (intrmask & mask), sync);
}


/** Set interrupt triggering mode (the same for all events).
 *
 * @param[in] devid     Symbolic device Id
 * @param[in] trigmode  Interrupt triggering mode when an event occurs
 * @param[in] sync      Immediate change
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigInt_TrigMode(CS42XX_DEVID devid, CS42XX_TRIGMODE trigmode,
		CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  msb = 0, lsb = 0;
	CYG_BYTE  intrmask = 0x00;

	/*
	 * Triggering modes are defined in two registers, to be read by
	 * "bit-column".
	 * E.g. if event triggering mode for events is '0bxy' (like 0b10),
	 * the two registers will contain:
	 *                  [Event1] [Event2] ...
	 * MSbits register     x1       x2    ...
	 * LSbits register     y1       y2    ...
	 *
	 * For easier use, all events are triggered on the same behaviour, i.e.
	 * x=x1=x2=... and y=y1=y2=... (no need to be more complex).
	 */

	// Create a mask so that all events are affected in once
	if (CS42XX_Supp_HasADC(devid))
	{
		intrmask |= CS42XX_REGM_INT_ADCCLKERR
				 | CS42XX_REGM_INT_ADCOVFL | CS42XX_REGM_INT_ADCUNDFL;
	}
	// else ADC is not supported.
	if (CS42XX_Supp_HasDAC(devid))
	{
		intrmask |= CS42XX_REGM_INT_DACCLKERR;
	}
	// else DAC is not supported.

	// Get the current registers contents
	msb = CS42XX_PEEK_(devid, CS42XX_REG_INTMODE_MSB, &err);
	if (ENOERR == err)
	{
		lsb = CS42XX_PEEK_(devid, CS42XX_REG_INTMODE_LSB, &err);
		if (ENOERR == err)
		{
			// Set the mask on MSbit(s) and LSbit(s)
			msb = ((trigmode >> 1)? intrmask: 0x00);
			lsb = ((trigmode &  1)? intrmask: 0x00);

			err = CS42XX_POKE(devid, CS42XX_REG_INTMODE_MSB, msb);
			if (ENOERR == err)
			{
				err = CS42XX_POKE(devid, CS42XX_REG_INTMODE_LSB, lsb);
				if ((ENOERR == err) && sync)
				{
					err = CS42XX_COMMIT(devid);
				}
				// else failure during 2nd POKE.
			}
			// else failure during 1st POKE.
		}
		// else failure during 2nd PEEK.
	}
	// else failure during 1st PEEK.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                            POWER CONTROL                                *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Freeze or enable changes on control ports of codec.
 * This command is always synchronous (i.e. if enabled, cache is flushed).
 * 
 * @param[in] devid   Symbolic device Id
 * @param[in] freeze  true if control ports shall be frozen, else false
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_FreezePorts(CS42XX_DEVID devid, cyg_bool freeze)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_POWER_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_POWER_FREEZE, freeze);
		err = CS42XX_POKE_COMMIT(devid, CS42XX_REG_POWER_CTRL, reg);
	}
	// else failure reading data.

	return err;
}


/** Set microphone preamplifier block in low-power mode.
 * 
 * @param[in] devid   Symbolic device Id
 * @param[in] lowpow  true if block shall enter low-power mode
 * @param[in] sync    Immediate change
 * 
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetPowerDown_Mic(CS42XX_DEVID devid,
		cyg_bool lowpow, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_POWER_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_POWER_MIC, lowpow);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_POWER_CTRL, reg, sync);
	}
	// else failure reading data.

	return err;
}


/** Set ADC pair in low-power mode.
 * 
 * @param[in] devid   Symbolic device Id
 * @param[in] lowpow  true if ADC shall enter low-power mode
 * @param[in] sync    Immediate change
 * 
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetPowerDown_ADC(CS42XX_DEVID devid,
		cyg_bool lowpow, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_POWER_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_POWER_ADC, lowpow);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_POWER_CTRL, reg, sync);
	}
	// else failure reading data.

	return err;
}


/** Set DAC pair in low-power mode.
 * 
 * @param[in] devid   Symbolic device Id
 * @param[in] lowpow  true if DAC shall enter low-power mode
 * @param[in] sync    Immediate change
 * 
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetPowerDown_DAC(CS42XX_DEVID devid,
		cyg_bool lowpow, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_POWER_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_POWER_DAC, lowpow);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_POWER_CTRL, reg, sync);
	}
	// else failure reading data.

	return err;
}


/** Set device in power-down mode (direct register access).
 * This command is always synchronous (i.e. if enabled, cache is flushed).
 * 
 * @param[in] devid    Symbolic device Id
 * @param[in] powdown  true if device shall enter power-down mode
 * @param[in] sync     Immediate change
 * 
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetPowerDown(CS42XX_DEVID devid, cyg_bool powdown)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_POWER_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_POWER_DEVICE, powdown);
		err = CS42XX_POKE_COMMIT(devid, CS42XX_REG_POWER_CTRL, reg);
	}
	// else failure reading data.

	return err;
}


/** Power the device up (blocking function).
 * During power-up, audio (input, output) is muted. Application has to unmute
 * input and output (if wished) when this function returns.
 *
 * @warning  When successful, and if the function is called from a thread,
 *           it blocks during about 200ms for the power up to be
 *           effective...
 *
 * @param[in] devid  Symbolic device Id
 *
 * @retval  ENOERR  Success
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_PowerUp(CS42XX_DEVID devid)
{
	Cyg_ErrNo err = ENOERR;

	// Mute input and output, if feature is supported
	err = CS42XX_Mute_ADC(devid, true, CS42XX_CACHE_NOSYNC);
	if ((ENOERR == err) || (-EDEVNOSUPP == err))
	{
		err = CS42XX_Mute_DAC(devid, true, CS42XX_CACHE_NOSYNC);
	}
	if ((ENOERR == err) || (-EDEVNOSUPP == err))
	{
		CS42XX_COMMIT(devid);
	}
	// else failure during previous operations: do not commit.

	if (ENOERR == err)
	{
		// Power device up
		err = CS42XX_SetPowerDown(devid, false);
		if (ENOERR == err)
		{
			// Wait for device to be powered up
			CS42XX_DELAY_MS(200);
		}
		// else power up failed.
	}
	// else failure during previous operations: cancel power up.

	return err;
}


/** Power the device down (blocking function).
 * Audio (input, output) is muted before the power down.
 *
 * @warning  When successful, this function is blocking during at least
 *           320ms...
 *
 * @param[in] devid  Symbolic device Id
 *
 * @retval  ENOERR  Success
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_PowerDown(CS42XX_DEVID devid)
{
	Cyg_ErrNo err = ENOERR;

	// Mute input and output, if feature is supported
	err = CS42XX_Mute_ADC(devid, true, CS42XX_CACHE_NOSYNC);
	if ((ENOERR == err) || (-EDEVNOSUPP == err))
	{
		err = CS42XX_Mute_DAC(devid, true, CS42XX_CACHE_NOSYNC);
	}
	if ((ENOERR == err) || (-EDEVNOSUPP == err))
	{
		err = CS42XX_COMMIT(devid);
		// Wait for mute to be effective
		CS42XX_DELAY_MS(20);
	}
	// else failure during previous operations: do not commit.

	if (ENOERR == err)
	{
		// Power device down
		err = CS42XX_SetPowerDown(devid, true);
		if (ENOERR == err)
		{
			// Wait for device to be powered up
			CS42XX_DELAY_MS(300);
		}
		// else power down failed.
	}
	// else failure during previous operations: cancel power down.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                    SETTINGS OF DIGITAL INTERFACES                       *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get format on digital output (related to ADC).
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Frame format
 ****************************************************************************/
CS42XX_DIF
CS42XX_GetFrameFormat_Out(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_DIF dif;
	CYG_BYTE   reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, perr);
		dif = CS42XX_GETVAL(reg, CS42XX_REGM_ADC_DIF_M, CS42XX_REGM_ADC_DIF_SHF);
	}
	else
	{
		// No ADC on that device
		dif = CS42XX_DIF_LEFT24;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return dif;
}


/** Get format on digital input (related to DAC).
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Frame format
 ****************************************************************************/
CS42XX_DIF
CS42XX_GetFrameFormat_In(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_DIF dif;
	CYG_BYTE   reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, perr);
		dif = CS42XX_GETVAL(reg, CS42XX_REGM_DAC_DIF_M, CS42XX_REGM_DAC_DIF_SHF);
	}
	else
	{
		// No DAC on that device
		dif = CS42XX_DIF_LEFT24;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return dif;
}


/** Get if device is master of digital output interface.
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   Device is master on digital output interface
 * @retval false  Device is slave on digital output interface
 ****************************************************************************/
cyg_bool
CS42XX_IsMaster_Out(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool ismaster;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, perr);
		ismaster = ((reg & CS42XX_REGM_ADC_MASTER)? true: false);
	}
	else
	{
		// No ADC on that device
		ismaster = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return ismaster;
}


/** Get if device is master of digital input interface.
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   Device is master on digital input interface
 * @retval false  Device is slave on digital input interface
 ****************************************************************************/
cyg_bool
CS42XX_IsMaster_In(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool ismaster;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, perr);
		ismaster = ((reg & CS42XX_REGM_DAC_MASTER)? true: false);
	}
	else
	{
		// No DAC on that device
		ismaster = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return ismaster;
}


/** Get the MCLK1 frequency divider (related to output interface).
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  The MCLK1 frequency divider (digital output)
 ****************************************************************************/
CS42XX_MCLK_DIV
CS42XX_GetMCLKFreq_Out(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_MCLK_DIV mclkdiv;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_MCLK_FREQ, perr);
		mclkdiv = CS42XX_GETVAL(reg, CS42XX_REGM_MCLK1_FREQ_M,
				CS42XX_REGM_MCLK1_FREQ_SHF);
	}
	else
	{
		// No ADC on that device
		mclkdiv = CS42XX_MCLK_DIV_1;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return mclkdiv;
}


/** Get the MCLK2 frequency divider (related to input interface).
 *
 * @warning  In synchronous mode, this value MAY NOT BE the real MCLK
 *           frequency, since the device uses MCLK1 for both digital
 *           interfaces.
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  The MCLK2 frequency divider (digital input)
 ****************************************************************************/
CS42XX_MCLK_DIV
CS42XX_GetMCLKFreq_In(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_MCLK_DIV mclkdiv;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_MCLK_FREQ, perr);
		mclkdiv = CS42XX_GETVAL(reg, CS42XX_REGM_MCLK2_FREQ_M,
				CS42XX_REGM_MCLK2_FREQ_SHF);
	}
	else
	{
		// No DAC on that device
		mclkdiv = CS42XX_MCLK_DIV_1;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return mclkdiv;
}


/** Get if sample rates of digital output (ADC) and digital input (DAC) are
 *  asynchronous.
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   Asynchronous mode: each interface has its own MCLK
 * @retval false  Both sample rates are derived from MCLK1 (MCLK2 is useless)
 ****************************************************************************/
cyg_bool
CS42XX_IsAsyncMode(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool async;
	CYG_BYTE reg;

	if (CS42XX_Supp_CanClockAsync(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, perr);
		async = ((reg & CS42XX_REGM_SIGSEL_ASYNC)? true: false);
	}
	else
	{
		// Device does not support asynchronous mode
		async = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
	}

	return async;
}


/** Get sample rate of digital output (related to ADC).
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Configured sample rate of digital output
 ****************************************************************************/
CS42XX_SAMPLERATE
CS42XX_GetSampleRate_Out(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_SAMPLERATE samplerate;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, perr);
		samplerate = CS42XX_GETVAL(reg, CS42XX_REGM_ADC_FM_M,
				CS42XX_REGM_ADC_FM_SHF);
	}
	else
	{
		// No ADC on that device
		samplerate = CS42XX_SAMPLERATE_SINGLE;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
	}

	return samplerate;
}


/** Get sample rate of digital input (related to DAC).
 *
 * If perr is not NULL, it will point to an error status:
 * - ( ENOERR)      Success
 * - (-ENODEV)      No such device
 * - (-EIO)         I/O error
 * - (-EDEVNOSUPP)  No DAC on that device
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Configured sample rate of digital input
 ****************************************************************************/
CS42XX_SAMPLERATE
CS42XX_GetSampleRate_In(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_SAMPLERATE samplerate;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, perr);
		samplerate = CS42XX_GETVAL(reg, CS42XX_REGM_DAC_FM_M,
				CS42XX_REGM_DAC_FM_SHF);
	}
	else
	{
		// No DAC on that device
		samplerate = CS42XX_SAMPLERATE_SINGLE;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
	}

	return samplerate;
}


/** Set format on digital output (related to ADC).
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] format  Format of frames on the interface
 * @param[in] sync    Immediate change
 * 
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigFrameFormat_Out(CS42XX_DEVID devid,
		CS42XX_DIF format, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETVAL(reg, format,
					CS42XX_REGM_ADC_DIF_M, CS42XX_REGM_ADC_DIF_SHF);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_ADC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No ADC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set format on digital input (related to DAC).
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] format  Format of frames on the interface
 * @param[in] sync    Immediate change
 * 
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigFrameFormat_In(CS42XX_DEVID devid,
		CS42XX_DIF format, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETVAL(reg, format,
					CS42XX_REGM_DAC_DIF_M, CS42XX_REGM_DAC_DIF_SHF);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_DAC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No DAC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set digital output interface of device in either master or slave mode.
 * In master mode, LRCK and SCLK are controlled by audio codec.
 * 
 * @param[in] devid   Symbolic device Id
 * @param[in] master  Master (true) or slave (false) codec interface
 * @param[in] sync    Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigMaster_Out(CS42XX_DEVID devid,
		CS42XX_DIFMS master, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;
	
	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_ADC_MASTER, master);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_ADC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No ADC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set digital input interface of device in either master or slave mode.
 * In master mode, LRCK and SCLK are controlled by audio codec.
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] master  Master (true) or slave (false) codec interface
 * @param[in] sync    Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigMaster_In(CS42XX_DEVID devid,
		CS42XX_DIFMS master, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_DAC_MASTER, master);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_DAC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No DAC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set MCLK(1,2) frequency dividers (MCLK1: output, MCLK2: input).
 *
 * @warning  If synchronous mode is selected, the MCLK2 divider will be the
 *           same as MCLK1 divider, whatever the 'divin' value given here.
 *
 * @note  If the device supports only ADC or DAC (thus has only one MCLK
 *        frequency), only the related divider will be taken in account.
 *        No -EDEVNOSUPP is returned.
 *        E.g.  for an ADC-only device, only divout will be used as MCLK
 *              parameter; and vice versa.
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] divout  Divider for MCLK1 (related to output interface)
 * @param[in] divin   Divider for MCLK2 (related to input interface)
 * @param[in] sync    Immediate change
 *
 * @retval  ENOERR  Success
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 * @retval -EIO     I/O error
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigMCLKFreq(CS42XX_DEVID devid,
		CS42XX_MCLK_DIV divout, CS42XX_MCLK_DIV divin,
		CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_MCLK_FREQ, &err);
	if (ENOERR == err)
	{
		if ((divout < CS42XX_MCLK_DIV_NB) &&
			(divin  < CS42XX_MCLK_DIV_NB))
		{
			if (CS42XX_Supp_HasADC(devid))
			{
				CS42XX_SETVAL(reg, divout, CS42XX_REGM_MCLK1_FREQ_M,
						CS42XX_REGM_MCLK1_FREQ_SHF);
			}
			// else device does not implement MCLK1.

			if (CS42XX_Supp_HasDAC(devid))
			{
				CS42XX_SETVAL(reg, divin, CS42XX_REGM_MCLK2_FREQ_M,
						CS42XX_REGM_MCLK2_FREQ_SHF);
			}
			// else device does not implement MCLK2.

			// Commit register changes
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_MCLK_FREQ,
						reg, sync);
		}
		else
		{
			// Invalid MCLK divider values
			err = -EINVAL;
		}
	}
	// else failure retrieving data.

	return err;
}


/** Set the relation between sample rates of digital output (ADC) and digital
 *  input (DAC).
 *
 * @param[in] devid  Symbolic device Id
 * @param[in] async  true to allow asynchronous sample rates between
 *                   interfaces
 * @param[in] sync   Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigAsyncMode(CS42XX_DEVID devid,
		cyg_bool async, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;
	
	if (CS42XX_Supp_CanClockAsync(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_SIGSEL_ASYNC, async);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_SIGNALSEL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// Device does not support asynchronous mode
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set sample rate of digital output (related to ADC).
 *
 * @param[in] devid  Symbolic device Id
 * @param[in] fm     Functional mode (determining sample rate)
 * @param[in] sync   Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigSampleRate_Out(CS42XX_DEVID devid,
		CS42XX_SAMPLERATE fm, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, &err);
		if (ENOERR == err)
		{
			if (fm < CS42XX_SAMPLERATE_NB)
			{
				CS42XX_SETVAL(reg, fm, CS42XX_REGM_ADC_FM_M, CS42XX_REGM_ADC_FM_SHF);
				err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADC_CTRL,
							reg, sync);
			}
			else
			{
				err = -EINVAL;
			}
		}
		// else failure retrieving data.
	}
	else
	{
		// No ADC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set sample rate of digital input (related to DAC).
 *
 * @param[in] devid  Symbolic device Id
 * @param[in] fm     Functional mode (determining sample rate)
 * @param[in] sync   Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigSampleRate_In(CS42XX_DEVID devid,
		CS42XX_SAMPLERATE fm, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;
	
	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, &err);
		if (ENOERR == err)
		{
			if (fm < CS42XX_SAMPLERATE_NB)
			{
				CS42XX_SETVAL(reg, fm, CS42XX_REGM_DAC_FM_M, CS42XX_REGM_DAC_FM_SHF);
				err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DAC_CTRL,
							reg, sync);
			}
			else
			{
				err = -EINVAL;
			}
		}
		// else failure retrieving data.
	}
	else
	{
		// No DAC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Set sample rate for both digital I/O.
 * 
 * @param[in] devid  Symbolic device Id
 * @param[in] fm     Functional mode (determining sample rate)
 * @param[in] sync   Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  At least 1 feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigSampleRate(CS42XX_DEVID devid,
		CS42XX_SAMPLERATE fm, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;

	err = CS42XX_ConfigSampleRate_Out(devid, fm, false);
	if (ENOERR == err)
	{
		err = CS42XX_ConfigSampleRate_In(devid, fm, sync);
	}
	// else the first sample rate configuration failed.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                 MUTE                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get mute state of ADC digital output.
 *
 * If perr is not NULL, it will point to an error status:
 * - ( ENOERR)      Success
 * - (-ENODEV)      No such device
 * - (-EIO)         I/O error
 * - (-EDEVNOSUPP)  No ADC on that device
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   ADC digital output is muted
 * @retval false  ADC digital output is unmuted
 ****************************************************************************/
cyg_bool
CS42XX_IsMuted_ADC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool muted;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, perr);
		muted = ((reg & CS42XX_REGM_ADC_MUTE)? true: false);
	}
	else
	{
		// No ADC on that device.
		muted = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return muted;
}


/** Get mute state of DAC analog output.
 *
 * If perr is not NULL, it will point to an error status:
 * - ( ENOERR)      Success
 * - (-ENODEV)      No such device
 * - (-EIO)         I/O error
 * - (-EDEVNOSUPP)  No DAC on that device
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   DAC analog output is muted
 * @retval false  DAC analog output is unmuted
 ****************************************************************************/
cyg_bool
CS42XX_IsMuted_DAC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool muted;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, perr);
		muted = ((reg & CS42XX_REGM_DAC_MUTE)? true: false);
	}
	else
	{
		// No DAC on that device.
		muted = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no report of error status.
	}

	return muted;
}


/** Mute or unmute ADC digital outputs (signal caught from ADC inputs).
 *
 * The ADCLEVELCHGPOLICY policy describes how to apply "Mute" on the
 * input signal (digitized in audio codec).
 *
 * @param devid  Symbolic device Id
 * @param mute   "Mute" status requested
 * @param sync   Immediate change
 * 
 * @retval  ENOERR      Success
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Mute_ADC(CS42XX_DEVID devid, cyg_bool mute, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;
	
	if (CS42XX_Supp_HasADC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_ADC_MUTE, mute);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_ADC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No ADC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Mute or unmute DAC analog outputs.
 * If "Mute" is set, the /MUTE line will be active as well.
 *
 * The DACLEVELCHGPOLICY policy describes how to apply "Mute" on the analog
 * output signal.
 *
 * @param devid  Symbolic device Id
 * @param mute   "Mute" status requested
 * @param sync   Immediate change
 * 
 * @retval  ENOERR      Success
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Feature not supported by this device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Mute_DAC(CS42XX_DEVID devid, cyg_bool mute, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;
	
	if (CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, &err);
		if (ENOERR == err)
		{
			CS42XX_SETBITMASK(reg, CS42XX_REGM_DAC_MUTE, mute);
			err = CS42XX_POKE_COMMITCOND(devid,
					CS42XX_REG_DAC_CTRL, reg, sync);
		}
		// else failure retrieving data.
	}
	else
	{
		// No DAC on that device
		err = -EDEVNOSUPP;
	}

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                             INPUT GAIN                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get input PGA gain on channel A.
 *
 * @param devid   Symbolic device Id
 *
 * @return  (Signed) input PGA gain on channel A
 ****************************************************************************/
cyg_int8
CS42XX_GetGain_ADCA(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_uint8 gain6b;

	gain6b = (CS42XX_PEEK_(devid, CS42XX_REG_ADCA_PGA, perr)
				& CS42XX_REGM_ADCA_PGAGAIN);
	// If the 6-bit value is negative, extend sign to 8 bits.
	if (gain6b & CS42XX_REGM_ADCA_PGAGAIN_NEG)
	{
		gain6b |= ~CS42XX_REGM_ADCA_PGAGAIN;
	}
	// else value is positive.

	return (cyg_int8)gain6b;
}


/** Get input PGA gain on channel B.
 *
 * @param devid   Symbolic device Id
 *
 * @return  (Signed) input PGA gain on channel B
 ****************************************************************************/
cyg_int8
CS42XX_GetGain_ADCB(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_uint8 gain6b;

	gain6b = (CS42XX_PEEK_(devid, CS42XX_REG_ADCB_PGA, perr)
				& CS42XX_REGM_ADCB_PGAGAIN);
	// If the 6-bit value is negative, extend sign to 8 bits.
	if (gain6b & CS42XX_REGM_ADCB_PGAGAIN_NEG)
	{
		gain6b |= ~CS42XX_REGM_ADCB_PGAGAIN;
	}
	// else value is positive.

	return (cyg_int8)gain6b;
}


/** Set PGA gain (of ADC) for channel A.
 * Gain is between -12 dB and +12 dB (other values are reserved). Invalid
 * gain results in no gain setting (and an error occurs).
 *
 * The audio codec receives a 6-bit signed value for PGA gain.
 * 
 * The ADCLEVELCHGPOLICY policy describes how to apply the new volume on
 * input signal (digitized in audio codec).
 * 
 * @param devid  Symbolic device Id
 * @param gain   (Signed) gain of PGA for this channel
 * @param sync   Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetGain_ADCA(CS42XX_DEVID devid,
		cyg_int8 gain, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADCA_PGA, &err);
	if (ENOERR == err)
	{
		if ((gain >= CS42XX_PGAGAIN_MIN) &&
			(gain <= CS42XX_PGAGAIN_MAX))
		{
			// (Keep reserved bits, if any)
			CS42XX_SETVAL(reg, ((cyg_uint8)gain), CS42XX_REGM_ADCA_PGAGAIN, 0);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADCA_PGA,
						reg, sync);
		}
		else
		{
			err = -EINVAL;
		}
	}
	// else failure retrieving data.

	return err;
}


/** Set PGA gain (of ADC) for channel B.
 * Gain is between -12 dB and +12 dB (other values are reserved). Invalid
 * gain results in no gain setting (and an error occurs).
 *
 * The audio codec receives a 6-bit signed value for PGA gain.
 * 
 * The ADCLEVELCHGPOLICY policy describes how to apply the new volume on
 * input signal (digitized in audio codec).
 * 
 * @param devid  Symbolic device Id
 * @param gain   (Signed) gain of PGA for this channel
 * @param sync   Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetGain_ADCB(CS42XX_DEVID devid,
		cyg_int8 gain, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADCB_PGA, &err);
	if (ENOERR == err)
	{
		if ((gain >= CS42XX_PGAGAIN_MIN) &&
			(gain <= CS42XX_PGAGAIN_MAX))
		{
			// (Keep reserved bits on, if any)
			CS42XX_SETVAL(reg, ((cyg_uint8)gain), CS42XX_REGM_ADCB_PGAGAIN, 0);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADCB_PGA,
						reg, sync);
		}
		else
		{
			err = -EINVAL;
		}
	}
	// else failure retrieving data.

	return err;
}


/** Set PGA gain (of ADC) for channels A and B.
 * Gain is between -12 dB and +12 dB (other values are reserved). Invalid
 * gain results in no gain setting (and an error occurs).
 *
 * The audio codec receives a 6-bit signed value for PGA gain.
 * 
 * The ADCLEVELCHGPOLICY policy describes how to apply the new volume on
 * input signal (digitized in audio codec).
 * 
 * @param devid  Symbolic device Id
 * @param gaina  (Signed) gain of PGA for channel A
 * @param gainb  (Signed) gain of PGA for channel B
 * @param sync   Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetGain_ADC(CS42XX_DEVID devid,
		cyg_int8 gaina, cyg_int8 gainb, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;

	err = CS42XX_SetGain_ADCA(devid, gaina, false);
	if (ENOERR == err)
	{
		err = CS42XX_SetGain_ADCB(devid, gainb, sync);
	}
	// else failure during first operation.

	return err;
}


/** Increase input PGA gain on both channels.
 *
 * @param devid  Symbolic device Id
 * @param incr   Increment value for PGA gain
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetGain_ADC_Up(CS42XX_DEVID devid,
		cyg_uint8 incr, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	cyg_int8  gaina, gainb;
	
	gaina = CS42XX_GetGain_ADCA(devid, &err);
	if (ENOERR == err)
	{
		gainb = CS42XX_GetGain_ADCB(devid, &err);
	}
	// else failure during first data retrieval.

	if (ENOERR == err)
	{
		if (gaina > CS42XX_PGAGAIN_MAX - incr)
		{
			gaina = CS42XX_PGAGAIN_MAX;
		}
		else
		{
			gaina += incr;
		}
		
		if (gainb > CS42XX_PGAGAIN_MAX - incr)
		{
			gainb = CS42XX_PGAGAIN_MAX;
		}
		else
		{
			gainb += incr;
		}
		
		err = CS42XX_SetGain_ADC(devid, gaina, gainb, sync);
	}
	// else failure during first operations.

	return err;
}


/** Decrease input PGA gain on both channels.
 *
 * @param devid  Symbolic device Id
 * @param decr   Decrement value for PGA gain
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetGain_ADC_Down(CS42XX_DEVID devid,
		cyg_uint8 decr, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	cyg_int8  gaina, gainb;
	
	gaina = CS42XX_GetGain_ADCA(devid, &err);
	if (ENOERR == err)
	{
		gainb = CS42XX_GetGain_ADCB(devid, &err);
	}
	// else failure during first data retrieval.

	if (ENOERR == err)
	{
		if (gaina < CS42XX_PGAGAIN_MIN + decr)
		{
			gaina = CS42XX_PGAGAIN_MIN;
		}
		else
		{
			gaina -= decr;
		}
		
		if (gainb < CS42XX_PGAGAIN_MIN + decr)
		{
			gainb = CS42XX_PGAGAIN_MIN;
		}
		else
		{
			gainb -= decr;
		}
		
		err = CS42XX_SetGain_ADC(devid, gaina, gainb, sync);
	}
	// else failure during first operations.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                            OUTPUT VOLUME                                *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get output volume on channel A.
 * This function translates attenuation (in device register) into a volume
 * (for application), so that volume 0 means maximal attenuation.
 *
 * @param devid   Symbolic device Id
 *
 * @return  Output volume on channel A
 ****************************************************************************/
cyg_uint8
CS42XX_GetVolume_DACA(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	return (0xFF - (CS42XX_PEEK_(devid, CS42XX_REG_DACA_VOL, perr)
				& CS42XX_REGM_DACA_VOL));
}


/** Get output volume on channel B.
 * This function translates attenuation (in device register) into a volume
 * (for application), so that volume 0 means maximal attenuation.
 *
 * @param devid   Symbolic device Id
 *
 * @return  Output volume on channel B
 ****************************************************************************/
cyg_uint8
CS42XX_GetVolume_DACB(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	return (0xFF - (CS42XX_PEEK_(devid, CS42XX_REG_DACB_VOL, perr)
				& CS42XX_REGM_DACB_VOL));
}


/** Set output volume on channel A.
 * This function translates volume (for application side) into attenuation
 * (in device register), so that volume 0 means maximal attenuation.
 * Decreasing volume of 1 step means applying -0.5 dB more to the signal;
 * increasing volume of 1 step means applying -0.5 dB less to the signal.
 *
 * The DACLEVELCHGPOLICY policy describes how to apply the new volume on
 * analog output signal.
 *
 * @param devid   Symbolic device Id
 * @param volume  Final volume to be applied to this DAC output channel
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetVolume_DACA(CS42XX_DEVID devid,
		cyg_uint8 volume, CS42XX_CACHESYNC sync)
{
	return CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DACA_VOL,
				(0xFF - volume), sync);
}


/** Set output volume on channel B.
 * This function translates volume (for application side) into attenuation
 * (in device register), so that volume 0 means maximal attenuation.
 * Decreasing volume of 1 step means applying -0.5 dB more to the signal;
 * increasing volume of 1 step means applying -0.5 dB less to the signal.
 *
 * The DACLEVELCHGPOLICY policy describes how to apply the new volume on
 * analog output signal.
 *
 * @param devid   Symbolic device Id
 * @param volume  Final volume to be applied to this DAC output channel
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetVolume_DACB(CS42XX_DEVID devid,
		cyg_uint8 volume, CS42XX_CACHESYNC sync)
{
	return CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DACB_VOL,
				(0xFF - volume), sync);
}


/** Set output volume on channels A and B.
 * A volume of 0 means minimal signal, while 0xFF means maximal signal level.
 * Decreasing volume of 1 step means applying -0.5 dB more to the signal;
 * increasing volume of 1 step means applying -0.5 dB less to the signal.
 *
 * The DACLEVELCHGPOLICY policy describes how to apply the new volume on
 * analog output signal.
 *
 * @param devid   Symbolic device Id
 * @param vola    Final volume to be applied to DAC output channel A
 * @param volb    Final volume to be applied to DAC output channel B
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetVolume_DAC(CS42XX_DEVID devid,
		cyg_uint8 vola, cyg_uint8 volb, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;

	err = CS42XX_SetVolume_DACA(devid, vola, false);
	if (ENOERR == err)
	{
		err = CS42XX_SetVolume_DACB(devid, volb, sync);
	}
	// else failure during first operation.

	return err;
}


/** Increase output volume on both channels.
 * Increasing volume of 1 step means applying +0.5 dB more to the signal.
 *
 * @param devid  Symbolic device Id
 * @param incr   Increment value for volume level
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetVolume_DAC_Up(CS42XX_DEVID devid,
		cyg_uint8 incr, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	cyg_uint8 vola, volb;

	vola = CS42XX_GetVolume_DACA(devid, &err);
	if (ENOERR == err)
	{
		volb = CS42XX_GetVolume_DACB(devid, &err);
	}
	// else failure during first data retrieval.

	if (ENOERR == err)
	{
		if (vola > CS42XX_VOLUME_MAX - incr)
		{
			vola = CS42XX_VOLUME_MAX;
		}
		else
		{
			vola += incr;
		}
		
		if (volb > CS42XX_VOLUME_MAX - incr)
		{
			volb = CS42XX_VOLUME_MAX;
		}
		else
		{
			volb += incr;
		}
		
		err = CS42XX_SetVolume_DAC(devid, vola, volb, sync);
	}
	// else failure during first operations.

	return err;
}


/** Decrease output volume on both channels.
 * Decreasing volume of 1 step means applying -0.5 dB more to the signal.
 *
 * @param devid  Symbolic device Id
 * @param decr   Decrement value for volume level
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetVolume_DAC_Down(CS42XX_DEVID devid,
		cyg_uint8 decr, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	cyg_uint8 vola, volb;
	
	vola = CS42XX_GetVolume_DACA(devid, &err);
	if (ENOERR == err)
	{
		volb = CS42XX_GetVolume_DACB(devid, &err);
	}
	// else failure during first data retrieval.

	if (ENOERR == err)
	{
		if (vola < CS42XX_VOLUME_MIN + decr)
		{
			vola = CS42XX_VOLUME_MIN;
		}
		else
		{
			vola -= decr;
		}
		
		if (volb < CS42XX_VOLUME_MIN + decr)
		{
			volb = CS42XX_VOLUME_MIN;
		}
		else
		{
			volb -= decr;
		}
		
		err = CS42XX_SetVolume_DAC(devid, vola, volb, sync);
	}
	// else failure during first operations.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                          LEVEL CHANGE POLICY                            *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/**
 * The CS42XX audio codec implements policies to deal with level changes,
 * i.e. when signal has to be attenuated or muted.
 *
 * Available policies are:
 * - Apply changes directly on signal.
 * - "Soft Ramp": increase or decrease level by 1/8 dB steps, to reach
 *   final level.
 * - "Zero Crossing": apply changes on level only when signal crosses the
 *   zero line, to minimize audible artifacts.
 * - Combined "Soft Ramp" and "Zero Crossing".
 */



/** Get current policy about level changes on ADC.
 *
 * @param devid   Symbolic device Id
 *
 * @return  Policy applied for level changes on ADC
 ****************************************************************************/
CS42XX_LVLCHGPOLICY
CS42XX_GetLCPolicy_ADC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CYG_BYTE reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_INCTRL, perr);
	return CS42XX_GETVAL(reg, CS42XX_REGM_ADC_LVLCHGPOLICY_M,
				CS42XX_REGM_ADC_LVLCHGPOLICY_SHF);
}


/** Get current policy about level changes on DAC.
 *
 * @param devid   Symbolic device Id
 *
 * @return  Policy applied for level changes on DAC
 ****************************************************************************/
CS42XX_LVLCHGPOLICY
CS42XX_GetLCPolicy_DAC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CYG_BYTE reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_OUTCTRL, perr);
	return CS42XX_GETVAL(reg, CS42XX_REGM_DAC_LVLCHGPOLICY_M,
				CS42XX_REGM_DAC_LVLCHGPOLICY_SHF);
}


/** Configure policy to apply to level changes on ADC.
 *
 * @param devid   Symbolic device Id
 * @param policy  Policy to apply
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigLCPolicy_ADC(CS42XX_DEVID devid,
		CS42XX_LVLCHGPOLICY policy, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_INCTRL, &err);
	if (ENOERR == err)
	{
		if (policy < CS42XX_LVLCHGPOLICY_NB)
		{
			CS42XX_SETVAL(reg, policy, CS42XX_REGM_ADC_LVLCHGPOLICY_M,
						CS42XX_REGM_ADC_LVLCHGPOLICY_SHF);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADC_INCTRL,
						reg, sync);
		}
		else
		{
			err = -EINVAL;
		}
	}
	// else failure during data retrieval.

	return err;
}


/** Configure policy to apply to level changes on DAC.
 *
 * @param devid   Symbolic device Id
 * @param policy  Policy to apply
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigLCPolicy_DAC(CS42XX_DEVID devid,
		CS42XX_LVLCHGPOLICY policy, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_OUTCTRL, &err);
	if (ENOERR == err)
	{
		if (policy < CS42XX_LVLCHGPOLICY_NB)
		{
			CS42XX_SETVAL(reg, policy, CS42XX_REGM_DAC_LVLCHGPOLICY_M,
						CS42XX_REGM_DAC_LVLCHGPOLICY_SHF);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DAC_OUTCTRL,
						reg, sync);
		}
		else
		{
			err = -EINVAL;
		}
	}
	// else failure during data retrieval.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                           IN-CODEC FILTERS                              *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get if high-pass filter is activated on ADC input.
 *
 * @param devid   Symbolic device Id
 *
 * @retval true   Filter activated
 * @retval false  Filter not activated
 ****************************************************************************/
cyg_bool
CS42XX_IsFiltered_ADC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool filtered = false;
	CYG_BYTE reg;

	// Warning, bit=1 means "filter deactivated": invert the logic
	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, perr);
	filtered = ((reg & CS42XX_REGM_ADC_DISABLEFILTER)? false: true);

	return filtered;
}


/** Get if de-emphasis filter is activated on DAC output.
 *
 * @param devid   Symbolic device Id
 *
 * @retval true   Filter activated
 * @retval false  Filter not activated
 ****************************************************************************/
cyg_bool
CS42XX_IsFiltered_DAC(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool filtered = false;
	CYG_BYTE reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, perr);
	filtered = ((reg & CS42XX_REGM_DAC_DEEMPH)? true: false);

	return filtered;
}


/** Activate or deactivate high-pass filter on ADC input.
 *
 * @param devid   Symbolic device Id
 * @param filter  true if filter shall be activated
 * @param sync    Immediate change
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetFilter_ADC(CS42XX_DEVID devid,
		cyg_bool filter, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_CTRL, &err);
	if (ENOERR == err)
	{
		// Warning, bit=1 means "filter deactivated": invert the logic
		CS42XX_SETBITMASK(reg, CS42XX_REGM_ADC_DISABLEFILTER, !filter);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADC_CTRL,
					reg, sync);
	}
	// else failure during data retrieval.

	return err;
}


/** Activate or deactivate de-emphasis filter on DAC output.
 *
 * @param devid   Symbolic device Id
 * @param filter  true if filter shall be activated
 * @param sync    Immediate change
 * 
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SetFilter_DAC(CS42XX_DEVID devid,
		cyg_bool filter, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_DAC_CTRL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_DAC_DEEMPH, filter);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_DAC_CTRL,
					reg, sync);
	}
	// else failure during data retrieval.

	return err;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                           SIGNAL SELECTION                              *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Get the analog input used for ADC.
 *
 * @param devid   Symbolic device Id
 *
 * @return  Input used for ADC
 ****************************************************************************/
CS42XX_AINSEL
CS42XX_GetADCInput(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CYG_BYTE reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_INCTRL, perr);
	return CS42XX_GETVAL(reg, CS42XX_REGM_ADC_SEL_M, CS42XX_REGM_ADC_SEL_SHF);
}


/** Get the source of signal played on auxiliary output.
 *
 * If perr is not NULL, its content will contain the error code:
 * - ( ENOERR)      Success
 * - (-ENODEV)      No such device
 * - (-EIO)         I/O error
 * - (-EDEVNOSUPP)  No auxiliary output on that device
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @return  Source signal for auxiliary output
 ****************************************************************************/
CS42XX_AOUTSEL
CS42XX_GetAuxOut(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	CS42XX_AOUTSEL aoutsel;
	CYG_BYTE       reg;

	if (CS42XX_Supp_HasAuxOutput(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, perr);
		aoutsel = CS42XX_GETVAL(reg, CS42XX_REGM_SIGSEL_AOUTSEL_M,
					CS42XX_REGM_SIGSEL_AOUTSEL_SHF);
	}
	else
	{
		// No auxiliary output on that device.
		aoutsel = CS42XX_AOUTSEL_HIGHZ;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no return of error.
	}

	return aoutsel;
}


/** Get if "Loop mode" is activated.
 * 
 * If perr is not NULL, it will point to an error status:
 * - ( ENOERR)      Success
 * - (-ENODEV)      No such device
 * - (-EIO)         I/O error
 * - (-EDEVNOSUPP)  Either no DAC or no ADC on that device
 *
 * @note  If -EDEVNOSUPP is returned, you can consider that the return
 *        value (false) is however correct, since loop mode is not active
 *        (and will never be).
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] perr  Returned error status (can be NULL)
 *
 * @retval true   Loop mode activated
 * @retval false  Loop mode deactivated
 ****************************************************************************/
cyg_bool
CS42XX_GetLoopMode(CS42XX_DEVID devid, Cyg_ErrNo *perr)
{
	cyg_bool loopmode;
	CYG_BYTE reg;

	if (CS42XX_Supp_HasADC(devid) && CS42XX_Supp_HasDAC(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, perr);
		loopmode = ((reg & CS42XX_REGM_SIGSEL_LOOPBACK)? true: false);
	}
	else
	{
		// There is not both ADC and DAC: loop mode cannot exist.
		loopmode = false;
		if (NULL != perr)
		{
			*perr = -EDEVNOSUPP;
		}
		// else no return of error.
	}

	return loopmode;
}


/** Select the appropriate analog input for ADC.
 *
 * @param[in] devid   Symbolic device Id
 * @param[in] source  Source of signal
 * @param(in] sync    Immediate change
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         I/O error
 * @retval -EDEVNOSUPP  Source reference out of range for that device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigADCInput(CS42XX_DEVID devid,
		CS42XX_AINSEL source, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_ADC_INCTRL, &err);
	if (ENOERR == err)
	{
		if (source >= CS42XX_AINSEL_NB)
		{
			// Invalid input source reference (whatever the device).
			err = -EINVAL;
		}
		else if (source >= CS42XX_Supp_GetNbAnalogInputs(devid))
		{
			// Invalid input source reference for that device.
			err = -EDEVNOSUPP;
		}
		else
		{
			// Parameters are correct, set ADC input.
			CS42XX_SETVAL(reg, source, CS42XX_REGM_ADC_SEL_M,
						CS42XX_REGM_ADC_SEL_SHF);
			err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_ADC_INCTRL,
						reg, sync);
		}
	}
	// else failure during data retrieval.
	
	return err;
}


/** Switch the analog input for ADC; mute input during switch.
 *
 * @warning  Function is blocking...
 *
 * @param devid   Symbolic device Id
 * @param source  New source of signal
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_SwitchADCInput(CS42XX_DEVID devid, CS42XX_AINSEL source)
{
	Cyg_ErrNo err;

	// Mute input
	err = CS42XX_Mute_ADC(devid, true, CS42XX_CACHE_SYNC);
	if (ENOERR == err)
	{
		// Wait for mute to be effective
		CS42XX_DELAY_MS(20);
		
		// Switch input
		err = CS42XX_ConfigADCInput(devid, source, CS42XX_CACHE_NOSYNC);
		if (ENOERR == err)
		{
			// Unmute input
			err = CS42XX_Mute_ADC(devid, false, CS42XX_CACHE_SYNC);
		}
		// else failure during ADC input setting (mute is useless).
	}
	// else failure during data retrieval.
	
	return err;
}


/** Select the signal to be played on auxiliary output.
 *
 * @param devid   Symbolic device Id
 * @param source  Source of signal
 * @param sync    Immediate change
 *
 * @retval  ENOERR      Operation successful
 * @retval -EINVAL      Invalid parameters
 * @retval -ENODEV      No such device
 * @retval -EIO         Communication with device failed
 * @retval -EDEVNOSUPP  Device does not support auxiliary output
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigAuxOut(CS42XX_DEVID devid,
		CS42XX_AOUTSEL source, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	if (CS42XX_Supp_HasAuxOutput(devid))
	{
		reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, &err);
		if (ENOERR == err)
		{
			if (source < CS42XX_AOUTSEL_NB)
			{
				CS42XX_SETVAL(reg, source, CS42XX_REGM_SIGSEL_AOUTSEL_M,
							CS42XX_REGM_SIGSEL_AOUTSEL_SHF);
				err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_SIGNALSEL,
							reg, sync);
			}
			else
			{
				err = -EINVAL;
			}
		}
		// else failure during data retrieval (err informed).
	}
	else
	{
		// Not supported by device.
		err = -EDEVNOSUPP;
	}

	return err;
}


/** Activate or deactivate "Loop mode".
 * In "Loop mode", the digital signal converted from ADC is sent to DAC for
 * restitution; thus, the analog output is directly issued from the analog
 * input.
 * 
 * @param devid     Symbolic device Id
 * @param loopmode  true to activate loop mode
 * @param sync      Immediate change
 *
 * @return  ErrNo
 ****************************************************************************/
Cyg_ErrNo
CS42XX_ConfigLoopMode(CS42XX_DEVID devid,
		cyg_bool loopmode, CS42XX_CACHESYNC sync)
{
	Cyg_ErrNo err;
	CYG_BYTE  reg;

	reg = CS42XX_PEEK_(devid, CS42XX_REG_SIGNALSEL, &err);
	if (ENOERR == err)
	{
		CS42XX_SETBITMASK(reg, CS42XX_REGM_SIGSEL_LOOPBACK, loopmode);
		err = CS42XX_POKE_COMMITCOND(devid, CS42XX_REG_SIGNALSEL,
					reg, sync);
	}
	// else failure during data retrieval.

	return err;
}


