//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_evt.c
 * @brief  Management of events triggered by CS42XX devices.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *
 * This code has been mainly inspired from the GPIO button driver
 * (by Ivan Djelic, Parrot S.A.).
 *
 * @note  This file is specific to Parrot5(+) architecture, due to access to
 *        core registers.
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/devs/cs42xx_regs.h>
#include <cyg/devs/cs42xx_hal.h>
#include <cyg/devs/cs42xx_evt.h>


#ifdef CYGPKG_DEVICES_CS42XX_USEINTR


/*
 * Types and structures.
 ************************/


/** Type for the storage of event callback and its triggering conditions.
 *
 * @typedef CS42XX_EVT_ITEM
 */
typedef struct CS42XX_EVT_ITEM_TAG CS42XX_EVT_ITEM;
struct CS42XX_EVT_ITEM_TAG
{
	CS42XX_DEVID    devid;  ///< Symbolic device Id
	CS42XX_EVT_MASK events; ///< Bit-fielded event(s)
	CS42XX_EVT_CB   cback;  ///< Function called when one event occurs
};



/*
 * Variables.
 *************/


/** Table of callbacks. */
static CS42XX_EVT_ITEM CS42XXEvtTable[CS42XX_EVT_NBCALLBACKS];
static cyg_mutex_t     CS42XXEvtMutex; ///< Mutex of CS42XXEvtTable.



/*
 * Definition of functions.
 ***************************/



/** Update interrupt mask according to events to be caught.
 *
 * @retval  ENOERR      Success
 * @retval -ENXIO       Bad configuration of driver
 * @retval -EDEVNOTSUP  Device is unable to trigger events
 ****************************************************************************/
static Cyg_ErrNo
CS42XX_Evt_UpdateIntr(void)
{
	Cyg_ErrNo    err = ENOERR;
	CS42XX_DEVID devid;
	CS42XX_EVT_MASK maskcurrent, maskupdate[CS42XX_NBDEVICES];
	int i;

	// Initialize update process
	for (devid = 0; devid < CS42XX_NBDEVICES; devid++)
	{
		maskupdate[devid] = 0;
	}

	// Compute the interrupt mask for each device
	cyg_drv_mutex_lock(&CS42XXEvtMutex);
	for (i = 0; i < CS42XX_EVT_NBCALLBACKS; i++)
	{
		CS42XX_EVT_ITEM *evtdesc;

		evtdesc = &CS42XXEvtTable[i];
		if ((evtdesc->events) && (NULL != evtdesc->cback))
		{
			// Consider these events have to trigger an intr. from
			// this device
			maskupdate[evtdesc->devid] |= evtdesc->events;
		}
		// else slot is empty (or not valid): ignore it.
	}
	cyg_drv_mutex_unlock(&CS42XXEvtMutex);

#ifdef CS42XX_USECACHE

	// Update the interrupt mask of devices.
	// With cache enabled, it is possible to avoid updating a register
	// that finally does not change.
	for(devid = 0;
		(ENOERR == err) && (devid < CS42XX_NBDEVICES);
		devid++)
	{
		// Get current mask value
		maskcurrent = CS42XX_GetIntMaskVal(devid, &err);
		if (ENOERR == err)
		{
			// Compare current mask with the new one
			if (maskupdate[devid] != maskcurrent)
			{
				// Update interrupt mask on device
				err = CS42XX_ConfigInt_MaskVal(devid, maskupdate[devid],
						CS42XX_CACHE_SYNC);
			}
			// else interrupt mask is already OK.
		}
		// else impossible to read the interrupt mask of this device:
		// abort operation.
	}

#else // !CS42XX_USECACHE

	// Update the interrupt mask of devices.
	// Since register cache is disabled, interrupt mask is updated for
	// all devices.
	for(devid = 0;
		(ENOERR == err) && (devid < CS42XX_NBDEVICES);
		devid++)
	{
		// Update interrupt mask on device
		err = CS42XX_ConfigInt_MaskVal(devid, maskupdate[devid],
				CS42XX_CACHE_SYNC);
	}

#endif // CS42XX_USECACHE

	return err;
}


/** Associate a callback to a set of events occurring on the given device.
 *
 * If the callback was already used for this same device, events are added
 * to this entry.
 *
 * @param devid   Symbolic device Id
 * @param events  Event(s) on this device associated to this callback
 * @param cback   Callback
 *
 * @retval  ENOERR      Success
 * @retval -EINVAL      Invalid parameters
 * @retval -ENOMEM      No slot available for this callback
 * @retval -ENXIO       Bad configuration of driver
 * @retval -EDEVNOTSUP  Device is unable to trigger events
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Evt_Register(CS42XX_DEVID devid, CS42XX_EVT_MASK events,
		CS42XX_EVT_CB cback)
{
	static cyg_bool  initialized = false;

	Cyg_ErrNo        err = ENOERR;
	CS42XX_EVT_ITEM *evtdesc;
	int i, emptyentry = -1;

	// In case of first call, initialize the "Event" and "HAL" layers
	if (!initialized)
	{
		err = CS42XX_Evt_Init();
		if (ENOERR == err)
		{
			initialized = true;
		}
		// else initialization failed (reason in err).
	}
	// else initialization is already OK.

	if (ENOERR == err)
	{
		// Check validity of parameters
		if ((devid < CS42XX_NBDEVICES) && (events) && (NULL != cback))
		{
			cyg_drv_mutex_lock(&CS42XXEvtMutex);
			
			// Look for an already existing entry
			err = -ENOENT;
			for (i = 0; (i < CS42XX_EVT_NBCALLBACKS) && (-ENOENT == err); i++)
			{
				evtdesc = &CS42XXEvtTable[i];
				if ((cback == evtdesc->cback) && (devid == evtdesc->devid))
				{
					// Entry already present: add new events to the current set
					evtdesc->events |= (events & CS42XX_EVT_MASK_ALL);
					err = -ENOERR;
				}
				else if ((emptyentry < 0) && (NULL == evtdesc->cback))
				{
					emptyentry = i;
				}
				// else this entry has not to be processed.
			}
			
			// If no entry existed already, add a new one
			if (-ENOENT == err)
			{
				if (emptyentry < 0)
				{
					// No empty space found
					err = -ENOMEM;
				}
				else
				{
					// Store a new entry
					CS42XXEvtTable[emptyentry].devid  = devid;
					CS42XXEvtTable[emptyentry].cback  = cback;
					CS42XXEvtTable[emptyentry].events =
						(events & CS42XX_EVT_MASK_ALL);
					err = ENOERR;
				}
			}
			// else an entry has already been processed in previous loop.

			cyg_drv_mutex_unlock(&CS42XXEvtMutex);
		}
		else
		{
			err = -EINVAL;
		}
	}
	// else initialization was required and it failed.

	if (ENOERR == err)
	{
		// Update the interrupt mask of devices, if required
		err = CS42XX_Evt_UpdateIntr();
	}
	// else operation already failed.

	return err;
}


/** Unregister an event callback for any device.
 *
 * All entries are looked for that callback on that device.
 * If such a callback is found, it is removed (callback replaced by NULL).
 *
 * @param devid   Symbolic device Id
 * @param cback   Callback to be unregistered
 *
 * @retval  ENOERR      Success
 * @retval -ENOENT      Entry not found (kind of success, though)
 * @retval -EINVAL      Invalid parameters
 * @retval -EDEVNOTSUP  Device is unable to trigger events
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Evt_UnregisterCallback(CS42XX_EVT_CB cback)
{
	Cyg_ErrNo        err = -EINVAL;
	CS42XX_EVT_ITEM *evtdesc;
	int i;

	if (NULL != cback)
	{
		cyg_drv_mutex_lock(&CS42XXEvtMutex);

		// Look for an entry containing the callback
		err = -ENOENT;
		for (i = 0; i < CS42XX_EVT_NBCALLBACKS; i++)
		{
			evtdesc = &CS42XXEvtTable[i];
			if (evtdesc->cback == cback)
			{
				evtdesc->cback = NULL;
				err = ENOERR;
			}
			// else this entry has not to be processed.
		}

		cyg_drv_mutex_unlock(&CS42XXEvtMutex);
	}
	// else parameter is invalid.

	if (ENOERR == err)
	{
		// Update the interrupt mask of devices, if required
		err = CS42XX_Evt_UpdateIntr();
	}
	// else operation already failed.

	return err;
}


/** Stop catching some events on the given device.
 *
 * All entries are looked for these events.
 * If an entry finally contains no triggering event, it is removed.
 *
 * @param devid   Symbolic device Id
 * @param events  Event(s) on this device not to be caught anymore
 *
 * @retval  ENOERR      Success
 * @retval  ENOENT      Entry not found (kind of success, though)
 * @retval -EDEVNOTSUP  Device is unable to trigger events
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Evt_UnregisterEvents(CS42XX_DEVID devid, CS42XX_EVT_MASK events)
{
	Cyg_ErrNo        err = ENOENT;
	CS42XX_EVT_ITEM *evtdesc;
	int i;

	cyg_drv_mutex_lock(&CS42XXEvtMutex);

	for (i = 0; i < CS42XX_EVT_NBCALLBACKS; i++)
	{
		evtdesc = &CS42XXEvtTable[i];
		if ((evtdesc->devid == devid) &&
			(evtdesc->events & events))
		{
			// Remove the events so that they don't trigger callback anymore
			evtdesc->events &= ~events;
			if (!(evtdesc->events))
			{
				// No more trigger: remove the callback
				evtdesc->cback = NULL;
			}
			// else some events are still to be caught: keep callback.

			err = ENOERR;
			// Do not break the loop! so that other entries can be parsed too.
		}
		// else this entry has not to be processed.
	}

	cyg_drv_mutex_unlock(&CS42XXEvtMutex);

	if (ENOERR == err)
	{
		// Update the interrupt mask of devices, if required
		err = CS42XX_Evt_UpdateIntr();
	}
	// else operation already failed.

	return err;
}


/** Process the interrupt caught, by reading status registers then calling
 *  the appropriate callbacks.
 *
 * This function should be registered by the CS42XX_HAL module.
 ****************************************************************************/
void
CS42XX_Evt_ProcessIntr(void)
{
	Cyg_ErrNo    err;
	CS42XX_DEVID devid;
	int i;

	// Read the interrupt register of all devices related to that
	// interrupt pin
	for (devid = 0; devid < CS42XX_NBDEVICES; devid++)
	{
		CS42XX_REG intstatus =
				CS42XX_PEEK_NOW_(devid, CS42XX_REG_INTSTATUS, &err);
		// (Warning, the content of interrupt register in this device is
		//  reset by a read, so its content is now not accessible anymore.)

		if (intstatus)
		{
			cyg_drv_mutex_lock(&CS42XXEvtMutex);
			
			// Call the appropriate functions depending on event occurred
			for (i = 0; i < CS42XX_EVT_NBCALLBACKS; i++)
			{
				CS42XX_EVT_ITEM *evtdesc;
				CS42XX_EVT_MASK  events;

				evtdesc = &CS42XXEvtTable[i];
				events  = (intstatus & evtdesc->events);
				if ((events) && (devid == evtdesc->devid) &&
					(NULL != evtdesc->cback))
				{
					// Call the hook with appropriate parameters
					(evtdesc->cback)(devid, events);
				}
				// else slot is either empty or does not fit with current
				// event.
				
				// In all cases, step to next slot: multiple callbacks can
				// be called for a same device.
			}

			cyg_drv_mutex_unlock(&CS42XXEvtMutex);
		}
		// else no interrupt flag on that device.
	}
}

#endif // CYGPKG_DEVICES_CS42XX_USEINTR


/** Initialize this submodule.
 *
 * @retval  ENOERR      Success
 * @retval -ENXIO       Bad configuration of driver
 * @retval -EDEVNOTSUP  Device is unable to trigger events
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Evt_Init(void)
{
	Cyg_ErrNo err = ENOERR;

#ifdef CYGPKG_DEVICES_CS42XX_USEINTR
	int i;

	// Initialize the table of event callbacks
	cyg_drv_mutex_init(&CS42XXEvtMutex);
	for (i = 0; i < CS42XX_EVT_NBCALLBACKS; i++)
	{
		CS42XXEvtTable[i].devid  = (CS42XX_DEVID)0;
		CS42XXEvtTable[i].events = 0;
		CS42XXEvtTable[i].cback  = NULL;
	}

	// Install the interrupt handler
	err = CS42XX_HAL_InstallIntr(
				(CS42XX_HAL_PROCESSINTR_CB)CS42XX_Evt_ProcessIntr);
#endif // CYGPKG_DEVICES_CS42XX_USEINTR

	return err;
}


