//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_supp.c
 * @brief  Supported devices and capabilities of each one.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *  
 * @author David Tylski
 * @date   Mar. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <string.h>
#include <cyg/devs/cs42xx_regs.h>
#include <cyg/devs/cs42xx_supp.h>



/*
 * Defines.
 ***********/


// Capabilities
#define CS42XX_SUPP_FLAG_ADC		(1<<0) ///< ADC capability.
#define CS42XX_SUPP_FLAG_DAC		(1<<1) ///< DAC capability.
#define CS42XX_SUPP_FLAG_INDEPCLK	(1<<2) ///< ADC and DAC can be clocked by different signals.
#define CS42XX_SUPP_FLAG_INTR		(1<<3) ///< Device can drive an interrupt pin.
#define CS42XX_SUPP_FLAG_AUXOUT		(1<<4) ///< Device can drive an auxiliary output.
#define CS42XX_SUPP_FLAG_TXCTRL		(1<<5) ///< IEC60958-3 transmit control is supported.

/** Number of couples (chiptype,revision) supported by driver. */
#define CS42XX_SUPP_NBCHIPTYPES		9



/*
 * Types.
 *********/


/** Structure to store the list of supported devices and their properties.
 * This is local to this part of driver.
 *
 * @typedef CS42XX_CHIPDESCR
 */
typedef struct CS42XX_SUPP_CHIPDESCR_TAG CS42XX_SUPP_CHIPDESCR;
struct CS42XX_SUPP_CHIPDESCR_TAG
{
	CS42XX_REG lastreg;      ///< Offset of last register in device.
	CYG_BYTE   capabilities; ///< Capabilities (mask of bytes).
	cyg_uint8  nbainputs;    ///< Number of AIN analog inputs (r09:2-0).
	cyg_uint8  nbdinputs;    ///< Number of SDIN digital inputs (r06:7).
};

/** Structure to associate chip Id, readable name, and description of
 *  capabilities.
 *
 * This type is local to this part of driver.
 *
 * @typedef CS42XX_SUPP_CHIPTYPE
 */
typedef struct CS42XX_SUPP_CHIPTYPE_TAG CS42XX_SUPP_CHIPTYPE;
struct CS42XX_SUPP_CHIPTYPE_TAG
{
	CYG_BYTE                     id;    ///< Chip type and version.
	const char                  *name;  ///< Designation of chip.
	const CS42XX_SUPP_CHIPDESCR *descr; ///< Description of chip capabilities.
};


typedef struct CS42XX_DEVICEINFO_TAG CS42XX_DEVICEINFO;
struct CS42XX_DEVICEINFO_TAG
{
	cyg_io_handle_t             handle;   ///< Handle, eCos side.
	CS42XX_DEVID                devid;    ///< Device Id, driver internal.
	const CS42XX_SUPP_CHIPTYPE *chiptype; ///< Chip type.
};



/*
 * Constants.
 *************/


/** Description of audio codecs and their specificities.
 *
 * The list shall be ordered the most specific device first (from chip id
 * point of view).
 *
 * This table shall be parsed as soon as the driver knows the device
 * version number (chip part number and revision).
 */
static const CS42XX_SUPP_CHIPDESCR cCS42XXSuppChipDescription[] =
{
	// CS4245 audio codecs:
	{
		.lastreg      = CS42XX_REG_LAST_CS4245,
		.capabilities = CS42XX_SUPP_FLAG_ADC |
						CS42XX_SUPP_FLAG_DAC |
						CS42XX_SUPP_FLAG_INDEPCLK |
		                CS42XX_SUPP_FLAG_INTR |
		                CS42XX_SUPP_FLAG_AUXOUT,
		.nbdinputs    = 1,     // No selection of digital input
		.nbainputs    = 7,     // Microphone + Stereo AIN1 to AIN6
	},
	// CS4265 audio codecs:
	{
		.lastreg      = CS42XX_REG_LAST_CS4265,
		.capabilities = CS42XX_SUPP_FLAG_ADC |
						CS42XX_SUPP_FLAG_DAC |
						CS42XX_SUPP_FLAG_TXCTRL,
		.nbdinputs    = 2,      // Selection between SDIN1 and SDIN2
		.nbainputs    = 2,      // Either microphone or stereo line input
	},
	// CS5345 audio ADC:
	{
		.lastreg      = CS42XX_REG_LAST_CS5345,
		.capabilities = CS42XX_SUPP_FLAG_ADC,
		.nbdinputs    = 1,      // Selection between SDIN1 and SDIN2
		.nbainputs    = 1,      // Either microphone or stereo line input
	},
};

/** List of chips supported by this driver. */
static const CS42XX_SUPP_CHIPTYPE cCS42XXSuppChipType[CS42XX_SUPP_NBCHIPTYPES] =
{
	{
		.id     = 0xC1,
		.name   = "CS4245 rev. A",
		.descr  = &cCS42XXSuppChipDescription[0],
	},{
		.id     = 0xC2,
		.name   = "CS4245 rev. B/C0",
		.descr  = &cCS42XXSuppChipDescription[0],
	},{
		.id     = 0xC3,
		.name   = "CS4245 rev. C1",
		.descr  = &cCS42XXSuppChipDescription[0],
	},
	{
		.id     = 0xC4,
		.name   = "CS4245 rev. D0",
		.descr  = &cCS42XXSuppChipDescription[0],
	},
	{
		.id     = 0xD1,
		.name   = "CS4265 rev. A",
		.descr  = &cCS42XXSuppChipDescription[1],
	},{
		.id     = 0xD2,
		.name   = "CS4265 rev. B/C0",
		.descr  = &cCS42XXSuppChipDescription[1],
	},{
		.id     = 0xD3,
		.name   = "CS4265 rev. C1",
		.descr  = &cCS42XXSuppChipDescription[1],
	},{
		.id     = 0xE1,
		.name   = "CS5345 rev. A",
		.descr  = &cCS42XXSuppChipDescription[2],
	},{
		.id     = 0xE2,
		.name   = "CS5345 rev. B/C0",
		.descr  = &cCS42XXSuppChipDescription[2],
	},{
		.id     = 0xE3,
		.name   = "CS5345 rev. C1",
		.descr  = &cCS42XXSuppChipDescription[2],
	}
};



/*
 * Variables.
 *************/


/** Storage of all pieces of information relative to devices detected. */
static CS42XX_DEVICEINFO CS42XXDeviceInfo[CS42XX_NBDEVICES];



/*
 * Declaration of local functions.
 **********************************/

// Get a pointer to the capabilities of chip.
const CS42XX_SUPP_CHIPDESCR * CS42XX_Supp_GetChipDescr(CS42XX_DEVID);

// Find a chip Id in the list of supported chips.
const CS42XX_SUPP_CHIPTYPE * CS42XX_Supp_GetChipType(CYG_BYTE);



/*
 * Definition of functions.
 ***************************/



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                      MANAGEMENT OF CAPABILITIES                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Get the address of the last register of the device memory.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @return  Address of the last register
 ****************************************************************************/
CS42XX_REG
CS42XX_Supp_GetLastReg(CS42XX_DEVID devid)
{
	CS42XX_REG lastreg = CS42XX_REG_LAST;  // Default value
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		lastreg = descr->lastreg;
	}
	// else description is not found.

	return lastreg;
}


/** Get if this device implements an ADC.
 *
 * @param[in] devid  Symbolic device Id
 *
 * @retval true   Device implements an ADC
 * @retval false  Device has no ADC
 ****************************************************************************/
cyg_bool
CS42XX_Supp_HasADC(CS42XX_DEVID devid)
{
	cyg_bool hasadc = false;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		hasadc = (descr->capabilities & CS42XX_SUPP_FLAG_ADC);
	}
	// else description is not found.

	return hasadc;
}


/** Get if this device implements a DAC.
 *
 * @param[in] devid  Symbolic device Id
 *
 * @retval true   Device implements a DAC
 * @retval false  Device has no DAC
 ****************************************************************************/
cyg_bool
CS42XX_Supp_HasDAC(CS42XX_DEVID devid)
{
	cyg_bool hasdac = false;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		hasdac = (descr->capabilities & CS42XX_SUPP_FLAG_DAC);
	}
	// else description is not found.

	return hasdac;
}


/** Get if this device can handle two asynchronous master clocks for I/O.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @retval true    Device can be clocked by two asynchronous master clocks
 * @retval false   Device must be driven by a single master clock
 ****************************************************************************/
cyg_bool
CS42XX_Supp_CanClockAsync(CS42XX_DEVID devid)
{
	cyg_bool canclockasync = false;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		canclockasync = (descr->capabilities & CS42XX_SUPP_FLAG_INDEPCLK);
	}
	// else description is not found.

	return canclockasync;
}


/** Get if auxiliary output is present on this device.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @retval true    Device has an auxiliary output
 * @retval false   Device has no auxiliary output
 ****************************************************************************/
cyg_bool
CS42XX_Supp_HasAuxOutput(CS42XX_DEVID devid)
{
	cyg_bool hasauxoutput = false;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		hasauxoutput = (descr->capabilities & CS42XX_SUPP_FLAG_AUXOUT);
	}
	// else description is not found.

	return hasauxoutput;
}


/** Get if an interrupt output pin is present on this device.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @retval true    Device can drive an interrupt output pin
 * @retval false   Device has no interrupt output pin
 ****************************************************************************/
cyg_bool
CS42XX_Supp_HasIntrOutput(CS42XX_DEVID devid)
{
	cyg_bool hasintroutput = false;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		hasintroutput = (descr->capabilities & CS42XX_SUPP_FLAG_INTR);
	}
	// else description is not found.

	return hasintroutput;
}


/** Get the number of analog inputs on device.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @return     Number of analog inputs on device
 * @retval 0   No description found, or no analog input
 ****************************************************************************/
cyg_uint8
CS42XX_Supp_GetNbAnalogInputs(CS42XX_DEVID devid)
{
	cyg_uint8 nb = 0;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		nb = descr->nbainputs;
	}
	// else description is not found.

	return nb;
}


/** Get the number of digital inputs on device.
 *
 * @param[in] devid   Symbolic device Id
 *
 * @return     Number of digital inputs on device
 * @retval 0   No description found, or no digital input
 ****************************************************************************/
cyg_uint8
CS42XX_Supp_GetNbDigitalInputs(CS42XX_DEVID devid)
{
	cyg_uint8 nb = 0;
	const CS42XX_SUPP_CHIPDESCR *descr;

	descr = CS42XX_Supp_GetChipDescr(devid);
	if (NULL != descr)
	{
		nb = descr->nbdinputs;
	}
	// else description is not found.

	return nb;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                       MANAGEMENT OF ALL DEVICES                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



/** Find a chip Id in the list of supported chips, and return a pointer to
 *  the specificities of this chip.
 *
 * @param[in] chipid   Chip Id (got from the 1st register of device)
 *
 * @return        Pointer to the description of chip
 * @retval NULL   Unknown chip Id
 ****************************************************************************/
const CS42XX_SUPP_CHIPTYPE *
CS42XX_Supp_GetChipType(CYG_BYTE chipid)
{
	const CS42XX_SUPP_CHIPTYPE *pchip = NULL;
	int i;

	for (i = 0; (NULL == pchip) && (i < CS42XX_SUPP_NBCHIPTYPES); i++)
	{
		if (chipid == cCS42XXSuppChipType[i].id)
		{
			pchip = &cCS42XXSuppChipType[i];
		}
		// else this is not the chip Id we are looking for.
	}

	return pchip;
}


/** Get the name of a chip, i.e. codec type and revision.
 *
 * @param[in]  devid   Symbolic device Id
 * @param[out] buf     Buffer to store chip name
 * @param[in]  buflen  Length of buffer
 *
 * @retval  ENOERR      Success, buffer filled in
 * @retval -EINVAL      Invalid parameters
 * @retval -EDEVNOSUPP  Device not supported
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Supp_GetChipName(CS42XX_DEVID devid, char *buf, cyg_count8 buflen)
{
	Cyg_ErrNo err = -EINVAL;
	const char *str = NULL;

	if ((devid < CS42XX_NBDEVICES) && (NULL != buf))
	{
		if (NULL != CS42XXDeviceInfo[devid].chiptype)
		{
			str = CS42XXDeviceInfo[devid].chiptype->name;
		}
		// else 'str' stays == NULL.

		if (NULL == str)
		{
			err = -EDEVNOSUPP;
		}
		else
		{
			strncpy(buf, str, buflen);
			err = ENOERR;
		}
	}
	// else invalid parameters.

	return err;
}


/** Get a pointer to the capabilities of chip.
 *
 * @param devid  Symbolic device Id
 *
 * @return       Pointer to the description of capabilities
 * @retval NULL  Description not found
 ****************************************************************************/
const CS42XX_SUPP_CHIPDESCR *
CS42XX_Supp_GetChipDescr(CS42XX_DEVID devid)
{
	const CS42XX_SUPP_CHIPTYPE  *chiptype;
	const CS42XX_SUPP_CHIPDESCR *descr = NULL;

	if (devid < CS42XX_NBDEVICES)
	{
		chiptype = CS42XXDeviceInfo[devid].chiptype;
		if (NULL != chiptype)
		{
			descr = chiptype->descr;
		}
		// else chip type is not filled.
	}
	// else device Id is incorrect.

	return descr;
}


/** Initialize driver according to the specificities of this device.
 * The chip Id is read from the first register of device, then according
 * to it, capabilities are looked into a short database.
 *
 * @retval  ENOERR   Success
 * @retval -ENODEV   No such device
 * @retval -EIO      Error while reading from device memory
 ****************************************************************************/
Cyg_ErrNo
CS42XX_Supp_Init(CS42XX_DEVID devid)
{
	Cyg_ErrNo err;
	CYG_BYTE  chipid;

	if (devid >= CS42XX_NBDEVICES)
	{
		// Invalid device
		err = -ENODEV;
	}
	else
	{
		// Get the part number Id and revision of the connected codec,
		// and find the description in database.
		chipid = CS42XX_GetChipId(devid, &err);
		if (ENOERR == err)
		{
			CS42XXDeviceInfo[devid].chiptype = CS42XX_Supp_GetChipType(chipid);
		}
		else
		{
			CS42XXDeviceInfo[devid].chiptype = NULL;
		}
		CS42XXDeviceInfo[devid].devid  = devid; // FIXME
		CS42XXDeviceInfo[devid].handle = NULL;  // FIXME
	}

	return err;
}

