//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_testi2c.c
 * @brief  Tests applicable on CS42XX driver - I�C part.
 *
 * @author David Tylski
 * @date   Mar. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */


#include <cyg/devs/cs42xx_api.h>

#include <cyg/devs/test/cs42xx_test.h>
#include <cyg/devs/test/cs42xx_testhmi.h>
#include <cyg/devs/test/cs42xx_testi2c.h>


// I�C testing software is conditioned by CS42XX_TEST_SUPPORT_I2C.
#ifdef CS42XX_TEST_SUPPORT_I2C



/*
 * Defines.
 ***********/

#define CS42XX_TESTI2C_THRPRIORITY			4
#define CS42XX_TESTI2C_STACKSIZE			4096


/*
 * Types.
 *********/

/** Structure for the storage of data related to I�C testing.
 * @typedef CS42XX_TESTI2C_DATA
 */
typedef struct CS42XX_TESTI2C_DATA_TAG CS42XX_TESTI2C_DATA;
struct CS42XX_TESTI2C_DATA_TAG
{
	CS42XX_TEST_ID testcurr;  ///< Id of test currently running.
	CS42XX_TEST_ID testreq;   ///< Id of future test.
	cyg_mutex_t    testmutex; ///< Mutex for testcurr and testreq.

	cyg_thread     threadobj;                       ///< Thread object.
	cyg_handle_t   thread;                          ///< Handle of thread.
	char           stack[CS42XX_TESTI2C_STACKSIZE]; ///< Stack.

	cyg_mbox       mailboxobj; ///< Mailbox object.
	cyg_handle_t   mailbox;    ///< Handle of mailbox.
};



/*
 * Variables.
 *************/

/** Data related to I�C testing. */
static CS42XX_TESTI2C_DATA CS42XXTestI2CData;



/*
 * Definition of functions.
 ***************************/



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                I�C TESTS                                *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** I�C hardware test: read all registers from device.
 *
 * This function reads all the device memory (whose length is determined
 * using the chip Id) using two different methods:
 * - Reading one register after the other,
 * - Reading all registers using the same I�C frame (MAP auto-increment
 *   functionality of audio codec).
 *
 * If all is successful, nothing is posted.
 * When an error is encountered, its value is posted to stderr and local
 * test process is stopped.
 *
 * @retval CYGNUM_TEST_PASS   Test passed (*perrno == ENOERR)
 * @retval CYGNUM_TEST_FAIL   Test failed (error occurred, see *perrno)
 ****************************************************************************/
static Cyg_test_code
CS42XX_Test_I2C_Read(void)
{
	Cyg_test_code result = CYGNUM_TEST_NA;
	Cyg_ErrNo     err = -EINVAL;
	CS42XX_REG    reg, chipid;
	CYG_BYTE      regval[CS42XX_REG_NBMAX];
	CS42XX_DEVID  devid;

	devid = CS42XX_Test_GetDeviceId();

	// Retrieve the chip Id into register cache (to determine the length
	// of device memory, if not already done)
	chipid = CS42XX_PEEK_(devid, CS42XX_REG_CHIPID, &err);

	if (ENOERR != err)
	{
		CYG_TEST_FAIL("I�C Read: chip Id");
		CS42XX_TEST_PERROR(err);
	}
	else
	{
		// Read all registers in order, one by one.
		for(reg = CS42XX_REG_FIRST;
			((ENOERR == err) && (reg < CS42XX_REG_LASTOF(devid)));
			reg++)
		{
			err = CS42XX_I2C_Read(devid, reg, &regval[reg-1], 1);
		}

		if (ENOERR != err)
		{
			CYG_TEST_FAIL("I�C Read: 1 by 1");
			diag_printf("\r\nReg %s", reg);
			CS42XX_TEST_PERROR(err);
		}
		else
		{
			// Read all registers in once.
			err = CS42XX_I2C_Read(devid, CS42XX_REG_FIRST, &regval[0],
					CS42XX_REG_NBOF(devid));
			if (ENOERR != err)
			{
				CYG_TEST_FAIL("I�C Read: all in once");
				CS42XX_TEST_PERROR(err);
			}
		}
	}

	// Score the test
	result = ((ENOERR == err)? CYGNUM_TEST_PASS: CYGNUM_TEST_FAIL);

	// Run the test only once.
	CS42XX_TestI2C_Run(CS42XX_TEST_I2C_NONE);

	return result;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                        MANAGEMENT OF I�C TESTS                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Thread testing I�C communication with device(s).
 *
 * According to its variable configuration, this thread starts and stops
 * tests specific to I�C management.
 *
 * @param data  (Not used)
 ****************************************************************************/
static void
CS42XX_TestI2C_Thread(CYG_ADDRWORD data)
{
	Cyg_test_code testresult;

	while(1)
	{
		switch(CS42XXTestI2CData.testcurr)
		{

		case CS42XX_TEST_I2C_READ:
			testresult = CS42XX_Test_I2C_Read();
			break;

		//case CS42XX_TEST_NONE:
		default:
			// No test: block until a new test is asked
			{
				void *ptr;
				ptr = cyg_mbox_get(CS42XXTestI2CData.mailbox);

				cyg_mutex_lock(&CS42XXTestI2CData.testmutex);
				CS42XXTestI2CData.testcurr = *(CS42XX_TEST_ID *)ptr;
				cyg_mutex_unlock(&CS42XXTestI2CData.testmutex);
			}
			testresult = CYGNUM_TEST_NA;
			break;
		}

		// Post test results
		if (CYGNUM_TEST_NA != testresult)
		{
			CS42XX_TestHMI_PostResult(testresult);
		}
	}
	// Loop infinitely.
}


/** Look for pending messages in mailbox; if any, take the first message
 *  and store it as the new test Id.
 *
 * This function is not blocking.
 *
 * @retval true   A message has been found and stored as new test Id
 * @retval false  No message pending
 ****************************************************************************/
static cyg_bool
CS42XX_TestI2C_TestMailbox(void)
{
	cyg_bool newmsg = false;
	void *ptr;

	ptr = cyg_mbox_tryget(CS42XXTestI2CData.mailbox);
	if (NULL != ptr)
	{
		// New message available
		newmsg = true;

		// Store it as a new test Id
		cyg_mutex_lock(&CS42XXTestI2CData.testmutex);
		CS42XXTestI2CData.testcurr = *(CS42XX_TEST_ID *)ptr;
		cyg_mutex_unlock(&CS42XXTestI2CData.testmutex);
	}

	return newmsg;
}


/** Select the I�C test to be run.
 *
 * @param[in] testid  I�C test to be run
 ****************************************************************************/
 void
CS42XX_TestI2C_Run(CS42XX_TEST_ID testid)
{
	// Post a new command to the thread
	CS42XXTestI2CData.testreq = testid;
	cyg_mbox_put(CS42XXTestI2CData.mailbox, &CS42XXTestI2CData.testreq);
}


/** Configure the I�C part of CS42XX driver testing software.
 *
 * @retval true   Success
 * @retval false  Failure during initialization
 ****************************************************************************/
 cyg_bool
CS42XX_TestI2C_Init(void)
{
	cyg_bool rv = true;

	// Initialize global variables
	CS42XXTestI2CData.testcurr = CS42XX_TEST_NONE;
	CS42XXTestI2CData.testreq  = CS42XX_TEST_NONE;
	cyg_mutex_init(&CS42XXTestI2CData.testmutex);

	// Create a mailbox to command I�C testing thread
	cyg_mbox_create(&CS42XXTestI2CData.mailbox,
					&CS42XXTestI2CData.mailboxobj);

	// Start the I�C testing thread
	cyg_thread_create(CS42XX_TESTI2C_THRPRIORITY,
			CS42XX_TestI2C_Thread, 0, "CS42XX I�C testing",
			&CS42XXTestI2CData.stack[0], CS42XX_TESTI2C_STACKSIZE,
			&CS42XXTestI2CData.thread, &CS42XXTestI2CData.threadobj);
	cyg_thread_resume(CS42XXTestI2CData.thread);

	return rv;
}



#endif // CS42XX_TEST_SUPPORT
