/**
 ***************************************************************************
 * @file tda7540.h
 *
 * @brief TDA7540 AM/FM Car Radio Tuner IC.
 *
 * Copyright (C) 2007 Parrot S.A.
 *
 * @author	Yoan Besson
 * @date	03/2007
 ***************************************************************************
 */

#ifndef _TDA7540_
#define _TDA7540_

#include <cyg/io/io.h>
#include <cyg/io/i2c.h>

#define TDA7540_AUTO_INC		0x40
#define TDA7540_BUFFER_SIZE		40
#define TDA7540_TV2_BUFFER_SIZE	10

#define TDA7540_USE_EEPROM

/** Set a multibit value in integer.
 * E.g. TDA7540_SETVAL(x,0x06,0x07) writes ".....110" in x.
 ****************************************************************************/
#define TDA7540_SETVAL(reg,val,valmask) ( (reg) = ((reg) & ~(valmask)) | (val) )
/** Get a multibit value in integer.
 * E.g. TDA7540_GETVAL(x,y,0x07) return ".....110" in y.
 ****************************************************************************/
#define TDA7540_GETVAL(reg,val,valmask) ( (val) = ((reg) & valmask) )

typedef enum TDA7540_UPDATE
{
	TDA7540_NO_UPDATE,	///< .
	TDA7540_WRITE_REG,	///< .
} TDA7540_UPDATE;

typedef enum TDA7540_ADDRESSES
{
	TDA7540_CHARGE_PUMP,		///< [0]Pll charge pump high and low current.
	TDA7540_PLL_PC_LOW,			///< [1]Pll program counter low  byte.
	TDA7540_PLL_PC_HIGH,		///< [2]Pll program counter high byte.
	TDA7540_TV1,				///< [3]Tuning Voltage 1.
	TDA7540_TV2,				///< [4]Tuning Voltage 2.
	TDA7540_IFC_CONTROL1,		///< [5]Lock detect;IF counter status;Mute&AGC;Tsample.
	TDA7540_IFC_CONTROL2,		///< [6]Error window; center freqency.
	TDA7540_AM_CONTROL,			///< [7]AM seek stop constant; multipath pin; AM prescaler; weather band TV offset; TV tracking in pll; local mode.
	TDA7540_QUALITY_ISS,		///< [8]ISS filter;ISS filter bandwidth;weather band ISS filter charge/discharge current.
	TDA7540_QUALITY_AC,			///< [9]Adjacent Channel high pass; AC gain; AC narrow band; AC wide band.
	TDA7540_QUALITY_MP,			///< [10]Multipath thereshold;Multipath gain;Multipath bandpass freq.
	TDA7540_QUALITY_DEV,		///< [11]ISS fixed bandwidth 80K; dev ratio; dev theres bandwidth dev charge current xxuA, xuA.
	TDA7540_SOFT_MUTE,			///< [12]Mute enable; Mute depth; Mute start point.
	TDA7540_AC_MUTE,			///< [13]Adjacent Channel mute filter bandwidth; AC mute theres; AC mute depth.
	TDA7540_VCO_PLLREF,			///< [14]PLL for FM/AM; PLL reference freq; VCO 0 degree/180 degree VCO divider.
	TDA7540_FM_AGC,				///< [15]KAGC thereshold; IF AGC thereshold;RF thereshold.
	TDA7540_AM_AGC,				///< [16]DAGC thereshold; WAGC thereshold.
	TDA7540_FM_DEMO,			///< [17]spike cancelation; demo adjust.
	TDA7540_SMETER,				///< [18]Smeter offset; slope; overdeviation current.
	TDA7540_IF_GAIN_XTAL,		///< [19]IF Gain; Xtal adjust; Xtal limitation.
	TDA7540_TANK_ADJ,			///< [20]IFT1; IFT2.
	TDA7540_IQ_MIXER,			///< [21]IQ mixer; softmute refer offset.
	TDA7540_AM_IF_NB,			///< [22]Int xus; detection; thereshold.
	TDA7540_SSTOP,				///< [23]SStop.
	TDA7540_STD1_AM_NOISE,		///< [24]Stereo VCO adjust; Audio noise blanker; audio AM noise blanker.
	TDA7540_STD2_IN_GAIN,		///< [25]In gain; de-emphasis; noise blanker peak charge; mono-stereo.
	TDA7540_STD3_NB,			///< [26]Deviation adjust; audio noise blanker.
	TDA7540_STD4_HICUT,			///< [27]hiCut depth;VHCH; VHCL.
	TDA7540_STD5_BLEND,			///< [28]Quality detect; audio noise field strength; stereo blend.
	TDA7540_STD6_MP,			///< [29]Multipath detector gain; charge current; band pass gain.
	TDA7540_STD7_ROLL_OFF,		///< [30]Level gain; roll off.
	TDA7540_STD8_AFS,			///< [31]AFS control ablility, AM audio noise blanker setting.
	TDA7540_STD9_AM_CORNER_FREQ,///< [32]Deemphasis, AM corner freq.
	TDA7540_TEST_TU1,			///< [33]Testing 1.
	TDA7540_TEST_TU2,			///< [34]Testing 2.
	TDA7540_TEST_TU3,			///< [35]Testing 3.
	TDA7540_TEST_TU4,			///< [36]Testing 4.
	TDA7540_TEST_TU5,			///< [37]Testing 5.
	TDA7540_TEST_STD,			///< [38]Testing 6.
	TDA7540_STD_SPIKE			///< [39]Spike canceler.
} TDA7540_ADDRESSES;

typedef enum TDA7540_TV2_ADDRESSES
{
	TDA7540_TV2_EUR_LOW,		///< [0]Tuning voltage Low		(Eur).
	TDA7540_TV2_EUR_MID,		///< [1]Tuning voltage Middle	(Eur).
	TDA7540_TV2_EUR_HIGH,		///< [2]Tuning voltage High	(Eur).
	TDA7540_TV2_OIRT_JP_LOW,	///< [3]Tuning voltage Low		(OIRT/Jap).
	TDA7540_TV2_OIRT_JP_MID,	///< [4]Tuning voltage Middle	(OIRT/Jap).
	TDA7540_TV2_OIRT_JP_HIGH,	///< [5]Tuning voltage High	(OIRT/Jap).
	TDA7540_TV2_EUR_PT1,		///< [6]Tuning voltage (Eur).
	TDA7540_TV2_EUR_PT2,		///< [7]Tuning voltage (Eur).
	TDA7540_TV2_EUR_PT3,		///< [8]Tuning voltage (Eur).
	TDA7540_TV2_EUR_PT4,		///< [9]Tuning voltage (Eur).
} TDA7540_TV2_ADDRESSES;

typedef enum TDA7540_MASKS
{
	TDA7540_MASK_CPUMP_FMSEEK		= 0x80,	///< [0]FMSEEK state.
	TDA7540_MASK_CPUMP_CURMODE		= 0x40, ///< [0]Selected current (high or low).
	TDA7540_MASK_CPUMP_LO_CUR		= 0x30, ///< [0]Pll charge pump low current.
	TDA7540_MASK_CPUMP_HI_CUR		= 0x0F, ///< [0]Pll charge pump high current.
									
	TDA7540_MASK_PLLCOUNTER_LO		= 0xFF, ///< [1]Pll program counter low  byte.
	TDA7540_MASK_PLLCOUNTER_HI		= 0xFF, ///< [2]Pll program counter high byte.
									
	TDA7540_MASK_TV1				= 0xFF,	///< [3]Tuning Voltage 1.
	TDA7540_MASK_TV2				= 0xFF,	///< [4]Tuning Voltage 2.
									
	TDA7540_MASK_IFCC_LOCKSTATE		= 0x80,	///< [5]Lock detect.
	TDA7540_MASK_IFCC_MUTEAGC		= 0x40,	///< [5]Mute&AGC.
	TDA7540_MASK_IFCC_IFCMODE		= 0x30,	///< [5]IF counter mode.
	TDA7540_MASK_IFCC_IFCSTATE		= 0x08,	///< [5]IF counter status.
	TDA7540_MASK_IFCC_TSAMPLE		= 0x07,	///< [5]Tsample.
									
	TDA7540_MASK_IFCC_FCENTER		= 0x1F,	///< [6]Center frequency.
	TDA7540_MASK_IFCC_EW			= 0xE0,	///< [6]Error window.
									
	TDA7540_MASK_AMCTRL_LOCAL_MODE	= 0x80, ///< [7]Local mode.
	TDA7540_MASK_AMCTRL_TVPLL_MODE	= 0x40,	///< [7]TV1/2 PLL mode.
	TDA7540_MASK_AMCTRL_ADD_OFFSTV	= 0x30,	///< [7]Additional TV1/2 offset.
	TDA7540_MASK_AMCTRL_PRESCALER	= 0x0C, ///< [7]Prescaler ratio.
	TDA7540_MASK_AMCTRL_PIN24		= 0x02,	///< [7]PIN24 usage.
	TDA7540_MASK_AMCTRL_AGCTC		= 0x01,	///< [7]AGC time constant.

	TDA7540_MASK_QUAL_ISSF_CURRENT	= 0xE0, ///< [8]Discharge current, charge current mid and narrow.
	TDA7540_MASK_QUAL_ISSF_30KHZ	= 0x08,	///< [8]Switch ISS filter freq. for weatherband.
	TDA7540_MASK_QUAL_ISSF_FREQ		= 0x04,	///< [8]Switch ISS filter frequency.
	TDA7540_MASK_QUAL_ISSF_SWITCH	= 0x02, ///< [8]Switch ISS filter.
	TDA7540_MASK_QUAL_ISSF_CTRL		= 0x01,	///< [8]ISS filter control state.

	TDA7540_MASK_PLLBAND			= 0xC0,	///< [14]PLL for FM/AM.
	TDA7540_MASK_PLLREF				= 0x38,	///< [14]PLL reference freq.
									
	TDA7540_MASK_SMETER_OD			= 0x80,	///< [18]S-METER overdeviation correction status (ON/OFF).
	TDA7540_MASK_SMETER_OD_CMAX		= 0x40,	///< [18]S-METER overd. correct current max.
	TDA7540_MASK_SMETER_OD_SLOP		= 0x20,	///< [18]S-METER slope mode (1V/decade, 1.5V/decade).
	TDA7540_MASK_SMETER_OFSG		= 0x10,	///< [18]S-METER offset sign.
	TDA7540_MASK_SMETER_OFSL		= 0x0F,	///< [18]S-METER offset slide.
									
	TDA7540_MASK_SWCTRL_SSTOP		= 0xF0,	///< [18]SSTOP reference (IFC, H&FSU>x.xV).
	TDA7540_MASK_SWCTRL_OUTPUT		= 0x02,	///< [18]S1 output state (emitter/open collector).
	TDA7540_MASK_SWCTRL_OUT_LVL		= 0x01,	///< [18]S1 output voltage level (LOW/HIGH).
									
	TDA7540_MASK_STD_INGAIN			= 0xC0,	///< [25]In gain.
	TDA7540_MASK_STD_DEEMPH			= 0x20,	///< [25]De-emphasis.
	TDA7540_MASK_STD_PTHRESHOLD		= 0x10,	///< [25]Pilot threshold high or low.
	TDA7540_MASK_STD_NB_PEAKC		= 0x08,	///< [25]Noise blanker peak charge.
	TDA7540_MASK_STD_MONOSTEREO		= 0x04,	///< [25]Mono/stereo.
	TDA7540_MASK_STD_MSTATE			= 0x02,	///< [25]State (ON/OFF) when muted.
	TDA7540_MASK_STD_MUTE			= 0x01,	///< [25](Un)mute stereo decoder.
} TDA7540_MASKS;

typedef enum TDA7540_BAND
{
	TDA7540_BAND_UNDEFINED,
	TDA7540_BAND_FM_EUROPE,
	TDA7540_BAND_FM_AUSTRALIA,
	TDA7540_BAND_FM_MIDLEEAST,
	TDA7540_BAND_FM_JAPAN,
	TDA7540_BAND_FM_AMERICA1,
	TDA7540_BAND_FM_AMERICA2,
	TDA7540_BAND_FM_AMERICA3,
	TDA7540_BAND_FM_AMERICA4,
	TDA7540_BAND_MW_EUROPE,
	TDA7540_BAND_MW_AUSTRALIA,
	TDA7540_BAND_MW_MIDLEEAST,
	TDA7540_BAND_MW_JAPAN,
	TDA7540_BAND_MW_AMERICA1,
	TDA7540_BAND_MW_AMERICA2,
	TDA7540_BAND_MW_AMERICA3,
	TDA7540_BAND_MW_AMERICA4,
	TDA7540_BAND_LW_EUROPE,
	TDA7540_BAND_FM_OIRT,
	TDA7540_BAND_SW1,
	TDA7540_BAND_SW2,
	TDA7540_BAND_SW3,
	TDA7540_BAND_SW4,
} TDA7540_BAND;

typedef enum TDA7540_FM_BAND
{
	// FM1 [87.5MHz - 108MHz]
	TDA7540_FM_92_0MHZ	= 92000,	///< For Charge Pump.
	TDA7540_FM_98_6MHZ	= 98600,	///< For Charge Pump.
	TDA7540_FM_106_9MHZ	= 106900,	///< For Charge Pump.
	TDA7540_FM_108_0MHZ	= 108000,	///< For Charge Pump and Multi-Point.
	TDA7540_FM_89_5MHZ	= 89500,	///< For Multi-Point.
	TDA7540_FM_92_7MHZ	= 92700,	///< For Multi-Point.
	TDA7540_FM_96_1MHZ	= 96100,	///< For Multi-Point.
	TDA7540_FM_99_6MHZ	= 99600,	///< For Multi-Point.
	TDA7540_FM_103_1MHZ	= 103100,	///< For Multi-Point.
	TDA7540_FM_106_5MHZ	= 106500,	///< For Multi-Point.
	// JP [76MHz - 90MHz]
	TDA7540_FM_80_0MHZ	= 80000,	///< For Charge Pump.
	TDA7540_FM_83_0MHZ	= 83000,	///< For Charge Pump.
	TDA7540_FM_87_0MHZ	= 87000,	///< For Charge Pump.
	TDA7540_FM_90_0MHZ	= 90000,	///< For Charge Pump and Multi-Point.
	TDA7540_FM_81_0MHZ	= 81000,	///< For Multi-Point.
	TDA7540_FM_86_0MHZ	= 86000,	///< For Multi-Point.
	// OIRT [66MHz - 74MHz]
	TDA7540_FM_74_0MHZ	= 74000,	///< For Charge Pump and Multi-Point.
	TDA7540_FM_67_0MHZ	= 67000,	///< For Multi-Point.
	TDA7540_FM_71_0MHZ	= 71000,	///< For Multi-Point.
	// MW1/LW [522KHz - 1610KHz]
	TDA7540_AM_720KHZ		= 720,		///< For Charge Pump.
	TDA7540_AM_999KHZ		= 999,		///< For Charge Pump.
	TDA7540_AM_1404KHZ	= 1404,		///< For Charge Pump.
	TDA7540_FM_1610KHZ	= 1610,		///< For Charge Pump.
	// SW1-4 [5.0MHz - 15.0MHz]
	TDA7540_FM_15MHZ		= 15000,	///< For Charge Pump.
} TDA7540_FM_BAND;

//////////////////////////////////////////////////////////////////////////[0]

typedef enum TDA7540_FM_HICURRENT
{
	TDA7540_FM_HICURRENT_0mA,
	TDA7540_FM_HICURRENT_0_5mA,
	TDA7540_FM_HICURRENT_1mA,
	TDA7540_FM_HICURRENT_1_5mA,
	TDA7540_FM_HICURRENT_2mA,
	TDA7540_FM_HICURRENT_2_5mA,
	TDA7540_FM_HICURRENT_3mA,
	TDA7540_FM_HICURRENT_3_5mA,
	TDA7540_FM_HICURRENT_4mA,
	TDA7540_FM_HICURRENT_4_5mA,
	TDA7540_FM_HICURRENT_5mA,
	TDA7540_FM_HICURRENT_5_5mA,
	TDA7540_FM_HICURRENT_6mA,
	TDA7540_FM_HICURRENT_6_5mA,
	TDA7540_FM_HICURRENT_7mA,
	TDA7540_FM_HICURRENT_7_5mA,
} TDA7540_FM_HICURRENT;

typedef enum TDA7540_LOWCURRENT
{
	TDA7540_LOWCURRENT_0uA		= 0x00,
	TDA7540_LOWCURRENT_50uA		= 0x10,
	TDA7540_LOWCURRENT_100uA	= 0x20,
	TDA7540_LOWCURRENT_150uA	= 0x30,
} TDA7540_LOWCURRENT;

typedef enum TDA7540_CURRENT_MODE
{
	TDA7540_LOWCURRENT_MODE		= 0x00,
	TDA7540_HIGHCURRENT_MODE	= 0x40,
} TDA7540_CURRENT_MODE;

typedef enum TDA7540_FM_SEEK_STATE
{
	TDA7540_FM_SEEK_OFF	= 0x00,
	TDA7540_FM_SEEK_ON	= 0x80,
} TDA7540_FM_SEEK_STATE;

//////////////////////////////////////////////////////////////////////////[1][2]

//////////////////////////////////////////////////////////////////////////[3]

//////////////////////////////////////////////////////////////////////////[4]

//typedef enum TDA7540_MULTI_POINT
//{
//	TDA7540_MP_UNDEFINED = 0,
//	TDA7540_MP_875_895,
//	TDA7540_MP_896_927,
//	TDA7540_MP_928_961,
//	TDA7540_MP_760_809 = 4,
//	TDA7540_MP_810_859,
//	TDA7540_MP_860_900,
//	TDA7540_MP_650_669 = 4,
//	TDA7540_MP_670_709,
//	TDA7540_MP_710_740,
//	TDA7540_MP_962_996,
//	TDA7540_MP_997_1031,
//	TDA7540_MP_1032_1065,
//	TDA7540_MP_1065_108,
//} TDA7540_MULTI_POINT;

//////////////////////////////////////////////////////////////////////////[5]

typedef enum TDA7540_LOCKSTATE
{
	TDA7540_LOCK_DISABLE	= 0x00,
	TDA7540_LOCK_ENABLE		= 0x80,
} TDA7540_LOCKSTATE;

typedef enum TDA7540_MUTEAGC
{
	TDA7540_MUTEAGC_DISABLE	= 0x00,
	TDA7540_MUTEAGC_ENABLE	= 0x40,
} TDA7540_MUTEAGC;

typedef enum TDA7540_IFCMODE
{
	TDA7540_IFCMODE_FM			= 0x10,
	TDA7540_IFCMODE_AM_450KHZ	= 0x20,
	TDA7540_IFCMODE_AM_10_7MHZ	= 0x30,
} TDA7540_IFCMODE;

typedef enum TDA7540_IFCSTATE
{
	TDA7540_IFC_DISABLE	= 0x00,
	TDA7540_IFC_ENABLE	= 0x08,
} TDA7540_IFCSTATE;

typedef enum TDA7540_TSAMPLE
{
	TDA7540_TSAMPLE_FM20_48_AM128	= 0x00,
	TDA7540_TSAMPLE_FM10_24_AM64	= 0x01,
	TDA7540_TSAMPLE_FM5_12_AM32		= 0x02,
	TDA7540_TSAMPLE_FM2_56_AM16		= 0x03,
	TDA7540_TSAMPLE_FM1_28_AM8		= 0x04,
	TDA7540_TSAMPLE_FM0_64_AM4		= 0x05,
	TDA7540_TSAMPLE_FM0_32_AM2		= 0x06,
	TDA7540_TSAMPLE_FM0_16_AM1		= 0x07,
} TDA7540_TSAMPLE;

//////////////////////////////////////////////////////////////////////////[6]

//center freqency and delta window
typedef enum TDA7540_IFC_FCENTER
{
	TDA7540_IFC_CENTER_449KHZ		= 0x00,
	TDA7540_IFC_CENTER_10_60625MHZ	= 0x00,
	TDA7540_IFC_CENTER_450KHZ		= 0x01,
	TDA7540_IFC_CENTER_10_61250MHZ	= 0x01,
	TDA7540_IFC_CENTER_451KHZ		= 0x02,
	TDA7540_IFC_CENTER_10_61875MHZ	= 0x02,
	TDA7540_IFC_CENTER_452KHZ		= 0x03,
	TDA7540_IFC_CENTER_10_62500MHZ	= 0x03,
	TDA7540_IFC_CENTER_453KHZ		= 0x04,
	TDA7540_IFC_CENTER_10_63125MHZ	= 0x04,
	TDA7540_IFC_CENTER_454KHZ		= 0x05,
	TDA7540_IFC_CENTER_10_63750MHZ	= 0x05,
	TDA7540_IFC_CENTER_455KHZ		= 0x06,
	TDA7540_IFC_CENTER_10_64375MHZ	= 0x06,
	TDA7540_IFC_CENTER_456KHZ		= 0x07,
	TDA7540_IFC_CENTER_10_65000MHZ	= 0x07,
	TDA7540_IFC_CENTER_457KHZ		= 0x08,
	TDA7540_IFC_CENTER_10_65625MHZ	= 0x08,
	TDA7540_IFC_CENTER_458KHZ		= 0x09,
	TDA7540_IFC_CENTER_10_66250MHZ	= 0x09,
	TDA7540_IFC_CENTER_459KHZ		= 0x0A,
	TDA7540_IFC_CENTER_10_66875MHZ	= 0x0A,
	TDA7540_IFC_CENTER_460KHZ		= 0x0B,
	TDA7540_IFC_CENTER_10_67500MHZ	= 0x0B,
	TDA7540_IFC_CENTER_461KHZ		= 0x0C,
	TDA7540_IFC_CENTER_10_68125MHZ	= 0x0C,
	TDA7540_IFC_CENTER_462KHZ		= 0x0D,
	TDA7540_IFC_CENTER_10_68750MHZ	= 0x0D,
	TDA7540_IFC_CENTER_463KHZ		= 0x0E,
	TDA7540_IFC_CENTER_10_69375MHZ	= 0x0E,
	TDA7540_IFC_CENTER_464KHZ		= 0x0F,
	TDA7540_IFC_CENTER_10_7MHZ		= 0x0F,
	// TO BE COMPLETED...
} TDA7540_IFC_FCENTER;

typedef enum TDA7540_IFC_EW
{
	TDA7540_IFC_EW_FM6_25_AM1K	= 0x60,
	TDA7540_IFC_EW_FM12_5_AM2K	= 0x80,
	TDA7540_IFC_EW_FM25_AM4K	= 0xA0,
	TDA7540_IFC_EW_FM50_AM8K	= 0xC0,
	TDA7540_IFC_EW_FM100_AM16K	= 0xE0,
} TDA7540_IFC_EW;

//////////////////////////////////////////////////////////////////////////[7]
typedef enum TDA7540_AMCTRL_AGCTC
{
	TDA7540_AMCTRL_AGCTC_NORMAL		= 0x00,
	TDA7540_AMCTRL_AGCTC_AMSEEKSTOP	= 0x01,
} TDA7540_AMCTRL_AGCTC;

typedef enum TDA7540_AMCTRL_PIN24
{
	TDA7540_AMCTRL_PIN24_FM_MULTIPATH	= 0x00,
	TDA7540_AMCTRL_PIN24_AM_STEREO		= 0x02,
} TDA7540_AMCTRL_PIN24;

typedef enum TDA7540_AMCTRL_PRESCALER
{
	TDA7540_AMCTRL_PRESCALER_RATIO10= 0x00,
	TDA7540_AMCTRL_PRESCALER_RATIO8	= 0x04,
	TDA7540_AMCTRL_PRESCALER_RATIO6	= 0x08,
	TDA7540_AMCTRL_PRESCALER_RATIO4	= 0x0C,
} TDA7540_AMCTRL_PRESCALER;

typedef enum TDA7540_AMCTRL_ADD_OFFSTV
{
	TDA7540_AMCTRL_ADD_OFFSTV12_DIS		= 0x00,
	TDA7540_AMCTRL_ADD_OFFSTV1_3_75		= 0x10,
	TDA7540_AMCTRL_ADD_OFFSTV2_3_75		= 0x20,
	TDA7540_AMCTRL_ADD_OFFSTV12_3_75	= 0x30,
} TDA7540_AMCTRL_ADD_OFFSTV;

typedef enum TDA7540_AMCTRL_TV_PLL_MODE
{
	TDA7540_AMCTRL_TV_PLL_MODE_TRACKING	= 0x00,
	TDA7540_AMCTRL_TV_PLL_MODE_INDEP	= 0x40,
} TDA7540_AMCTRL_TV_PLL_MODE;

typedef enum TDA7540_AMCTRL_LOCAL_MODE
{
	TDA7540_AMCTRL_LOCAL_MODE_DISABLE	= 0x00,
	TDA7540_AMCTRL_LOCAL_MODE_ENABLE	= 0x80,
} TDA7540_AMCTRL_LOCAL_MODE;

//////////////////////////////////////////////////////////////////////////[8]
typedef enum TDA7540_QUAL_ISSF_CTRL
{
	TDA7540_QUAL_ISSF_CTRL_ON	= 0x00,
	TDA7540_QUAL_ISSF_CTRL_OFF	= 0x01,
} TDA7540_QUAL_ISSF_CTRL;

typedef enum TDA7540_QUAL_ISSF_SWITCH
{
	TDA7540_QUAL_ISSF_SWITCH_OFF= 0x00,
	TDA7540_QUAL_ISSF_SWITCH_ON	= 0x02,
} TDA7540_QUAL_ISSF_SWITCH;

typedef enum TDA7540_QUAL_ISSF_FREQ
{
	TDA7540_QUAL_ISSF_120KHZ	= 0x00,
	TDA7540_QUAL_ISSF_80KHZ		= 0x04,
} TDA7540_QUAL_ISSF_FREQ;

typedef enum TDA7540_QUAL_ISSF_30KHZ
{
	TDA7540_QUAL_ISSF_30KHZ_OFF	= 0x00,
	TDA7540_QUAL_ISSF_30KHZ_ON	= 0x08,
} TDA7540_QUAL_ISSF_30KHZ;

typedef enum TDA7540_QUAL_ISSF_CURRENT
{
	TDA7540_QUAL_ISSF_1_74_124	= 0x00,
	TDA7540_QUAL_ISSF_3_72_122	= 0x20,
	TDA7540_QUAL_ISSF_5_70_120	= 0x40,
	TDA7540_QUAL_ISSF_7_68_118	= 0x60,
	TDA7540_QUAL_ISSF_9_66_116	= 0x80,
	TDA7540_QUAL_ISSF_11_64_114	= 0xA0,
	TDA7540_QUAL_ISSF_13_62_112	= 0xC0,
	TDA7540_QUAL_ISSF_15_60_110	= 0xE0,
} TDA7540_QUAL_ISSF_CURRENT;

//////////////////////////////////////////////////////////////////////////[9]

//////////////////////////////////////////////////////////////////////////[10]

//////////////////////////////////////////////////////////////////////////[11]

//////////////////////////////////////////////////////////////////////////[12]

//////////////////////////////////////////////////////////////////////////[13]

//////////////////////////////////////////////////////////////////////////[14]

//For reference freqency
typedef enum TDA7540_PLL_REFERENCE
{
	TDA7540_PLL_REFERECE_50KHz	= 0x20,
	TDA7540_PLL_REFERECE_25KHz	= 0x28,
	TDA7540_PLL_REFERECE_10KHz	= 0x30,
	TDA7540_PLL_REFERECE_9KHz	= 0x38,
	TDA7540_PLL_REFERECE_2KHz	= 0x00
} TDA7540_PLL_REFERENCE;

typedef enum TDA7540_PLL_BAND
{
	TDA7540_PLL_BAND_FM	= 0x00,
	TDA7540_PLL_BAND_AM	= 0x40,
	TDA7540_PLL_BAND_AM_LPF	= 0xC0,
} TDA7540_PLL_BAND;

//////////////////////////////////////////////////////////////////////////[15]

//////////////////////////////////////////////////////////////////////////[16]

//////////////////////////////////////////////////////////////////////////[17]

//////////////////////////////////////////////////////////////////////////[18]

// S-METER offset slide.
typedef enum TDA7540_SMETER_OFFSET_SL
{
	TDA7540_PLL_SMETER_OFSL_0dB		= 0x00,
	TDA7540_PLL_SMETER_OFSL_1dB		= 0x01,
	TDA7540_PLL_SMETER_OFSL_2dB		= 0x02,
	TDA7540_PLL_SMETER_OFSL_3dB		= 0x03,
	TDA7540_PLL_SMETER_OFSL_4dB		= 0x04,
	TDA7540_PLL_SMETER_OFSL_5dB		= 0x05,
	TDA7540_PLL_SMETER_OFSL_6dB		= 0x06,
	TDA7540_PLL_SMETER_OFSL_7dB		= 0x07,
	TDA7540_PLL_SMETER_OFSL_8dB		= 0x08,
	TDA7540_PLL_SMETER_OFSL_9dB		= 0x09,
	TDA7540_PLL_SMETER_OFSL_10dB	= 0x10,
	TDA7540_PLL_SMETER_OFSL_11dB	= 0x11,
	TDA7540_PLL_SMETER_OFSL_12dB	= 0x12,
	TDA7540_PLL_SMETER_OFSL_13dB	= 0x13,
	TDA7540_PLL_SMETER_OFSL_14dB	= 0x14,
	TDA7540_PLL_SMETER_OFSL_15dB	= 0x15
} TDA7540_SMETER_OFFSET_SL;

// S-METER offset sign.
typedef enum TDA7540_SMETER_OFFSET_SIGN
{
	TDA7540_PLL_SMETER_OF_SIGN_M	= 0x00,
	TDA7540_PLL_SMETER_OF_SIGN_P	= 0x10,
} TDA7540_SMETER_OFFSET_SIGN;

// S-METER slope mode (1V/decade, 1.5V/decade).
typedef enum TDA7540_SMETER_OVERDEV_SLOPE
{
	TDA7540_PLL_SMETER_OD_SLOPE_1V		= 0x00,
	TDA7540_PLL_SMETER_OD_SLOPE_1_5V	= 0x20,
} TDA7540_SMETER_OVERDEV_SLOPE;

// S-METER overd. correct current max.
typedef enum TDA7540_SMETER_OVERDEV_CMAX
{
	TDA7540_PLL_SMETER_OD_CMAX_45uA	= 0x00,
	TDA7540_PLL_SMETER_OD_CMAX_90uA	= 0x40,
} TDA7540_SMETER_OVERDEV_CMAX;

// S-METER overdeviation correction status (ON/OFF).
typedef enum TDA7540_SMETER_OVERDEV
{
	TDA7540_PLL_SMETER_OVERDEV_OFF	= 0x00,
	TDA7540_PLL_SMETER_OVERDEV_ON	= 0x80,
} TDA7540_SMETER_OVERDEV;

//////////////////////////////////////////////////////////////////////////[19]

//////////////////////////////////////////////////////////////////////////[20]

//////////////////////////////////////////////////////////////////////////[21]

//////////////////////////////////////////////////////////////////////////[22]

//////////////////////////////////////////////////////////////////////////[23]

// Output S1 / S2.
typedef enum TDA7540_SWCTRL_OUTPUT_SEL
{
	TDA7540_SWCTRL_OUTPUT_SEL_S1	= 0x00,
	TDA7540_SWCTRL_OUTPUT_SEL_S2	= 0x01,
} TDA7540_SWCTRL_OUTPUT_SEL;

// Output voltage level (LOW/HIGH).
typedef enum TDA7540_SWCTRL_SOUTPUT_VOLT
{
	TDA7540_SWCTRL_OUT_VOLTAGE_LOW	= 0x00,
	TDA7540_SWCTRL_OUT_VOLTAGE_HIGH	= 0x01,
} TDA7540_SWCTRL_SOUTPUT_VOLT;

// S1 output state (emitter/open collector).
typedef enum TDA7540_SWCTRL_SOUTPUT
{
	TDA7540_SWCTRL_OUT_EMITTER		= 0x00,
	TDA7540_SWCTRL_OUT_OPEN_COLL	= 0x01,
} TDA7540_SWCTRL_SOUTPUT;

// SSTOP reference (IFC, H&FSU>x.xV).
typedef enum TDA7540_SWCTRL_SSTOP
{
	TDA7540_SWCTRL_SSTOP_FSU_EQUAL	= 0x00,
	TDA7540_SWCTRL_SSTOP_FSU_0_89V	= 0x01,
	TDA7540_SWCTRL_SSTOP_FSU_1_16V	= 0x02,
	TDA7540_SWCTRL_SSTOP_FSU_1_43V	= 0x03,
	TDA7540_SWCTRL_SSTOP_FSU_1_70V	= 0x04,
	TDA7540_SWCTRL_SSTOP_FSU_1_97V	= 0x05,
	TDA7540_SWCTRL_SSTOP_FSU_2_24V	= 0x06,
	TDA7540_SWCTRL_SSTOP_FSU_2_51V	= 0x07,
	TDA7540_SWCTRL_SSTOP_FSU_2_78V	= 0x08,
	TDA7540_SWCTRL_SSTOP_FSU_3_05V	= 0x09,
	TDA7540_SWCTRL_SSTOP_FSU_3_32V	= 0x10,
	TDA7540_SWCTRL_SSTOP_FSU_3_59V	= 0x11,
	TDA7540_SWCTRL_SSTOP_FSU_3_86V	= 0x12,
	TDA7540_SWCTRL_SSTOP_FSU_4_13V	= 0x13,
	TDA7540_SWCTRL_SSTOP_FSU_4_40V	= 0x14,
	TDA7540_SWCTRL_SSTOP_FSU_4_67V	= 0x15
} TDA7540_SWCTRL_SSTOP;

//////////////////////////////////////////////////////////////////////////[24]

//////////////////////////////////////////////////////////////////////////[25]

typedef enum TDA7540_STD_MUTE
{
	TDA7540_STD_MUTE_DISABLE	= 0x00,
	TDA7540_STD_MUTE_ENABLE		= 0x01,
} TDA7540_STD_MUTE;

typedef enum TDA7540_STD_MUTED_STATE
{
	TDA7540_STD_ON_WHEN_MUTED	= 0x00,
	TDA7540_STD_OFF_WHEN_MUTED	= 0x02,
} TDA7540_STD_MUTED_STATE;

typedef enum TDA7540_STD_STEREOMONO
{
	TDA7540_STD_MONO	= 0x00,
	TDA7540_STD_AUTO	= 0x04,
} TDA7540_STD_STEREOMONO;

typedef enum TDA7540_STD_NB_PEAKCHARGE
{
	TDA7540_STD_NBPC_LO	= 0x00,
	TDA7540_STD_NBPC_HI	= 0x08,
} TDA7540_STD_NB_PEAKCHARGE;

typedef enum TDA7540_STD_PILOT_THRESHOLD
{
	TDA7540_STD_PT_HI	= 0x00,
	TDA7540_STD_PT_LO	= 0x10,
} TDA7540_STD_PILOT_THRESHOLD;

typedef enum TDA7540_STD_DEEMPHASIS
{
	TDA7540_STD_DEEMPH_50us	= 0x00,
	TDA7540_STD_DEEMPH_75us	= 0x20,
} TDA7540_STD_DEEMPHASIS;

typedef enum TDA7540_STD_INGAIN
{
	TDA7540_STD_INGAIN_3_75dB	= 0x00,
	TDA7540_STD_INGAIN_2_5dB	= 0x10,
	TDA7540_STD_INGAIN_1_25dB	= 0x20,
	TDA7540_STD_INGAIN_0dB		= 0x30,
} TDA7540_STD_INGAIN;

//////////////////////////////////////////////////////////////////////////[26]

//////////////////////////////////////////////////////////////////////////[27]

//////////////////////////////////////////////////////////////////////////[28]

//////////////////////////////////////////////////////////////////////////[29]

//////////////////////////////////////////////////////////////////////////[30]

//////////////////////////////////////////////////////////////////////////[31]

//////////////////////////////////////////////////////////////////////////[32]

//////////////////////////////////////////////////////////////////////////[33]

//////////////////////////////////////////////////////////////////////////[34]

//////////////////////////////////////////////////////////////////////////[35]

//////////////////////////////////////////////////////////////////////////[36]

//////////////////////////////////////////////////////////////////////////[37]

//////////////////////////////////////////////////////////////////////////[38]

//////////////////////////////////////////////////////////////////////////[39]

void TDA7540_Init(cyg_i2c_bus* busTDA, cyg_i2c_bus* busEEPROM, cyg_uint16 addrEEPROM);
void TDA7540_Uninit(void);
		
int TDA7540_Write(cyg_uint8 addr, const cyg_uint8* buf, cyg_uint8 len);
void TDA7540_Read(cyg_uint8* buf);
void TDA7540_LoadEEPROM(void);
cyg_bool TDA7540_ReadEEPROM(cyg_uint8 addr, cyg_uint8* buf, cyg_uint8 len);
void	TDA7540_WriteEEPROM(cyg_uint8 addr, const cyg_uint8* buf, cyg_uint8 len);
cyg_bool TDA7540_CheckEEPROM(void);
cyg_bool TDA7540_CheckI2cComm(void);

void test(void);


int TDA7540_I2CUpdate(cyg_uint8 inStart,cyg_uint8 inStop);

//////////////////////////////////////////////////////////////////////////[0]
void TDA7540_SetCPCurrent(TDA7540_FM_HICURRENT locur, TDA7540_FM_HICURRENT hicur, TDA7540_UPDATE update);
void TDA7540_SetCPCurrentMode(TDA7540_CURRENT_MODE mode, TDA7540_UPDATE update);
void TDA7540_SetCPFMSeek(TDA7540_FM_SEEK_STATE state, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[1][2]
void TDA7540_SetCounter(TDA7540_BAND inBand, cyg_uint32 inFrequency, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[4]
void TDA7540_SetTV2Value(cyg_uint8 tv2, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[5]
void TDA7540_SetIFCLockDetector(TDA7540_LOCKSTATE state, TDA7540_UPDATE update);
void TDA7540_SetIFCMuteAndAGC(TDA7540_MUTEAGC muteagc, TDA7540_UPDATE update);
void TDA7540_SetIFCMode(TDA7540_IFCMODE ifcmode, TDA7540_UPDATE update);
void TDA7540_SetIFCState(TDA7540_IFCSTATE state, TDA7540_UPDATE update);
void TDA7540_SetIFCSampleTime(TDA7540_TSAMPLE tsample, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[6]
void TDA7540_SetIFCFCenter(TDA7540_IFC_FCENTER fcenter, TDA7540_UPDATE update);
void TDA7540_SetIFCErrorWindow(TDA7540_IFC_EW ew, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[7]
void TDA7540_SetAMControlAGCTimeConst(TDA7540_AMCTRL_AGCTC agctc, TDA7540_UPDATE update);
void TDA7540_SetAMControlPin24(TDA7540_AMCTRL_PIN24 usage, TDA7540_UPDATE update);
void TDA7540_SetAMControlPrescaler(TDA7540_AMCTRL_PRESCALER prescaler, TDA7540_UPDATE update);
void TDA7540_SetAMControlAdditTVOffset(TDA7540_AMCTRL_ADD_OFFSTV offtv, TDA7540_UPDATE update);
void TDA7540_SetAMControlPllMode(TDA7540_AMCTRL_TV_PLL_MODE mode, TDA7540_UPDATE update);
void TDA7540_SetAMControlLocalMode(TDA7540_AMCTRL_LOCAL_MODE state, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[8]
void TDA7540_SetQualISSFilterCtrl(TDA7540_QUAL_ISSF_CTRL state, TDA7540_UPDATE update);
void TDA7540_SetQualISSFilterSwitch(TDA7540_QUAL_ISSF_SWITCH state, TDA7540_UPDATE update);
void TDA7540_SetQualISSFilterFreq(TDA7540_QUAL_ISSF_FREQ freq, TDA7540_UPDATE update);
void TDA7540_SetQualISSFilter30KHzWB(TDA7540_QUAL_ISSF_30KHZ state, TDA7540_UPDATE update);
void TDA7540_SetQualISSFilterCurrent(TDA7540_QUAL_ISSF_CURRENT value, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[14]
void TDA7540_SetPllReference(TDA7540_PLL_BAND pllband, TDA7540_PLL_REFERENCE pllref, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[18]
void TDA7540_SetSMeterOffsetSL(TDA7540_SMETER_OFFSET_SL slide, TDA7540_SMETER_OFFSET_SIGN sign, TDA7540_UPDATE update);
void TDA7540_SetSMeterOverdevSlope(TDA7540_SMETER_OVERDEV_SLOPE slope, TDA7540_UPDATE update);
void TDA7540_SetSMeterOverdevCMax(TDA7540_SMETER_OVERDEV_CMAX cmax, TDA7540_UPDATE update);
void TDA7540_SetSMeterOverdev(TDA7540_SMETER_OVERDEV state, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[23]
void TDA7540_SetSwCtrlSStopRef(TDA7540_SWCTRL_SSTOP ref, TDA7540_UPDATE update);
void TDA7540_SetSwCtrlOutput(TDA7540_SWCTRL_OUTPUT_SEL output, TDA7540_SWCTRL_SOUTPUT state, TDA7540_UPDATE update);
void TDA7540_SetSwCtrlOutputVoltage(TDA7540_SWCTRL_OUTPUT_SEL output, TDA7540_SWCTRL_SOUTPUT_VOLT mode, TDA7540_UPDATE update);

//////////////////////////////////////////////////////////////////////////[25]
void TDA7540_MuteStereoDecoder(TDA7540_STD_MUTE state, TDA7540_UPDATE update);
void TDA7540_SetDeEmphasis(TDA7540_STD_DEEMPHASIS value, TDA7540_UPDATE update);

void TDA7540_SetRegister(cyg_uint8 reg, cyg_uint8 mask, cyg_uint8 val, TDA7540_UPDATE update);
cyg_uint8 TDA7540_GetRegister(cyg_uint8 reg, cyg_uint8 mask);

cyg_int8 TDA7540_GetTV2Value(cyg_uint8 reg);


#endif /*_TDA7540_*/
