/**
 ****************************************************************************
 * Driver for Wolfson WM8731 stereo audio codec.
 *
 * @file   include/WM8731_regs.h
 * @brief  Wolfson WM8731 Registers Description.
 *
 * Code is based on the WM8731.pdf
 *
 * @author Marc Flambard
 * @date   2007-05-24
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#ifndef WM8731_H
#define WM8731_H

#include<cyg/infra/cyg_type.h>
#include<cyg/error/codes.h>
#include<cyg/io/i2c.h>
#include<pkgconf/devices_wm8731.h>
#include<cyg/devs/wm8731_regs.h>

/*
 * Defines.
 ***********/


/** Number of concurrent devices. */
#define WM8731_NBDEVICES	CYGPKG_DEVICES_WM8731_NBDEVICES


/** Minimal in gain: -34.5dB. step:1.5dB */
#define WM8731_INVOL_MIN		0
/** Maximal in gain: +12dB. */
#define WM8731_INVOL_MAX		0x1F
/** in gain for 0dB. */
#define WM8731_INVOL_0DB		0x17

/** Minimal output volume: -73dB. step: 1dB */
#define WM8731_HPVOL_MIN		0x30
/** Maximal output volume: +6dB. */
#define WM8731_HPVOL_MAX		0x7F
/** Output volume for 0dB. */
#define WM8731_HPVOL_0DB		0x79

/** Store dc offset when High Pass Filter disabled.
 * @typedef WM8731_HPOR
 */
typedef enum WM8731_INSEL_TAG
{
	WM8731_INSEL_LINE   = 0,  ///< clear offset
	WM8731_INSEL_MIC    = 1   ///< store offset
} WM8731_INSEL;

/** Side Tone Attenuation.
 * @typedef WM8731_SIDEATT
 */
typedef enum WM8731_SIDEATT_TAG
{
	WM8731_SIDEATT_M6DB = 0x00, ///< -6dB.
	WM8731_SIDEATT_M9DB,        ///< -9dB.
	WM8731_SIDEATT_M12DB,       ///< -12dB.
	WM8731_SIDEATT_M15DB        ///< -15dB.
} WM8731_SIDEATT;

/** De-emphasis Control.
 * @typedef WM8731_DEEMP
 */
typedef enum WM8731_DEEMP_TAG
{
	WM8731_DEEMP_DISABLE = 0x00, ///< Disable.
	WM8731_DEEMP_32KHZ,          ///< 32kHz.
	WM8731_DEEMP_441KHZ,         ///< 44.1kHz.
	WM8731_DEEMP_48KHZ           ///< 48kHz.
} WM8731_DEEMP;

/** Store dc offset when High Pass Filter disabled.
 * @typedef WM8731_HPOR
 */
typedef enum WM8731_HPOR_TAG
{
	WM8731_HPOR_CLEAR   = 0, ///< clear offset
	WM8731_HPOR_STORE   = 1   ///< store offset
} WM8731_HPOR;

/** Digital interface format.
 * @typedef WM8731_FORMAT
 */
typedef enum WM8731_FORMAT_TAG
{
	WM8731_FORMAT_RIGHTJUST = 0x00, ///< MSB-First, right justified
	WM8731_FORMAT_LEFTJUST,         ///< MSB-First, left justified
	WM8731_FORMAT_I2S,              ///< I2S Format, MSB-First left-1 justified
	WM8731_FORMAT_DSP               ///< DSP Mode, frame sync + 2 data packed words
} WM8731_FORMAT;

/** Input Audio Data Bit Length Select.
 * @typedef WM8731_IWL
 */
typedef enum WM8731_IWL_TAG
{
	WM8731_IWL_16B = 0x00,  ///< 16 bits
	WM8731_IWL_20B,         ///< 20 bits
	WM8731_IWL_24B,         ///< 24 bits
	WM8731_IWL_32B          ///< 32 bits
} WM8731_IWL;

/** Power down control.
 * @typedef WM8731_POWER
 */
typedef struct WM8731_POWER_TAG
{
	cyg_bool lineinpd;
	cyg_bool micpd;
	cyg_bool adcpd;
	cyg_bool dacpd;
	cyg_bool outpd;
	cyg_bool oscpd;
	cyg_bool clkoutpd;
	cyg_bool poweroff;
} WM8731_POWER;

/** Structure to set all global configuration in once.
 * @typedef WM8731_CONFIG
 */
typedef struct WM8731_CONFIG_TAG
{
	cyg_bool adchpd;
	WM8731_DEEMP deemp;
	cyg_bool dacmu;
	cyg_bool hpor;
	WM8731_POWER power;
	WM8731_FORMAT format;
	WM8731_IWL iwl;
	cyg_bool lrp;
	cyg_bool lrswap;
	cyg_bool ms;
	cyg_bool bclkinv;
	cyg_bool usb_norm;
	cyg_bool bosr;
	cyg_int8 sr;
	cyg_bool clkidiv2;
	cyg_bool clkodiv2;
	cyg_bool active;
} WM8731_CONFIG;

/** Structure to set all global settings in once.
 * @typedef WM8731_CONFIG
 */
typedef struct WM8731_SETTINGS_TAG
{
	cyg_int8 linvol;
	cyg_bool linmute;
	cyg_bool lrinboth;
	cyg_int8 rinvol;
	cyg_bool rinmute;
	cyg_bool rlinboth;
	cyg_int8 lhpvol;
	cyg_int8 lhpvol_old;	//volume before mute
	cyg_bool lhpmute;
	cyg_bool lzcen;
	cyg_bool lrhpboth;
	cyg_int8 rhpvol;
	cyg_int8 rhpvol_old;	//volume before mute
	cyg_bool rhpmute;
	cyg_bool rzcen;
	cyg_bool rlhpboth;
	cyg_bool micboost;
	cyg_bool mutemic;
	WM8731_INSEL insel;
	cyg_bool bypass;
	cyg_bool dacsel;
	cyg_bool sidetone;
	WM8731_SIDEATT sideatt;
} WM8731_SETTINGS;

/** type of device ID.
 * @typedef WM8731_DEVID
 */
typedef cyg_uint8 WM8731_DEVID;

/**
 * Functions prototypes.
 */

Cyg_ErrNo WM8731_Init(const WM8731_DEVID devid, const cyg_i2c_bus *bus, const cyg_bool ad0);
Cyg_ErrNo WM8731_Reset(const WM8731_DEVID devid);
Cyg_ErrNo WM8731_Config_Get(const WM8731_DEVID devid, WM8731_CONFIG *config_target);
Cyg_ErrNo WM8731_Config_Set(const WM8731_DEVID devid, const WM8731_CONFIG *config_source);
Cyg_ErrNo WM8731_Settings_Get(const WM8731_DEVID devid, WM8731_SETTINGS *settings_target);
Cyg_ErrNo WM8731_Settings_Set(const WM8731_DEVID devid, const WM8731_SETTINGS *settings_source);
Cyg_ErrNo WM8731_GetLInVol(const WM8731_DEVID devid, cyg_uint8 *vol);
Cyg_ErrNo WM8731_SetLInVol(const WM8731_DEVID devid, const cyg_uint8 vol);
Cyg_ErrNo WM8731_GetLInMute(const WM8731_DEVID devid, cyg_bool *mute);
Cyg_ErrNo WM8731_SetLInMute(const WM8731_DEVID devid, const cyg_bool mute);
Cyg_ErrNo WM8731_GetRInVol(const WM8731_DEVID devid, cyg_uint8 *vol);
Cyg_ErrNo WM8731_SetRInVol(const WM8731_DEVID devid, const cyg_uint8 vol);
Cyg_ErrNo WM8731_GetRInMute(const WM8731_DEVID devid, cyg_bool *mute);
Cyg_ErrNo WM8731_SetRInMute(const WM8731_DEVID devid, const cyg_bool mute);
Cyg_ErrNo WM8731_GetLHpVol(const WM8731_DEVID devid, cyg_uint8 *vol);
Cyg_ErrNo WM8731_SetLHpVol(const WM8731_DEVID devid, const cyg_uint8 vol);
Cyg_ErrNo WM8731_GetLHpMute(const WM8731_DEVID devid, cyg_bool *mute);
Cyg_ErrNo WM8731_SetLHpMute(const WM8731_DEVID devid, const cyg_bool mute);
Cyg_ErrNo WM8731_GetLHpZcen(const WM8731_DEVID devid, cyg_bool *zcen);
Cyg_ErrNo WM8731_SetLHpZcen(const WM8731_DEVID devid, const cyg_bool zcen);
Cyg_ErrNo WM8731_GetRHpVol(const WM8731_DEVID devid, cyg_uint8 *vol);
Cyg_ErrNo WM8731_SetRHpVol(const WM8731_DEVID devid, const cyg_uint8 vol);
Cyg_ErrNo WM8731_GetRHpMute(const WM8731_DEVID devid, cyg_bool *mute);
Cyg_ErrNo WM8731_SetRHpMute(const WM8731_DEVID devid, const cyg_bool mute);
Cyg_ErrNo WM8731_GetRHpZcen(const WM8731_DEVID devid, cyg_bool *zcen);
Cyg_ErrNo WM8731_SetRHpZcen(const WM8731_DEVID devid, const cyg_bool zcen);
Cyg_ErrNo WM8731_Power_Config_Get(const WM8731_DEVID devid, WM8731_POWER *config_target);
Cyg_ErrNo WM8731_Power_Config_Set(const WM8731_DEVID devid, const WM8731_POWER *config_source);

#endif
