/**
 ****************************************************************************
 * Driver for Wolfson WM8731 stereo audio codec.
 *
 * @file   include/WM8731_regs.h
 * @brief  Wolfson WM8731 Registers Description.
 *
 * Code is based on the WM8731.pdf
 *
 * @author Marc Flambard
 * @date   2007-05-24
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#ifndef WM8731_REGS_H
#define WM8731_REGS_H

/**
 * @defgroup WM8731_regs Register address
 *
 * @{
 */

#define WM8731_R0_LIN            0x00       ///< Left Line Input
#define WM8731_R1_RIN            0x01       ///< Right Line Input
#define WM8731_R2_LHP            0x02       ///< Left Headphone Out
#define WM8731_R3_RHP            0x03       ///< Right Headphone Out
#define WM8731_R4_A_AUDIO_PATH   0x04       ///< Analog audio path control
#define WM8731_R5_D_AUDIO_PATH   0x05       ///< Digital audio path control
#define WM8731_R6_POWER          0x06       ///< Power down control
#define WM8731_R7_D_AUDIO_FRMT   0x07       ///< Digital Audio Interface Format
#define WM8731_R8_SAMPLING       0x08       ///< Sampling Control
#define WM8731_R9_ACTIVE         0x09       ///< Active control
#define WM8731_R10_RESET         0x0F       ///< Reset control

/** @}
 *
 * @defgroup WM8731_regs_descr  Register description
 *
 * Here is the description of register contents, i.e. what does each bit
 * mean in each register.
 *
 * @{
 */

/**
 * @defgroup wm8731_regs_r0  R0(0x00): Left Line Input
 * @{
 */
#define WM8731_R_LRINBOTH       (1<<8)        ///< Left and Right Chanel Control
#define WM8731_R_LINMUTE        (1<<7)        ///< Left Input PGA Analogue Mute
#define WM8731_R_LINVOL(x)      (x)&1F        ///< Left Input PGA Volume level
/** @}
 * @defgroup wm8731_regs_r1  R1(0x01):  Right Line Input
 * @{
 */
#define WM8731_R_RLINBOTH       (1<<8)         ///< Left and Right Chanel Control
#define WM8731_R_RINMUTE        (1<<7)         ///< Right Input PGA Analogue Mute
#define WM8731_R_RINVOL(x)      (x)&1F         ///< Right Input PGA Volume level
/** @}
 * @defgroup wm8731_regs_r2  R2(0x02): Left Headphone Out
 * @{
 */
#define WM8731_R_LRHPBOTH       (1<<8)         ///< Left and Right Chanel Control
#define WM8731_R_LZCEN          (1<<7)         ///< Left Headphone Output Zero Cross Enable
#define WM8731_R_LHPVOL(x)      (x)&7F         ///< Left Output Volume level
/** @}
 * @defgroup wm8731_regs_r3  R3(0x03): Right Headphone Out
 * @{
 */
#define WM8731_R_RLHPBOTH       (1<<8)         ///< Left and Right Chanel Control
#define WM8731_R_RZCEN          (1<<7)         ///< Right Headphone Output Zero Cross Enable
#define WM8731_R_RHPVOL(x)      (x)&7F         ///< Right Output Volume level
/** @}
 * @defgroup wm8731_regs_r4  R4(0x04): Analog audio path control
 * @{
 */
#define WM8731_R_MICBOOST       (1<<0)         ///< Microphone Input Level Boost
#define WM8731_R_MUTEMIC        (1<<1)         ///< Mic Input Mute to ADC
#define WM8731_R_INSEL          (1<<2)         ///< Microphone/Line Input Select to ADC
#define WM8731_R_BYPASS         (1<<3)         ///< Bypass Switch
#define WM8731_R_DACSEL         (1<<4)         ///< DAC Select
#define WM8731_R_SIDETONE       (1<<5)         ///< Side Tone Switch
#define WM8731_R_SIDEATT(x)     (((x)&3)<<6)   ///< Side Tone Attenuation
/** @}
 * @defgroup wm8731_regs_r5  R5(0x05): Digital audio path control
 * @{
 */
#define WM8731_R_ADCHPD         (1<<0)         ///< ADC High Pass Filter disable
#define WM8731_R_DEEMP(x)       (((x)&3)<<1)   ///< De-emphasis Control
#define WM8731_R_DACMU          (1<<3)         ///< DAC Soft Mute Control
#define WM8731_R_HPOR           (1<<4)         ///< Store dc offset when High Pass Filter disabled
/** @}
 * @defgroup wm8731_regs_r6  R6(0x06): Power down control
 * @{
 */
#define WM8731_R_LINEINPD       (1<<0)         ///< Line Input Power Down
#define WM8731_R_MICPD          (1<<1)         ///< Microphone Input an Bias Power down
#define WM8731_R_ADCPD          (1<<2)         ///< ADC Power Down
#define WM8731_R_DACPD          (1<<3)         ///< DAC Power Down
#define WM8731_R_OUTPD          (1<<4)         ///< Ouputs Power Down
#define WM8731_R_OSCPD          (1<<5)         ///< Oscillator Power Down
#define WM8731_R_CLKOUTPD       (1<<6)         ///< CLKOUT power down
#define WM8731_R_POWEROFF       (1<<7)         ///< POWEROFF mode
/** @}
 * @defgroup wm8731_regs_r7  R7(0x07): Digital Audio Interface Format
 * @{
 */
#define WM8731_R_FORMAT(x)      (((x)&3)<<0)   ///< Audio Data Format Select
#define WM8731_R_IWL(x)         (((x)&3)<<2)   ///< Input Audio Data Bit Length Select
#define WM8731_R_LRP            (1<<4)         ///< DACLRC phase control
#define WM8731_R_LRSWAP         (1<<5)         ///< DAC Left Right Clock Swap
#define WM8731_R_MS             (1<<6)         ///< Master Slave Mode Control
#define WM8731_R_BCLKINV        (1<<7)         ///< Bit Clock Invert
/** @}
 * @defgroup wm8731_regs_r8  R8(0x08): Sampling Control
 * @{
 */
#define WM8731_R_USB_NORM       (1<<0)         ///< Mode Select USB/Normal
#define WM8731_R_BOSR           (1<<1)         ///< Base Over-Sampling Rate
#define WM8731_R_SR(x)          (((x)&7)<<2)   ///< ADC and DAC sample rate control
#define WM8731_R_CLKIDIV2       (1<<6)         ///< Core Clock divider select
#define WM8731_R_CLKODIV2       (1<<7)         ///< CLKOUT divider select
/** @}
 * @defgroup wm8731_regs_r9  R9(0x09): Active control
 * @{
 */
#define WM8731_R_ACTIVE         (1<<0)         ///< Activate Interface
/** @}
 * @defgroup wm8731_regs_r10  R10(0x0F): Reset control
 * @{
 */
#define WM8731_R_RESET          (0x00)         ///< Reset Device
/** @} */
// !wm8731_regs_descr
/** @} */

#endif
