#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "config.h"

#define U16 unsigned short
#define U32 unsigned int

#define BAUDRATE_NB 20
int valid_baud_rate_tab[BAUDRATE_NB][2] = {
		{B50, 50},
		{B75, 75},
		{B110, 110},
		{B134, 134},
		{B150, 150},
		{B200, 200},
		{B300, 300},
		{B600, 600},
		{B1200, 1200},
		{B1800, 1800},
		{B2400, 2400},
		{B4800, 4800},
		{B9600, 9600},
		{B19200, 19200},
		{B38400, 38400},
		{B57600, 57600},
		{B115200, 115200},
		{B230400, 230400},
		{B460800, 460800},
		{B921600, 921600}
};

static unsigned char buffer[BUFFER_SIZE+2]; // buffer + CRC


void print_usage(void);

/*
 * Fletcher's checksum
 * @param data  Pointer to the data to be summed
 * @param len  Length in 16-bit words
 * @return 16 bits checksum
 */
U16 checksum(U16 *data, U16 len)
{
	U32 sum1 = 0xffff, sum2 = 0xffff;

	while (len) {
		unsigned int tlen = len > 360 ? 360 : len;
		len -= tlen;
		do {
			sum1 += *data++;
			sum2 += sum1;
		} while (--tlen);
		sum1 = (sum1 & 0xffff) + (sum1 >> 16);
		sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	}
	/* Second reduction step to reduce sums to 16 bits */
	sum1 = (sum1 & 0xffff) + (sum1 >> 16);
	sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	return sum2 << 16 | sum1;
}

int main(int argc, char ** argv)
{
	int fd;
	struct termios tio;
	char buf[255];
	int baudrate;
	int ret;
	int i;
	static int nb_try = 0;
	int read_size;
	U16 crc;
	unsigned char c;

	// basic check on parameters
	if (argc < 3 || argc > 4) {
		print_usage();
	} else if (argc == 4 && strcmp("-f", argv[3])) {
		print_usage();
	}

	// get the baudrate from the command line
	baudrate = atoi(argv[2]);

	// search the correct baudrate flag
	i = 0;
	while (i < BAUDRATE_NB && valid_baud_rate_tab[i][1] != baudrate)
			i++;
	if (i >= BAUDRATE_NB) {
			fprintf(stderr, "baudrate not found\n");
			return -1;
	}
	baudrate = valid_baud_rate_tab[i][0];

	// open the device
	fd = open(argv[1], O_RDWR | O_NOCTTY ); 
	if (fd < 0) {
		perror(argv[1]);
		exit(-1);
	}

	bzero(&tio, sizeof(tio));
	if (argc == 4 && !strcmp("-f", argv[3])) {
		tio.c_cflag = baudrate | CRTSCTS | CS8 | CLOCAL | CREAD;
	} else {
		tio.c_cflag = baudrate | /*CRTSCTS |*/ CS8 | CLOCAL | CREAD;
	}
	tio.c_iflag = IGNPAR;
	tio.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	tio.c_lflag = 0;

	tio.c_cc[VTIME]    = 10;//0;   /* inter-character timer unused */
	tio.c_cc[VMIN]     = 1;//BUFFER_SIZE+2;   /* blocking read until X chars received */

	tcflush(fd,TCIFLUSH);
	tcsetattr(fd,TCSANOW,&tio);

	printf("test_p5_board lunch ...\n");

	/*
	while (1) {
		ret = read(fd, buffer, BUFFER_SIZE+1);
		fwrite(buffer, 1, ret, stdout);
	}*/

	while(1) {
#ifdef SEND_DATA_TO_HOST
		// read (BUFFER_SIZE+2) byte from tty
		read_size = 0;
		while (read_size != BUFFER_SIZE+2) {
				ret = read(fd, buffer+read_size, BUFFER_SIZE+2-read_size);
				if (ret < 0) {
					fprintf(stderr, "error on read: %d\n", ret);
					exit(-1);
				}
				read_size += ret;
		}
		
		// verify the checksum
		crc = checksum((U16 *) buffer, BUFFER_SIZE/2);
		if (crc != *(U16*) &buffer[BUFFER_SIZE]) {
			fprintf(stderr, "checksum error %x %x\n", crc, *(U16*) &buffer[BUFFER_SIZE]);
			exit(-1);
		}
		printf("%d: CRC OK\n", nb_try);
#endif

#ifdef RECEIVE_DATA_ON_P5_BOARD
		// send (BUFFER_SIZE+2) to the P5 dev board
		for (i = 0; i < BUFFER_SIZE; i++)
				buffer[i] = rand();
		*(U16 *) &buffer[BUFFER_SIZE] = checksum((U16*) buffer, BUFFER_SIZE/2);
		ret = write(fd, buffer, BUFFER_SIZE+2);
		if (ret < 0) {
				fprintf(stderr, "write error\n");
				exit(-1);
		}
		printf("write OK\n");
#endif
		nb_try++;
	}
	
	return 0;
}


void print_usage(void)
{
		fprintf(stderr, "\n");
		fprintf(stderr, "Usage: test_p5_board device baudrate [-f]\n");
		fprintf(stderr, "\n");
		fprintf(stderr, "Options:\n");
		fprintf(stderr, "\t-f: flow control with RTS/CTS\n");
		fprintf(stderr, "\n");
		fprintf(stderr, "typical options : /dev/ttyS0 38400\n");
		fprintf(stderr, "\n");
		exit(-1);
}
