//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  fatfs_boot.c
* @brief FAT boot basic functions
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     florent.bayendrian@parrot.fr
* @date       2007-07-28
*******************************************************************************
*/
#include <pkgconf/system.h>
#include <pkgconf/hal.h>
#include <pkgconf/io_fileio.h>
#include <pkgconf/fs_fat.h>

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_trac.h>        // tracing macros
#include <cyg/infra/cyg_ass.h>         // assertion macros
#include <cyg/crc/crc.h>

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include <stdlib.h>
#include <string.h>

#include <cyg/infra/diag.h>
#include <cyg/fileio/fileio.h>
#include <cyg/io/io.h>
#include <cyg/io/config_keys.h>
#include <blib/blib.h>
#include <cyg/fs/fatfs.h>

#include "fatfs.h"
#include <cyg/hal/hal_cache.h>

#define FATFS_BOOT_NODE_POOL_SIZE 2 // max simultaneously open files at boot time
static fatfs_node_t fatfs_boot_node_pool_base[FATFS_BOOT_NODE_POOL_SIZE];

static fatfs_disk_t fatfs_boot_disk;
#define FATFS_BOOT_MAX_BLOCK_SIZE 512
static cyg_uint8 fatfs_boot_sector[FATFS_BOOT_MAX_BLOCK_SIZE]
        CYGBLD_ATTRIB_ALIGN(HAL_DCACHE_LINE_SIZE);
static cyg_uint32 fatfs_boot_part_start;
static cyg_uint32 fatfs_boot_part_size;


#define MBR_SIG_ADDR  0x1FE   // signature address
#define MBR_SIG_BYTE0 0x55    // signature first byte value
#define MBR_SIG_BYTE1 0xAA    // signature second byte value
#define MBR_PART_ADDR 0x1BE   // first partition address
#define MBR_PART_SIZE 0x10    // partition size

#define READ_DWORD(_data_, _val_)                           \
    do {                                                    \
        _val_ = *((cyg_uint8 *)_data_)           |          \
                *(((cyg_uint8 *)_data_)+1) << 8  |          \
                *(((cyg_uint8 *)_data_)+2) << 16 |          \
                *(((cyg_uint8 *)_data_)+3) << 24;           \
    } while (0)




/**
 * Read data in a block (same interface as cyg_blib_read)
 *
 * @param priv not used
 * @param buf data buffer ptr
 * @param len number of bytes to read
 * @param bnum block number
 * @param pos position inside the block
 *
 * @return Cyg_ErrNo
 */
Cyg_ErrNo fatfs_boot_disk_read(void       *priv,
                               void       *buf,
                               cyg_uint32 *len,
                               cyg_uint32  bnum,
                               cyg_uint32  pos)
{
    fatfs_disk_t *disk = &fatfs_boot_disk;
    cyg_uint8    *bbuf = (cyg_uint8 *) buf;
    cyg_int32     size = *len;
    Cyg_ErrNo     ret  = ENOERR;
    cyg_uint32    block_size = fatfs_boot_disk.disk_info.block_size;


    if (pos >= block_size) {
        bnum += pos/block_size;
        pos   = pos & (block_size - 1);
    }

    if (pos == 0 && (size%block_size) == 0) {
        cyg_uint32 bsize = size/block_size;
        return disk->boot_bread(bbuf, bnum+fatfs_boot_part_start, bsize);
    }

    CYG_ASSERT(block_size <= FATFS_BOOT_MAX_BLOCK_SIZE,
               "block_size not supported");

    while (size > 0) {
        cyg_uint32 csize;

        if ((size + pos) > block_size)
            csize = block_size - pos;
        else
            csize = size;

        if (bnum > fatfs_boot_part_size)
            return EINVAL;
        ret = disk->boot_bread(fatfs_boot_sector, bnum+fatfs_boot_part_start, 1);
        if (ret != ENOERR)
            break;

        memcpy((void *)bbuf, (void *)(fatfs_boot_sector+pos), csize);
        bbuf += csize;
        size -= csize;
        pos   = 0;
        bnum++;

    }
    *len -= size;
    return ret;
}


void fatfs_boot_read_partition(cyg_uint8 * data, cyg_uint8 *part_id,
                               cyg_uint32 *part_start, cyg_uint32 *part_size)
{
    CYG_ASSERT(data && part_id && part_start && part_size, "NULL pointer!");
    *part_id = data[4];
    READ_DWORD(&data[8], *part_start);
    READ_DWORD(&data[12], *part_size);
}

/**
 * Init the boot search for open
 *
 * @param boot_bread block read function
 * @param part_num -0 for the all device, 1 to 4 for the primary partitions
 * @param blocks_num number of block of the device
 * @param block_size block size of the device (usually 512)
 * @param pos position inside the block
 *
 * @return Cyg_ErrNo
 */
Cyg_ErrNo fatfs_boot_init(fatfs_boot_bread_cb_t boot_bread,
                          cyg_uint32 part_num,
                          cyg_uint32 blocks_num,
                          cyg_uint32 block_size)
{
    cyg_uint8 part_id, *pdata;
    fatfs_disk_t       *disk = &fatfs_boot_disk;
    fatfs_dir_entry_t   root_dentry;
    Cyg_ErrNo           err = ENOERR;

    CYG_ASSERT(part_num <= 4, "invalid partition number");
    CYG_ASSERT(block_size <= FATFS_BOOT_MAX_BLOCK_SIZE,
               "block_size not supported");
    CYG_ASSERT(FATFS_BOOT_NODE_POOL_SIZE <= FATFS_NODE_POOL_SIZE,
               "FATFS_NODE_POOL_SIZE too small");

    memset(&root_dentry, 0, sizeof(fatfs_dir_entry_t));
    if (part_num > 0) {
        // read and parse the MBR
        err = boot_bread(fatfs_boot_sector, 0, 1);
        if (err != ENOERR) {
            err = EIO;
            goto end;
        }

        if ((MBR_SIG_BYTE0 != fatfs_boot_sector[MBR_SIG_ADDR+0]) ||
            (MBR_SIG_BYTE1 != fatfs_boot_sector[MBR_SIG_ADDR+1])) {
            err = EINVAL;
            goto end;
        }
        pdata = &fatfs_boot_sector[MBR_PART_ADDR+MBR_PART_SIZE*(part_num-1)];
        fatfs_boot_read_partition(pdata,
                                  &part_id,
                                  &fatfs_boot_part_start,
                                  &fatfs_boot_part_size);
        if (!part_id) {
            err = EINVAL;
            goto end;
        }
        // check partition table consistency
        if ((fatfs_boot_part_start > blocks_num) ||
            (fatfs_boot_part_size  > blocks_num) ||
            (fatfs_boot_part_size+fatfs_boot_part_start > blocks_num)) {
            err = EINVAL;
            goto end;
        }

    } else {
        // Take all the card as a FAT partition
        fatfs_boot_part_start = 0;
        fatfs_boot_part_size = blocks_num;
    }

    memset(disk, 0, sizeof(fatfs_disk_t));
    disk->node_pool_base = fatfs_boot_node_pool_base;
    disk->node_pool_size = FATFS_BOOT_NODE_POOL_SIZE;
    disk->bcache_mem = NULL;
    disk->disk_info.phys_block_size = disk->disk_info.block_size = block_size;
    disk->disk_info.blocks_num = blocks_num;
    disk->block_size_log2 = get_val_log2(block_size);
    disk->boot_bread = boot_bread;
    disk->disk_read = fatfs_boot_disk_read;
    disk->disk_write = NULL; // not used
    disk->read_only = true;

    init_fatfs_fds(disk);

    disk->dev_h = NULL;
    err = fatfs_init(disk);
    if (err != ENOERR) {
        goto end;
    }
    fatfs_node_cache_init(disk);
    fatfs_get_root_dir_entry(disk, &root_dentry);
    disk->root = fatfs_node_alloc(disk, &root_dentry);
    fatfs_node_ref(disk, disk->root);

end:
    return err;
}

Cyg_ErrNo fatfs_boot_open(const char * name, CYG_ADDRWORD *file_desc)
{
    fatfs_disk_t       *disk = &fatfs_boot_disk;
    fatfs_node_t       *node = NULL;
    cyg_dir             dir  = CYG_DIR_NULL;
    fatfs_fd_t         *fd;
    fatfs_dirsearch_t   ds;
    Cyg_ErrNo          err;

    init_dirsearch(&ds, disk, (fatfs_node_t *) dir, name);
    err = fatfs_find(&ds);
    if (err == ENOERR) {
        node = ds.node;
    } else {
        return err;
    }

    if (S_ISDIR(node->dentry.mode))
        return EISDIR;

    // Allocate file object private data and
    // make a reference to this file node
    fd = alloc_fatfs_fd(disk, node);
    if (NULL == fd)
        return EMFILE;

    fatfs_node_ref(disk, node);

    // Initialize the file object
    *file_desc    = (CYG_ADDRWORD) fd;
    return ENOERR;
}

int fatfs_boot_read(CYG_ADDRWORD file_desc, void *buf, cyg_uint32 *offset, cyg_uint32 count)
{
    fatfs_disk_t  *disk   = &fatfs_boot_disk;
    fatfs_fd_t    *fd     = (fatfs_fd_t *) file_desc;
    fatfs_node_t  *node   = fd->node;
    Cyg_ErrNo     err;

    // Adjust size to end of file if necessary
    if (count > node->dentry.size - *offset)
        count = node->dentry.size - *offset;

    err = fatfs_read_data(disk, &node->dentry, &fd->pos, buf, &count);
    if (err != ENOERR)
        return -1;

    // Update working vars
    *offset += count;

    return count;
}

void fatfs_boot_close(CYG_ADDRWORD file_desc)
{
    fatfs_disk_t  *disk   = &fatfs_boot_disk;
    fatfs_fd_t    *fd    = (fatfs_fd_t *) file_desc;
    fatfs_node_t  *node  = fd->node;

    fatfs_node_unref(disk, node);
    free_fatfs_fd(disk, fd);
}
