//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
******************************************************************************
* @file src/fatfs_robust.c
* @brief FAT FileSystem Robustness
*
* Copyright (C) 2008 Parrot S.A.
*
* @author     gregoire.etienne@parrot.com
* @date       2008-02-13
******************************************************************************
*/


#include "fatfs.h"
#include "fatfs_supp.h"
#include "fatfs_robust.h"

#include <cyg/crc/crc.h>

/** Check if a partition has been correctly unmounted
 *  check if disk has been correctly umounted, seeking for "UMOUNTED"
 *  string. Then writes "MOUNTED"
 *
 * @param disk A pointer to the disk
 * @return ENOERR if "UMOUNTED" string has been found
 */
int robust_magic_mount(fatfs_disk_t *disk)
{
    int     err = EINVAL;
    char    str[10];
    char    *bufsector;
    cyg_uint32 n_sector, lensec;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return EINVAL;
    }

    /*-----------------------------------------------------------------------*/
    /* Reading FAT0 CRC                                                      */
    /*-----------------------------------------------------------------------*/
    n_sector = RBST_FAT0CRC_SEC;
    lensec   = 1;
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);
    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        GET_DWORD(bufsector, disk->fat0_crc, RBST_FAT0CRC_OFFSET);
    }


    /*-----------------------------------------------------------------------*/
    /* Reading FAT1 CRC and magic                                            */
    /*-----------------------------------------------------------------------*/
    n_sector = RBST_FAT1CRC_SEC;
    lensec   = 1;
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);
    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        GET_BYTES(bufsector, str,    8     , RBST_MAGIC_OFFSET);
        GET_DWORD(bufsector, disk->fat1_crc, RBST_FAT1CRC_OFFSET);

        if (0 != strncmp(str, "UMOUNTED", 8))
        {// writing in this sector failed
            FAT_WARNING("FAT partition hasn't been correctly umounted\n");
            err = ENOENT;
        }
        else
        {
            sprintf( str, "MOUNTED ");
            SET_BYTES(bufsector, str, 8, RBST_MAGIC_OFFSET);
            err = ENOERR;
            FAT_WARNING("FAT partition has been correctly umounted\n");

            err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &lensec, n_sector);
            if (err != ENOERR)
            {// writing in this sector failed
                FAT_ERROR("cyg_io_write ERROR sector= %d\n", n_sector);
            }
        }
    }

    free(bufsector);
    return err;
}


/** Mark a partition as correctly unmounted writting a magic
 *  write "UMOUNTED" into disk
 *
 * @param disk A pointer to the disk
 * @return ENOERR if no error
 */
int robust_magic_umount(fatfs_disk_t *disk)
{
    int     err = EINVAL;
    char    str[10];
    char    *bufsector;
    cyg_uint32 n_sector = RBST_FAT1CRC_SEC;
    cyg_uint32 lensec   = 1;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return EINVAL;
    }

    // Now we're going to compute FAT1 crc
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);

    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        sprintf( str, "UMOUNTED");
        SET_BYTES(bufsector, str, 8, RBST_MAGIC_OFFSET);

        err = ENOERR;
        FAT_DEBUG("FAT partition has been correctly umounted\n");

        lensec = 1;
        n_sector = RBST_FAT1CRC_SEC;
        err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &lensec, n_sector);
        if (err != ENOERR)
        {// writing in this sector failed
            FAT_ERROR("cyg_io_write ERROR sector= %d\n", n_sector);
        }
    }

    free(bufsector);
    return err;
}


/** Set FAT0 crc value in Boot Sector.
 *
 * this function computes FAT0 crc value, stores it into boot sector
 *
 * @param disk A pointer to the disk
 * @param crc value to store
 * @return ENOERR if no error
 */
int robust_set_fat0_crc(fatfs_disk_t *disk, cyg_uint32 crc)
{
    int     err = EINVAL;
    char    *bufsector;
    cyg_uint32 n_sector = RBST_FAT0CRC_SEC;
    cyg_uint32 lensec  = 1;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return 0;
    }

    // Now we're going to compute FAT0 crc
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);
    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        SET_DWORD(bufsector, crc, RBST_FAT0CRC_OFFSET);

        err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &lensec, n_sector);

        if (err != ENOERR)
        {// writing in this sector failed
            FAT_ERROR("cyg_io_write ERROR sector= %d\n", n_sector);
        }
    }

    free(bufsector);
    return err;
}


/** Store FAT0crc value in Boot Sector.
 *
 * this function computes FAT0 crc value, stores it into boot sector
 *
 * @param disk A pointer to the disk
 * @param crc value to store
 * @return ENOERR if no error
 */
int robust_set_fat1_crc(fatfs_disk_t *disk, cyg_uint32 crc)
{
    int     err = EINVAL;
    char    *bufsector;
    cyg_uint32 n_sector = RBST_FAT1CRC_SEC;
    cyg_uint32 lensec   = 1;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return 0;
    }

    // Now we're going to compute FAT1 crc
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);

    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        SET_DWORD(bufsector, crc, RBST_FAT1CRC_OFFSET);

        err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &lensec, n_sector);

        if (err != ENOERR)
        {// writing in this sector failed
            FAT_ERROR("cyg_io_write ERROR sector= %d\n", n_sector);
        }
    }

    free(bufsector);
    return err;
}




/** Write FAT0 to FAT1.
 *
 * Make a backup of FAT0 to FAT1
 *
 * @param disk A pointer to the disk
 * @return ENOERR if no error
 */
int robust_fat0_to_fat1(fatfs_disk_t *disk)
{
    int err = ENOERR;
    void   *bufsector;

    cyg_uint32  FAT0_firstsec   = disk->fat_tbl_pos / disk->sector_size;
    cyg_uint32  FAT1_firstsec   = disk->fat_tbl_pos_backup / disk->sector_size;
    cyg_uint32  FAT_lengthsec   = disk->fat_tbl_size / disk->sector_size;

    cyg_uint32  src_sector      = FAT0_firstsec ;
    cyg_uint32  last_sector     = FAT0_firstsec+FAT_lengthsec;
    cyg_uint32  dst_sector      = FAT1_firstsec;
    cyg_uint32  one_sector      = 1;

    //allocate a buffer to work on a whole sector
    bufsector = malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return ENOSPC;
    }

    //
    do{
        err = cyg_io_bread(disk->dev_h, (void*)bufsector, &one_sector, src_sector);

        if (err != ENOERR)
        {// reading in this sector failed
            FAT_ERROR("cyg_io_bread ERROR sector= %d\n", src_sector);
            free(bufsector);
            return err;
        }
        else
        {
            err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &one_sector, dst_sector);
            if (err != ENOERR)
            {// writing in this sector failed
                FAT_ERROR("cyg_io_write ERROR sector= %d\n", dst_sector);
                free(bufsector);
                return err;
            }

        }

        src_sector++;
        dst_sector++;

    }while( src_sector < last_sector);
    free( bufsector);

    // Now we're going to set FAT0 crc
    err = robust_set_fat1_crc(disk, disk->fat0_crc);
    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("failed to set FAT1crc\n");
    }
    else
    {
        disk->fat1_crc = disk->fat0_crc;
    }

    FAT_DEBUG("exit : err = %d\n", err);

    return err;

}


/** Write FAT1 to FAT0. Restore
 *
 * @param disk A pointer to the disk
 * @return ENOERR if no error
 */
int robust_fat1_to_fat0(fatfs_disk_t *disk)
{
    int err = ENOERR;
    void   *bufsector;

    cyg_uint32  FAT0_firstsec   = disk->fat_tbl_pos / disk->sector_size;
    cyg_uint32  FAT1_firstsec   = disk->fat_tbl_pos_backup / disk->sector_size;
    cyg_uint32  FAT_lengthsec   = disk->fat_tbl_size / disk->sector_size;

    cyg_uint32  src_sector      = FAT1_firstsec ;
    cyg_uint32  last_sector     = FAT1_firstsec+FAT_lengthsec;
    cyg_uint32  dst_sector      = FAT0_firstsec ;
    cyg_uint32  one_sector      = 1;

    //allocate a buffer to work on a whole sector
    bufsector = malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return ENOSPC;
    }

    do{
        err = cyg_io_bread(disk->dev_h, (void*)bufsector, &one_sector, src_sector);

        if (err != ENOERR)
        {// reading in this sector failed
            FAT_ERROR("cyg_io_bread ERROR sector= %d\n", src_sector);
            free(bufsector);
            return err;
        }
        else
        {
            err = cyg_io_bwrite( disk->dev_h, (void*)bufsector, &one_sector, dst_sector);
            if (err != ENOERR)
            {// writing in this sector failed
                FAT_ERROR("cyg_io_write ERROR sector= %d\n", dst_sector);
                free(bufsector);
                return err;
            }

        }

        src_sector++;
        dst_sector++;

    }while( src_sector < last_sector);
    free( bufsector);

    // Now we're going to set FAT0 crc
    err = robust_set_fat0_crc(disk, disk->fat0_crc);

    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("failed to set FAT0crc\n");
    }
    else
    {
        disk->fat0_crc = disk->fat0_crc;
    }

    FAT_DEBUG("exit : err = %d\n", err);

    return err;
}

/** Compute a FAT table CRC
 *
 * @param hdl                A handler to the disk
 * @param first_sector       first sector of the FAT table
 * @param fat_length_sectors number of sectors of the FAT table
 * @param bytespersector     number of bytes per sector
 * @return CRC value
 */

cyg_uint32 robust_calc_fatcrc(  cyg_io_handle_t hdl,
                                cyg_uint32      first_sector,
                                cyg_uint32      fat_length_sectors,
                                cyg_uint32      bytespersector )
{
    int err;
    char   *bufsector;
    unsigned int  crc       = 0;
    unsigned int  length    = 0;
    cyg_uint32  n_sector    = first_sector ;
    cyg_uint32  last_sector = first_sector+fat_length_sectors;
    cyg_uint32  one_sector  = 1;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(bytespersector);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return 0;
    }

    //initializing crc computing
    cyg_posix_crc_start(&crc, &length);

    do{
        err = cyg_io_bread(hdl, (void*)bufsector, &one_sector, n_sector);

        if (err != ENOERR)
        {
            FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
            cyg_posix_crc_end(&crc, &length);
            free(bufsector);
            return crc; // anyway, in this case crc can't be true
        }

        n_sector++;
        cyg_posix_crc_cont(&crc, &length, (char*)bufsector, bytespersector);

    }while( n_sector < last_sector);

    cyg_posix_crc_end(&crc, &length);

    free( bufsector);

    return crc;
}



/** Check FAT tables validity.
 *
 * @param disk A pointer to the disk
 * @return ENOERR if no error
 */
int robust_fat_check(fatfs_disk_t *disk)
{
    int err = EINVAL;
    cyg_uint32  FAT0_crc; // First FAT table CRC
    cyg_uint32  FAT1_crc; // Second FAT table CRC

    // FAT table check and restore if needed
    FAT0_crc = robust_calc_fatcrc(  disk->dev_h,
                                    disk->fat_tbl_pos/disk->sector_size,
                                    disk->fat_tbl_size/disk->sector_size,
                                    disk->sector_size );
    if(FAT0_crc != disk->fat0_crc)
    {
        /*-------------------------------------------------------------------*/
        /* FAT0 crc is false                                                 */
        /*      if   FAT1 crc is correct then restore FAT0 from FAT1         */
        /*      else fail                                                    */
        /*-------------------------------------------------------------------*/
        FAT_WARNING("FAT0 crc is false :read=0x%08x, computed=0x%08x\n",
                    disk->fat0_crc, FAT0_crc);
        FAT1_crc = robust_calc_fatcrc(  disk->dev_h,
                                        disk->fat_tbl_pos_backup/disk->sector_size,
                                        disk->fat_tbl_size/disk->sector_size,
                                        disk->sector_size );

        if(FAT1_crc != disk->fat1_crc){
            FAT_ERROR("FAT1 crc is false :read=0x%08x, computed=0x%08x\n",
                        disk->fat1_crc, FAT1_crc);
            FAT_ERROR("Can't restore FAT table\n");
            err = EINVAL;
        }else{
            FAT_DEBUG("Restore FAT0 from FAT1\n");
            err = robust_fat1_to_fat0(disk);
            if(err != ENOERR){
                FAT_ERROR("Restoring failed\n");
            }
        }
    }
    else
    {
        /*-------------------------------------------------------------------*/
        /* FAT0 crc is correct                                               */
        /*      if   FAT1 is correct and up to date then everything's fine   */
        /*      else restore FAT1 from FAT0                                  */
        /*-------------------------------------------------------------------*/
        FAT_DEBUG("FAT0 is correct: crc = 0x%08x\n", FAT0_crc);

        FAT1_crc = robust_calc_fatcrc(  disk->dev_h,
                                        disk->fat_tbl_pos_backup/disk->sector_size,
                                        disk->fat_tbl_size/disk->sector_size,
                                        disk->sector_size );
        if(FAT1_crc != disk->fat1_crc) // FAT1 is not up to date
        {// FAT1 needs to be updated
            FAT_WARNING("FAT1 crc is false :read=0x%08x, computed=0x%08x\n",
                        disk->fat1_crc, FAT1_crc);
            err = robust_fat0_to_fat1(disk);
            if(err != ENOERR) {
                FAT_ERROR("Restoring failed\n");
            }else{
                FAT_WARNING("Restored FAT1 from FAT0\n");
            }
        }else{
            FAT_DEBUG("FAT1 is correct: crc = 0x%08x\n", FAT1_crc);
            err = ENOERR;


        }
    }
    return err;
}


/** Clean FAT partition after being restored.
 *
 * @param disk A pointer to the disk
 * @param cluster start cluster
 * @return ENOERR if no error
 */
int robust_remove_dirty_entries(fatfs_disk_t *disk, cyg_uint32 cluster)
{
    int err = EINVAL;
    fat_raw_dir_entry_t raw_dentry;
    fatfs_data_pos_t    pos;

    // Initialize position
    pos.cluster      = cluster;
    pos.cluster_snum = 0;
    pos.cluster_pos  = 0;

    while (true)
    {
        // Read next dir entry
        err = read_next_raw_dentry(disk, &pos, &raw_dentry);
        if (err != ENOERR)
           return err;


        /* Entry is deleted or zeroed                                        */
        /*-------------------------------------------------------------------*/
        if ( DENTRY_IS_DELETED(&raw_dentry) || DENTRY_IS_ZERO(&raw_dentry))
        {
            pos.cluster_pos += DENTRY_SIZE;
        }
        else
        {
            int entry_type CYGBLD_ATTRIB_UNUSED;
            cyg_uint32 tentry;
            cyg_uint32 file_first_cluster = raw_dentry.cluster | (raw_dentry.cluster_HI << 16);

            err = read_tentry(disk, file_first_cluster, &tentry);
            if (err != ENOERR)
                return err;

            /* Cluster pointed by this entry is free -> DELETE ENTRY         */
            /*---------------------------------------------------------------*/
            entry_type = get_tentry_type(disk, tentry);
            if(   (  (file_first_cluster == 0 ) || (tentry == TENTRY_FREE) )
               && (  ( 0 != strncmp("..", raw_dentry.name, 2) ) )  ) //cause ".." dir can point on cluster 0 in FAT16
            {
#ifdef CYGCFG_FS_FAT_USE_VFAT
                FAT_WARNING("CLEAN:free entry name=%s c=%08x, s=%08x, p=%08x\n",raw_dentry.lname,
                            pos.cluster, pos.cluster_snum, pos.cluster_pos);
#else
                FAT_WARNING("CLEAN:free entry name=%s c=%08x, s=%08x, p=%08x\n",raw_dentry.name,
                            pos.cluster, pos.cluster_snum, pos.cluster_pos);
#endif // CYGCFG_FS_FAT_USE_VFAT
                // Delete dentry
                raw_dentry.name[0] = 0xE5;
                raw_dentry.size    = 0;
                raw_dentry.cluster = 0;
                err = write_raw_dentry(disk, &pos, &raw_dentry);
                if (err != ENOERR)
                    return err;
            }
            /* Entry is a directory -> GET INTO AND CLEAN                    */
            /*---------------------------------------------------------------*/
            else if( DENTRY_IS_DIR(&raw_dentry) )
            {
                if(  (0 != strncmp("..", raw_dentry.name, 2) )
                   &&(0 != strncmp(".", raw_dentry.name, 1) )  )
                {
                    robust_remove_dirty_entries(disk,
                                 raw_dentry.cluster | (raw_dentry.cluster_HI << 16));
                }
            }
#ifdef CYGCFG_FS_FAT_USE_VFAT
            err = go_to_next_raw_entry(disk, &pos, &raw_dentry, CO_NONE);
            if (err != ENOERR)
                return err;
#else //CYGCFG_FS_FAT_USE_VFAT
            pos.cluster_pos += DENTRY_SIZE;
#endif
        }

        if (err != ENOERR)
            return (err == EEOF ? ENOERR : err);
    }
    return err;
}



/** Synchronize FAT tables
 *
 * FAT tables are synchronized only if there is no opened file.
 *
 * @param disk A pointer to the disk
 * @return ENOERR if no error
 */
int robust_fat_sync(fatfs_disk_t *disk)
{
    cyg_uint32 FAT0_crc;
    int err = ENOERR;

    FAT_API("%d live nodes\n", nb_live_nodes);

    // FAT0: up to date but invalid
    // FAT1: old and valid
    FAT0_crc = robust_calc_fatcrc(  disk->dev_h,
                                    disk->fat_tbl_pos/disk->sector_size,
                                    disk->fat_tbl_size/disk->sector_size,
                                    disk->sector_size );

    err = robust_set_fat0_crc(disk, FAT0_crc); // update FAT0 crc
    // FAT0: up to date and valid
    // FAT1: old and valid
    disk->fat0_crc = FAT0_crc;

    err = robust_fat0_to_fat1(disk);            // update FAT1 from FAT0
    // FAT0: up to date and valid
    // FAT1: up to date and valid

    return err;
}
