#ifndef CYGONCE_FATFS_FATFS_ROBUST_H
#define CYGONCE_FATFS_FATFS_ROBUST_H

/**
******************************************************************************
* @file include/fatfs_robust.h
* @brief FAT FileSystem Robustness header
*
* Copyright (C) 2008 Parrot S.A.
*
* @author     gregoire.etienne@parrot.com
* @date       2008-02-13
******************************************************************************
*/

#include "fatfs_supp.h"

#define RBST_FAT0CRC_SEC       (1)   //! Sector Number where Fat0 crc is stored 
#define RBST_FAT1CRC_SEC       (2)   //! Sector Number where Fat1 crc is stored
#define RBST_MAGIC_OFFSET      (0x4) //! Offset where magic is written
#define RBST_FAT0CRC_OFFSET    (0x0) //! Offset where Fat0 crc is written
#define RBST_FAT1CRC_OFFSET    (0x0) //! Offset where Fat1 crc is written

int robust_magic_mount          (fatfs_disk_t *disk);
int robust_magic_umount         (fatfs_disk_t *disk);
int robust_fat_check            (fatfs_disk_t *disk);
int robust_remove_dirty_entries (fatfs_disk_t *disk, cyg_uint32 cluster);
int robust_fat_sync             (fatfs_disk_t *disk);

cyg_uint32 robust_calc_fatcrc(   cyg_io_handle_t hdl,
                                 cyg_uint32      first_sector, 
                                 cyg_uint32      fat_length_sectors,
                                 cyg_uint32      bytespersector );

int robust_fat1_to_fat0         (fatfs_disk_t *disk);
int robust_fat0_to_fat1         (fatfs_disk_t *disk);

int robust_set_fat0_crc         (fatfs_disk_t *disk, cyg_uint32 crc);
int robust_set_fat1_crc         (fatfs_disk_t *disk, cyg_uint32 crc);



/*! \page fatfs_robust FAT Partition Robustness

<b>FAT Partition Robustness</b> is an optional patch specifically made to
guarantee FAT partition coherency over flash memory. 
<p>
The original file allocation table (FAT) file system enabled file modification 
operations to be interrupted before completion. In this way, actions such as 
sudden power loss or sudden removal of a storage card frequently resulted in 
data loss and file system corruption. By making file operations 
transaction-safe, "FATRobust" stabilizes the file system and ensures that the 
file system is not corrupted when an interruption occurs.

"FATRobust" works with two copies of the FAT table, a FAT0 table in which 
current  operations are being conducted and a FAT1 table that is a stable copy 
of the last known good FAT. Modifications to FAT1 are not committed until all 
elements in a transaction have completed successfully. If a transaction fails 
to complete properly, the disk is set in the same state as it was before the 
transaction began. After all transactions are successfully completed, the FAT0  
table is copied to FAT1.


<p>
This patch can be enabled/disabled through the configtool:
<div><img src="fatfs_config.png" alt="FAT package configuration"></div>
<p>
Nota Bene:
<ul>
<li>This patch implies FAT standard modification so a partition can be read 
under any OS but must only be written through eCos.
<li>This patch implies lower performances, FAT16 small partitions (less than 
64Mo) are much more suitable  
</ul>

<hr>

\section fatrob_howto Howto

This example is made with a sdcard.
<br>

First of all, <b>the partition must be formatted as "robust"</b>, passing 
<tt>true</tt> as fifth parameter of <tt>fatfs_format()</tt>. This will compute
crc values for both FAT0 and FAT1.

\code
    n = cyg_io_lookup("/dev/sdcard/0", &sdcard);
    if (n != 0) {
        diag_printf("lookup error\n");
        assert( n == 0);
    }

    n = fatfs_format( sdcard,
                        1,      // partition number
                        false,  // erase only fat table
                        0,      // nbr of sectors per cluster
                        true);  // format as fat robust
   if (n != 0) {
      diag_printf("(EE) format /dev/sdcard/1 \n");
      assert( n == 0);
   }
   else {
      diag_printf("(  ) format /dev/sdcard/1 \n");
   }
\endcode
 
<b>Once this partition is mounted, "robust" functionnality must be activated.
</b> It will check magic key and both crc. 
\code
    n = mount("/dev/sdcard/1", "/", "fatfs");

    if (n != 0) {
        diag_printf("mount error\n");
        assert( n == 0);
    }
    else 
    {
        int len, ret;
        cyg_bool robust_part;

        robust_part = 1;
        len = sizeof(robust_part);
        ret = cyg_fs_setinfo("/", FS_INFO_FATFS_ROBUST, &robust_part, len);
        if (ret != ENOERR) {
            diag_printf("partition can't be set as robust\n");
            assert(0);
        }
    }
\endcode

<b>Now this partition can be used as usual.</b>

*/

#endif //CYGONCE_FATFS_FATFS_ROBUST_H
