#include "fatfs_supp.h"
#include "fatfs_robust.h"

// -------------------------------------------------------------------------
// disk_cluster_write()
// Writes data to disk at specified cluster position.

static __inline__ int
disk_cluster_write(fatfs_disk_t *disk,
                   void         *buf,
                   cyg_uint32   *len,
                   cyg_uint32    cluster,
                   cyg_uint32    cluster_pos)
{
    cyg_uint32 block, block_pos;

    if (disk->read_only)
        return EPERM;

    cluster_to_block_pos(disk, cluster, cluster_pos, &block, &block_pos);
    CYG_ASSERT(disk->disk_write, "NULL pointer!!");

    return disk->disk_write(&disk->blib, buf, len, block, block_pos);
}

// -------------------------------------------------------------------------
// jdays_to_gdate()
// Converts juilan days into gregorian date.

static void
jdays_to_gdate(cyg_uint32 jd, int *day, int *month, int *year)
{
    cyg_uint32 l, n, i, j;

    l = jd + 68569;
    n = (4 * l) / 146097;
    l = l - (146097 * n + 3) / 4;
    i = (4000 * (l + 1)) / 1461001;
    l = l - (1461 * i) / 4 + 31;
    j = (80 * l) / 2447;
    *day = l - (2447 * j) / 80;

    l = j / 11;
    *month = j + 2 - (12 * l);
    *year = 100 * (n - 49) + i + l;
}

// -------------------------------------------------------------------------
// date_unix2dos()
// Converts unix timestamp to dos time and date.

static void
date_unix2dos(time_t  unix_timestamp,
              cyg_uint16 *dos_time,
              cyg_uint16 *dos_date)
{
    cyg_uint32 jd;
    cyg_uint16 dtime, ddate;
    int        hour, min, sec;
    int        day, month, year;

    hour = (unix_timestamp / 3600) % 24;
    min  = (unix_timestamp / 60) % 60;
    sec  =  unix_timestamp % 60;

    jd = JD_1_JAN_1970 + unix_timestamp / (3600 * 24);
    jdays_to_gdate(jd, &day, &month, &year);

    if (year < 1980)
        year = 1980;

    dtime = (hour << 11) | (min << 5) | (sec >> 1);
    ddate = ((year - 1980) << 9) | (month << 5) | day;

    if (NULL != dos_time) *dos_time = dtime;
    if (NULL != dos_date) *dos_date = ddate;
}


/** Format disk into FAT32.
 *
 *  -------------------------------------------
 * | BOOT REGION                               |
 * | size = BPB_RSVDSECCNT x BPB_BYTSPERSEC    |
 *  -------------------------------------------
 * | FAT                                       |
 * | size = BPB_FATSz32 x BPB_BYTSPERSEC       |
 *  -------------------------------------------
 * | FAT (backup)                              |
 * | size = BPB_FATSz32 x BPB_BYTSPERSEC       |
 *  -------------------------------------------
 * | CLUSTER 2                                 |
 * | size = BPB_BYTSPERSEC                     |
 *  -------------------------------------------
 * | CLUSTER 3                                 |
 * | size = BPB_BYTSPERSEC                     |
 *  -------------------------------------------
 */


#define BPB_NUMFATS           2 //! (0x010) Number of FAT

#define FAT32_BPB_RSVDSECCNT 32 //! (0x00E) Count of reserved for the boot region
#define BPB_FSINFOSEC         4 //! (0x030)FS Info Sector
#define BPB_BKBOOTSEC         6 //! (0x032) Boot Backup Sector

#define FAT32_ENTRY_LAST     0x0FFFFFF8
#define FAT32_ENTRY_RSVD     0x0FFFFFFF

Cyg_ErrNo
fatfs_format_fat32( cyg_io_handle_t dev,
                    cyg_uint32      first_sector,
                    cyg_uint32      BPB_TotSec,
                    cyg_uint32      BPB_BytsPerSec,
                    cyg_uint32      sectors_per_cluster,
                    cyg_bool        erase,
                    cyg_bool        robust)
{
    char        str[0x20];
    cyg_uint8   *buffer;
    cyg_uint32  len;
    cyg_uint32  n_sector;
    cyg_uint32  erase_until_sect;
    cyg_uint32  BPB_FATSz;      // Size of ONE File Allocation Table
    cyg_uint32  BPB_SecPerClus; // Number of sectors per cluster
    cyg_uint32  FSI_FreeCount;  // Number of free clusters
#ifdef CYGCFG_FS_FAT_ROBUST
    cyg_uint32  FAT0_crc = 0;       // First FAT table CRC
#endif
    int err;

    FAT_DEBUG(" Input params:\n \
                    first sector = %d\n \
                    NbTotSec     = %d sectors\n \
                    BytsPerSec   = %d bytes\n \
                    SecPerClus   = %d sectors\n \
                    erase        = %d\n \
                    robust       = %d\n",
                first_sector,BPB_TotSec,BPB_BytsPerSec,sectors_per_cluster, erase,robust );

    /*****************************************************************************************/
    /* Determination of the number of sectors per cluster                                    */
    /*****************************************************************************************/
    /* For FAT32, try to do the same as M$'s format command:
     * fs size < 256M: 0.5k clusters
     * fs size <   8G: 4k clusters
     * fs size <  16G: 8k clusters
     * fs size <  32G: 16k clusters
     * fs size >= 32G: 32k clusters
     */
    if (!sectors_per_cluster)
    {
        /* if      ( BPB_TotSec <   0x80000) BPB_SecPerClus =  1;
        else */
        // using 4k clusters min to optimize performances
        if      ( BPB_TotSec < 0x1000000) BPB_SecPerClus =  8;
        else if ( BPB_TotSec < 0x2000000) BPB_SecPerClus = 16;
        else if ( BPB_TotSec < 0x4000000) BPB_SecPerClus = 32;
        else                              BPB_SecPerClus = 64;
    } else {
        if ( (sectors_per_cluster != 1 ) &&
             (sectors_per_cluster != 2 ) &&
             (sectors_per_cluster != 4 ) &&
             (sectors_per_cluster != 8 ) &&
             (sectors_per_cluster != 16) &&
             (sectors_per_cluster != 32) &&
             (sectors_per_cluster != 64)   )
            return EINVAL;
        else
            BPB_SecPerClus = sectors_per_cluster;
    }

    {   //Computation coming from mkdosfs tool
        cyg_uint32 fatdata = BPB_TotSec - FAT32_BPB_RSVDSECCNT;
        cyg_uint32 clust32 = 2 + (((cyg_uint64) fatdata*BPB_BytsPerSec + BPB_NUMFATS*8) /
                             ((cyg_uint32)BPB_SecPerClus*BPB_BytsPerSec + BPB_NUMFATS*4));

        BPB_FATSz = ((clust32*4 + BPB_BytsPerSec) - 1 ) / BPB_BytsPerSec;
    }

    // Computation of the number of data cluster
    FSI_FreeCount = (BPB_TotSec - (2*BPB_FATSz + FAT32_BPB_RSVDSECCNT )) / BPB_SecPerClus;
     //Adjusting the number of sectors used
    BPB_TotSec = FAT32_BPB_RSVDSECCNT + 2*BPB_FATSz + FSI_FreeCount*BPB_SecPerClus;
    // First data cluster is set as used
    FSI_FreeCount--;

    buffer = (cyg_uint8*)malloc(BPB_BytsPerSec);
    if(buffer == 0)
        return ENOSPC;

    /*****************************************************************************************/
    /* Erasing the partition area                                                            */
    /*****************************************************************************************/
    if(erase)
    {   //! Erasing the whole partition to 0
        erase_until_sect = first_sector+BPB_TotSec;
    }
    else
    {   //! Erasing only Boot, FAT regions and the first data cluster
        erase_until_sect = first_sector+FAT32_BPB_RSVDSECCNT+2*BPB_FATSz + BPB_SecPerClus;
    }

    memset(buffer, 0, BPB_BytsPerSec);
    len = 1;

    for(n_sector = first_sector; n_sector < (erase_until_sect); n_sector++)
    {
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }
    }

/****************************************************************************/
/* NOTA BENE: FAT Region are written before boot sector cause we want to be */
/*            able to compute FAT Region CRC and stores it in boot sector   */
/****************************************************************************/

    /*****************************************************************************************/
    /* FAT0 Region                                                                           */
    /*****************************************************************************************/
    memset(buffer, 0, BPB_BytsPerSec);

    SET_DWORD(buffer, FAT32_ENTRY_LAST,     0x00);
    SET_DWORD(buffer, FAT32_ENTRY_RSVD,     0x04);
    SET_DWORD(buffer, FAT32_ENTRY_LAST,     0x08);

    n_sector = first_sector+FAT32_BPB_RSVDSECCNT;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }
#ifdef CYGCFG_FS_FAT_ROBUST
    if(robust == true)
    {
        FAT0_crc = robust_calc_fatcrc(  dev,
                                        (first_sector+FAT32_BPB_RSVDSECCNT),
                                        BPB_FATSz,
                                        BPB_BytsPerSec );
    }
#endif

    /*****************************************************************************************/
    /* FAT1 Region                                                                           */
    /*****************************************************************************************/
    err = cyg_io_bwrite( dev, (void*)buffer, &len, (n_sector+BPB_FATSz) );
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }


    /*****************************************************************************************/
    /* BOOT SECTOR                                                                           */
    /*****************************************************************************************/
    memset(buffer, 0, BPB_BytsPerSec);

    SET_DWORD(buffer,     0x009000EB,     0x00); //!BS_jmpBoot    : Jump instruction boot code
    sprintf( str, "MSWIN4.1");
    SET_BYTES(buffer,            str,  8, 0x03); //!BS_OEMName    : "MSWIN4.1" is a default value
    SET_WORD (buffer, BPB_BytsPerSec,     0x0B); //!BPB_BytsPerSec: 512 for maximum compatibility
    SET_BYTE (buffer, BPB_SecPerClus,     0x0D); //!BPB_SecPerClus:
    SET_WORD (buffer, FAT32_BPB_RSVDSECCNT,     0x0E); //!BPB_RsvdSecCnt: typically 32 for FAT32 volumes
    SET_BYTE (buffer, BPB_NUMFATS   ,     0x10); //!BPB_NumFATs   : Must be 2
    SET_WORD (buffer,         0x0000,     0x11); //!BPB_RootEntCnt: 0 for FAT32 volumes
    SET_WORD (buffer,         0x0000,     0x13); //!BPB_TotSec16  : 0 for FAT32 volumes
    SET_BYTE (buffer,           0xF8,     0x15); //!BPB_Media     : 0xF8 for non-removable media
    SET_WORD (buffer,         0x0000,     0x16); //!BPB_FATSz16   : 0 for FAT32 volumes
    SET_WORD (buffer,         0x003C,     0x18); //!BPB_SecPerTrk : TODO
    SET_WORD (buffer,         0x0004,     0x1A); //!BPB_NumHeads  : TODO
    SET_DWORD(buffer,     0x00000000,     0x1C); //!BPB_HiddSec   : TODO
    SET_DWORD(buffer, BPB_TotSec    ,     0x20); //!BPB_TotSec32  :

    /*****************************************************************************************/
    /* FAT32 Struct Starting at offset 36                                                    */
    /*****************************************************************************************/
    SET_DWORD(buffer,  BPB_FATSz    ,     0x24);   //!BPB_FATSz32   : Size of one FAT (in sectors)
    SET_WORD (buffer,         0x0000,     0x28);   //!BPB_ExtFlags  : TODO
    SET_WORD (buffer,         0x0000,     0x2A);   //!BPB_FSVer     : TODO
    SET_DWORD(buffer,     0x00000002,     0x2C);   //!BPB_RootClust : Default value for FAT32 volumes
    SET_WORD (buffer,  BPB_FSINFOSEC,     0x30);   //!BPB_FSInfo    : Default value for FAT32 volumes
    SET_WORD (buffer,  BPB_BKBOOTSEC,     0x32);   //!BPB_BkBootSec : Default value for FAT32 volumes
    SET_DWORD(buffer,     0x00000000,     0x34);   //!BPB_Reserved  : Must be set to 0
    SET_DWORD(buffer,     0x00000000,     0x38);   //!BPB_Reserved  : Must be set to 0
    SET_DWORD(buffer,     0x00000000,     0x3B);   //!BPB_Reserved  : Must be set to 0
    SET_BYTE (buffer,           0x00,     0x40);   //!BPB_DrvNum    : TODO
    SET_BYTE (buffer,           0x00,     0x41);   //!BPB_Reserved  : Must be set to 0
    SET_BYTE (buffer,           0x29,     0x42);   //!BPB_BootSig   : Must be 0x29
    SET_DWORD(buffer,     0x00000000,     0x43);   //!BPB_VolID     : TODO
    sprintf( str, "NO NAME    ");
    SET_BYTES(buffer,            str, 11, 0x47);   //!BPB_VolLab    : default
    sprintf( str, "FAT32   ");
    SET_BYTES(buffer,            str,  8, 0x52);  //!BPB_FilSysType:

    SET_DWORD(buffer,     0xAA550000,     0x1FC); //!FSI_TrailSig (+0x1FC):

    // Writing Boot Sector
    len      = 1;
    n_sector = first_sector + 0;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }

    // Writing Boot Sector Backup
    n_sector = first_sector + BPB_BKBOOTSEC;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }

#ifdef CYGCFG_FS_FAT_ROBUST
    /*****************************************************************************************/
    /* FAT32 Robust Info */
    /*****************************************************************************************/
    if(robust == true)
    {
        memset(buffer, 0, BPB_BytsPerSec);
        n_sector = first_sector + FAT32_BPB_RSVDSECCNT;
        FAT0_crc = robust_calc_fatcrc(  dev,
                                        n_sector,
                                        BPB_FATSz,
                                        BPB_BytsPerSec );

        SET_DWORD(buffer, FAT0_crc      ,     RBST_FAT0CRC_OFFSET); //!FAT0crc
        n_sector = first_sector + RBST_FAT0CRC_SEC;
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }

        SET_DWORD(buffer, FAT0_crc      ,     RBST_FAT1CRC_OFFSET); //!FAT1crc
        sprintf( str, "UMOUNTED");
        SET_BYTES(buffer,            str,  8, RBST_MAGIC_OFFSET); //!FATROB magic:
        n_sector = first_sector + RBST_FAT1CRC_SEC;
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }
    }
#endif //CYGCFG_FS_FAT_ROBUST

    /*****************************************************************************************/
    /* FAT32 FSInfo Sector                                                                   */
    /*****************************************************************************************/
    memset(buffer, 0, BPB_BytsPerSec);

    SET_DWORD(buffer,   0x41615252,     0x000); //!FSI_LeadSig  (+0x000):
    SET_DWORD(buffer,   0x61417272,     0x1E4); //!FSI_StrucSig (+0x1E4):
    SET_DWORD(buffer,FSI_FreeCount,     0x1E8); //!FSI_FreeCount(+0x1E8):
    SET_DWORD(buffer,   0x00000002,     0x1EC); //!FSI_NextFree (+0x1EC):
    SET_DWORD(buffer,   0xAA550000,     0x1FC); //!FSI_TrailSig (+0x1FC):

    len = 1;
    n_sector = first_sector + BPB_FSINFOSEC;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }

    free(buffer);

    FAT_DEBUG(" PartInfo:\n \
                    BPB_TotSec  = %d sectors\n \
                    BytsPerSec  = %d bytes\n \
                    SecPerClus  = %d sectors\n \
                    NbRsvdSec   = %d sectors\n \
                    FATSize     = %d sectors\n",
                BPB_TotSec, BPB_BytsPerSec, BPB_SecPerClus,
                FAT32_BPB_RSVDSECCNT, BPB_FATSz);

    return ENOERR;
}



/** Format disk into FAT16.
 *
 *  -------------------------------------------
 * | BOOT REGION                               |
 * | size = BPB_RSVDSECCNT x BPB_BYTSPERSEC    |
 *  -------------------------------------------
 * | FAT                                       |
 * | size = BPB_FATSz x BPB_BYTSPERSEC         |
 *  -------------------------------------------
 * | FAT (backup)                              |
 * | size = BPB_FATSz x BPB_BYTSPERSEC         |
 *  -------------------------------------------
 * | CLUSTER 0 (ROOT DIR)                      |
 * | size = 32 * ROOTENTRYCNT                  |
 *  -------------------------------------------
 * | CLUSTER 1                                 |
 * | size = BPB_BYTSPERSEC                     |
 *  -------------------------------------------
 */

#define FAT16_BPB_RSVDSECCNT   3 //! (0x00E) Count of reserved for the boot region

#define FAT16_ENTRY_LAST    0xFFF8
#define FAT16_ENTRY_RSVD    0xFFFF

Cyg_ErrNo
fatfs_format_fat16( cyg_io_handle_t dev,
                    cyg_uint32      first_sector,
                    cyg_uint32      BPB_TotSec,
                    cyg_uint32      BPB_BytsPerSec,
                    cyg_bool        erase,
                    cyg_bool        robust )
{
    char        str[0x20];
    int         err;
    cyg_uint8   *buffer;
    cyg_uint32  len;
    cyg_uint32  n_sector;
    cyg_uint32  erase_until_sect;
    cyg_uint32  nb_data_clust CYGBLD_ATTRIB_UNUSED; // Number of data clusters
    cyg_uint32  BPB_SecPerClus =  4;    // Number of sectors per cluster
    cyg_uint32  BPB_FATSz;              // Size of ONE File Allocation Table
    cyg_uint32  BPB_RootEntCnt = 512;   // RootDir length = 512 entries = 512*32 bytes = 0x4000
#ifdef CYGCFG_FS_FAT_ROBUST
    cyg_uint32  FAT0_crc = 0;           // First FAT table CRC
#endif

    FAT_DEBUG(" Input params:\n \
                    first sector = %d\n \
                    NbTotSec     = %d sectors\n \
                    BytsPerSec   = %d bytes\n \
                    erase        = %d\n \
                    robust       = %d\n",
                first_sector,BPB_TotSec,BPB_BytsPerSec, erase,robust );

    /*****************************************************************************************/
    /* Determination of the number of sectors per cluster                                    */
    /*****************************************************************************************/
    {   //Computation coming from mkdosfs tool
        cyg_uint32 fatdata = BPB_TotSec - ((BPB_RootEntCnt*32) / BPB_BytsPerSec) - FAT16_BPB_RSVDSECCNT;
        cyg_uint32 clust16 = 2 + (((cyg_uint64)fatdata*BPB_BytsPerSec + BPB_NUMFATS*4) /
                                  ((cyg_uint32)BPB_SecPerClus*BPB_BytsPerSec + BPB_NUMFATS*2));

        BPB_FATSz = ((clust16*2 + BPB_BytsPerSec) - 1 ) / BPB_BytsPerSec;
    }

    // Computation of the number of data cluster
    nb_data_clust = (BPB_TotSec - (2*BPB_FATSz) - FAT16_BPB_RSVDSECCNT) / BPB_SecPerClus;
    //BPB_TotSec = FAT16_BPB_RSVDSECCNT + (2*BPB_FATSz) + (nb_data_clust*BPB_SecPerClus) ;

    buffer = (cyg_uint8*)malloc(BPB_BytsPerSec);
    if(buffer == 0)
        return ENOSPC;

    /*****************************************************************************************/
    /* Erasing the partition area                                                            */
    /*****************************************************************************************/
    if(erase)
    {   //! Erasing the whole partition to 0
        erase_until_sect = first_sector + BPB_TotSec;
    }
    else
    {   //! Erasing only Boot, FAT regions and the root dir cluster
        erase_until_sect = first_sector + FAT16_BPB_RSVDSECCNT +
                            2*BPB_FATSz + (BPB_RootEntCnt * 32 / BPB_BytsPerSec);
    }

    memset(buffer, 0, BPB_BytsPerSec);
    len = 1;

    for(n_sector = first_sector; n_sector < (erase_until_sect); n_sector++)
    {
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }
    }


    /*****************************************************************************************/
    /* FAT Sector                                                                            */
    /*****************************************************************************************/
    memset(buffer, 0, BPB_BytsPerSec);

    SET_DWORD(buffer, FAT16_ENTRY_LAST,     0x00);
    SET_DWORD(buffer, FAT16_ENTRY_RSVD,     0x02);

    n_sector = first_sector + FAT16_BPB_RSVDSECCNT;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }

    n_sector = first_sector + FAT16_BPB_RSVDSECCNT + BPB_FATSz;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
    if (err != ENOERR)
    {
        free(buffer);
        return err;
    }

#ifdef CYGCFG_FS_FAT_ROBUST
    if(robust == true)
    {
        n_sector = first_sector + FAT16_BPB_RSVDSECCNT;
        FAT0_crc = robust_calc_fatcrc(  dev,
                                        n_sector,
                                        BPB_FATSz,
                                        BPB_BytsPerSec );

        SET_DWORD(buffer, FAT0_crc      ,     RBST_FAT0CRC_OFFSET); //!FAT0crc
        n_sector = first_sector + RBST_FAT0CRC_SEC;
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }

        SET_DWORD(buffer, FAT0_crc      ,     RBST_FAT1CRC_OFFSET); //!FAT1crc
        sprintf( str, "UMOUNTED");
        SET_BYTES(buffer,            str,  8, RBST_MAGIC_OFFSET); //!FATROB magic:
        n_sector = first_sector + RBST_FAT1CRC_SEC;
        err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector );
        if (err != ENOERR)
        {
            free(buffer);
            return err;
        }
    }
#endif //CYGCFG_FS_FAT_ROBUST


    /*****************************************************************************************/
    /* BOOT SECTOR                                                                           */
    /*****************************************************************************************/
    memset(buffer, 0, BPB_BytsPerSec);

    SET_DWORD(buffer,           0x009000EB,     0x00); //!BS_jmpBoot    : Jump instruction boot code
    sprintf( str, "MSWIN4.1");
    SET_BYTES(buffer, str                 ,  8, 0x03); //!BS_OEMName    : "MSWIN4.1" is a default value
    SET_WORD (buffer, BPB_BytsPerSec      ,     0x0B); //!BPB_BytsPerSec: 512 for maximum compatibility
    SET_BYTE (buffer, BPB_SecPerClus      ,     0x0D); //!BPB_SecPerClus:
    SET_WORD (buffer, FAT16_BPB_RSVDSECCNT,     0x0E); //!BPB_RsvdSecCnt: typically 0 for FAT16 volumes
    SET_BYTE (buffer, BPB_NUMFATS         ,     0x10); //!BPB_NumFATs   : Must be 2
    SET_WORD (buffer, BPB_RootEntCnt      ,     0x11); //!BPB_RootEntCnt: for compatibility it is set to cluster size
    if(BPB_TotSec <= 0xFFFF)
    {
    SET_WORD (buffer, BPB_TotSec          ,     0x13); //!BPB_TotSec16  : 0 for FAT32 volumes
    SET_DWORD(buffer,           0x00000000,     0x20); //!BPB_TotSec32  :
    }
    else
    {
    SET_WORD (buffer,               0x0000,     0x13); //!BPB_TotSec16  : 0 for FAT32 volumes
    SET_DWORD(buffer, BPB_TotSec          ,     0x20); //!BPB_TotSec32  :
    }
    SET_BYTE (buffer,                 0xF8,     0x15); //!BPB_Media     : 0xF8 for non-removable media
    SET_WORD (buffer, BPB_FATSz           ,     0x16); //!BPB_FATSz16   : 0 for FAT32 volumes
    SET_WORD (buffer,               0x003C,     0x18); //!BPB_SecPerTrk : TODO
    SET_WORD (buffer,               0x0004,     0x1A); //!BPB_NumHeads  : TODO
    SET_DWORD(buffer,           0x00000000,     0x1C); //!BPB_HiddSec   : TODO


    /*****************************************************************************************/
    /* FAT16 Struct Starting at offset 36                                                    */
    /*****************************************************************************************/
    SET_BYTE (buffer,           0x00,     0x24); //!BPB_DrvNum    : TODO
    SET_BYTE (buffer,           0x00,     0x25); //!BPB_Reserved  : Must be set to 0
    SET_BYTE (buffer,           0x29,     0x26); //!BPB_BootSig   : Must be 0x29
    SET_DWORD(buffer,     0x00000000,     0x27); //!BPB_VolID     : TODO
    sprintf( str, "NO NAME    ");
    SET_BYTES(buffer,            str, 11, 0x2B); //!BPB_VolLab    : default
    sprintf( str, "FAT16   ");
    SET_BYTES(buffer,            str,  8, 0x36); //!BPB_FilSysType:
    SET_DWORD(buffer,     0xAA550000,     0x1FC); //!FSI_TrailSig (+0x1FC):

    // Writing Boot Sector
    len      = 1;
    n_sector = first_sector + 0;
    err = cyg_io_bwrite( dev, (void*)buffer, &len, n_sector);

    free(buffer);

    FAT_DEBUG(" PartInfo:\n \
                    BPB_TotSec = %d sectors\n \
                    BytsPerSec = %d bytes\n \
                    SecPerClus = %d sectors\n \
                    NbRsvdSec  = %d sectors\n \
                    FATSize    = %d sectors\n \
                    RootEntCnt = %d entries(32bytes)\n",
                BPB_TotSec, BPB_BytsPerSec, BPB_SecPerClus,
                FAT16_BPB_RSVDSECCNT, BPB_FATSz, BPB_RootEntCnt );

    return err;
}



/** Writes the FAT32 FileSystem Info.
 *
 * @param disk A pointer to the disk
 * @return ENOERR in any case
 */
int fatfs_get_fat32_fsinfo(fatfs_disk_t *disk,
                           cyg_uint32   *total_clusters,
                           cyg_uint32   *free_clusters)
{
    cyg_uint8 data[0x5A];
    cyg_uint32 len;
    cyg_uint32 freecount_clust;  // 0x1E8
    cyg_uint32 fbr_bytes_per_sec;
    cyg_uint32 fbr_sec_per_fat;
    cyg_uint32 fbr_fs_info_sec;
    int err = ENOERR;

    len = 0x5A;
    err = disk_read(disk, (void*)data, &len, 0);
    if (err != ENOERR)
        return err;

    GET_WORD(data,  fbr_bytes_per_sec,  0x0B);
    GET_WORD(data,  fbr_sec_per_fat,    0x16);

    // This is a quick check for FAT12/16 or FAT32 boot record.
    // The sec_per_fat field must be 0 on FAT32, since this
    // field plays a crucial role in detection of the FAT type
    // (12,16,32) it is quite safe to make this assumption.
    if (0 == fbr_sec_per_fat)
    {
         int FSInfo_address;

         GET_WORD(data,  fbr_fs_info_sec,    0x30);

         // Read the FAT32 FSInfo Sector if we're this is a FAT32
         FSInfo_address = fbr_bytes_per_sec * fbr_fs_info_sec;
         len = 0x8;
         err = disk_read(disk, (void*)data, &len, FSInfo_address+0x1E8);
         if (err != ENOERR)
            return err;

         GET_DWORD(data, freecount_clust,0x0);

         *total_clusters = disk->fat_tbl_nents - 2;
         *free_clusters  = freecount_clust;
      }
      else
      {
         err = EINVAL;
      }

    return err;
}


/** Writes the FAT32 FileSystem Info.
 *
 * @param disk A pointer to the disk
 * @return ENOERR in any case
 */
int fatfs_write_fat32_fsinfo(fatfs_disk_t   *disk,
                             cyg_uint32     freecount_clust,
                             cyg_uint32     nextfree_clust)
{
    cyg_uint8 data[0x5A];
    cyg_uint32 len;
    cyg_uint32 fbr_bytes_per_sec;
    cyg_uint32 fbr_sec_per_fat;
    cyg_uint32 fbr_fs_info_sec;
    int err;

    len = 0x5A;
    err = disk_read(disk, (void*)data, &len, 0);
    if (err != ENOERR)
        return err;

    GET_WORD(data,  fbr_bytes_per_sec,  0x0B);
    GET_WORD(data,  fbr_sec_per_fat,    0x16);

    // This is a quick check for FAT12/16 or FAT32 boot record.
    // The sec_per_fat field must be 0 on FAT32, since this
    // field plays a crucial role in detection of the FAT type
    // (12,16,32) it is quite safe to make this assumption.
    if (0 == fbr_sec_per_fat)
    {
      int FSInfo_address;

      GET_WORD(data,  fbr_fs_info_sec,    0x30);

      // Read the FAT32 FSInfo Sector if we're this is a FAT32
      FSInfo_address = fbr_bytes_per_sec * fbr_fs_info_sec;

      SET_DWORD(data, freecount_clust,0x0);
      if(nextfree_clust != 0)
        SET_DWORD(data, nextfree_clust, 0x4);

      len = 0x8;
      err = disk_write(disk, (void*)data, &len, FSInfo_address+0x1E8 );
    }

    return ENOERR;
}

// -------------------------------------------------------------------------
// set_tentry_type()
// Sets the type of FAT table entry.

static void
set_tentry_type(fatfs_disk_t *disk, cyg_uint32 *entry, cyg_uint32 type)
{
    switch (disk->fat_type)
    {
        case FATFS_FAT12:
            switch (type)
            {
                case TENTRY_FREE:     *entry = 0x0000; return;
                case TENTRY_LAST:     *entry = 0x0FF8; return;
                case TENTRY_RESERVED: *entry = 0x0FF0; return;
                case TENTRY_BAD:      *entry = 0x0FF7; return;
                default:
                    CYG_ASSERT(false, "Unknown tentry type");
            }
            break;

        case FATFS_FAT16:
            switch (type)
            {
                case TENTRY_FREE:     *entry = 0x0000; return;
                case TENTRY_LAST:     *entry = 0xFFF8; return;
                case TENTRY_RESERVED: *entry = 0xFFF0; return;
                case TENTRY_BAD:      *entry = 0xFFF7; return;
                default:
                    CYG_ASSERT(false, "Unknown tentry type");
            }
            break;

        case FATFS_FAT32:
            switch (type)
            {
                case TENTRY_FREE:     *entry = 0x00000000; return;
                case TENTRY_LAST:     *entry = 0x0FFFFFF8; return;
                case TENTRY_RESERVED: *entry = 0x0FFFFFF0; return;
                case TENTRY_BAD:      *entry = 0x0FFFFFF7; return;
                default:
                    CYG_ASSERT(false, "Unknown tentry type");
            }
            break;

        default:
            CYG_ASSERT(false, "Unknown FAT type");
    }
}

static cyg_uint8 *erase_cluster_data = NULL;

void
init_erase_function(void)
{
    if (erase_cluster_data) {
        free(erase_cluster_data);
        erase_cluster_data = calloc(BYTES_PER_CLUSTER_MIN, sizeof(cyg_uint8));
    }
}

static int
erase_cluster(fatfs_disk_t *disk, cyg_uint32 cluster)
{
    cyg_uint32 pos     = 0;
    cyg_uint32 len_tot = 32 * (disk->cluster_size >> 5);
    cyg_uint32 len_to_write = BYTES_PER_CLUSTER_MIN;
    int err;

    do {
        if (erase_cluster_data != NULL) {
            err = disk_cluster_write(disk, (void*)erase_cluster_data,
                    &len_to_write, cluster, pos);
        } else {
            err = ENOERR;
        }
        pos += BYTES_PER_CLUSTER_MIN;
        if (err != ENOERR)
            return err;
     } while(pos < len_tot);

    CYG_TRACE1(TCL, "erasing cluster=%d", cluster);

    return ENOERR;
}

// -------------------------------------------------------------------------
// mark_cluster()
// Marks cluster (sets the cluster's FAT table entry to given type).

static int
mark_cluster(fatfs_disk_t *disk, cyg_uint32 cluster, cyg_uint32 type)
{
    cyg_uint32 tentry;

    set_tentry_type(disk, &tentry, type);
    return write_tentry(disk, cluster, &tentry);
}

// -------------------------------------------------------------------------
// find_next_free_cluster()
// Finds first free cluster starting from given cluster.
// If none is available free_cluster is set to 0.
// If CO_MARK_LAST is set in opts the found cluster is marked as LAST.
// If CO_ERASE_NEW is set in opts the found cluster is erased.

int
find_next_free_cluster(fatfs_disk_t   *disk,
                       cyg_uint32      start_cluster,
                       cyg_uint32     *free_cluster,
                       cluster_opts_t  opts)
{
    cyg_uint32 c, tentry;
    int        err;


    if (start_cluster == 0 )
    {
      /*--------------------------------------------------------*/
      /* if the aim of this call is to browse the cluster table */
      /* from, cluster 0 we gonna use an optimized function     */
      /*--------------------------------------------------------*/
      c = 1;
      *free_cluster = 0;

      CYG_TRACE1(TCL, "starting at cluster=%d", c);

      // Search from the starting cluster to the end of FAT and
      // from start of FAT to the starting cluster
      read_n_tentry_init(disk);

      while (c != start_cluster)
      {
         // Check for end of FAT
         if (c >= disk->fat_tbl_nents)
         {
            c = 2;
            if (c >= start_cluster)
                break;
         }

         err = read_n_tentry(disk, c, &tentry);

         if (err != ENOERR)
            return err;

         if (TENTRY_FREE == get_tentry_type(disk, tentry))
         {
            CYG_TRACE1(TCL, "found free cluster=%d", c);
            *free_cluster = c;

            if (opts & CO_MARK_LAST)
                err = mark_cluster(disk, c, TENTRY_LAST);

            if ((err == ENOERR) && (opts & CO_ERASE_NEW))
                err = erase_cluster(disk, c);

            return err;
        }
        c++;
      }//while
    }//if (start_cluster == 0 )
    else
    {
      /*--------------------------------------------------------*/
      /* if the aim of this call is to browse few clusters      */
      /* twe don't need to use an optimized function            */
      /*--------------------------------------------------------*/
      if (start_cluster < 2) c = 2;
      else                   c = start_cluster + 1;

      *free_cluster = 0;

      CYG_TRACE1(TCL, "starting at cluster=%d", c);

      // Search from the starting cluster to the end of FAT and
      // from start of FAT to the starting cluster
      while (c != start_cluster)
      {
         // Check for end of FAT
         if (c >= disk->fat_tbl_nents)
         {
            c = 2;
            if (c >= start_cluster)
                break;
         }

         err = read_tentry(disk, c, &tentry);

         if (err != ENOERR)
            return err;

         if (TENTRY_FREE == get_tentry_type(disk, tentry))
         {
            CYG_TRACE1(TCL, "found free cluster=%d", c);

            *free_cluster = c;

            if (opts & CO_MARK_LAST)
                err = mark_cluster(disk, c, TENTRY_LAST);

            if ((err == ENOERR) && (opts & CO_ERASE_NEW))
                err = erase_cluster(disk, c);

            return err;
         }
         c++;
      }//while
   }//else

    // No free clusters found
    FAT_WARNING("!!! no free clusters found\n");

    return ENOSPC;
}


// -------------------------------------------------------------------------
// free_cluster_chain()
// Marks all clusters FREE from given cluster to the last cluster in chain.

static int
free_cluster_chain(fatfs_disk_t *disk, cyg_uint32 start_cluster)
{
    cyg_uint32 c, next_c, tentry;
    bool       last;
    int        err;

    CYG_TRACE1(TCL, "start cluster=%d", start_cluster);

    c = next_c = start_cluster;
    last = false;
    while (!last)
    {
        err = read_tentry(disk, c, &tentry);
        if (err != ENOERR)
            return err;

        switch (get_tentry_type(disk, tentry))
        {
            case TENTRY_LAST:
                // Last cluster in chain
                last = true;
                break;
            case TENTRY_REGULAR:
                // Get next cluster in chain
                next_c = get_tentry_next_cluster(disk, tentry);
                break;
            default:
                FAT_WARNING("!!! inconsistant FAT tentry=%x c=%d\n",
                                tentry, c);
                return EIO;
        }

        // Set the current tentry to FREE
        set_tentry_type(disk, &tentry, TENTRY_FREE);
        err = write_tentry(disk, c, &tentry);
        if (err != ENOERR)
            return err;

        // Next cluster in chain
        c = next_c;
    }

    CYG_TRACE1(TCL, "last cluster=%d", c);

    return ENOERR;
}

/** Browse root dir to set the volume label.
 *
 * @param disk   A pointer to the disk
 * @param buf    A pointer to the volume id name buffer

 * @return EEOF if no volume id is set
 */
int fatfs_set_label( fatfs_disk_t   *disk,
                     char           *buf)
{
    fat_raw_dir_entry_t raw_cdentry;
    fatfs_data_pos_t    pos;
    int                 err = ENOERR;

    err = fatfs_find_label(disk, &raw_cdentry, &pos);

    /*-----------------------------------------------------------------------*/
    /* A volume label entry already exist                                    */
    /*-----------------------------------------------------------------------*/
    if(err == ENOERR)
    {
#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( DENTRY_IS_LNDIR(&raw_cdentry) ){
            sprintf( raw_cdentry.lname, "%s", buf );
		    FAT_INFO("Volume label(ln) rw : %s\n", buf);
        }
        else
#endif
        {
            int i_char, len;
            len = strlen(buf);

            for(i_char=0;i_char<len;i_char++){
                if(buf[i_char] > 0x7F) buf[i_char] = 0x2A; // 0x2A = '*'
            }

            if(len <= 8){
                snprintf( raw_cdentry.name, len+1, "%s",buf);
                snprintf( raw_cdentry.ext , SEXT_MAX_CHAR, "   ");
            }else{
                snprintf( raw_cdentry.name, SNAME_MAX_CHAR, "%s", &(buf[0]));
                snprintf( raw_cdentry.ext , SEXT_MAX_CHAR , "%s", &(buf[8]) );
            }
            FAT_INFO("Volume label(sh) rw : %s\n", buf);
        }

        err = write_raw_dentry(disk, &pos, &raw_cdentry);
    }
    /*-----------------------------------------------------------------------*/
    /* No volume label entry found                                           */
    /*-----------------------------------------------------------------------*/
    else if (err == EEOF)
    {
        int i_char, len;
        time_t time;
        len = strlen(buf);

        err = ENOERR; // reset error code for upper layers (ENOSPC is valid)

        for(i_char=0;i_char<len;i_char++){
            if(buf[i_char] > 0x7F) buf[i_char] = 0x2A; // 0x2A = '*'
        }

        FAT_INFO("Volume label(sh) w : %s\n", buf);

        if(len <= 8){
            snprintf( raw_cdentry.name, len+1, "%s",buf);
            snprintf( raw_cdentry.ext, SEXT_MAX_CHAR, "   ");
        }else{
            snprintf( raw_cdentry.name, SNAME_MAX_CHAR, "%s", &(buf[0]));
            snprintf( raw_cdentry.ext , SEXT_MAX_CHAR , "%s", &(buf[8]) );
        }
        raw_cdentry.attr = S_FATFS_VOLUME;
        raw_cdentry.nb_entries = 1 ;

        time = cyg_timestamp();

        date_unix2dos(time, &raw_cdentry.crt_time, &raw_cdentry.crt_date);
        date_unix2dos(time, NULL                 , &raw_cdentry.acc_date);
        date_unix2dos(time, &raw_cdentry.wrt_time, &raw_cdentry.wrt_date);

        raw_cdentry.crt_sec_100 = 0;
        raw_cdentry.size        = 0;
        raw_cdentry.nt_reserved = 0;
        raw_cdentry.cluster     = 0;
        raw_cdentry.cluster_HI  = 0;
        raw_cdentry.wrt_time    = 0;
        raw_cdentry.wrt_date    = 0;

        err = write_raw_dentry(disk, &pos, &raw_cdentry);
    }
    /*-----------------------------------------------------------------------*/
    /* No more space left to write volume label entry                        */
    /*-----------------------------------------------------------------------*/
    else if (err == ENOSPC)
    {

    }

    return err;
}


#ifdef CYGCFG_FS_FAT_USE_VFAT
/** Translate an UTF8 string into an UTF16 string.
 *
 *  An UTF8 char can be encodded on 4 bytes with 21 significant bits
 *  An UTF16 char can be encodded on 2 words with 20 significant bits
 *
 * @param tab_utf8 A pointer to the utf8 string
 * @param tab_utf16 A pointer to the utf16 string
 * @param nb_bytes_utf8 number of utf8 characters
 * @return the number of utf16 bytes or 0 upon error
 */
cyg_uint32 xlate_utf8_to_utf16( cyg_uint8  *tab_utf8,
                                cyg_uint16 *tab_utf16,
                                cyg_uint32 nb_bytes_utf8 )
{
    cyg_uint32      nb_bytes_utf16 = 0;
    const cyg_uint8 *ip;

    ip = tab_utf8;

    while (*ip && nb_bytes_utf8 > 0)
    {
        if( *ip <= 0x7F )
        {
            tab_utf16[nb_bytes_utf16] = (cyg_uint8)*ip;
            nb_bytes_utf8 -= 1;
            ip = &(ip[1]);
        }
        else if(  ( (*ip   & 0xE0) == UTF8_2_BYTES_MASK )
                &&( (ip[1] & 0xC0) == 0x80 ) )
        {// Then we need 2 bytes
            tab_utf16[nb_bytes_utf16]   = ( (ip[0] & 0x1F)<<6) | ( ip[1] & 0x3F );
            nb_bytes_utf8 -= 2;
            ip = &(ip[2]);
        }
        else if(  ( (*ip   & 0xF0) == UTF8_3_BYTES_MASK )
                &&( (ip[1] & 0xC0) == 0x80 )
                &&( (ip[2] & 0xC0) == 0x80 ) )
        {// Then we need 3 bytes
            tab_utf16[nb_bytes_utf16] = ( (ip[0] & 0x0F)<<12) | ( (ip[1] & 0x3F)<<6) | ( ip[2] & 0x3F );
            nb_bytes_utf8 -= 3;
            ip = &(ip[3]);
        }
        else if(  ( (*ip   & 0xF8) == UTF8_4_BYTES_MASK )
                &&( (ip[1] & 0xC0) == 0x80 )
                &&( (ip[2] & 0xC0) == 0x80 )
                &&( (ip[3] & 0xC0) == 0x80 ))
        {// Then we need 3 bytes
            cyg_uint32 ucs4 = ( (ip[0]&0x07)<<18) | ( (ip[1]&0x3F)<<12) | ( (ip[2]&0x3F)<< 6) | (ip[3]&0x3F);
            tab_utf16[nb_bytes_utf16] = ((ucs4>>10)&0x3FF) | UTF16_MASK_SURR_HIG;
            nb_bytes_utf16++;

            if(nb_bytes_utf16 > UTF16_LNAME_MAX_CHAR )
            {   // error
                return 0;
            }

            tab_utf16[nb_bytes_utf16] = (ucs4&0x3FF) | UTF16_MASK_SURR_LOW;
            nb_bytes_utf8 -= 4;
            ip = &(ip[4]);
        }
        else
        {   // error
            return 0;
        }

        nb_bytes_utf16++;

        if(nb_bytes_utf16 > UTF16_LNAME_MAX_CHAR )
        {   // error
            return 0;
        }
    }
    tab_utf16[nb_bytes_utf16] = 0;


    return nb_bytes_utf16;
}
#endif


#ifdef CYGCFG_FS_FAT_USE_VFAT
/** Writes dir entry to disk even if they are between 2 clusters(can't be more than 2).
 *
 * @param disk    A pointer to the disk
 * @param pos     A pointer to the structure describing the current position
 * @param dentry  A pointer to source buffer
 * @param pt_len  Contains the lentgh of the buffer
 * @return ENOERR upon success, or an error code
 */

static int write_raw_long_dentry( fatfs_disk_t     *disk,
                                 fatfs_data_pos_t  *pos,
                                 cyg_uint8         *pt_data,
                                 cyg_uint32        *pt_len ,
                                 fat_raw_dir_entry_t *dentry )
{
   cyg_uint16 next_pos;
   cyg_uint32 len_to_write;   //can't be more than 21*0x20=0x2A0 but 32bits is due to disk_cluster_write
   cyg_uint16 offset;

   fatfs_data_pos_t  pos_tmp;

   int err              = ENOERR;

   pos_tmp.cluster      = pos->cluster;
   pos_tmp.cluster_snum = pos->cluster_snum;
   pos_tmp.cluster_pos  = pos->cluster_pos;

   next_pos             = pos_tmp.cluster_pos + *pt_len;
   len_to_write         = *pt_len;
   offset               = 0;

    if ((pos->cluster!=0) && (next_pos > disk->cluster_size) )
    {
        // if there is too much entries to fit in the current cluster
        // Extend the chain for one cluster
        len_to_write = disk->cluster_size - pos_tmp.cluster_pos;

        // we are not writting to the FAT12/16 root directory
        err = disk_cluster_write(disk, (void*)pt_data, &len_to_write, pos_tmp.cluster, pos_tmp.cluster_pos);

        if (err != ENOERR)
            return err;

        //First we try to get the next cluster
        // If there is no next cluster, we gonna add one to the end of chain
        err = get_next_cluster(disk, &pos_tmp, CO_EXTEND);

        if (err != ENOERR)
                return err;

        offset       = len_to_write;
        len_to_write = next_pos - disk->cluster_size;
        next_pos     = len_to_write;
    }
    else if((pos->cluster==0) && (next_pos > disk->fat_root_dir_size))
    {
        // The entry can't be written in the root directory
        err = ENOSPC;
        return err;
    }


   // Check if we are writting to the FAT12/16 root directory
   if (0 == pos_tmp.cluster)
      err = disk_write(disk, (void*)&(pt_data[offset]), &len_to_write, disk->fat_root_dir_pos + pos_tmp.cluster_pos);
   else
      err = disk_cluster_write(disk, (void*)&(pt_data[offset]), &len_to_write, pos_tmp.cluster, pos_tmp.cluster_pos);

   return err;
}
#endif // CYGCFG_FS_FAT_USE_VFAT

/** Writes dir entry from disk.
 *
 * @param disk A pointer to the disk
 * @param pos ---
 * @param dentry A pointer to long dir name structure
 * @return ENOERR upon success, or an error code
 */

int write_raw_dentry(fatfs_disk_t        *disk,
                 fatfs_data_pos_t    *pos,
                 fat_raw_dir_entry_t *dentry)
{
    cyg_uint32 len;
    int err;

    if( DENTRY_IS_DELETED(dentry) )
    {
    FAT_INFO("entry info cl=0x%04x-%04x, name=DELETED, size=%d\tpos(0x%08x,0x%08x,0x%08x)\n",
                dentry->cluster_HI, dentry->cluster,
                dentry->size,
                pos->cluster, pos->cluster_snum, pos->cluster_pos);
    }else{
    FAT_INFO("entry info cl=0x%04x-%04x, name=%s, ext=%s,size=%d\tpos(0x%08x,0x%08x,0x%08x)\n",
                dentry->cluster_HI, dentry->cluster,
                dentry->name, dentry->ext, dentry->size,
                pos->cluster, pos->cluster_snum, pos->cluster_pos);
    }

#ifdef CYGCFG_FS_FAT_USE_VFAT

    if(dentry->nb_entries == 0)
    {   // it means that the utf8 string is wrong formatted
        return EINVAL;
    }

   if( DENTRY_IS_LNDIR(dentry) )
   {
      //------------------------------------------------
      // Writes a Long Dir Entry
      // Now, we know it's a long dir entry, we need to use
      // a larger memory buffer depending on the nbr_of_le
      //------------------------------------------------
      cyg_uint8  i_le;           // index of long entries
      cyg_uint8  i_byte;         // index of bytes within a long entry
      cyg_uint8  ldir_chksum;    // short entry chksum
      char       *pt_entryblock; // pointer to a single long entry
      cyg_uint16 *utf16_name   ; // pointer to a single long entry

      len = (dentry->nb_entries)*DENTRY_SIZE;
      memset(disk->wr_data, 0, len);
      utf16_name = (cyg_uint16*)(dentry->lname);

      // sets the LDIR_ord
     if(DENTRY_IS_DELETED(dentry))
     {
         for(i_le = 0; i_le < dentry->nb_entries ; i_le++)
            SET_BYTE(&(disk->wr_data[i_le*DENTRY_SIZE]), 0xE5, 0x00); // mark as deleted and nothing more
     }
     else
     {
        // the first thing to do is to fill the short entry
        // in order to be able to compute the checksum
        pt_entryblock = &(disk->wr_data[(dentry->nb_entries-1)*DENTRY_SIZE ]);
        SET_BYTES(pt_entryblock, dentry->name,     8, 0x00);
        SET_BYTES(pt_entryblock, dentry->ext,      3, 0x08);
        SET_BYTE(pt_entryblock,  dentry->attr,        0x0B);
        SET_BYTE(pt_entryblock,  dentry->nt_reserved, 0x0C);
        SET_BYTE(pt_entryblock,  dentry->crt_sec_100, 0x0D);
        SET_WORD(pt_entryblock,  dentry->crt_time,    0x0E);
        SET_WORD(pt_entryblock,  dentry->crt_date,    0x10);
        SET_WORD(pt_entryblock,  dentry->acc_date,    0x12);
        SET_WORD(pt_entryblock,  dentry->cluster_HI,  0x14);
        SET_WORD(pt_entryblock,  dentry->wrt_time,    0x16);
        SET_WORD(pt_entryblock,  dentry->wrt_date,    0x18);
        SET_WORD(pt_entryblock,  dentry->cluster,     0x1A);
        SET_DWORD(pt_entryblock, dentry->size,        0x1C);

        // computation of the short entry checksum
        ldir_chksum = ChkSum( &(disk->wr_data[(dentry->nb_entries-1)*DENTRY_SIZE ]) );

        for(i_le = 1; i_le < dentry->nb_entries ; i_le++)
        {
            cyg_uint16 name_offset;

            pt_entryblock = &(disk->wr_data[((dentry->nb_entries-1)-i_le)*DENTRY_SIZE]);

            if( i_le == (dentry->nb_entries-1) )
               SET_BYTE(pt_entryblock, (LAST_LONG_ENTRY | i_le ), 0x00);
            else
               SET_BYTE(pt_entryblock, i_le, 0x00);

            SET_BYTE(pt_entryblock, 0x0F,          0x0B);// sets the LDIR_attr
            SET_BYTE(pt_entryblock, 0x00,          0x0C);// sets the LDIR_type
            SET_BYTE(pt_entryblock, ldir_chksum,   0x0D);// sets the LDIR_chksum
            SET_WORD(pt_entryblock, 0x0000,        0x1A);// sets the LDIR_fstcluslo

            name_offset = (i_le-1)*CHAR_PER_LE;

            SET_WORD(pt_entryblock, utf16_name[0+name_offset], 0x01);
            SET_WORD(pt_entryblock, utf16_name[1+name_offset], 0x03);
            SET_WORD(pt_entryblock, utf16_name[2+name_offset], 0x05);
            SET_WORD(pt_entryblock, utf16_name[3+name_offset], 0x07);
            SET_WORD(pt_entryblock, utf16_name[4+name_offset], 0x09);
            SET_WORD(pt_entryblock, utf16_name[5+name_offset], 0x0E);
            SET_WORD(pt_entryblock, utf16_name[6+name_offset], 0x10);
            SET_WORD(pt_entryblock, utf16_name[7+name_offset], 0x12);
            SET_WORD(pt_entryblock, utf16_name[8+name_offset], 0x14);
            SET_WORD(pt_entryblock, utf16_name[9+name_offset], 0x16);
            SET_WORD(pt_entryblock, utf16_name[10+name_offset],0x18);
            SET_WORD(pt_entryblock, utf16_name[11+name_offset],0x1C);
            SET_WORD(pt_entryblock, utf16_name[12+name_offset],0x1E);
      }// end for(i_le = 1; i_le < nbr_of_le ; i_le++)

      cyg_uint8 nb_spare_char, offset;
      offset = 0 ;
      nb_spare_char = dentry->nb_utf16_chars % 13 ;// strlen returns +1 char

      if( nb_spare_char != 0 )
      {
         switch(nb_spare_char){
            case  1:{offset=0x03;break;}
            case  2:{offset=0x05;break;}
            case  3:{offset=0x07;break;}
            case  4:{offset=0x09;break;}
            case  5:{offset=0x0E;break;}
            case  6:{offset=0x10;break;}
            case  7:{offset=0x12;break;}
            case  8:{offset=0x14;break;}
            case  9:{offset=0x16;break;}
            case 10:{offset=0x18;break;}
            case 11:{offset=0x1C;break;}
            case 12:{offset=0x1E;break;}
            default:{FAT_ERROR("nb_spare_char=%d\n", nb_spare_char);}// it's impossible to reach
         }

         // filename termination on entry 0
         pt_entryblock = &(disk->wr_data[0]);
         SET_WORD(pt_entryblock,  0x0000,    offset); // it's the end of filename
         for(i_byte=offset+1; i_byte<0x20; i_byte++)
         {
            if( (i_byte==0x03)||(i_byte==0x05)||(i_byte==0x07)||(i_byte==0x09)
              ||(i_byte==0x0E)||(i_byte==0x10)||(i_byte==0x12)||(i_byte==0x14)
              ||(i_byte==0x16)||(i_byte==0x18)||(i_byte==0x1C)||(i_byte==0x1E) )
            {
               SET_WORD(pt_entryblock,  0xFFFF,    i_byte); // padding until the '\0'
            }
         }// end for
      } //end - if( nb_spare_char != 0 )
   } // end else

   err = write_raw_long_dentry(disk, pos, (void*)disk->wr_data, &len, dentry);

   if (err != ENOERR)
       return err;

   }
   else
#endif //CYGCFG_FS_FAT_USE_VFAT
   {
      //-------------------------------------------
      // Writes a Short Dir Entry
      //-------------------------------------------
      SET_BYTES(disk->wr_data, dentry->name,     8, 0x00);
      SET_BYTES(disk->wr_data, dentry->ext,      3, 0x08);
      SET_BYTE(disk->wr_data,  dentry->attr,        0x0B);
      SET_BYTE(disk->wr_data,  dentry->nt_reserved, 0x0C);
      SET_BYTE(disk->wr_data,  dentry->crt_sec_100, 0x0D);
      SET_WORD(disk->wr_data,  dentry->crt_time,    0x0E);
      SET_WORD(disk->wr_data,  dentry->crt_date,    0x10);
      SET_WORD(disk->wr_data,  dentry->acc_date,    0x12);
      SET_WORD(disk->wr_data,  dentry->cluster_HI,  0x14);
      SET_WORD(disk->wr_data,  dentry->wrt_time,    0x16);
      SET_WORD(disk->wr_data,  dentry->wrt_date,    0x18);
      SET_WORD(disk->wr_data,  dentry->cluster,     0x1A);
      SET_DWORD(disk->wr_data, dentry->size,        0x1C);

      len = DENTRY_SIZE;

      // Check if we are writting to the FAT12/16 root directory
      if (0 == pos->cluster)
         err = disk_write(disk, (void*)disk->wr_data, &len, disk->fat_root_dir_pos + pos->cluster_pos);
      else
         err = disk_cluster_write(disk, (void*)disk->wr_data, &len, pos->cluster, pos->cluster_pos);

      if (err != ENOERR)
         return err;
   }

#if TDE
    print_raw_dentry(dentry);
#endif

    return ENOERR;
}

// -------------------------------------------------------------------------
// raw_dentry_set_deleted()
// Sets the dentry filename first char to 0xE5 (ie deleted).

static __inline__ void
raw_dentry_set_deleted(fatfs_disk_t *disk, fat_raw_dir_entry_t *dentry)
{
   dentry->name[0] = 0xE5;
}


// -------------------------------------------------------------------------
// set_raw_dentry_filename()
// Sets the filename of given dir entry.
static void
set_raw_dentry_filename(fat_raw_dir_entry_t *raw_dentry,
                        const char          *name,
                        int                  namelen,
                        fatfs_disk_t        *disk,
                        fatfs_dir_entry_t   *dentry)
{
    int         i, nidx, begext;
    const char *cname;
    char       *cptr;

    // Special case check
    if ('.' == name[0])
    {
        if ('\0' == name[1])
        {
            strcpy(raw_dentry->name, ".       ");
            strcpy(raw_dentry->ext,  "   ");
            return;
        }
        else if ('.' == name[1] && '\0' == name[2])
        {
            strcpy(raw_dentry->name, "..      ");
            strcpy(raw_dentry->ext,  "   ");
            return;
        }
    }

#ifdef CYGCFG_FS_FAT_USE_VFAT
   cname = name;
   cptr  = raw_dentry->lname;

   while ( *cname != '\0' )
   {
        *cptr++ = *cname++;
   }
   *cptr = '\0';

   CYG_TRACE1(TDE, "vfat name='%s' ", name);

#endif //CYGCFG_FS_FAT_USE_VFAT

    namelen = strlen(name);

    if (0 == namelen){
        namelen = 9999;
    }

    begext = 8;
    for (i = 0; i < namelen; i++)
    {
        if(name[i] == '.') begext = i;
    }

    nidx  = 0;
    cname = name;
    cptr  = raw_dentry->name;
    for (i = 0; (i<begext)&&(i<8); i++)
    {
        if (   (*cname != '\0')
            && (nidx++ < begext) )
        {
            if (*cname < 0x80){
                *cptr++ = toupper(*cname++);
            }else if(*cname != '.'){
                *cptr++ = '_';
                cname++;
            }else{
                cname++;
            }
        }
        else
            *cptr++ = ' ';
    }

    *cptr = '\0';
    cname = &name[begext];
    while (*cname != '.' && *cname != '\0' && nidx++ < namelen)
        cname++;

    if ('.' == *cname && nidx++ < namelen)
        cname++;

    cptr = raw_dentry->ext;
    for (i = 0; i < 3; i++)
    {
        if (*cname != '.' && *cname != '\0' && nidx++ < namelen)
        {
            if (*cname < 0x80)
                *cptr++ = toupper(*cname++);
            else
            {
            *cptr++ = '_';
            cname++;
            }
        }
        else
        {
            *cptr++ = ' ';
        }
    }
    *cptr = '\0';

#ifdef CYGCFG_FS_FAT_USE_VFAT
    {
      fat_raw_dir_entry_t  raw_cdentry;
      fatfs_data_pos_t     pos;
      char                 buf_name[SNAME_MAX_CHAR];
      int                  err   = ENOERR;
      int                  nb_files = 0; // count the number of files with the same 8 first chars

      pos.cluster      = dentry->parent_cluster;
      pos.cluster_pos  = 0;
      pos.cluster_snum = 0;

      while (true)
      {
         err = read_next_raw_dentry(disk, &pos, &raw_cdentry);

         if ( (err != ENOERR) || DENTRY_IS_ZERO(&raw_cdentry) )
            break;

         if ( !DENTRY_IS_DELETED(&raw_cdentry) )
         {
            if ( (0 == strncmp(raw_dentry->name, raw_cdentry.name, SNAME_MAX_CHAR))
               &&(0 == strncmp(raw_dentry->ext , raw_cdentry.ext , SEXT_MAX_CHAR))
               &&(0 != strcmp (raw_dentry->lname , raw_cdentry.lname)) )
            {
               nb_files ++;

               memcpy(buf_name,raw_dentry->name,SNAME_MAX_CHAR);

               if     (nb_files < 10)     raw_dentry->name[6] = '\0';
               else if(nb_files < 100)    raw_dentry->name[5] = '\0';
               else if(nb_files < 1000)   raw_dentry->name[4] = '\0';
               else if(nb_files < 10000)  raw_dentry->name[3] = '\0';
               else if(nb_files < 100000) raw_dentry->name[2] = '\0';
               else                       break; // Error Case

               sprintf(buf_name, "%s~%d", raw_dentry->name, nb_files);
               memcpy(raw_dentry->name, buf_name, SNAME_MAX_CHAR);

               //Let's recount the number of files with raw_dentry->name
               pos.cluster      = dentry->parent_cluster;
               pos.cluster_pos  = 0;
               pos.cluster_snum = 0;
            }
         }

         if ( DENTRY_IS_DELETED(&raw_cdentry) )
            pos.cluster_pos += DENTRY_SIZE;
         else
            err = go_to_next_raw_entry(disk, &pos, &raw_cdentry, CO_NONE);

         if (err != ENOERR)
                break;
      }
   }
#endif
    FAT_INFO("filename='%s' namelen=%d dos name='%s' dos ext='%s'\n",
                    name, namelen, raw_dentry->name, raw_dentry->ext);
}

// -------------------------------------------------------------------------
// get_free_raw_dentry()
// Gets free dir entry slot searching from given position extending the
// directory if needed. If an deleated entry is found it is reused.

static int
get_free_raw_dentry(fatfs_disk_t     *disk,
                    fatfs_data_pos_t *pos,
                    int              namelen)
{
    fat_raw_dir_entry_t raw_dentry;
    fatfs_data_pos_t    cpos;
    int                 err = ENOERR;

    cpos = *pos;

    // If we are reading the root dir on FAT32 we have
    // to correct the position to the root dir cluster
    if (FATFS_FAT32 == disk->fat_type && 0 == cpos.cluster)
        cpos.cluster = disk->fat_root_dir_cluster;

    CYG_TRACE3(TDE, "cluster=%d snum=%d pos=%d",
                    pos->cluster, pos->cluster_snum, pos->cluster_pos);

    while (true)
    {
        // FAT12/16 root dir check
        if (0 == cpos.cluster)
        {
            if (cpos.cluster_pos >= disk->fat_root_dir_size)
                err = ENOSPC;
        }
        else
        {
            // Change cluster if needed
            if (cpos.cluster_pos >= disk->cluster_size)
                err = get_next_cluster(disk, &cpos, CO_EXTEND | CO_ERASE_NEW);
        }

        if (err != ENOERR)
            return err;

        err = read_raw_dentry(disk, &cpos, &raw_dentry);
        if (err == EEOF)
        {   // in this case, it means that the entry is located on several clusters
            // but this cluster is the last of the chain -> it's not a valid entry
            // it could for exemple happen when a disk is formatted
            *pos = cpos;
            return ENOERR;
        }
        else if (err != ENOERR)
            return err;

        if (DENTRY_IS_DELETED(&raw_dentry))
        {
#ifdef CYGCFG_FS_FAT_USE_VFAT
            cyg_uint16        nb_entries_needed;
            fatfs_data_pos_t  cpos_new;

            memcpy(&cpos_new, &cpos, sizeof(fatfs_data_pos_t));
            // determine the number of long entry needed
            nb_entries_needed = namelen / 13;
            if( (namelen % 13) != 0 )
               nb_entries_needed++;// if nb_of_char is a multiple of 13, there is no termination
            nb_entries_needed++;// to count the short entry

            if(nb_entries_needed > nb_consecutive_free_entries( disk, &cpos_new) )
            {
               memcpy(&cpos, &cpos_new, sizeof(fatfs_data_pos_t));
            }
            else
#endif //CYGCFG_FS_FAT_USE_VFAT
            {
               CYG_TRACE3(TDE, "deleted dentry at cluster=%d snum=%d pos=%d",
                            cpos.cluster, cpos.cluster_snum, cpos.cluster_pos);
               *pos = cpos;
               return ENOERR;
            }
        }
        else if (DENTRY_IS_ZERO(&raw_dentry))
        {
            CYG_TRACE3(TDE, "zero dentry at cluster=%d snum=%d pos=%d",
                            cpos.cluster, cpos.cluster_snum, cpos.cluster_pos);

            *pos = cpos;
            return ENOERR;
        }
#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( !DENTRY_IS_DELETED(&raw_dentry) )
         {
            err = go_to_next_raw_entry(disk, &cpos, &raw_dentry, CO_EXTEND | CO_ERASE_NEW);
            if (err != ENOERR)
               return err;
         }
#else
      cpos.cluster_pos += DENTRY_SIZE;
#endif
    }
}


// -------------------------------------------------------------------------
// dentry_to_raw()
// Converts dir entry to raw FAT dir entry.

static void
dentry_to_raw( fatfs_disk_t         *disk,
               fatfs_dir_entry_t    *dentry,
               fat_raw_dir_entry_t  *raw_dentry)
{

#ifdef CYGCFG_FS_FAT_USE_VFAT
    cyg_uint32 nb_bytes_utf16;
    cyg_uint32 nb_bytes_utf8;
#endif //CYGCFG_FS_FAT_USE_VFAT

    set_raw_dentry_filename(raw_dentry, dentry->filename, 0, disk, dentry);

#ifdef CYGCFG_FS_FAT_USE_VFAT
    if(  ( 0 != strncmp("." , raw_dentry->name, 1) )
       &&( 0 != strncmp("..", raw_dentry->name, 2) ) )
    {
        FAT_DEBUG(" name=%s\n", raw_dentry->lname);

        // Sets this raw entry as a long name entry
        raw_dentry->lndattr = 0x0F;

        /*-----------------------------------------------------------------------*/
        /* Determine the number of entries needed                                */
        /*-----------------------------------------------------------------------*/
        nb_bytes_utf8 = strlen(dentry->filename);

        raw_dentry->nb_entries     = nb_bytes_utf8 / 13;

        if( (nb_bytes_utf8 % 13) != 0 )
        {
            raw_dentry->nb_entries++;// if nb_of_char is a multiple of 13, there is no termination
        }
        raw_dentry->nb_entries++;// to count the short entry


        /*-----------------------------------------------------------------------*/
        /* Translate filename from utf8 to utf16                                 */
        /*-----------------------------------------------------------------------*/
        nb_bytes_utf16 = xlate_utf8_to_utf16( (cyg_uint8*)(raw_dentry->lname),
                                              wr_utf16, nb_bytes_utf8 );

        // determine the number of long entry
        raw_dentry->nb_utf16_chars = nb_bytes_utf16;

        if(nb_bytes_utf16 == 0){ // it means that the utf8 string is wrong formatted
            raw_dentry->nb_entries = 0;
            return;
        }else{
            memcpy(raw_dentry->lname, wr_utf16, sizeof(cyg_uint16)*nb_bytes_utf16 );
        }
    }
    else
    {// we're dealing with "." or "..", then a short entry is enough
        raw_dentry->nb_entries = 1 ;
        raw_dentry->lndattr = 0x0;
    }

#else
    raw_dentry->nb_entries = 1 ;
#endif //CYGCFG_FS_FAT_USE_VFAT

    if (__stat_mode_DIR == dentry->mode)
        raw_dentry->attr = S_FATFS_DIR;
    else
        raw_dentry->attr = S_FATFS_ARCHIVE;

#ifdef CYGCFG_FS_FAT_USE_ATTRIBUTES
    raw_dentry->attr = dentry->attrib;
#endif // CYGCFG_FS_FAT_USE_ATTRIBUTES

    date_unix2dos(dentry->ctime, &raw_dentry->crt_time, &raw_dentry->crt_date);
    date_unix2dos(dentry->atime, NULL,                  &raw_dentry->acc_date);
    date_unix2dos(dentry->mtime, &raw_dentry->wrt_time, &raw_dentry->wrt_date);

    raw_dentry->crt_sec_100 = 0; //FIXME
    raw_dentry->size        = dentry->size;
    raw_dentry->nt_reserved = dentry->priv_data;
    raw_dentry->cluster     = dentry->cluster & 0xFFFF;
    raw_dentry->cluster_HI  = dentry->cluster >> 16;
}


// -------------------------------------------------------------------------
// write_data()
// Writes data to given position.

static int
write_data(fatfs_disk_t     *disk,
           void             *data,
           cyg_uint32       *len,
           fatfs_data_pos_t *pos)
{
    cyg_uint8   *buf   = (cyg_uint8 *) data;
    cyg_uint32   size  = *len;
    int          err   = ENOERR;

    CYG_TRACE4(TDO, "len=%d cluster=%d snum=%d pos=%d",
                    *len, pos->cluster, pos->cluster_snum,
                    pos->cluster_pos);

    while (size > 0)
    {
        cyg_uint32 csize;

        // Check if we are still inside current cluster
        if ((pos->cluster!=0)&&(pos->cluster_pos >= disk->cluster_size     ))
        {
            // Get next cluster of file, if at the last
            // cluster try to extend the cluster chain
            err = get_next_cluster(disk, pos, CO_EXTEND);
            if (err != ENOERR)
                goto out;
        }
        else if ((pos->cluster==0)&&(pos->cluster_pos >= disk->fat_root_dir_size))
        {
            err = ENOSPC;
            goto out;
        }

        // Adjust the data chunk size to be read to the cluster boundary
        if (size > (disk->cluster_size - pos->cluster_pos))
            csize = disk->cluster_size - pos->cluster_pos;
        else
            csize = size;

        CYG_TRACE4(TDO, "-- len=%d cluster=%d snum=%d pos=%d",
                        csize, pos->cluster, pos->cluster_snum,
                        pos->cluster_pos);

        err = disk_cluster_write(disk, (void*)buf, &csize,
                                 pos->cluster, pos->cluster_pos);

        if (err != ENOERR)
            goto out;

        // Adjust running variables

        buf              += csize;
        pos->cluster_pos += csize;
        size             -= csize;
    }

out:
    *len -= size;

    CYG_TRACE1(TDO, "total len=%d", *len);



    return err;
}

// -------------------------------------------------------------------------
// init_dir_entry()
// Initializes attributes of a new dir entry.

static void
init_dir_entry(fatfs_dir_entry_t *dentry,
               const char        *name,
               int                namelen,
               mode_t             mode,
               cyg_uint32         parent_cluster,
               cyg_uint32         first_cluster,
               fatfs_data_pos_t  *pos)
{
    if (0 == namelen)
        namelen = 12;

    strncpy(dentry->filename, name, namelen);
    dentry->filename[namelen] = '\0';

    dentry->mode  = mode;

#ifdef CYGCFG_FS_FAT_USE_ATTRIBUTES
    if (S_ISDIR(dentry->mode))
        dentry->attrib = S_FATFS_DIR;
    else
        dentry->attrib = S_FATFS_ARCHIVE;
#endif // CYGCFG_FS_FAT_USE_ATTRIBUTES

    dentry->ctime =
    dentry->atime =
    dentry->mtime = cyg_timestamp();

    dentry->priv_data      = 0;
    dentry->size           = 0;
    dentry->cluster        = first_cluster;
    dentry->parent_cluster = parent_cluster;
    dentry->disk_pos       = *pos;
}

// -------------------------------------------------------------------------
// is_root_dir_entry()
// Check if the given dir entry is the root dir entry.

static __inline__ bool
is_root_dir_entry(fatfs_dir_entry_t *dentry)
{
    return ('\0' == dentry->filename[0] && 0 == dentry->cluster);
}

// -------------------------------------------------------------------------
// fatfs_initpos()
// Initializes position to the start of the given file.

int
fatfs_initpos(fatfs_disk_t      *disk,
              fatfs_dir_entry_t *file,
              fatfs_data_pos_t  *pos)
{
    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);
    CYG_CHECK_DATA_PTRC(pos);

    pos->cluster      = file->cluster;
    pos->cluster_snum = 0;
    pos->cluster_pos  = 0;

    return ENOERR;
}


// -------------------------------------------------------------------------
// fatfs_write_dir_entry()
// Writes dir entry to disk.

int
fatfs_write_dir_entry(fatfs_disk_t *disk, fatfs_dir_entry_t *dentry)
{
    fat_raw_dir_entry_t raw_dentry;
    int                 err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dentry);

    dentry_to_raw(disk, dentry, &raw_dentry);
    err = write_raw_dentry(disk, &dentry->disk_pos, &raw_dentry);
    return err;
}

// -------------------------------------------------------------------------
// fatfs_delete_file()
// Marks dir entry as deleted and frees its cluster chain.

int
fatfs_delete_file(fatfs_disk_t *disk, fatfs_dir_entry_t *file)
{
    fat_raw_dir_entry_t raw_dentry;
    int                 err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);

    if (is_root_dir_entry(file))
        return EINVAL;

    CYG_TRACE1(TDE, "filename='%s'", file->filename);

    dentry_to_raw(disk, file, &raw_dentry);

    // Check if we are about to delete a directory
    if (DENTRY_IS_DIR(&raw_dentry))
    {
        fat_raw_dir_entry_t raw_cdentry;
        fatfs_data_pos_t    pos;
        int                 i = 0;

        fatfs_initpos(disk, file, &pos);

        CYG_TRACE0(TDE, "got directory");

        // Count number of entries in this dir
        while (true)
        {
            err = read_next_raw_dentry(disk, &pos, &raw_cdentry);

            if (EEOF == err)
                break;
            else if (err != ENOERR)
                return err;

#ifdef CYGCFG_FS_FAT_USE_VFAT
           if ( DENTRY_IS_DELETED(&raw_cdentry) )
               pos.cluster_pos += DENTRY_SIZE;
           else
            {
               err = go_to_next_raw_entry(disk, &pos, &raw_cdentry, CO_NONE);

               if (EEOF == err)        break;
               else if (err != ENOERR) return err;
            }
#else
            pos.cluster_pos += DENTRY_SIZE;
#endif

            i++;
        }
        CYG_TRACE1(TDE, "child count=%d", i);

        // Check if the dir is empty (except '.' and '..')
        if (i > 2)
            return EPERM;
    }

    // Free file clusters
    free_cluster_chain(disk, file->cluster);
    raw_dentry_set_deleted(disk, &raw_dentry);
    err = write_raw_dentry(disk, &file->disk_pos, &raw_dentry);
    return err;
}

// -------------------------------------------------------------------------
// fatfs_create_file()
// Creates a new file.

int
fatfs_create_file(fatfs_disk_t      *disk,
                  fatfs_dir_entry_t *dir,
                  const char        *name,
                  int                namelen,
                  fatfs_dir_entry_t *dentry)
{
    fatfs_data_pos_t pos;
    int              err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dir);
    CYG_CHECK_DATA_PTRC(name);
    CYG_CHECK_DATA_PTRC(dentry);

    CYG_TRACE2(TDE, "filename='%s' parent='%s'", name, dir->filename);

    fatfs_initpos(disk, dir, &pos);

    // Get free dir entry in parent dir
    err = get_free_raw_dentry(disk, &pos, namelen);
    if (err != ENOERR)
        return err;

    // Create new file dir entry
    init_dir_entry(dentry,
                   name,
                   namelen,
                   __stat_mode_REG,
                   dir->cluster,
                   0,
                   &pos);

    err = fatfs_write_dir_entry(disk, dentry);
    if (err != ENOERR)
       return err;

    return ENOERR;
}

// -------------------------------------------------------------------------
// fatfs_create_dir()
// Creates a new directory.

int
fatfs_create_dir(fatfs_disk_t      *disk,
                 fatfs_dir_entry_t *dir,
                 const char        *name,
                 int                namelen,
                 fatfs_dir_entry_t *dentry)
{
    fatfs_dir_entry_t cdentry;
    fatfs_data_pos_t  pos;
    cyg_uint32        free_cluster;
    int               err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dir);
    CYG_CHECK_DATA_PTRC(name);
    CYG_CHECK_DATA_PTRC(dentry);

    CYG_TRACE2(TDE, "filename='%s' parent='%s'", name, dir->filename);

    // Get free cluster
    err = find_next_free_cluster(disk,
                                 0,
                                 &free_cluster,
                                 CO_MARK_LAST | CO_ERASE_NEW);

    if (err != ENOERR)
        return err;

    fatfs_initpos(disk, dir, &pos);

    // Get free dir entry in parent dir
    err = get_free_raw_dentry(disk, &pos, namelen);

    if (err != ENOERR)
        return err;

    // Create new dir entry
    init_dir_entry(dentry,
                   name,
                   namelen,
                   __stat_mode_DIR,
                   dir->cluster,
                   free_cluster,
                   &pos);

    err = fatfs_write_dir_entry(disk, dentry);

    if (err != ENOERR)
       return err;

    // Create '.' and '..' dir entries

    fatfs_initpos(disk, dentry, &pos);

    CYG_TRACE0(TDE, "Creating '.' entry");

    init_dir_entry(&cdentry,
                   ".",
                   0,
                   __stat_mode_DIR,
                   dentry->cluster,
                   dentry->cluster,
                   &pos);

    err = fatfs_write_dir_entry(disk, &cdentry);
    if (err != ENOERR)
        return err;

    pos.cluster_pos += DENTRY_SIZE;

    CYG_TRACE0(TDE, "Creating '..' entry");

    init_dir_entry(&cdentry,
                   "..",
                   0,
                   __stat_mode_DIR,
                   dentry->cluster,
                   dir->cluster,
                   &pos);

    err = fatfs_write_dir_entry(disk, &cdentry);
    if (err != ENOERR)
        return err;

    return ENOERR;
}

// -------------------------------------------------------------------------
// fatfs_trunc_file()
// Truncates a file to zero length.

int
fatfs_trunc_file(fatfs_disk_t *disk, fatfs_dir_entry_t *file)
{
    int err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);

    CYG_TRACE1(TDE, "file='%s'", file->filename);

    if (S_ISDIR(file->mode))
        return EINVAL;

    if (0 == file->size)
        return ENOERR;

    err = free_cluster_chain(disk, file->cluster);
    if (err != ENOERR)
        return err;

    // Update file attributes

    file->cluster = 0;
    file->size    = 0;
    file->mtime   =
    file->atime   = cyg_timestamp();

    return fatfs_write_dir_entry(disk, file);
}

// -------------------------------------------------------------------------
// fatfs_rename_file()
// Renames a file.

int
fatfs_rename_file(fatfs_disk_t      *disk,
                  fatfs_dir_entry_t *dir1,
                  fatfs_dir_entry_t *target,
                  fatfs_dir_entry_t *dir2,
                  const char        *name,
                  int                namelen)
{
    fat_raw_dir_entry_t raw_dentry;
    fatfs_data_pos_t    new_pos;
    int                 err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dir1);
    CYG_CHECK_DATA_PTRC(target);
    CYG_CHECK_DATA_PTRC(dir2);
    CYG_CHECK_DATA_PTRC(name);

    if (is_root_dir_entry(target))
        return EINVAL;

    strncpy(target->filename, name, namelen);
    target->filename[namelen] = '\0';

    // Moving around in same dir

    if (dir1 == dir2)
    {
        FAT_INFO("same dir\n");
        return fatfs_write_dir_entry(disk, target);
    }

    // Moveing around in different dirs
    fatfs_initpos(disk, dir2, &new_pos);

    // Get free dir entry in target dir
    err = get_free_raw_dentry(disk, &new_pos, namelen);
    if (err != ENOERR)
        return err;

    // Write file dentry to new location

    dentry_to_raw(disk, target, &raw_dentry);
    err = write_raw_dentry(disk, &new_pos, &raw_dentry);
    if (err != ENOERR)
        return err;

    // Delete dentry at old location
    raw_dentry_set_deleted(disk, &raw_dentry);
    raw_dentry.size    = 0;
    raw_dentry.cluster = 0;
    err = write_raw_dentry(disk, &target->disk_pos, &raw_dentry);
    if (err != ENOERR)
        return err;

    // Set file new position and parent cluster
    target->disk_pos       = new_pos;
    target->parent_cluster = dir2->cluster;

    // If we moved a directory, we also have to correct the '..' entry
    if ( S_ISDIR(target->mode) )
    {
        fat_raw_dir_entry_t raw_cdentry;
        fatfs_data_pos_t    pos;

        fatfs_initpos(disk, target, &pos);

        while (true)
        {
            err = read_next_raw_dentry(disk, &pos, &raw_cdentry);

            if (EEOF == err)
                return EINVAL; // This dir doesn't have the '..' entry,
                               // that means something is very wrong
            else if (err != ENOERR)
                return err;

            if (0 == strncmp("..", raw_cdentry.name, 2))
            {
                raw_cdentry.cluster = dir2->cluster;
                err = write_raw_dentry(disk, &pos, &raw_cdentry);
                if (err != ENOERR)
                    return err;
                break;
            }

#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( DENTRY_IS_DELETED(&raw_cdentry) )
            pos.cluster_pos += DENTRY_SIZE;
        else
        {
            err = go_to_next_raw_entry(disk, &pos, &raw_cdentry, CO_NONE);
            if (EEOF == err)
                return EINVAL; // This dir doesn't have the '..' entry,
                               // that means something is very wrong
            else if (err != ENOERR)
                return err;
        }
#else
         pos.cluster_pos += DENTRY_SIZE;
#endif

        }
    }

    return ENOERR;
}

// -------------------------------------------------------------------------
// fatfs_write_data()
// Writes data to disk.

int
fatfs_write_data(fatfs_disk_t      *disk,
                 fatfs_dir_entry_t *file,
                 fatfs_data_pos_t  *pos,
                 void              *data,
                 cyg_uint32        *len)
{
    int err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);
    CYG_CHECK_DATA_PTRC(data);
    CYG_CHECK_DATA_PTRC(len);
    CYG_CHECK_DATA_PTRC(pos);

    // Check if this file has a zero size and no first cluster
    if (0 == file->size && 0 == file->cluster)
    {
        cyg_uint32 free_cluster;

        FAT_DEBUG("new cluster for zero file\n");;
        err = find_next_free_cluster(disk, 0, &free_cluster, CO_MARK_LAST);

        if (err != ENOERR)
        {
            *len = 0;
            return err;
        }

        file->cluster = free_cluster;
        fatfs_initpos(disk, file, pos);
    }

    return write_data(disk, data, len, pos);
}
