//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/*
 * \file sniffs.c
 *
 * Single file flash filesystem
 * 
 * This file implements SNIFFS.
 * This is a very simple filesystem for NOR flash.
 * It can store only one file!
 * 
 * \author    frederic.lair@parrot.fr
 * \date      2006-02-22
 */


#include <pkgconf/system.h>
#include <pkgconf/hal.h>
#include <pkgconf/kernel.h>
#include <pkgconf/io_fileio.h>
#include <pkgconf/fs_sniffs.h>

#include <cyg/kernel/ktypes.h>         // base kernel types
#include <cyg/infra/cyg_trac.h>        // tracing macros
#include <cyg/infra/cyg_ass.h>         // assertion macros

#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include <stdlib.h>
#include <string.h>

#include <cyg/fileio/fileio.h>

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>

#ifdef CYGPKG_IO_FLASH
#include <pkgconf/io_flash.h>
#include <cyg/io/flash.h>
#endif

#ifdef CYGPKG_IO_NOR_FLASH
#include <pkgconf/io_nor_flash.h>
#include <cyg/io/nor_flash.h>
#endif



/*
            ------------------------------------------
           |      Debug output				           |
            ------------------------------------------
*/
#if CYGOPT_FS_SNIFFS_DEBUG_LEVEL
#include <cyg/infra/diag.h>			// HAL polled output

#define DBGPRINT(...) diag_printf("Func: %s, Line: %d, ",__FUNCTION__,__LINE__),diag_printf(__VA_ARGS__)
#else
#define DBGPRINT(...)

#endif // CYGOPT_FS_SNIFFS_DEBUG_LEVEL

#if CYGOPT_FS_SNIFFS_DEBUG_LEVEL >= 2
#define DBGPRINT2(...) diag_printf("Func: %s, Line: %d, ",__FUNCTION__,__LINE__),diag_printf(__VA_ARGS__)
#else
#define DBGPRINT2(...)
#endif



//==========================================================================
// Sizes derived from configuration

// -------------------------------------------------------------------------

// This value should not change,
// in order to ensure compatibility with previous versions
#define SNIFFS_MAX_NAME_SIZE   65

//==========================================================================
// Forward definitions

// Filesystem operations
static int sniffs_mount    ( cyg_fstab_entry *fste, cyg_mtab_entry *mte );
static int sniffs_umount   ( cyg_mtab_entry *mte );
static int sniffs_open     ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int mode,  cyg_file *fte );
static int sniffs_unlink   ( cyg_mtab_entry *mte, cyg_dir dir, const char *name );

static int sniffs_opendir  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             cyg_file *fte );
static int sniffs_chdir    ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             cyg_dir *dir_out );
static int sniffs_stat     ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             struct stat *buf);
static int sniffs_getinfo  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int key, void *buf, int len );
static int sniffs_setinfo  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int key, void *buf, int len );

// File operations
static int sniffs_fo_read      (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio);
static int sniffs_fo_write     (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio);
static int sniffs_fo_lseek     (struct CYG_FILE_TAG *fp, off_t *pos, int whence );
static int sniffs_fo_ioctl     (struct CYG_FILE_TAG *fp, CYG_ADDRWORD com,
                                CYG_ADDRWORD data);
static int sniffs_fo_fsync     (struct CYG_FILE_TAG *fp, int mode );        
static int sniffs_fo_close     (struct CYG_FILE_TAG *fp);
static int sniffs_fo_fstat     (struct CYG_FILE_TAG *fp, struct stat *buf );
static int sniffs_fo_getinfo   (struct CYG_FILE_TAG *fp, int key, void *buf, int len );
static int sniffs_fo_setinfo   (struct CYG_FILE_TAG *fp, int key, void *buf, int len );

// Directory operations
static int sniffs_fo_dirread      (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio);
static int sniffs_fo_dirlseek     (struct CYG_FILE_TAG *fp, off_t *pos, int whence );


//==========================================================================
// Filesystem table entries

// -------------------------------------------------------------------------
// Fstab entry.
// This defines the entry in the filesystem table.
// For simplicity we use _FILESYSTEM synchronization for all accesses since
// we should never block in any filesystem operations.

FSTAB_ENTRY( sniffs_fste, "sniffs", 0,
             CYG_SYNCMODE_FILE_FILESYSTEM|CYG_SYNCMODE_IO_FILESYSTEM,
             sniffs_mount,
             sniffs_umount,
             sniffs_open,
			 (cyg_fsop_unlink *)sniffs_unlink,
			 (cyg_fsop_mkdir *)cyg_fileio_seltrue,
			 (cyg_fsop_rmdir *)cyg_fileio_seltrue,
			 (cyg_fsop_rename *)cyg_fileio_enosys,
			 (cyg_fsop_link *)cyg_fileio_enosys,
             sniffs_opendir,
             sniffs_chdir,
             sniffs_stat,
             sniffs_getinfo,
             sniffs_setinfo);

// -------------------------------------------------------------------------
// File operations.
// This set of file operations are used for normal open files.

static cyg_fileops sniffs_fileops =
{
    sniffs_fo_read,
    sniffs_fo_write,
    sniffs_fo_lseek,
    sniffs_fo_ioctl,
    cyg_fileio_seltrue,
    sniffs_fo_fsync,
    sniffs_fo_close,
    sniffs_fo_fstat,
    sniffs_fo_getinfo,
    sniffs_fo_setinfo
};

// -------------------------------------------------------------------------
// Directory file operations.
// This set of operations are used for open directories. Most entries
// point to error-returning stub functions. Only the read, lseek and
// close entries are functional.

static cyg_fileops sniffs_dirops =
{
    sniffs_fo_dirread,
    (cyg_fileop_write *)cyg_fileio_enosys,
    sniffs_fo_dirlseek,
    (cyg_fileop_ioctl *)cyg_fileio_enosys,
    cyg_fileio_seltrue,
    (cyg_fileop_fsync *)cyg_fileio_enosys,
    sniffs_fo_close,
    (cyg_fileop_fstat *)cyg_fileio_enosys,
    (cyg_fileop_getinfo *)cyg_fileio_enosys,
    (cyg_fileop_setinfo *)cyg_fileio_enosys
};


typedef struct{
	void*       sniffs_base;
	cyg_uint32  sniffs_len;
	int         sniffs_blk_size;
}sniffs_config_t;

sniffs_config_t sniffs_config;

typedef cyg_uint32 sniffs_block;


/*!
 * Store the result of the last call to the function findbuffer_node().
 * This is used to improve position search speed.
 */
typedef struct {
	int          valid;
	off_t        pos;
	sniffs_block *block;
}sniffs_last_search_t;
sniffs_last_search_t sniffs_last_search;

//==========================================================================
// File node
// This data structure represents the file.

typedef struct
{
    mode_t              mode;           // node type
    cyg_ucount32        refcnt;         // open file/current dir references
    size_t              size;           // size of file in bytes
    sniffs_block        *data;          // Address of first file data block
	char                *name;
}sniffs_node;

sniffs_node file_node; 


/*!
 * \return The address of the file's first data block.
 */
sniffs_block* first_block(void)
{
    char *addr;
    
    addr = (char*)sniffs_config.sniffs_base + SNIFFS_MAX_NAME_SIZE;
    
    if((cyg_uint32)addr%sizeof(cyg_uint32))
    {
        addr += sizeof(cyg_uint32) - (cyg_uint32)addr%sizeof(cyg_uint32);
    }
    
    return (sniffs_block*)addr;
}



/*!
 * Find the next data block of a file.
 * \param block Current data block.
 * \return The address of the file's next data block.
 */
sniffs_block* next_block( sniffs_block * block )
{
	char *add;
	add = (char*)block + *block + sizeof(*block);

	if( (cyg_uint32)add%sizeof(cyg_uint32) )
		add += sizeof(cyg_uint32) - (cyg_uint32)add%sizeof(cyg_uint32);

	return (sniffs_block*)add;
}

/*!
 * Calculate the file size.
 * \return The file size in bytes.
 */
size_t get_file_size(void)
{
	size_t size;
	sniffs_block* data_block;

	// Calculate the file size
	size = 0;
	data_block = first_block();
	while( *data_block != 0xFFFFFFFF && *data_block  )
	{
		size += *data_block;
		data_block = next_block(data_block);
	}

	return size;
}

/*!
 * Delete flash area.
 */
int flash_delete( void* dest, int len )
{
	int status = ENOERR;
	void *err_addr;
    cyg_uint32 *ptr;
	int i, del_bytes;

#ifdef CYGHWR_IO_FLASH_BLOCK_LOCKING
	flash_unlock( dest, len, &err_addr);
#endif

	while( len > 0 )
	{
        // Compute delete size at this pass
		if( len > sniffs_config.sniffs_blk_size )
            del_bytes = sniffs_config.sniffs_blk_size;
		else
			del_bytes = len;

        // Check that this flash space is not empty
        ptr = dest;
		for( i=0; i<(del_bytes/sizeof(cyg_uint32)); i+=sizeof(cyg_uint32) )
		{
			if( *(ptr++) != 0xFFFFFFFF )
				break;
		}
        
        // If this flash block need to be erased
		if( i != del_bytes/sizeof(cyg_uint32) )
		{
            // Erase flash block
			status = flash_erase( dest, del_bytes, &err_addr );
			if(status)
				DBGPRINT( "Erase error: add = %p - %s\n", err_addr, flash_errmsg( status ));
        }

		len -= del_bytes;
        dest = (char*)dest + del_bytes;
	}

#ifdef CYGHWR_IO_FLASH_BLOCK_LOCKING
	flash_lock( dest, len, &err_addr);
#endif
	
	return status;
}

/*!
 * Program data in flash.
 * \param dest Address of flash to program, has to be aligned on 4.
 * \param src Address of source buffer.
 * \param len Size in bytes of the buffer to program.
 */
int flash_write( void* dest, void* src, int len )
{
	int status = 0;
	void *err_addr;
	cyg_uint32 val;


#ifdef CYGHWR_IO_FLASH_BLOCK_LOCKING
	flash_unlock( dest, len, &err_addr);
#endif
	
	// Aligned data
	if( !(((cyg_uint32)dest)%sizeof(cyg_uint32)) 
		&& !(((cyg_uint32)src)%sizeof(cyg_uint32)) 
		&& !(((cyg_uint32)len)%sizeof(cyg_uint32)) )
		status = flash_program( dest, src, len, &err_addr );
	// Unaligned data
	else
	{
		while( len > 0 )
		{
			if( len < sizeof(cyg_uint32) )
			{
				val = *(cyg_uint32*)dest;
				memcpy( &val, src, len );
			}
			else
			{
				memcpy( &val, src, sizeof(cyg_uint32) );
			}

			status = flash_program( dest, &val, sizeof(cyg_uint32), &err_addr );
			if(status)
				DBGPRINT( "Program error: add = %p - %s\n", err_addr, flash_errmsg( status ));
			
			len -= sizeof(cyg_uint32);
            dest = (char*)dest + sizeof(cyg_uint32);
            src = (char*)src + sizeof(cyg_uint32);
		}
	}

#ifdef CYGHWR_IO_FLASH_BLOCK_LOCKING
	flash_lock( dest, len, &err_addr);
#endif

	return status;
}

//==========================================================================
// Node buffer management

// -------------------------------------------------------------------------
// findbuffer_node()
// return a pointer to the data at the indicated file position, extending
// the buffer if required.

static int findbuffer_node( sniffs_node  *node, // node pointer
                            off_t pos,          // data position to get
                            sniffs_block **data_block, // returned pointer to data block header
                            cyg_uint8 **buffer, // returned buffer pointer
                            size_t *size,       // returned buffer size
                            cyg_bool alloc)     // extend allocation?
{
	off_t current_pos;

	*size = 0;
	current_pos = 0;

	if( sniffs_last_search.valid && pos >= sniffs_last_search.pos )
	{
		*data_block = sniffs_last_search.block;
		current_pos = sniffs_last_search.pos;
	}
	else
		*data_block = node->data;


    if( alloc && pos == node->size )
    {
		// Extend the file
		
		while( **data_block != 0xFFFFFFFF && **data_block )
		{
			*data_block = next_block(*data_block);
		}
		if( (cyg_uint32)*data_block+sizeof(cyg_uint32)+*size
			> (cyg_uint32)sniffs_config.sniffs_base+sniffs_config.sniffs_len )
		{
			sniffs_last_search.valid = 0;
			return ENOSPC;
		}

		*buffer = (cyg_uint8*)*data_block + sizeof(**data_block);
		*size = (cyg_uint32)sniffs_config.sniffs_base+sniffs_config.sniffs_len-(cyg_uint32)*data_block-sizeof(cyg_uint32);
	}
	else
	{
		while(**data_block != 0xFFFFFFFF && **data_block)
		{
			if( current_pos+**data_block <= pos )
			{
				current_pos += **data_block;
				*data_block = next_block(*data_block);
			}
			else
			{
				*buffer = (cyg_uint8*)*data_block + sizeof(**data_block) + pos - current_pos;
				*size = **data_block - pos + current_pos;
				break;
			}
		}
	}

	// Store search result
	sniffs_last_search.valid = 1;
	sniffs_last_search.pos = current_pos;
	sniffs_last_search.block = *data_block;

    return ENOERR;
}



//==========================================================================
// Filesystem operations

// -------------------------------------------------------------------------
// sniffs_mount()
// Process a mount request. This mainly creates a root for the
// filesystem.

static int sniffs_mount( cyg_fstab_entry *fste, cyg_mtab_entry *mte )
{
 	int err, status;
	void *flash_start, *flash_stop;
	int block_size, nb_blocks;	
   
	err = ENOERR;

	file_node.refcnt = 0;

    mte->root = (cyg_dir)&file_node;

	// Flash init
	status = flash_init( diag_printf );
	if( status )
	{
		DBGPRINT( "%s\n", flash_errmsg( status ) );
		err = EIO;
	}

	// Get limits
	status = flash_get_limits( (void*)0, &flash_start, &flash_stop );
	if( status )
	{
		DBGPRINT( "%s\n", flash_errmsg( status ) );
		err = EIO;
	}

	// Get blocks info
	status = flash_get_block_info( &block_size, &nb_blocks );
	if( status )
	{
		DBGPRINT( "%s\n", flash_errmsg( status ) );
		err = EIO;
	}
   
	// Set configuration data
	sniffs_config.sniffs_base = (char*)flash_start + CYGNUM_SNIFFS_OFFSET;
	sniffs_config.sniffs_len = CYGNUM_SNIFFS_LENGTH;
	sniffs_config.sniffs_blk_size = block_size;

    return err;
}

// -------------------------------------------------------------------------
// sniffs_umount()
// Unmount the filesystem. This will currently only succeed if the
// filesystem is empty.

static int sniffs_umount   ( cyg_mtab_entry *mte )
{
    sniffs_node *root = (sniffs_node *)mte->root;

    // Check for open/inuse root
    if( root->refcnt != 0 )
        return EBUSY;

    // Clear root pointer
    mte->root = CYG_DIR_NULL;

    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_open()
// Open a file for reading or writing.

static int sniffs_open     ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int mode,  cyg_file *file )
{
	sniffs_node* node = (sniffs_node *)mte->root;
    int err = ENOERR;
	char filename[SNIFFS_MAX_NAME_SIZE];
	char *char_ptr;
	int i,status;
 
	// Init stored search result
	sniffs_last_search.valid = 0;
	sniffs_last_search.pos = 0;
	sniffs_last_search.block = 0;

	// Extract the file name
	char_ptr = rindex( name, '/' );
	if( char_ptr )
		strncpy( filename, char_ptr+1, SNIFFS_MAX_NAME_SIZE );
	else
		strncpy( filename, name, SNIFFS_MAX_NAME_SIZE );
	filename[SNIFFS_MAX_NAME_SIZE-1] = '\0';
	
	DBGPRINT2("sniffs_open %s\n",filename);

	// Check if the file exists
	if( strncmp( filename, sniffs_config.sniffs_base, SNIFFS_MAX_NAME_SIZE ) )
		err = ENOENT;

	// If the file does not exists
    if(    ( err == ENOENT )
        && ( mode & O_CREAT ) )
    {
        // No node there, if the O_CREAT bit is set then we must
        // create it

        // Check if there is already a file
        for( i=0; i<SNIFFS_MAX_NAME_SIZE; i++ )
            if( ((char*)sniffs_config.sniffs_base)[i] != 0xFF )
                return ENOSPC;
        
        // Check that the flash is empty
        status = flash_delete( sniffs_config.sniffs_base, sniffs_config.sniffs_len);
        if( status )
            return EIO;

        // Program the file name
        status = flash_write( sniffs_config.sniffs_base, (char*)filename, strlen(filename)+1 );
        if( status )
            return EIO;
		
        // File size is 0
        node->size = 0;

        err = ENOERR;
    }
	// If the file exists
    else if( err == ENOERR )
    {
        // The file exists. If the O_CREAT and O_EXCL bits are set, we
        // must fail the open.
        if( (mode & (O_CREAT|O_EXCL)) == (O_CREAT|O_EXCL) )
            err = EEXIST;

		// Calculate the file size
		node->size = get_file_size();
    
        // If the O_TRUNC bit is set we must clean out the file data.
        if( err == ENOERR && (mode & O_TRUNC ) )
        {
            // Erase flash
            status = flash_delete( sniffs_config.sniffs_base, sniffs_config.sniffs_len);
            if( status )
                return EIO;

            // Program the file name
            status = flash_write( sniffs_config.sniffs_base, (char*)filename, strlen(filename)+1 );
            if( status )
                return EIO;
		
            // File size is 0
            node->size = 0;
        }
    }

    if( err != ENOERR ) return err;


	// Initialize the file node
	node->mode = __stat_mode_REG;
	node->data = first_block();
	node->name = sniffs_config.sniffs_base;

    node->refcnt++;       // Count successful open
    
    // Initialize the file object
    
    file->f_flag        |= mode & CYG_FILE_MODE_MASK;
    file->f_type        = CYG_FILE_TYPE_FILE;
    file->f_ops         = &sniffs_fileops;
    file->f_offset      = (mode&O_APPEND) ? node->size : 0;
    file->f_data        = (CYG_ADDRWORD)node;
    file->f_xops        = 0;

    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_unlink()
// Remove a file link from its directory.

static int sniffs_unlink   ( cyg_mtab_entry *mte, cyg_dir dir, const char *name )
{
	char filename[SNIFFS_MAX_NAME_SIZE];
	char *char_ptr;
	int status;
	sniffs_node* node = (sniffs_node *)mte->root;

	// Extract the file name
	char_ptr = rindex( name, '/' );
	if( char_ptr )
		strncpy( filename, char_ptr+1, SNIFFS_MAX_NAME_SIZE );
	else
		strncpy( filename, name, SNIFFS_MAX_NAME_SIZE );
	filename[SNIFFS_MAX_NAME_SIZE-1] = '\0';
	
	// Check if the file exists
	if( strncmp( filename, sniffs_config.sniffs_base, SNIFFS_MAX_NAME_SIZE ) )
		return ENOENT;

	// Do not delete the file if it is in use
	if( node->refcnt )
		return EBUSY;

	// Erase flash
	status = flash_delete( sniffs_config.sniffs_base, sniffs_config.sniffs_len);
	if( status )
		return EIO;
	
    return ENOERR;
}


// -------------------------------------------------------------------------
// sniffs_opendir()
// Open a directory for reading.

static int sniffs_opendir  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             cyg_file *file )
{
    // Initialize the file object, setting the f_ops field to a
    // special set of file ops.
    file->f_type        = CYG_FILE_TYPE_FILE;
    file->f_ops         = &sniffs_dirops;
    file->f_offset      = 0;
    file->f_data        = 0;
    file->f_xops        = 0;
        
    return ENOERR;

}

// -------------------------------------------------------------------------
// sniffs_chdir()
// Change directory support.

static int sniffs_chdir    ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             cyg_dir *dir_out )
{
    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_stat()
// Get struct stat info for named object.

static int sniffs_stat     ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             struct stat *buf)
{
	char filename[SNIFFS_MAX_NAME_SIZE];
	char *char_ptr;

	// If root directory
	if( name[0] == '\0' )
	{
		// Fill in the status
		buf->st_mode        = __stat_mode_DIR;
		buf->st_ino         = 0;
		buf->st_dev         = 0;
		buf->st_nlink       = 0;
		buf->st_uid         = 0;
		buf->st_gid         = 0;
		buf->st_size        = 0;
		buf->st_atime       = 0;
		buf->st_mtime       = 0;
		buf->st_ctime       = 0;
	}
	else
	{
		// Extract the file name
		char_ptr = rindex( name, '/' );
		if( char_ptr )
			strncpy( filename, char_ptr+1, SNIFFS_MAX_NAME_SIZE );
		else
			strncpy( filename, name, SNIFFS_MAX_NAME_SIZE );
		filename[SNIFFS_MAX_NAME_SIZE-1] = '\0';
		
		// Check if the file exists
		if( strncmp( filename, sniffs_config.sniffs_base, SNIFFS_MAX_NAME_SIZE ) )
			return ENOENT;
		
		// Fill in the status
		buf->st_mode        = __stat_mode_REG;
		buf->st_ino         = 0;
		buf->st_dev         = 0;
		buf->st_nlink       = 0;
		buf->st_uid         = 0;
		buf->st_gid         = 0;
		buf->st_size        = get_file_size();
		buf->st_atime       = 0;
		buf->st_mtime       = 0;
		buf->st_ctime       = 0;
	}

    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_getinfo()
// Getinfo. Currently only support pathconf().

static int sniffs_getinfo  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int key, void *buf, int len )
{
    return EINVAL;
}

// -------------------------------------------------------------------------
// sniffs_setinfo()
// Setinfo. Nothing to support here at present.

static int sniffs_setinfo  ( cyg_mtab_entry *mte, cyg_dir dir, const char *name,
                             int key, void *buf, int len )
{
    // No setinfo keys supported at present
    
    return EINVAL;
}


//==========================================================================
// File operations

// -------------------------------------------------------------------------
// sniffs_fo_read()
// Read data from the file.

static int sniffs_fo_read      (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
    sniffs_node *node = (sniffs_node *)fp->f_data;
    int i;
    off_t pos = fp->f_offset;
    ssize_t resid = uio->uio_resid;
	sniffs_block *data_block;

    // Loop over the io vectors until there are none left
    for( i = 0; i < uio->uio_iovcnt; i++ )
    {
        cyg_iovec *iov = &uio->uio_iov[i];
        char *buf = (char *)iov->iov_base;
        off_t len = iov->iov_len;

        // Loop over each vector filling it with data from the file.
        while( len > 0 && pos < node->size )
        {
            cyg_uint8 *fbuf;
            size_t bsize;
            off_t l = len;
            int err;

            // Get a pointer to the data at offset _pos_.
            err = findbuffer_node( node, pos, &data_block, &fbuf, &bsize, false );

            if( err != ENOERR )
                return err;

            // adjust size to end of file if necessary
            if( l > node->size-pos )
                l = node->size-pos;
            
            // adjust size to the amount of contiguous data we can see
            // at present.
            if( l > bsize )
                l = bsize;

            // copy data out
            memcpy( buf, fbuf, l );

            // Update working vars
            len -= l;
            buf += l;
            pos += l;
            resid -= l;
        }
    }

    // We successfully read some data,
	// update the file offset and transfer residue.

    uio->uio_resid = resid;
    fp->f_offset = pos;
    
    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_fo_write()
// Write data to file.

static int sniffs_fo_write     (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
    sniffs_node *node = (sniffs_node *)fp->f_data;
    off_t pos = fp->f_offset;
    ssize_t resid = uio->uio_resid;    
    int err = ENOERR;
    int i;
	sniffs_block *data_block;
	int status;

    // If the APPEND mode bit was supplied, force all writes to
    // the end of the file.
    if( fp->f_flag & CYG_FAPPEND )
        pos = fp->f_offset = node->size;
    
    ///// Check that pos is within current file size, or at the very end.
    // Check that pos is not less than zero
	if( pos < 0 /*|| pos > node->size*/ )
        return EINVAL;

    // Now loop over the iovecs until they are all done, or
    // we get an error.
    for( i = 0; i < uio->uio_iovcnt; i++ )
    {
        cyg_iovec *iov = &uio->uio_iov[i];
        char *buf = (char *)iov->iov_base;
        off_t len = iov->iov_len;

        // loop over the vector writing it to the file until it has
        // all been done.
        while( len > 0 )
        {
            cyg_uint8 *fbuf;
            size_t bsize;
            off_t l = len;
            
            err = findbuffer_node( node, pos, &data_block, &fbuf, &bsize, true );

            // Stop writing on error (ex: no more space)
            if( err != ENOERR )
                return err;
            
			// If this data block is already used, return an error
			if( *data_block != 0xFFFFFFFF )
				return EACCES;

            // adjust size to this block
            if( l > bsize )
                l = bsize;

			// Program data
			status = flash_write( fbuf, buf, l );
			if( status )
				return EIO;

			// Write new data block len
			status = flash_write( data_block, &l, sizeof(*data_block) );
			if( status )
				return EIO;

			// Update file size
			node->size += l;

            // Update working vars
            len -= l;
            buf += l;
            pos += l;
            resid -= l;
        }
    }

    // We wrote some data successfully, update
    // the file offset and transfer residue.
    uio->uio_resid = resid;
    fp->f_offset = pos;
    
    return err;
}

// -------------------------------------------------------------------------
// sniffs_fo_lseek()
// Seek to a new file position.

static int sniffs_fo_lseek     (struct CYG_FILE_TAG *fp, off_t *apos, int whence )
{
    sniffs_node *node = (sniffs_node *)fp->f_data;
    off_t pos = *apos;

    switch( whence )
    {
    case SEEK_SET:
        // Pos is already where we want to be.
        break;

    case SEEK_CUR:
        // Add pos to current offset.
        pos += fp->f_offset;
        break;

    case SEEK_END:
        // Add pos to file size.
        pos += node->size;
        break;

    default:
        return EINVAL;
    }

    // If pos is beyond the current end of the file, extend the file
    // with zero filled storage up to the new position.
    
    if( (pos > node->size) 
        && ( fp->f_flag & CYG_FWRITE ) )
    {
		diag_printf("sniffs_fo_lseek pos %d node->size %d\n",pos,node->size);
        off_t fpos = node->size;
        off_t len = pos-fpos;
		sniffs_block *data_block;
		int status;

        while( len > 0 )
        {
            cyg_uint8 *fbuf;
            size_t bsize;
            off_t l = len;
            int err;
			
			err = findbuffer_node( node, fpos, &data_block, &fbuf, &bsize, true );

            if( err != ENOERR )
                return err;

            // adjust size to this block
            if( l > bsize )
                l = bsize;

			// Write new data block len
			status = flash_write( data_block, &l, sizeof(*data_block) );
			if( status )
				return EIO;

            // Update working vars
            len -= l;
            fpos += l;
        }

        // We wrote some data successfully, increase its size appropriately.
		node->size = fpos;    
    }
    
    // Check that pos is still within current file size, or at the
    // very end.
    if( pos < 0 || pos > node->size )
        return EINVAL;

    // All OK, set fp offset and return new position.
    *apos = fp->f_offset = pos;
    
    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_fo_ioctl()
// Handle ioctls. Currently none are defined.

static int sniffs_fo_ioctl     (struct CYG_FILE_TAG *fp, CYG_ADDRWORD com,
                                CYG_ADDRWORD data)
{
    return EINVAL;
}

// -------------------------------------------------------------------------
// sniffs_fo_fsync().
// Force the file out to data storage.

static int sniffs_fo_fsync     (struct CYG_FILE_TAG *fp, int mode )
{
    // Data is always permanently where it belongs, nothing to do
    // here.
  
    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_fo_close()
// Close a file. We just decrement the refcnt and let it go away if
// that is all that is keeping it here.

static int sniffs_fo_close     (struct CYG_FILE_TAG *fp)
{
	sniffs_node *node = (sniffs_node *)fp->f_data;

	if( node )
		node->refcnt--;

    fp->f_data = 0;     // zero data pointer
    
    return ENOERR;
}

// -------------------------------------------------------------------------
//sniffs_fo_fstat()
// Get file status.

static int sniffs_fo_fstat     (struct CYG_FILE_TAG *fp, struct stat *buf )
{
    sniffs_node *node = (sniffs_node *)fp->f_data;

    // Fill in the status
    buf->st_mode        = node->mode;
    buf->st_ino         = (ino_t)node;
    buf->st_dev         = 0;
    buf->st_nlink       = 0;
    buf->st_uid         = 0;
    buf->st_gid         = 0;
    buf->st_size        = node->size;
    buf->st_atime       = 0;
    buf->st_mtime       = 0;
    buf->st_ctime       = 0;
    
    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_fo_getinfo()
// Get info. Currently only supports fpathconf().

static int sniffs_fo_getinfo   (struct CYG_FILE_TAG *fp, int key, void *buf, int len )
{
    return EINVAL;
}

// -------------------------------------------------------------------------
// sniffs_fo_setinfo()
// Set info. Nothing supported here.

static int sniffs_fo_setinfo   (struct CYG_FILE_TAG *fp, int key, void *buf, int len )
{
    // No setinfo key supported at present
    
    return ENOERR;
}


//==========================================================================
// Directory operations

// -------------------------------------------------------------------------
// sniffs_fo_dirread()
// Read a single directory entry from a file.

static int sniffs_fo_dirread      (struct CYG_FILE_TAG *fp, struct CYG_UIO_TAG *uio)
{
    struct dirent *ent = (struct dirent *)uio->uio_iov[0].iov_base;
	int i;

	// Only one entry per directory
	if( fp->f_offset > 0 )
		return ENOENT;

	// Check if there is a file
	for( i=0; i<SNIFFS_MAX_NAME_SIZE; i++ )
		if( ((char*)sniffs_config.sniffs_base)[i] != 0xFF )
			break;

	if( i == SNIFFS_MAX_NAME_SIZE )
		return ENOENT;

	strncpy( ent->d_name, sniffs_config.sniffs_base, SNIFFS_MAX_NAME_SIZE );
	ent->d_name[SNIFFS_MAX_NAME_SIZE-1] = '\0';

    uio->uio_resid = 0;
	fp->f_offset++;

    return ENOERR;
}

// -------------------------------------------------------------------------
// sniffs_fo_dirlseek()
// Seek directory to start.

static int sniffs_fo_dirlseek     (struct CYG_FILE_TAG *fp, off_t *pos, int whence )
{
    return ENOERR;
}

// -------------------------------------------------------------------------
// EOF sniffs.c
