//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  mklib.c
* @brief Make a symbolic table of exported symbols
*
* This host tool is used to generate a symbol table and a dummy shared library
* for exported symbols. It also generates a table of SWI instructions used for
* system calls.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-02-08
*******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdint.h>

#include "export.h"

struct symbol {
    uint32_t strtab_index;
    uint32_t switab_index;
};

// The following function comes from the binutils package

/* Standard ELF hash function.  Do not change this function; you will
   cause invalid hash tables to be generated.  */
static unsigned long
elf_hash (const char *namearg) {
    const unsigned char *name = (const unsigned char *) namearg;
    unsigned long h = 0;
    unsigned long g;
    int ch;

    while ((ch = *name++) != '\0') {
        h = (h << 4) + ch;
        if ((g = (h & 0xf0000000)) != 0) {
            h ^= g >> 24;
            /* The ELF ABI says `h &= ~g', but this is equivalent in
               this case and on some machines one insn instead of two.  */
            h ^= g;
        }
    }
    return h & 0xffffffff;
}

int
main(int argc, char *argv[]) {

    struct symbol *symtab;
    const char *name;
    syscall_mode_t mode;
    uint32_t nbucket, nchain, *bucket, *chain;
    uint32_t i, nb_symbols, strtab_size, switab_size, hash, link;

    nb_symbols = 0;
    strtab_size = 0;
    switab_size = 0;

    // count symbols
    for ( i = 0; export_table[i].mode != NONE; i++ ) {
        nb_symbols++;
    }

    if (( argc == 2 )&&( strcmp( argv[1], "-dummy" ) == 0 )) {
        // create a dummy library
        for ( i = 0; i < nb_symbols; i++ ) {
            printf( "void %s(void) {};\n", export_table[i].name );
        }
        return 0;
    }

    nbucket = nb_symbols/2 + 1;
    nchain = nb_symbols;

    // allocate tables
    symtab = (struct symbol *)malloc(nb_symbols*sizeof(struct symbol));
    bucket = (uint32_t *)calloc( sizeof(uint32_t), nbucket );
    chain = (uint32_t *)calloc( sizeof(uint32_t), nchain );

    assert(symtab && nbucket && bucket && chain);

    printf( "//\n// This file was automatically generated with mklib.c, "
            "DO NOT EDIT !!!\n//\n\n" );

    printf( "#include <elflib.h>\n\n" );

    // declare __check_xxx and __wrap_xxx symbols

    printf( "// System calls support functions\n");
    for ( i = 0; i < nb_symbols; i++ ) {

        mode = export_table[i].mode;

        if ((mode == SYSTEM_CHECK)||(mode == SYSTEM_CHECK_INDIRECT)) {
            printf( "extern void __check_%s(void);\n", export_table[i].name );
        }
        if ((mode == USER_WRAP)||(mode == SYSTEM_CHECK_INDIRECT)) {
            printf( "extern void __wrap_%s(void);\n", export_table[i].name );
        }
        if ((mode == USER)||(mode == SYSTEM_NOCHECK)) {
            printf( "extern void %s(void);\n", export_table[i].name );
        }
        if (mode == DATA) {
            printf( "extern int %s;\n", export_table[i].name );
        }
    }
    printf("\n");

    // build hash table & print SWI calls table
    printf( "// SWI function table\n");

    printf( "void switab(void) __attribute__ ((naked));\n" );
    printf( "void switab(void) {\n" );

    for ( i = 0; i < nb_symbols; i++ ) {
        mode = export_table[i].mode;
        name = export_table[i].name;

        if (( mode == SYSTEM_NOCHECK )||
            ( mode == SYSTEM_CHECK   )||
            ( mode == SYSTEM_CHECK_INDIRECT )) {

            printf( "    asm volatile ( \".global __swi_%s\" );\n", name );
            printf( "    asm volatile ( \"__swi_%s:\\n\\tswi\" );\n", name );
            printf( "    asm volatile ( \".word %s%s\" );\n",
                    (( mode == SYSTEM_CHECK )||
                     ( mode == SYSTEM_CHECK_INDIRECT ))? "__check_" : "",name);

            symtab[i].switab_index = switab_size;
            switab_size += 8;
        }
        // store index in string table
        symtab[i].strtab_index = strtab_size+1;
        strtab_size += strlen(name)+1;

        // add entry to hash table
        hash = elf_hash(name);
        if ( !bucket[hash % nbucket] ) {
            bucket[hash % nbucket] = i+1;
        }
        else {
            // append at end of chain
            link = bucket[hash % nbucket];
            while ( chain[link] ) {
                link = chain[link];
            }
            chain[link] = i+1;
        }
    }
    printf( "};\n\nconst int switab_bytes = %u;\n\n", (unsigned)switab_size );

    // print hash table
    printf( "\n// Hash table\n");
    printf( "#define NBUCKET (%u)\n#define NCHAIN  (%u)\n\n",
            (unsigned)nbucket, (unsigned)nchain);
    printf( "static const unsigned int export_bucket[NBUCKET] = {\n   " );

    for ( i = 0; i < nbucket; i++ ) {
        printf( " %u,", (unsigned)bucket[i] );
        if (((i+1) & 0xf ) == 0 ) {
            printf("\n   ");
        }
    }
    printf( "\n};\n\n");
    printf( "static const unsigned int export_chain[NCHAIN] = {\n   " );

    for ( i = 0; i < nchain; i++ ) {
        printf( " %u,", (unsigned)chain[i] );
        if (((i+1) & 0xf ) == 0 ) {
            printf("\n   ");
        }
    }
    printf( "\n};\n\n");

    // print string table
    printf( "// String table\n");
    printf( "#define STRTAB_SIZE (%u)\n\n", (unsigned)strtab_size );
    printf( "static const char export_strtab[STRTAB_SIZE] =\"" );

    for ( i = 0; i < nb_symbols; i++ ) {
        printf( "\\0\\\n%s", export_table[i].name );
    }
    printf("\";\n\n");

    // print symbol table

    printf( "// Symbol table\n");
    printf( "#define NB_SYMBOLS (%u)\n", (unsigned)nb_symbols );
    printf( "#define _FUNC      ELF32_ST_INFO(STB_GLOBAL, STT_FUNC)\n" );
    printf( "#define _DATA      ELF32_ST_INFO(STB_GLOBAL, STT_OBJECT)\n\n" );
    printf( "static const Elf32_Sym export_symtab[NB_SYMBOLS+1] = {\n" );

    // start with a dummy symbol
    printf("    { 0, 0, 0, ELF32_ST_INFO(STB_LOCAL, STT_NOTYPE), 0, 0 },\n");

    for ( i = 0; i < nb_symbols; i++ ) {
        // st_name
        printf( "    { %u,", (unsigned)symtab[i].strtab_index );
        // st_value
        switch ( export_table[i].mode ) {

        case DATA:
            printf( " (Elf32_Addr)&%s,", export_table[i].name );
            break;

        case USER:
            printf( " (Elf32_Addr)%s,", export_table[i].name );
            break;

        case SYSTEM_NOCHECK:
        case SYSTEM_CHECK:
            printf( " ((Elf32_Addr)switab)+%u,",
                    (unsigned)symtab[i].switab_index );
            break;

        case USER_WRAP:
        case SYSTEM_CHECK_INDIRECT:
            printf( " (Elf32_Addr)__wrap_%s,", export_table[i].name );
            break;

        case NONE:
            assert(0);
            break;

        }
        // st_size
        printf( " 0," );
        // st_info
        if ( export_table[i].mode == DATA ) {
            printf( " _DATA," );
        }
        else {
            printf( " _FUNC," );
        }
        // st_other
        printf( " 0," );
        // st_shndx
        printf( " SHN_ABS },\n" );
    }
    printf("};\n\n");

    // print elflib structure
    printf( "const struct elflib dynamo_export_lib = {\n" );
    printf( "   .status = ELFLIB_STATUS_LOADED,\n" );
    printf( "   .base = 0,\n" );
    printf( "   .strtab = export_strtab,\n" );
    printf( "   .strtab_len = STRTAB_SIZE,\n" );
    printf( "   .nbucket = NBUCKET,\n" );
    printf( "   .bucket = export_bucket,\n" );
    printf( "   .nchain = NCHAIN,\n" );
    printf( "   .chain = export_chain,\n" );
    printf( "   .symtab = export_symtab,\n" );
    printf( "};\n" );

    return 0;
}
