//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  check_cyg.c
* @brief System call checking routines.
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-03-01
*******************************************************************************
*/

#include <cyg/kernel/kapi.h>
#include <stdio.h>
#include <stdarg.h>

#include "common.h"

//! Copy memory from user space to system space
int
system_copy(void *dst, const void *src, unsigned int size) {
    int err;

    err = RANGE_GET_RO((CYG_ADDRESS)src, size);
    if ( !err ) {
        memcpy(dst, src, size);
        RANGE_RELEASE((CYG_ADDRESS)src, size);
    }
    return err;
}

// IMPORTANT NOTE: When __check_foo() is called, it uses the kernel stack,
// _not_ the user stack. User stack parameters may be retrieved using array
// swi_stack[SAV_SYSTEM_SP] in dynamo_thread_data_t.

//=============================================================================
// cyg_mutex_xxx calls
//=============================================================================

struct safe_mutex {
    cyg_mutex_t *key;
    cyg_mutex_t mutex;
};

//! We need a mutex pool to avoid corruption problems
static struct safe_mutex mutex_pool[DYNAMO_MAX_MUTEXES];

static struct safe_mutex *
find_safe_mutex( cyg_mutex_t *mutex ) {
    int i, index;
    struct safe_mutex *sf, *ret = NULL;

    // use mutex pointer as a hash key
    index = ((cyg_uint32)mutex) % DYNAMO_MAX_MUTEXES;

    // lookup mutex
    for ( i = 0; i < DYNAMO_MAX_MUTEXES; i++ ) {
        sf = &mutex_pool[(i+index) % DYNAMO_MAX_MUTEXES];
        if ( sf->key == mutex ) {
            ret = sf;
            break;
        }
    }
    return ret;
}

void
__check_cyg_mutex_init( cyg_mutex_t *mutex ) {
    int i, index;
    struct safe_mutex *sf, *ret = NULL;

    // use mutex pointer as a hash key
    index = ((cyg_uint32)mutex) % DYNAMO_MAX_MUTEXES;

    // find an empty slot
    for ( i = 0; ( i < DYNAMO_MAX_MUTEXES )&& !ret; i++ ) {
        sf = &mutex_pool[(i+index) % DYNAMO_MAX_MUTEXES];

        // allocate atomically
        cyg_scheduler_lock();
        if ( sf->key == NULL ) {
            ret = sf;
            ret->key = mutex;
            cyg_mutex_init(&ret->mutex);
        }
        cyg_scheduler_unlock();
    }
}

void
__check_cyg_mutex_destroy( cyg_mutex_t *mutex ) {
    struct safe_mutex *sm;

    sm = find_safe_mutex(mutex);
    if ( sm ) {
        cyg_scheduler_lock();
        if ( !sm->mutex.owner && !sm->mutex.queue.queue ) {
            cyg_mutex_destroy(&sm->mutex);
            sm->key = NULL;
        }
        else {
            DIAG( "mutex is owned or its wait queue is not empty !\n" );
        }
        cyg_scheduler_unlock();
    }
}

cyg_bool_t
__check_cyg_mutex_lock( cyg_mutex_t *mutex ) {
    struct safe_mutex *sm;
    cyg_bool_t ret = false;

    sm = find_safe_mutex(mutex);
    if ( sm ) {
        int overlocked;
        cyg_thread *self = (cyg_thread *)cyg_thread_self();

        // detect consistency problems
        cyg_scheduler_lock();
        overlocked = ( sm->mutex.locked && (self == sm->mutex.owner));
        cyg_scheduler_unlock();

        if ( !overlocked ) {
            ret = cyg_mutex_lock(&sm->mutex);
        }
        else {
            DIAG( "thread already owns this mutex !\n" );
        }
    }
    return ret;
}

cyg_bool_t
__check_cyg_mutex_trylock( cyg_mutex_t *mutex ) {
    struct safe_mutex *sm;
    cyg_bool_t ret = false;

    sm = find_safe_mutex(mutex);
    if ( sm ) {
        ret = cyg_mutex_trylock(&sm->mutex);
    }
    return ret;
}

void
__check_cyg_mutex_release( cyg_mutex_t *mutex ) {
    struct safe_mutex *sm;

    sm = find_safe_mutex(mutex);
    if ( sm ) {
        cyg_mutex_release(&sm->mutex);
    }
}

void
__check_cyg_mutex_unlock( cyg_mutex_t *mutex ) {
    struct safe_mutex *sm;

    sm = find_safe_mutex(mutex);
    if ( sm ) {
        int underlocked;
        cyg_thread *self = (cyg_thread *)cyg_thread_self();

        // detect consistency problems
        cyg_scheduler_lock();
        underlocked = ( !sm->mutex.locked || (self != sm->mutex.owner));
        cyg_scheduler_unlock();

        if ( !underlocked ) {
            cyg_mutex_unlock(&sm->mutex);
        }
        else {
            DIAG( "thread does not own this mutex !\n" );
        }
    }
}

//=============================================================================
// cyg_thread_xxx calls
//=============================================================================

cyg_handle_t SYSCALL(cyg_thread_create)(cyg_addrword_t      sched_info,
                                        cyg_thread_entry_t  *entry,
                                        cyg_addrword_t      entry_data,
                                        char                *name,
                                        void                *stack_base,
                                        cyg_ucount32        stack_size,
                                        cyg_thread          *thread);

cyg_handle_t
__check_cyg_thread_create(cyg_addrword_t      sched_info,
                          cyg_thread_entry_t  *entry,
                          cyg_addrword_t      entry_data,
                          char                *name,
                          void                *stack_base,
                          cyg_ucount32        stack_size,
                          cyg_thread          *thread)
{
    cyg_handle_t handle = 0;
    cyg_process *self = cyg_process_self();

    // no additional check should be needed
    cyg_process_add_thread(self, sched_info, entry, entry_data, name,
                           stack_base, stack_size, &handle, thread);
    return handle;
}

void
__wrap_cyg_thread_create(cyg_addrword_t      sched_info,
                         cyg_thread_entry_t  *entry,
                         cyg_addrword_t      entry_data,
                         char                *name,
                         void                *stack_base,
                         cyg_ucount32        stack_size,
                         cyg_handle_t        *handle,
                         cyg_thread          *thread)
{
    *handle = SYSCALL(cyg_thread_create)(sched_info, entry, entry_data,
                                         name, stack_base, stack_size, thread);
}

void
__check_cyg_thread_suspend(cyg_handle_t thread) {
    cyg_process *self = cyg_process_self();

    // make sure this thread belongs to our process
    if (self && (self == cyg_process_get((cyg_thread *)thread))) {
        cyg_thread_suspend(thread);
    }
}

void
__check_cyg_thread_release(cyg_handle_t thread) {
    cyg_process *self = cyg_process_self();

    // make sure this thread belongs to our process
    if (self && (self == cyg_process_get((cyg_thread *)thread))) {
        cyg_thread_release(thread);
    }
}

void
__check_cyg_thread_resume(cyg_handle_t thread) {
    cyg_process *self = cyg_process_self();

    // make sure this thread belongs to our process
    if (self && (self == cyg_process_get((cyg_thread *)thread))) {
        cyg_thread_resume(thread);
    }
}

void
__check_cyg_thread_kill(cyg_handle_t thread) {
    cyg_process *self = cyg_process_self();

    // make sure this thread belongs to our process
    if (self && (self == cyg_process_get((cyg_thread *)thread))) {
        cyg_thread_kill(thread);
    }
}

//=============================================================================
// diag_xxx calls
//=============================================================================

#define DIAG_PRINTF_MAXBUF (512)

int SYSCALL(diag_vsnprintf)(char *buf,size_t len, const char *fmt, va_list ap);

//! This will not be used
int
__check_diag_printf(const char *fmt, ...) {
    return 0;
}

int
__wrap_diag_printf(const char *fmt, ...)
{
    int ret;
    va_list ap;
    // FIXME: this is not thread safe
    static char buf[DIAG_PRINTF_MAXBUF];

    va_start(ap, fmt);
    // FIXME: we should use a user-mode sprintf
    ret = SYSCALL(diag_vsnprintf)(buf, DIAG_PRINTF_MAXBUF-1, fmt, ap);
    va_end(ap);

    SYSCALL(diag_write_string)(buf);

    return ret;
}

void
__check_diag_write_string(const char *psz) {
    // FIXME: we should copy the string to kernel stack
    diag_write_string(psz);
}
