/**
 ******************************************************************************
 * @file  swi_handler.S 
 * @brief SWI exception handler
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-05-26
 ******************************************************************************
 */

#include "../../../hal/arm/arch/current/arm.inc"
#include <pkgconf/hal_arm.h>
#include <pkgconf/dynamo.h>

// Dynamo thread data defines
#define SAV_R0          (0x00)
#define SAV_R1          (0x04)
#define SAV_R2          (0x08)
#define SAV_R3          (0x0c)
#define SAV_SVC_SPSR    (0x10)
#define SAV_SYSTEM_SP   (0x14)
#define SAV_SYSTEM_LR   (0x18)
#define SAV_SIZE        (0x20)

#define DACR_OFF        (SAV_SIZE)
#define ENTRY_OFF       (DACR_OFF+4)
#define PROCESS_OFF     (ENTRY_OFF+4)
#define USTACK_BASE_OFF (PROCESS_OFF+4)
#define USTACK_TOP_OFF  (USTACK_BASE_OFF+4)

#define KSTACK_OFF      (USTACK_TOP_OFF+4)
#define KSTACK_TOP_OFF  (KSTACK_OFF + CYGOPT_DYNAMO_KSTACK_SIZE)

/* How much params we are going to move from user to kernel stack:
 * default to 12 x 32 bits parameters, which should be enough */
#define USTACK_COPY_BLK (12*4)

        .code   32
        .text
        .section        ".text.swi_handler","ax"
        .global swi_handler

swi_handler:
        // if we came from a privileged mode, execute system call now
        mrs     ip,spsr
        and     ip,ip,#CPSR_MODE_BITS
        cmp     ip,#CPSR_USER_MODE
        ldrne   lr,[lr]                 // lr = function to call
        movnes  pc,lr                   // return to pre-SWI mode

        // we were called from user mode
        ldr     sp,.cyg_current_thread_table
        ldr     sp,[sp]                 // sp = current thread pointer
        add     sp,sp,#THREAD_SIZE      // sp points to extra thread data
        stmia   sp,{r0-r3}              // save a few working registers

        mov     r3,sp                   // put data pointer into common reg
        mrs     r0,spsr
        str     r0,[r3,#SAV_SVC_SPSR]   // save svc_spsr
        mov     r2,lr                   // r2 = swi instr address + 4

        // grant client access to all domains
        mrc     p15,0,r0,c3,c0,0
        ldr     r1,=CYGOPT_DYNAMO_DACR_MASK
        bic     r0,r0,r1                // r0 = DACR & ~M
        ldr     r1,=(CYGOPT_DYNAMO_DACR_MASK & 0x55555555)
        orr     r0,r0,r1
        mcr     p15,0,r0,c3,c0,0        // write back DACR

        // switch to SYSTEM mode (with interrupts still disabled)
        mrs     r0,cpsr
        orr     r0,r0,#CPSR_SYSTEM_MODE
        msr     cpsr,r0
        str     sp,[r3,#SAV_SYSTEM_SP]  // save system_sp
        str     lr,[r3,#SAV_SYSTEM_LR]  // save system_lr
	mov	r1,sp                   // save user stack

        // switch to kernel stack	
        add     sp,r3,#KSTACK_OFF
        add     sp,sp,#CYGOPT_DYNAMO_KSTACK_SIZE

	// copy params from user stack to kernel stack
	add	lr,r1,#USTACK_COPY_BLK
1:	ldr     r0,[lr,#-4]!
        str     r0,[sp,#-4]!
	cmp	lr,r1
        bne     1b

        // and re-enable interrupts
        mrs     r0,cpsr
        bic     r0,r0,#CPSR_IRQ_DISABLE|CPSR_FIQ_DISABLE
        msr     cpsr,r0
        
        // check that SWI instruction is within SWI table
        // if not, just ignore it and return
        ldr     r0,.switab              // r0 = SWI table base
        ldr     r1,.switab_bytes
        ldr     r1,[r1]                 // r1 = SWI table size in bytes
        add     r1,r0,r1
        cmp     r2,r0                   // SWI addr+4 <= SWI table base ?
        bls     return_from_call
        cmp     r2,r1                   // SWI addr+4 >= SWI table end ?
        bhs     return_from_call

        ldr     ip,[r2]                 // ip = system call address
        ldmia   r3,{r0-r3}              // restore registers
        mov     lr,pc
        mov     pc,ip                   // and do system call

        // we normally come back here, except if exit() was called
return_from_call:

        // @todo: implement kill signal

        // disable interrupts
        mrs     lr,cpsr
        orr     lr,lr,#CPSR_IRQ_DISABLE|CPSR_FIQ_DISABLE
        msr     cpsr,lr

        // restore user stack pointer, possibly faulty/unaligned
        ldr     sp,.cyg_current_thread_table
        ldr     sp,[sp]                 // current thread pointer
        add     sp,sp,#THREAD_SIZE      // sp points to extra thread data
        ldr     sp,[sp,#SAV_SYSTEM_SP]  // restore sp
        
        // switch back to SVC mode
        mrs     lr,cpsr
        bic     lr,lr,#CPSR_MODE_BITS
        orr     lr,lr,#CPSR_SUPERVISOR_MODE
        msr     cpsr,lr
        // get a pointer on thread extra data
        ldr     sp,.cyg_current_thread_table
        ldr     sp,[sp]                 // current thread pointer
        add     sp,sp,#THREAD_SIZE      // sp points to extra thread data

        // restore DACR
        stmia   sp,{r0-r2}              // save some registers
        ldr     r0,[sp,#DACR_OFF]       // get user-mode DACR
        ldr     r1,=CYGOPT_DYNAMO_DACR_MASK
        and     r0,r0,r1
        mrc     p15,0,r2,c3,c0,0        // r2 = current DACR
        bic     r2,r2,r1
        orr     r0,r0,r2                // r0 = new DACR
        mcr     p15,0,r0,c3,c0,0        // restore DACR
        ldmia   sp,{r0-r2}

        ldr     lr,[sp,#SAV_SVC_SPSR]   // get svc_spsr
        msr     spsr,lr                 // restore spsr
        ldr     lr,[sp,#SAV_SYSTEM_LR]  // get system_lr
        movs    pc,lr                   // and return from SWI

.cyg_current_thread_table:
        .word	cyg_current_thread_table

.switab:
        .word	switab

.switab_bytes:
        .word	switab_bytes
