//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  vm_alloc.c
* @brief Virtual memory allocation.
*
* This is a very basic virtual memory allocator. It allocates aligned chunks of
* MMU sections (1 Mbyte) using a simple first-fit algorithm.
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-02-06
*******************************************************************************
*/

#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/utils/utils.h>
#include <stdlib.h>

#include "vm_alloc.h"

#include <cyg/hal/hal_mmu.h>

#define MAP_BITS           (32)

#define IS_ALIGNED(_P,_S)  (((_P)&((_S)-1)) == 0 )
#define ALIGN_BASE(_P,_S)  ((_P)&~((_S)-1))
#define ALIGN_TOP(_P,_S)   (((_P)+(_S)-1)&~((_S)-1))

#define GET_BIT(_n)        ((vm.map[(_n) >> 5] >> ((_n) & 0x1f)) & 1)
#define SET_BIT(_n)        { vm.map[(_n) >> 5] |= 1 << ((_n) & 0x1f); }
#define CLR_BIT(_n)        { vm.map[(_n) >> 5] &= ~(1 << ((_n) & 0x1f)); }

struct vm_allocator {
    int init;
    CYG_ADDRESS base;
    cyg_uint32 *map;
    cyg_uint32 nmap;
    cyg_uint32 nsec;
};

static struct vm_allocator vm = { .init = 0 };

/** Allocator initialisation.
 *
 * This function is called to setup the VM allocator.
 *
 * @param base Physical RAM base address must be page-aligned
 * @param table A array of size @ref nb_32mb
 * @param nb_32mb The number of available chunks of 32 MBytes
 * @return 0 upon success, -1 otherwise
 */
int
vm_init( CYG_ADDRESS base, unsigned int *table, int nb_32mb ) {
    int i, err = -1;

    vm.base = ALIGN_TOP(base, MMU_SECTION_SIZE);

    if (( vm.base == base )&&
        ( nb_32mb > 0 )) {

        vm.map = table;
        vm.nmap = nb_32mb;
        vm.nsec = vm.nmap*MAP_BITS;

        for ( i = 0; i < vm.nmap; i++ ) {
            vm.map[i] = 0;
        }
        err = 0;
        vm.init = 1;
    }

    return err;
}

/** VM allocator.
 *
 * @param size Size to allocate
 * @return An aligned address if allocation was successful, 0 otherwise.
 */
CYG_ADDRESS
vm_allocate(size_t size) {
    int i, sections, in_free_block, index = 0;
    CYG_ADDRESS addr = 0;

    // convert into sections
    size = (size + MMU_SECTION_SIZE-1)/MMU_SECTION_SIZE;

    if ( vm.init && size ) {

        // look for a free block
        in_free_block = 0;
        sections = size;

        for ( i = 0; ( i < vm.nsec )&& sections; i++ ) {

            if ( in_free_block ) {
                // measuring free block
                if ( GET_BIT(i) ) {
                    // section is already allocated
                    in_free_block = 0;
                    sections = size;
                }
                else {
                    // section is free
                    sections--;
                }
            }
            else {
                // skip full chunks of 32 sections
                if ( !( i & 0x1f )&&( 0 == ~vm.map[i >> 5] )) {
                    i += MAP_BITS-1;
                }
                else {
                    if ( !GET_BIT(i) ) {
                        // section is free
                        in_free_block = 1;
                        index = i;
                        sections--;
                    }
                }
            }
        }
        if ( !sections ) {
            // mark sections as allocated
            for ( i = index; i < index+size; i++ ) {
                SET_BIT(i);
            }
            addr = vm.base + index*MMU_SECTION_SIZE;
        }
    }

    return addr;
}


/** VM slot deallocator.
 *
 * @param addr The slot address to free
 */
void
vm_free( CYG_ADDRESS addr, size_t size ) {
    int i, index, nsec;
    CYG_ADDRESS top, base;

    base = ALIGN_BASE( addr, MMU_SECTION_SIZE );
    top = ALIGN_TOP( addr+size, MMU_SECTION_SIZE );

    if ( vm.init &&
         ( base >= vm.base )&&
         ( top <= vm.base + vm.nsec*MMU_SECTION_SIZE )) {

        // mark sections as allocated
        index = (base-vm.base)/MMU_SECTION_SIZE;
        nsec = (top-base)/MMU_SECTION_SIZE;

        for ( i = index; i < index+nsec; i++ ) {
            CLR_BIT(i);
        }
    }
}
