/**
*******************************************************************************
* @file  include/rtmon.h
* @brief Real-Time Monitor
*
* This package provides a real-time monitoring tool.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2006-12-01
*******************************************************************************
*/
#ifndef CYGONCE_RTMON_H
#define CYGONCE_RTMON_H

// Event selection
#define RTMON_EVENT_THREAD_SWITCH   (1 << 1)
#define RTMON_EVENT_SCHED_LOCK      (1 << 2)
#define RTMON_EVENT_CHECK_DSR_CNT   (1 << 3)
#define RTMON_EVENT_PROFILING       (1 << 4)
#define RTMON_EVENT_USER            (1 << 5)
#define RTMON_EVENT_USER_U32        (1 << 6)
#define RTMON_EVENT_THREAD_STATE    (1 << 7)

// Module selection
#define RTMON_DRIVER_AAI            (1 << 1)
#define RTMON_DRIVER_SDCARD         (1 << 2)


#define RTMON_EVENT_NONE            (0)
#define RTMON_EVENT_ALL             (0xffffffff)

#define RTMON_IRQ_MAX               (32)

struct rtmon_cfg {
    cyg_uint32 isr_mask;
    cyg_uint32 dsr_mask;
    cyg_uint32 event_mask;
    cyg_uint32 driver_mask;
    cyg_uint32 dsr_latency[RTMON_IRQ_MAX];
    const char *logfile;
};

#define RTMON_ASSERT( _cond_, _fmt_, args... ) { \
    do {                                         \
        if ( !(_cond_)) {                        \
            rtmon_flush( _fmt_, ##args );        \
        }                                        \
    } while (0);                                 \
}

// RTMon control functions
void rtmon_start               (struct rtmon_cfg *cfg );
void rtmon_stop                (void);
void rtmon_flush               (const char *format, ...);
// RTMon log functions
void rtmon_log_driver_start    (cyg_uint32 user_event);
void rtmon_log_driver_stop     (cyg_uint32 user_event);
void rtmon_set_user_event_label(cyg_uint32 user_event, const char *name  );
void rtmon_log_user_start      (cyg_uint32 user_event);
void rtmon_log_user_stop       (cyg_uint32 user_event);
void rtmon_log_user_start_timed(cyg_uint32 user_event);
void rtmon_log_user_stop_timed (cyg_uint32 user_event, cyg_uint32 tlimit );

void rtmon_set_user_u32_label  (cyg_uint32 user_event, const char *name  );
void rtmon_log_user_u32        (cyg_uint32 user_event, cyg_uint32 value );

// RTMon API for getting event buffer without killing the app
cyg_uint32 rtmon_get_max_buffer_size(void);
cyg_uint32 rtmon_copy_log(void *log_data, const char *fmt, ...);

/*! \mainpage Real-Time Monitor


<center><h2>RTmon, un gnrateur de trace temps-rel pour eCos</h2></center>

<p>
<div align="center">
<hr>
\ref intro_sec <br>
\ref sec_step1 &nbsp;|&nbsp;
\ref sec_step2 <br>
\ref sec_step3 &nbsp;|&nbsp;
\ref sec_step4 &nbsp;|&nbsp;
\ref sec_step5 &nbsp;|&nbsp;
<hr>
</div>

\section intro_sec Introduction

Ce package a pour but de permettre de tracer prcisment le fonctionnement du 
noyau eCos (interruptions, excution des tches, etc.).

La mthode consiste  enregistrer en temps rel dans un buffer circulaire les 
vnements "intressants": ISR, DSR, variation du scheduler lock, vnements 
dfinis par l'utilisateur, etc. A un instant choisi, qui peut correspondre  
une condition prcise (par exemple, une routine de DSR qui s'excute trop 
longtemps aprs une ISR), ce buffer est envoy sur une UART (diag) et 
rcupre par un PC, qui peut la visualiser graphiquement.

Cet outil peut vous aider  rsoudre des problmes lis au temps-rel, ou 
simplement voir comment se comporte votre application et vrifier que tout 
fonctionne comme prvu.

<hr>


\section sec_evol Evolutions (HEAD & v3.0)
Cette section dcrit les volutions effectues sur la HEAD & v3.0 de eCos.

<h2>eCos HEAD</h2>
Voici les volutions apportes sur la HEAD de eCos

<h3>Association de labels avec les user events et u32</h3>
les fonctions suivantes permettent d'associer une label  une user event ou 
un u32 de manire  ce que ce label s'affiche dans GTKWave.

\code
   void rtmon_set_user_event_label(cyg_uint32 user_event, const char *name );
   void rtmon_set_user_u32_label  (cyg_uint32 user_event, const char *name );
\endcode

<h3>Log de l'tat des threads (RUNNING/SLEEPING)</h3>
NB:Pour en bnficier, ils faut avoir updat sur la HEAD le fichier <tt>sched.cxx</tt>
et le rpertoire <tt>SoftParrot5/ecos/ecos-current/packages/services/rtmon</tt>.

Une fois la trace obtenue, il suffit d'appeler rtd2lxt en lui passant l'option <tt>--thread-state</tt>
Pour traduire la trace, il suffit d'utiliser rtd2lxt:
\code
    $ rtd2lxt --thread-state -e ~/tftp/program.elf dump
\endcode

Dans GTKWave, pour chaque thread apparait une nouvelle ligne qui reprsente l'tat de la thread<br>
Le "RTMon report", comporte une nouvelle section qui prsente les statistiques concernant la dure
de l'tat RUNNING des threads.

<p>

<p>
<h2>eCos v3.0</h2>
<h3>Nouvelles fonctionnalits</h3>
Voici les volutions apportes sur eCos v3.0
<ul>
<li> le nombre de user events utilisables est de <b>255</b> contre 32 auparavant.
<li> log de la <b>priorit de la thread courante</b>.
<li> log des <b>inversions de priorit</b>
</ul>
<p>

<h3>Changement d'API</h3>
Voici les modifications de l'API RTMon sur eCos v3.0
<ul>
<li> le log des user events est actif uniquement si le champ event_mask de la
configuration est initialis avec <b>RTMON_EVENT_USER</b>(le champ user_mask est supprim). 
<b>255 user events</b> sont maintenant utilisables contre 32 auparavant
<li> le log des user values est actif uniquement si le champ event_mask de la
configuration est initialis avec <b>RTMON_EVENT_USER_U32</b>(toujours limit  32 valeurs).
<li> le champ module_mask de la structure de configuration est renomm en <b>driver_mask</b>
<li> le champ <b>user_mask</b> de la structure de configuration est supprim
</ul>

<hr>


\section sec_step1 Etape 1: Gnration des outils 

Pour gnrer les outils rtd2lxt et getrtd, suivre la procdure suivante:
\code
$ cd ecos-current/packages/services/rtmon/current/host
$ make
\endcode
NB:Sous Linux, il faut au pralable avoir install les packages 
"libbz2-dev", "libzlib-ruby", "zlib1g-dev" et "flex-old"

NB: eventuellement faire un "make clean" si les outils avaient dj t gnrs
prcdemment.
<p>
Pour gnrer gtkwave pour Linux :
\code
$ cd ecos-current/packages/services/rtmon/current/host/gtkwave
$ tar -xzf gtkwave-3.0.18-linux.tgz
...
$ cd gtkwave-3.0.18
$ ./configure
...
$ make
\endcode
<p>
Pour installer gtkwave pour Cygwin, dcompresser simplement l'archive 
gtkwave-3.0.18-cygwin.tgz.
<p>
Pour installer Kcachegrind (sous Linux): 
\code
$ apt-get install kcachegrind
\endcode
ou <a href="http://sourceforge.net/projects/wincachegrind/">
Wincachegrind</a> (sous Windows)
<p>
(voir \ref pg_instal pour plus de dtails)
<p>
<hr>


\section sec_step2 Etape 2: Configuration d'eCos

A l'aide du configtool, il faut inclure le package RTMon dans la configuration 
d'eCos.<br>
Adaptez la taille de la trace  vos besoin (la dure dpends principalement 
des informations traces: profiling, scheduler, ...)<br>
Verifiez que eCos et le code de votre application sont bien compils en -02.
<p>
(voir \ref pg_instal pour plus de dtails)
<p>
<hr>
\section sec_step3 Etape 3: Instrumentation du code

Instrumentez votre code en fonction de vos besoins. Le minimum tant de 
dfinir une configuration et d'appeler <tt>rtmon_start( )</tt>.
<p>
Ci-dessous la configuration minimale conseille:
\code
#include <cyg/rtmon/rtmon.h>

void cyg_user_start(void) 
{
    struct rtmon_cfg cfg;

    cfg.isr_mask    = RTMON_EVENT_ALL;
    cfg.dsr_mask    = RTMON_EVENT_ALL;
    cfg.event_mask  = RTMON_EVENT_THREAD_SWITCH |
                      RTMON_EVENT_USER          |
                      RTMON_EVENT_USER_U32      |
                      RTMON_EVENT_CHECK_DSR_CNT ;
    cfg.driver_mask = RTMON_EVENT_NONE;
    memset( cfg.dsr_latency, 0, sizeof(cyg_uint32)*RTMON_IRQ_MAX );
    cfg.dsr_latency[3]  = 80;     // P5 AAI (80 us latency)
    cfg.dsr_latency[21] = 5000;   // P5 TIMER3 (5 ms latency)
    cfg.logfile = NULL;

    // ....

    rtmon_start(&cfg);
}
\endcode
<p>
Il est important de configurer le champ <i>event_mask</i> selon ses besoins:
<ul><li>RTMON_EVENT_THREAD_SWITCH: log les taches actives.
    <li>RTMON_EVENT_USER         : active la trace des user events.
    <li>RTMON_EVENT_USER_U32     : active la trace des user values.
    <li>RTMON_EVENT_PROFILING    : active la fonction de profiling.
    <li>RTMON_EVENT_CHECK_DSR_CNT: vrifie que toute DSR poste est traite 
                                   avant l'ISR suivante.
    <li>RTMON_EVENT_SCHED_LOCK   : log les informations relatives au scheduler
</ul>

Ainsi que le <i>driver_mask</i> selon ses besoins:
<ul><li>RTMON_DRIVER_AAI: Active les traces du driver AAI.
</ul>
<p>
Il est galement trs intressant d'encadrer certaines portions de code 
(boucles, fonctions,...) en logguant un user event. De cette manire, gtkwave 
fournit une reprsentation du temps pass dans ces portions de code.

\code
#include <cyg/rtmon/rtmon.h>

    //...

   rtmon_set_user_event_label(1, "loop" ); // display label in GTKwave

   rtmon_log_user_start(1);

   for(i=0; i<=I_MAX; i++)
   {
       //traitement
   }
   rtmon_log_user_stop(1);
\endcode

<p>
Pour tre certain de rcuprer la trace  un moment prcis, il est possible 
placer un appel  <tt>rtmon_flush( )</tt> dans le code. Sinon il suffit 
d'interrompre l'execution du code avec GDB (Ctrl+C) et d'xcuter la
fonction <tt>rtmon_flush( )</tt> en tapant :
\code
    call rtmon_flush(0)
\endcode
<p>
(voir \ref pg_api pour plus de dtails)
<p>
<hr>

\section sec_step4 Etape 4: Obtention de la trace

Avant de lancer l'application, il faut ouvrir une console pour se mettre en 
attente de la rception du buffer RTMon. Le dump doit tre envoye sur 
l'entre standard de <tt>getrtd</tt>. Par exemple, si l'UART de diagnostic 
est relie au port srie d'une sonde BDI (et accessible sur le port 8 de 
son adresse IP), on pourrait utiliser:
<p>
<h3>Sous Linux:</h3>
\code
    $ netcat sonde_ip 8 | getrtd -o dump
\endcode
si elle est directement relie au port COM du pc, on peut utiliser:
\code
    $ cat /dev/ttyS0 | getrtd -o dump
\endcode

NB: pour configurer le port COM, il suffit de 
\code
    $ stty -F/dev/ttyS0 raw -echo 115200
\endcode

<p>
<h3>Sous Cygwin:</h3>
<span style="color: #d00">
Il faut avoir install le package <b>netcat</b> et la partition sur laquelle on 
enregistre le dump doit tre en <b>"binmode"</b>: </span>
\code
    $ nc sonde_ip 8 | getrtd -o dump
\endcode

<p>
(voir \ref pg_getdump pour plus de dtails)
<p>
<hr>

\section sec_step5 Etape 5: Traduction de la trace

Pour traduire la trace, il suffit d'utiliser rtd2lxt:
\code
    $ rtd2lxt -e ~/tftp/program.elf dump
\endcode
<p>
<tt>rtd2lxt</tt> gnre  partir du fichier <tt>dump</tt> trois fichiers en
sortie: <tt>dump.lxt</tt>, <tt>dump.sav</tt> et <tt>dump.prof</tt>:
<p>
<ul>
<li><tt>dump.lxt </tt>: trace au format GTKWave
<li><tt>dump.sav </tt>: mise en forme pour GTKWave (ordre des traces, 
labels)
<li><tt>dump.prof</tt>: mise en forme pour Kcachegrind
<li><tt>dump.html</tt>: rapport statistique RTMon de la trace
</ul>

<p>
Pour afficher la trace, lancez GtkWave:
\code
    $ gtkwave -A dump.lxt
\endcode


<p>
(voir \ref pg_getdump pour plus de dtails)
<p>
<hr>

\section sec_step6 Etape 6: Analyse de la trace

Le rapport gnr au format html permet d'avoir un rsum de la trace, 
notamment concernant les durs minimium, moyennes et maximum des ISR, DSR, Latence, ...
<p>
<span style="color: #d00">- Les mesures obtenues doivent tre vrifies sur 
GTKWave et explicables (les valeurs excessives sont repres en orange ou 
rouge (selon la criticit).<span>
<p>
GTKWave permet de visualiser graphiquement la trace est les vnements loggus.
<p>
Dans le cas d'une incohrence dans les traces, merci de sauvegarder le fichier 
elf et la trace dump correspondante et de me contacter  l'adresse suivante: 
gregoire.etienne@parrot.com 

*/


/*! \page pg_tools RTMon Tools

<h2>Outils d'analyse d'une trace temps-rel</h2>

<p>
<div align="center">
<hr>
\ref gtkwave_sec  &nbsp;|&nbsp;
\ref kcachegrind_sec  &nbsp;|&nbsp;
\ref rtmonreport_sec  &nbsp;|&nbsp;
<hr>
</div>

\section gtkwave_sec gtkwave

Cet outil de visualisation de traces est disponible sous Linux et CygWin.
Vous pouvez en rcuprer le code source, ou une version compile pour CygWin
dans le rpertoire <tt>rtmon/current/host/gtkwave</tt>. Lire le fichier
<tt>README.cygwin</tt> dans ce mme rpertoire pour plus d'informations.
<p>
Lorsque GTKWave est install, il vous suffit pour visualiser la trace
<tt>dump.lxt</tt> de taper:

\code
    $ gtkwave -f dump.lxt -a dump.sav
\endcode
ou
\code
    $ gtkwave -A dump.lxt
\endcode

Pour plus d'informations sur l'utilisation de GTKWave, voir la
<a
href="http://home.nc.rr.com/gtkwave/gtkwave.pdf">documentation</a>.

Voici un exemple d'affichage:
<p>
<div>
<img alt="screenshot GTKWave" src="gtkwave.png">
</div>

<hr>
\section rtmonreport_sec RTMon Report
L'outil rtd2lxt gnre un rapport statistique de la trace au format html.
<p>
Ce rapport indique pour chaque vecteur d'interruption le min, le max et 
la moyenne de la dure de chaque ISR et DSR ainsi que leur latence.

<p>
La couleur associe (jaune, orange ou rouge) prcise la criticit suppose 
(par exemple, une DSR de plus de 150us ne permet pas d'utiliser le HP ou 
les micros de l'AAI  une frquence de 48 kHz).

<p>
Les dures colores en orange ou rouge doivent toutes tre justifies en 
observant la trace sous gtkwave.


<p>
<div>
<img alt="screenshot RTMon report" src="report.png">
</div>

<p>
Ce rapport indique pour chaque vnement utilisateur:
<ul>
<li>la dure cumulative en us ainsi qu'en % par rapport  la dure de la trace.
Cette fonctionnalit est principalement utile pour avoir une ide du temps CPU
consomm par un algo. Toutefois, cette mesure peut se trouver fausse si la 
section de code instrumente est trop souvent prempte.
<li>le min, le max et la moyenne de la dure d'un vnement. Pour le min et le
max, le rapport prcise la localisation dans la trace. Cette focntionnalit
est utile pour dtecter des dures anormalement longues 
</ul>
<div>
<img alt="screenshot RTMon report" src="report02.png">
</div>


<hr>
\section kcachegrind_sec Kcachegrind ou Wincachegrind
Cet outil de profiling permet d'obtenir une reprsentation graphique du temps 
pass dans chaque fonction du code. Cet outil a une granularit de 50us, par 
consquent ne peut tre reprsentatif pour les fonctions dont la dure est 
bien plus courte que cette priode.

Pour analyser la trace, il suffit de taper la commande suivante ( le fichier 
"dump.prof" est gnr par l'outil rtd2lxt en mme temps que les fichiers 
.lxt et .sav.
\code
    $ kcachegrind dump.prof
\endcode

Voici un exemple d'applicatif de lecture/criture sur SDCard qui montre que 
sur 1,218 seconde, l'on passe en cumul 900 ms dans la fonction 
<tt>p5dev_go_idle</tt> (ce qui correspond  la tache idle de eCos):
<p>
<div>
<img alt="screenshot Kcachegrind" src="kcachegrind.png">
</div>

Note: il faut savoir que cette fonctionnalit de RTMon ncessite l'utilisation 
du Timer 2 du P5. Il reste donc les timers 0 et 1 de disponibles pour 
l'applicatif.


*/

/*! \page pg_instal RTMon instal

<center><h2>Installation et compilation des outils RTMon</h2></center>

<p>
<div align="center">
<hr>
\ref emb_ssec  &nbsp;|&nbsp;
\ref host_ssec
<hr>
</div>

\section emb_ssec Package RTMON pour eCos
Pour utiliser RTmon, il vous faut installer le package <b>rtmon</b>
(<tt>ecos-current/packages/services/rtmon</tt>):
<p>
<div><img src="rtmon01.png" alt="Installation de RTmon"></div>
<p>

choisir la taille en kilo-octets (<tt>CYGPKG_RTMON_LOGSIZE</tt>) du buffer
circulaire de trace (par dfaut 512 kB) et le nombre max de taches 
(<tt>CYGPKG_RTMON_NBTHD</tt>) qui seront cres au cours du programme(par 
dfaut 256). Les autres paramtres de configuration (filtre d'vnements, 
dlais) se configurent  l'excution.
<br>
\section host_ssec Gnration des outils PC
Pour gnrer les outils rtd2lxt et getrtd, suivre la procdure suivante:
\code
$ cd ecos-current/packages/services/rtmon/current/host
$ make
\endcode
NB:Sous Linux, il faut au pralable avoir install les packages 
"libbz2-dev", "libzlib-ruby", "zlib1g-dev" et "flex-old"

NB: Faire un "make clean" si les outils avaient dj t gnrs
prcdemment.
<p>
Pour gnrer gtkwave pour Linux, dans un premier temps, un des packages 
"libgtk1.2-dev" ou "libgtk2.0-dev" est requis, puis :
\code
$ cd ecos-current/packages/services/rtmon/current/host/gtkwave
$ tar -xzf gtkwave-3.0.18-linux.tgz
...
$ cd gtkwave-3.0.18
$ ./configure
...(rpondre 1 ou 2 suivant le package libgtk install)
$ make
\endcode
NB: .
<p>
Pour gnrer gtkwave pour Cygwin, dcompresser simplement l'archive 
gtkwave-3.0.18-cygwin.tgz.
<p><p>
Pour installer Kcachegrind (uniquement sous Linux pour le moment) :

\code
$ apt-get install kcachegrind
\endcode

*/





/*! \page pg_api RTMon API

<p>
<div align="center">
<hr>
\ref api_sec  &nbsp;|&nbsp;
\ref flush_sec
<hr>
</div>


\section api_sec API

Une fois le package compil dans eCos, l'API suivante est disponible en
incluant le header <tt>&lt;cyg/rtmon/rtmon.h&gt;</tt>:
<p>
<ul>
<li>\ref ssec_rtmon_stop
<li>\ref ssec_rtmon_flush
<li>\ref ssec_rtmon_log_user_start
<li>\ref ssec_rtmon_log_user_stop
<li>\ref ssec_rtmon_log_user_start_timed
<li>\ref ssec_rtmon_log_user_stop_timed
<li>\ref ssec_rtmon_log_user_u32
<li>\ref ssec_rtmon_assert
<li>\ref ssec_rtmon_log_caller
<li>\ref ssec_rtmon_log_pc
</ul>
<p>
<hr>
\subsection ssec_rtmon_start rtmon_start
<b>void rtmon_start( struct rtmon_cfg *cfg )</b>

<p>
Le moniteur se configure et se lance lors de l'excution  l'aide
de la fonction <tt>rtmon_start()</tt>,  laquelle la structure ci-dessous doit
tre passe:
<p>
\code
struct rtmon_cfg {
    cyg_uint32 isr_mask;
    cyg_uint32 dsr_mask;
    cyg_uint32 event_mask;
    cyg_uint32 driver_mask;
    cyg_uint32 dsr_latency[RTMON_IRQ_MAX];
    const char *logfile;
};
\endcode
<p>

Les champs de cette structure permettent de slectionner les vnements 
filtrer et les conditions d'arrt du moniteur:

<ul>
<li><i>isr_mask:</i> Ce masque de 32 bits permet de choisir pour quels vecteurs
   d'interruption les routines d'ISR doivent tre loggues. Ces vecteurs dpendent
   de la plateforme hardware; par exemple pour logguer les interruptions de
   l'UART2 du Parrot5, il faut inclure <tt>&lt;cyg/hal/p5dev.h&gt;</tt> et
   inclure dans le masque la valeur (<tt>1 << CYGNUM_HAL_INTERRUPT_UART2</tt>).
   <p>
   Pour logguer toutes les (resp. aucune) interruptions, il suffit d'utiliser le
   masque <tt>RTMON_EVENT_ALL</tt> (resp. <tt>RTMON_EVENT_NONE</tt>).
   <p>
<li><i>dsr_mask</i>: Ce masque de 32 bits permet de choisir pour quels vecteurs
   d'interruption les routines DSR doivent tre loggues.
   <p>
<li><i>event_mask</i>: Ce masque de 32 bits permet de choisir des
   vnements/variables  logguer:
      <ul>
      <li><tt>RTMON_EVENT_THREAD_SWITCH</tt>: Log des changements de taches.
      <li><tt>RTMON_EVENT_USER</tt>: Active la trace des user events.
      <li><tt>RTMON_EVENT_USER_U32</tt>: Active la trace des user values.
      <li><tt>RTMON_EVENT_SCHED_LOCK</tt>: Log des vnements relatifs au scheduler
         lock: variation de la valeur du lock et fonction appelante.
      <li><tt>RTMON_EVENT_CHECK_DSR_CNT</tt>: Vrification de la file d'attente
         DSR: si deux DSR sont successivement postes pour une mme interruption 
         des (i.e. si la premire DSR n'a pas t traite  temps), l'arrt 
         (flush) du moniteur se produit.
      <li><tt>RTMON_EVENT_PROFILING</tt>: Active la fonction de profiling qui trace 
         toutes les 50us la ligne de code executee.
      </ul>
   Comme prcdemment les masques <tt>RTMON_EVENT_ALL</tt> et
   <tt>RTMON_EVENT_NONE</tt> permettent de slectionner tous les vnements ou
   aucun des vnements.
   <p>

<li><i>driver_mask</i>: Ce masque de 32 bits permet de choisir les modules
    tracer. Ces traces sont implmentes dans certains packages et sont
   compils si le package RTMon est inclus.:
   <ul>
      <li><tt>RTMON_DRIVER_AAI</tt>: Log des vnements relatifs à l'AAI
   </ul>
   Comme prcdemment les masques <tt>RTMON_EVENT_ALL</tt> et
   <tt>RTMON_EVENT_NONE</tt> permettent de slectionner tous les modules ou
   aucun des modules.
   <p>

   <p>
<li><i>dsr_latency[]</i>: Ce tableau de valeur non-signes sur 32 bits permet
   de dfinir, pour chaque vecteur d'interruption, une latence limite au-del de
   laquelle l'arrt du moniteur est dclench. La latence est dfinie comme le
   dlai (en micro-secondes) entre la fin de la routine ISR et le dbut de la
   routine DSR. Une valeur de 0 quivaut  une latence infinie (i.e. jamais
   vrifi).
   <p>
<li><i>logfile</i>: Le nom d'un fichier dans lequel le moniteur crit la trace
   d'vnements au moment du flush.
   <span style="color: #d00">Note: cette fonction n'est pas encore
   disponible</span>
</ul>
<p>
Au moment de l'appel  <tt>rtmon_start()</tt>, le moniteur se met en route
et commence  enregistrer les vnements dans son buffer circulaire.
<p>


<p>
<hr>
\subsection ssec_rtmon_stop rtmon_stop
   <b>void rtmon_stop(void)</b>
   <p>
   Cette fonction permet d'arrter le moniteur. Celui-ci ne peut tre redmarr
   qu'en appelant la fonction <tt>rtmon_start()</tt>.
   <hr>
\subsection ssec_rtmon_flush rtmon_flush
   <b>void rtmon_flush(const char *format, ...)</b>
   <p>
   Cette fonction arrte le moniteur et envoie le buffer circulaire de trace
   sur la sortie diagnostic. La chaine de caractres <tt>format</tt> permet de
   passer un message et des arguments ( la manire de <tt>sprintf()</tt>) qui
   seront enregistrs dans la trace, normalement pour expliquer la raison de
   l'arrt.
   <hr>
\subsection ssec_rtmon_log_user_start rtmon_log_user_start
   <b>void rtmon_log_user_start(cyg_uint32 user_event)</b><p>
   Cette fonction permet de logguer le dbut d'un vnement utilisateur, numrot
   par le paramtre <tt>user_event</tt> (0 <= <tt>user_event</tt> < 256).
   <hr>
\subsection ssec_rtmon_log_user_stop rtmon_log_user_stop
   <b>void rtmon_log_user_stop( cyg_uint32 user_event)</b><p>
   Cette fonction permet de logguer la fin d'un vnement utilisateur, numrot
   par le paramtre <tt>user_event</tt> (0 <= <tt>user_event</tt> < 256).
   <hr>
\subsection ssec_rtmon_log_user_start_timed rtmon_log_user_start_timed
   <b>void rtmon_log_user_start_timed(cyg_uint32 user_event)</b><p>
   Cette fonction permet de logguer le dbut d'un vnement utilisateur
   chronomtr, numrot par le paramtre <tt>user_event</tt>
   (0 <= <tt>user_event</tt> < 256).
   <hr>
\subsection ssec_rtmon_log_user_stop_timed rtmon_log_user_stop_timed
   <b>void rtmon_log_user_stop_timed( cyg_uint32 user_event,
   cyg_uint32 tlimit )</b><p>
   Cette fonction permet de logguer la fin d'un vnement utilisateur
   chronomtr, numrot par le paramtre <tt>user_event</tt>
   (0 <= <tt>user_event</tt> < 256). Le paramtre
   <tt>tlimit</tt> dfinit une latence limite en micro-secondes. Si le dlai entre
   le dbut et la fin de l'vnement est suprieur  cette limite, le moniteur
   est arrt.
   <hr>
\subsection ssec_rtmon_log_user_u32 rtmon_log_user_u32
   <b>void rtmon_log_user_u32( cyg_uint32 user_event, cyg_uint32 value )</b>
   <p>
   Cette fonction permet de logguer la valeur d'une variable 32 bits,
   numrote par le paramtre <tt>user_event</tt> (0 <= <tt>user_event</tt> < 32).
   <hr>
\subsection ssec_rtmon_set_user_event_label rtmon_set_user_event_label
   <b>void rtmon_set_user_event_label(cyg_uint32 user_event, const char *name )</b>
   Cette fonction associe un label  un vnement. Ce label sera affich dans 
   GTKwave<hr>
\subsection ssec_rtmon_set_user_u32_label rtmon_set_user_u32_label
   <b>void rtmon_set_user_u32_label(cyg_uint32 user_event, const char *name )</b>
   Cette fonction associe un label  une valeur u32. Ce label sera affich dans 
   GTKwave<hr>
\subsection ssec_rtmon_assert RTMON_ASSERT
   <b>#define RTMON_ASSERT( cond, fmt, args... ) ...</b><p>
   Cette macro permet de vrifier tester une condition et d'arrter le moniteur
   si cette condition choue. La chaine de caractres <tt>fmt</tt> et les
   paramtres <tt>args</tt> permettent de passer un message explicatif.
   <hr>
\subsection ssec_rtmon_log_caller RTMON_LOG_CALLER
   <b>#define RTMON_LOG_CALLER( ) </b><p>
   Cette macro permet de logger (<tt>fonction()[fichier:ligne]</tt>)
   l'appelant de la fonction ou est place cette macro.
   NB: Cette macro peut-tre utilise autant de fois que ncessaire. 
   L'affichage dans GtkWave conserve le dernier appelant loggu jusqu'au 
   prochain log.<hr>
\subsection ssec_rtmon_log_pc RTMON_LOG_PC
   <b>#define RTMON_LOG_PC( ) </b><p>
   Cette macro permet de logger (<tt>fonction()[fichier:ligne]</tt>) diffrents 
   passages dans le code.
   NB: Cette macro peut-tre utilise autant de fois que ncessaire. L'affichage 
   dans GtkWave conserve le dernier point de passage logg jusqu'au prochain 
   log.<hr>


\section flush_sec Conditions d'arret du moniteur

Le moniteur cesse d'enregistrer les vnements lorsque l'une des conditions
suivantes est remplie:

<ul>
<li>Une latence limite de DSR ou utilisateur a t dpasse
<li>Une assertion <tt>RTMON_ASSERT()</tt> dans le code applicatif a chou
<li>La fonction <tt>rtmon_flush()</tt> a directement t appele
</ul>

Il envoie alors sur la sortie <i>diagnostic</i> son buffer de trace, qui peut
tre rcupr sur un PC pour exploitation.
<p>
La sortie serie <i>diagnostic</i> est par dfaut l'UART0 mais peut-tre 
redirige vers une autre UART en utilisant configtool:<br>
<tt>eCos HAL -> ARM architecture -> ARM ARM9 architecture -> Parrot5/5+ 
development board -> Diagnostic serial port</tt>

*/




/*! \page pg_getdump RTMon Get Dump

<center><h2>Obtention de la trace RTMon</h2></center>

<p>
<div align="center">
<hr>
\ref sec_getrtd  &nbsp;|&nbsp;
\ref sec_rtd2lxt
<hr>
</div>

\section sec_getrtd getrtd :

Cet outil permet de rcuprer et stocker dans un fichier la trace envoye sur
l'UART de diagnostic.

\code
getrtd: Download and save an RTmon dump
usage: getrtd [-h] [-o <filename> [-l]] <infile>
 -h             : Display this help
 -o <filename>  : Write dump to <filename> instead of stdout
 -l             : Run in (infinite) loop mode and save each new dump in <filenameXXXX> where XXXX is a time stamp and <filename> was set with option -o
\endcode

Le dump doit tre envoye sur l'entre standard de <tt>getrtd</tt>. Par
exemple, si l'UART de diagnostic est relie au port srie d'une sonde BDI
(et accessible sur le port 8 de son adresse IP), on pourrait utiliser:

<h3>Sous Linux:</h3>
\code
    $ netcat sonde_ip 8 | getrtd -o dump
\endcode
si elle est directement relie au port COM du pc, on peut utiliser:
\code
    $ cat /dev/ttyS0 | getrtd -o dump
\endcode

NB: dans le 2e cas,il convient de configurer le port COM 
\code
    $ stty -F/dev/ttyS0 raw -echo 115200
\endcode

<p>
<h3>Sous Cygwin:</h3>

<span style="color: #d00">
Il faut avoir install le package <b>netcat</b> et la partition sur laquelle on 
enregistre le dump doit tre en <b>"binmode"</b>(pour viter une translation 
automatique LF -> CRLF)</span>
\code
    $ nc sonde_ip 8 | getrtd -o dump
\endcode
Le type de la partition est donn par la commande <tt>mount</tt>
Exemple:
\code
   $ mount
   C:\Cygwin\bin on /usr/bin type system (binmode)
   C:\Cygwin\lib on /usr/lib type system (binmode)
   C:\Cygwin on / type system (binmode)
   d: on /ecos-d type user (textmode)
   ...
\endcode


\section sec_rtd2lxt rtd2lxt
Cet outil permet de convertir le format interne (RTD) de trace utilis par
RTmon au format LXT utilis par l'outil
<a href="http://home.nc.rr.com/gtkwave/">GTKWave</a>. Si le fichier ELF
correspondant au code ayant produit la trace est disponible, il peut servir
 afficher des informations symboliques (appels de fonctions).

<tt>rtd2lxt</tt> gnre  partir du fichier <tt>dump</tt> 4 fichiers en
sortie: <tt>dump.html</tt>, <tt>dump.lxt</tt>, <tt>dump.sav</tt> et <tt>dump.prof</tt>:
<p>
<ul>
<li><tt>dump.lxt </tt>: trace au format GTKWave
<li><tt>dump.sav </tt>: mise en forme pour GTKWave (ordre des traces, 
labels)
<li><tt>dump.prof</tt>: mise en forme pour Kcachegrind
<li><tt>dump.html</tt>: rapport statistique RTMon de la trace
</ul>
<p>

\code
Usage: rtd2lxt [-v] [-e <exefile>] [--start-time <n>] [--end-time <n>] <dumpfile> [<lxtfile> <savefile>]
 -h               : Display this help
 -e <filename>    : ELF file corresponding to this <dumpfile>
 --start-time <n> : dump will be processed from time stamp <n>(usec)
 --end-time   <n> : dump will be processed until time stamp <n>(usec)
 --thread-state   : process thread state informations 
\endcode

Il est possible de limiter l'analyse du dump  un intervalle de la trace en 
prcisant une date de dbut et(/ou) une date de fin. avec les options 
<tt>--start-time</tt> et <tt>--end-time</tt>

L'option <tt>--thread-state</tt> permet d'afficher les informations relatives 
 ltat des threads

<p>
<u>Exemple</u>:
<p>
\code
rtd2lxt -e ~/tftp/program.elf dump
\endcode



On peut 


*/





#endif /* CYGONCE_RTMON_H */
